/* 
 * auxapp_trace.c
 *
 * 
 *
 */
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
#include <stdio.h>
#include <stdlib.h>
#include <bfd.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include <assert.h>


#ifdef DEBUG
# error remove this code
#else
#define DBPRINT(sw,format,args...) do {                                         \
        if (sw) {                                                               \
                fprintf (stderr, "%s:%s - " format,                    \
                         __FILE__, __FUNCTION__, ## args);        \
        }                                                                       \
} while (0)
#endif




uint32_t binary_desc_init (const char *binary_full_path, uint32_t base_address);
int32_t binary_desc_finish (uint32_t);
int32_t get_symbol_info_str (uint32_t bd, uint32_t eip, char *info, uint32_t len);


typedef struct _binary_desc_tag {
	uint32_t base_address;
	bfd *abfd;
	asection *text_section;
	uint32_t text_start;
	uint32_t text_end;
	asymbol ** symbol_table;
} binary_desc;

static int32_t get_symbols (binary_desc * bd);
static uint32_t is_initialized = 0;


/*****************************************************************************/
/* this function must be called once before any call to 
 * auxapp_get_symbol_info is called. if the function fails
 * then get-symbol-info should not be called
 * accepts: binary_full_path - full path to the binary whose
 *	symbols are to be looked for
 * returns: 0 on success, -1 on failure
 */
/*****************************************************************************/
uint32_t
binary_desc_init (const char *binary_full_path, uint32_t base_address)
{
	int32_t retval = -1;
	binary_desc * bd = NULL;

	/* XXX is this enough or do we need to protect for MT */
	if ( ! is_initialized ) {
		bfd_init ();
		is_initialized = 1;
	}

	bd = malloc (sizeof (*bd));
	if (bd == NULL) {
		DBPRINT (0, "malloc failed\n");
		goto out;
	}
	memset (bd, 0, sizeof (*bd));

	/* open the binary */
	bd->abfd = bfd_openr (binary_full_path, "elf32-i386");
	if (!bd->abfd) {
		DBPRINT (0, "can't open %s: %s\n",
			 binary_full_path,
			 bfd_errmsg (bfd_get_error ()));
		goto out;
	}

	if (!bfd_check_format (bd->abfd, bfd_object)) {
		DBPRINT (0, "can't load %s: %s\n",
			 binary_full_path,
			 bfd_errmsg (bfd_get_error ()));
		goto out;
	}

	bd->base_address = base_address;

	/* get descriptor of the text section */
	bd->text_section = bfd_get_section_by_name (bd->abfd, ".text");
	if (bd->text_section == NULL) {
		DBPRINT (0, "bfd_get_section_by_name failed: %s\n",
			 bfd_errmsg (bfd_get_error ()));
		goto out;
	}
	bd->text_start = bd->text_section->vma + bd->base_address;
	bd->text_end = bd->text_start + bd->text_section->size;

	/* get the symbols */
	if (get_symbols (bd)) {
		DBPRINT (0, "get_symbols failed\n");
		goto out;
	}

	retval = 0;
out:
	if (retval && bd) {
		free (bd);
		bd = NULL;
	}
	return (uint32_t)bd;
}

/*****************************************************************************/
/*****************************************************************************/
int32_t
binary_desc_finish (uint32_t bdh)
{
	binary_desc * bd = (binary_desc *)bdh;
	bool retval = true;

	if (bd) {
		if (bd->symbol_table) {
			free (bd->symbol_table);
			bd->symbol_table = NULL;
		}
		if (bd->abfd) {
			retval = bfd_close (bd->abfd);
			bd->abfd = NULL;
		}
		free (bd);
	}
	return retval ? 0 : -1;

}

/*****************************************************************************/
/*****************************************************************************/
int32_t
get_symbol_info_str (uint32_t bdh, uint32_t pc, char *info_str, uint32_t len)
{
	binary_desc * bd = (binary_desc *)bdh;
	int32_t retval = -1;

	DBPRINT (0, "pc=%x text_start=%x text_end=%x\n", pc, bd->text_start, bd->text_end);

	if ((bd->text_section != NULL) && (pc >= bd->text_start) && (pc < bd->text_end)) {
		const char *pc_filename = (const char *) 0;
		const char *pc_function = (const char *) 0;
		uint32_t pc_linenum = 0;
		int32_t ret;
		int32_t j, found;
		const char * sym_name, * file_name;
		uint32_t sym_addr, prev_sym_addr = 0;
	       	int32_t sym_index = -1, prev_sym_index = -1;


		/* get the line info */

		ret = bfd_find_nearest_line (bd->abfd, bd->text_section,
					   (struct bfd_symbol **) 0,
					   pc - bd->text_start,
					   &pc_filename, &pc_function,
					   &pc_linenum);
		DBPRINT ( !ret, "can't find line. pc=%x: %s\n", pc, bfd_errmsg (bfd_get_error ()));
		sym_name = (pc_function && pc_function[0] != '\0') ? pc_function : "???";

		/* get the symbol info */


		/* symbols are not sorted - need to run over all of them */
		for (found = j = 0; ! found && bd->symbol_table[j] != NULL; j ++) {
			/* look only for functions in text section */
			if ((bd->symbol_table[j]->section->index == bd->text_section->index) &&
					(bd->symbol_table[j]->flags & BSF_FUNCTION)) {
				sym_addr = bd->symbol_table[j]->value +
					bd->text_section->vma +
					bd->base_address;
				if (sym_addr == pc) {
					sym_index = j;
					found = 1;
				} else if ((sym_addr < pc) && (sym_addr >= prev_sym_addr)) {
					prev_sym_addr = sym_addr;
					prev_sym_index = j;
				}
			}
		}

		if ( ! found ) {
			sym_index = prev_sym_index;
		}
		if (sym_index < 0) {
			DBPRINT (0, "cannot find symbol. pc=%x\n", pc);
			goto out;
		}


		sym_addr = bd->symbol_table[sym_index]->value +
			bd->text_section->vma +
			bd->base_address;
		sym_name = bfd_asymbol_name (bd->symbol_table[sym_index]);
		file_name = (pc_filename && pc_filename[0] != '\0') ?  pc_filename : "???";

		/* print the info */

		snprintf (info_str, len, "    at %s+%d(%s:%d)@0x%x:ia32",
				sym_name, pc - sym_addr,
				strrchr (pc_filename, '/') ?
						strrchr (pc_filename, '/') + 1 :
						file_name,
				pc_linenum, pc);
		retval = 0;
	} else {
		DBPRINT (0, "pc is not in text section. pc=%x text-section=%x-%x\n",
				pc, bd->text_start, bd->text_end);
		goto out;
	}
out:
	if (retval != 0)
		snprintf (info_str, len, "    - no symbol - :ia32");
	return retval;
}

static int32_t
get_symbols (binary_desc * bd)
{
	int32_t retval = -1;
	long storage_needed;
	long number_of_symbols;
	
	storage_needed = bfd_get_symtab_upper_bound (bd->abfd);
	
	if (storage_needed < 0) {
		DBPRINT (0, "bfd_get_symtab_upper_bound failed: %s\n",
				bfd_errmsg (bfd_get_error ()));
		goto out;
	}
	
	if (storage_needed == 0) {
		DBPRINT (0, "no symbols\n");
		retval = 0;
		goto out;
	}
	bd->symbol_table = (asymbol **) malloc (storage_needed);
	if (bd->symbol_table == NULL) {
		DBPRINT (0, "malloc failed\n");
		goto out;
	}
	number_of_symbols = bfd_canonicalize_symtab (bd->abfd, bd->symbol_table);
	
	if (number_of_symbols < 0) {
		DBPRINT (0, "bfd_canonicalize_symtab failed: %s\n",
			bfd_errmsg (bfd_get_error ()));
		goto out;
	}
	
	retval = 0;
out:
	return retval;
}


