/* 
 * jni32_args.c
 *
 * 
 *
 */
 /*
 *
 * Copyright (C) 2005       Stanislav Polevic <stanislav.v.polevic@intel.com>
 * Copyright (C) 2005       Vasily Levchenko <vasily.v.levchenko@intel.com>
 * Copyright (C) 2005       Alexander Alexeev <alexander.n.alexeev@intel.com>
 * Copyright (C) 2005       Valery Ushakov <valery.e.ushakov@intel.com>
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
#include <jni.h>
#include <stdint.h>
#include <stdlib.h>

#include <ia32el_mme_api.h>
#include <jni32_args.h>
#include <jni32_stack.h>
#include <jni32_refmap.h>
#include <jni32_diag.h>

#define INVALID_SIGNATURE "Invalid signature letter"

static int32_t extract_array_args_worker(ia32el_context32 *ec32, const struct jni32_method_info *method, uint32_t sp32, jvalue *args);
static int32_t __extract_stack_args(uint32_t sp, jvalue *res, const struct jni32_method_info *method);
static int32_t extract_valist_args_worker(ia32el_context32 *ec32, const struct jni32_method_info *method, uint32_t sp32, jvalue *args);
static int32_t extract_stack_args_worker(ia32el_context32 *ec32, const struct jni32_method_info *method, uint32_t sp32, jvalue *args);

/*
 * copy value from parameter jvalue array to result array
 */
static int32_t
extract_array_args_worker(ia32el_context32 *ec32 UNUSED,
				const struct jni32_method_info *method,
				uint32_t sp32, jvalue *args)
{
	size_t i;
	jvalue *arr;
	ptr32_t jobj;


	assert(method != NULL);
	assert(args != NULL);
	assert(method->argc == strlen(method->sig));

	jobj = btmix_pop_jobject(&sp32);
	arr = (jvalue *) I2P(jobj);

	for (i = 0; i < method->argc; ++i) {
		switch (method->sig[i]) {
		case JNI_BOOL_CHAR:
			args[i].z = arr[i].z;
		break;
		case JNI_BYTE_CHAR:
			args[i].b = arr[i].b;
		break;
		case JNI_CHAR_CHAR:
			args[i].c = arr[i].c;
		break;
		case JNI_SHORT_CHAR:
			args[i].s = arr[i].s;
		break;
		case JNI_INT_CHAR:
			args[i].i = arr[i].i;
		break;
		case JNI_LONG_CHAR:
			args[i].j = arr[i].j;
		break;
		case JNI_FLOAT_CHAR:
			args[i].f = arr[i].f;
		break;
		case JNI_DOUBLE_CHAR:
			args[i].d = arr[i].d;
		break;
		case JNI_REF_CHAR:
			args[i].l = jni32_get_ref(arr[i].i);
		break;
		default:
			MME_THROW_EXCEPTION("invalid signature");
			
			assert(!INVALID_SIGNATURE);
			return -1;
		break;
		}
	}

	return 0;
}


/*
 * auxiliary function, pulls out arguments consecution
 * from arbitrary memory part
 */
static int32_t
__extract_stack_args(uint32_t sp, jvalue *res,
			const struct jni32_method_info *method)
{
	size_t i;


	for (i = 0; i < method->argc; ++i) {
		switch (method->sig[i]) {
		case JNI_BOOL_CHAR:
			res[i].z = btmix_pop_jboolean(&sp);
			break;
		case JNI_BYTE_CHAR:
			res[i].b = btmix_pop_jbyte(&sp);
			break;
		case JNI_CHAR_CHAR:
			res[i].c = btmix_pop_jchar(&sp);
			break;
		case JNI_SHORT_CHAR:
			res[i].s = btmix_pop_jshort(&sp);
			break;
		case JNI_INT_CHAR:
			res[i].i = btmix_pop_jint(&sp);
			break;
		case JNI_LONG_CHAR:
			res[i].j = btmix_pop_jlong(&sp);
			break;
		case JNI_FLOAT_CHAR:				/* ... causes widening! */
			res[i].f = btmix_pop_jdouble(&sp);
			break;
		case JNI_DOUBLE_CHAR:
			res[i].d = btmix_pop_jdouble(&sp);
			break;
		case JNI_REF_CHAR:
			res[i].l = jni32_demarshal_ref(&sp);
			break;
		default:
			MME_THROW_EXCEPTION("invalid signature");
			assert(!INVALID_SIGNATURE);
			return -1;
		break;
		}
	}

	return 0;
}

/*
 * pulls out valist pointer from stack and uses it as SP
 */
static int32_t
extract_valist_args_worker(ia32el_context32 *ec32 UNUSED,
				const struct jni32_method_info *method,
				uint32_t sp32, jvalue *args)
{
	uint32_t valist_sp =	btmix_pop_jobject(&sp32); /* XXX: uwe: _pop_ptr? */
	return (__extract_stack_args(valist_sp, args, method));
}


/*
 * pass sp32 for pulling arguments out
 */
static int32_t
extract_stack_args_worker(ia32el_context32 *ec32 UNUSED,
				const struct jni32_method_info *method,
				uint32_t sp32, jvalue *args)
{
	return ( __extract_stack_args(sp32, args, method));
}


#define EXTRACT_ARGS_WORKER(worker) do {\
	int32_t status;\
\
	status = (*worker)(ec32, method, sp32, args);\
	return args;\
} while (0)


/*
 * Returns array of arguments retrieved from stack.
 */
jvalue *
retrieve_args_stack(ia32el_context32 *ec32,
					const struct jni32_method_info *method,
					uint32_t sp32, jvalue *args)
{
	EXTRACT_ARGS_WORKER(extract_stack_args_worker);
}


/*
 * Returns array of arguments retreived from jvalue array.
 */
jvalue *
retrieve_args_array(ia32el_context32 *ec32,
					const struct jni32_method_info *method,
					uint32_t sp32, jvalue *args)
{
	EXTRACT_ARGS_WORKER(extract_array_args_worker);
}


/*
 * Returns array of arguments retreived from va_list.
 */
jvalue *
retrieve_args_valist(ia32el_context32 *ec32,
					const struct jni32_method_info *method,
					uint32_t sp32, jvalue *args)
{
	EXTRACT_ARGS_WORKER(extract_valist_args_worker);
}
