/* 
 * jni32_diag.c
 *
 * 
 *
 */
 /*
 *
 * Copyright (C) 2005       Stanislav Polevic <stanislav.v.polevic@intel.com>
 * Copyright (C) 2005       Vasily Levchenko <vasily.v.levchenko@intel.com>
 * Copyright (C) 2005       Alexander Alexeev <alexander.n.alexeev@intel.com>
 * Copyright (C) 2005       Valery Ushakov <valery.e.ushakov@intel.com>
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
#include <jni.h>
#include <mme_internal.h>
#include <jni32_diag.h>
#include <jni32_jnienv.h>

struct error_class_cache {
	const char *name;
	jclass clazz;
};

static void jni32_throw_error_impl (JNIEnv *, enum exceptions, const char *);

static struct error_class_cache throw_class_name[] = {
	[JNI32_UNKNOW] = {"java/lang/UnknownError", NULL},
	[JNI32_INTERNAL] = {"java/lang/InternalError", NULL},
	[JNI32_UNSATISFIED_LINK] = {"java/lang/UnsatisfiedLinkError", NULL},
	[JNI32_STACK_OVERFLOW] = {"java/lang/StackOverflowError", NULL},
	[JNI32_LINKAGE] = {"java/lang/LinkageError", NULL},
	[JNI32_OUT_OF_MEMORY] = {"java/lang/OutOfMemoryError", NULL},
};

/*
 * When jnimix encouters an error this function will get called to
 * report it to the java world.
 *
 * XXX: TODO: Preallocate OutOfMemoryError reference?
 */
static void
jni32_throw_error_impl (JNIEnv * env, enum exceptions exception_num, const char *message)
{
	jclass clazz;

	assert (exception_num < JNI32_NEXCEPTIONS);

	/* Is some other execption already being thrown? */
	if (__PREDICT_FALSE ((*env)->ExceptionCheck (env)))
		return;

	clazz = throw_class_name[exception_num].clazz;
	if (__PREDICT_FALSE (clazz == NULL)) {
		jclass clocal;

		clocal = (*env)->FindClass (env, throw_class_name[exception_num].name);
		if (__PREDICT_FALSE (clocal == NULL)) {
			/* FindClass has thrown an Error, so we don't have to */
			return;
		}

		clazz = (*env)->NewGlobalRef (env, clocal);
		(*env)->DeleteLocalRef (env, clocal);

		throw_class_name[exception_num].clazz = clazz;
	}

	(*env)->ThrowNew (env, clazz, message);
}

void
jni32_throw_error (enum exceptions exception_num, const char *message)
{
	JNIEnv *env;

	env = jni32_jnienv_get_current ();
	if (__PREDICT_FALSE (env == NULL)) {
		assert (!"JNIEnv is NULL");
		return;
	}

	jni32_throw_error_impl (env, exception_num, message);
}

void
jni32_throw_init_error (JavaVM * vm, enum exceptions exception_num, const char *message)
{
	int result;
	union {
		JNIEnv *env;
		void * ptr;
	} pc;


	assert (vm != NULL);

	result = (*vm)->GetEnv (vm, &pc.ptr, JNI_VERSION_1_1);

	if (__PREDICT_FALSE (result != JNI_OK)) {
		assert (exception_num < JNI32_NEXCEPTIONS);
		fprintf (stderr, "%s: %s\n",
				throw_class_name[exception_num].name + 10, /* - java/lang/ */
			 	message);
		return;
	}

	jni32_throw_error_impl (pc.env, exception_num, message);
}
