/* 
 * jni32_hash_table.c
 *
 * 
 *
 */
 /*
 *
 * Copyright (C) 2005       Stanislav Polevic <stanislav.v.polevic@intel.com>
 * Copyright (C) 2005       Vasily Levchenko <vasily.v.levchenko@intel.com>
 * Copyright (C) 2005       Alexander Alexeev <alexander.n.alexeev@intel.com>
 * Copyright (C) 2005       Valery Ushakov <valery.e.ushakov@intel.com>
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
#include <stdlib.h>
#include <string.h>

#include <mme_internal.h>
#include <jni32_hash_table.h>
#include <jni32_diag.h>

struct data_pair {
	key_type key;
	val_type value;
};

struct content_elem;

struct content_elem {
	struct data_pair data;
	struct content_elem *next;
};

struct _jni32_hash_table_t {
	struct content_elem **lists;
	hash_func hash_fptr;
	size_t lists_size; 
};

/*
 *
 */
jni32_hash_table HIDDEN
jni32_hash_table_create(size_t lists_size, hash_func fptr)
{
	struct _jni32_hash_table_t *table;

	assert(fptr);

	table = (jni32_hash_table) malloc(sizeof(struct _jni32_hash_table_t));
	if (__PREDICT_FALSE(table == NULL)) {
		MME_THROW_EXCEPTION("malloc failed");
		return NULL;
	}

	table->lists = (struct content_elem **)
					malloc(lists_size * sizeof(struct _jni32_hash_table_t*));

	if (__PREDICT_FALSE(table->lists == NULL)) {
 		MME_THROW_EXCEPTION("malloc failed");
		return NULL;
	}

	table->lists_size = lists_size;

	bzero(table->lists, lists_size * sizeof(struct _jni32_hash_table_t*));
	table->hash_fptr = fptr;

	return table;
}

/*
 *
 */
void HIDDEN
jni32_hash_table_destroy(const jni32_hash_table table)
{
	struct _jni32_hash_table_t *hash_table = table;
	struct content_elem *list_iter;
	struct content_elem *del_elem;
	size_t i;

	if (__PREDICT_TRUE(hash_table != NULL)) {
		for (i = 0; i < hash_table->lists_size; ++i) {

			list_iter = hash_table->lists[i];

			while (list_iter != NULL) {
				del_elem = list_iter;
				list_iter = list_iter->next;
				free(del_elem);
			}

		}

		free(hash_table->lists);
		free(hash_table);
	}
}

/*
 *
 */
jni32_hash_table_res HIDDEN
jni32_set_hash_table_value(const jni32_hash_table table,
							const key_type key, const val_type value)
{
	struct _jni32_hash_table_t *hash_table = table;
	size_t list_index;
	struct content_elem *list_iter;
	struct content_elem *new_elem;

	list_index = hash_table->hash_fptr(key);
	list_iter = hash_table->lists[list_index];

#if defined(BTMIX_DEBUG)
	while (list_iter != NULL) {
		ASSERT(key != list_iter->data.key);
		list_iter = list_iter->next;
	}
#endif

	new_elem = malloc(sizeof(struct content_elem));
	if (__PREDICT_FALSE(new_elem == NULL)) {
		MME_THROW_EXCEPTION("malloc failed");
		return JNI32_HASH_TABLE_ERROR;
	}
	new_elem->data.key = key;
	new_elem->data.value = value;

	new_elem->next = hash_table->lists[list_index];
	hash_table->lists[list_index] = new_elem;

	return JNI32_HASH_TABLE_KEY_NEW;
}

/*
 *
 */
jni32_hash_table_res HIDDEN
jni32_get_hash_table_value(const jni32_hash_table table, const key_type key, val_type *value)
{
	struct _jni32_hash_table_t *hash_table = table;
	size_t list_index;
	struct content_elem *list_iter;

	list_index = hash_table->hash_fptr(key);
	list_iter = hash_table->lists[list_index];

	while (list_iter != NULL) {
		if (key == list_iter->data.key) {
			*value = list_iter->data.value;
			return JNI32_HASH_TABLE_KEY_EXIST;
		}
		list_iter = list_iter->next;
	}
	return JNI32_HASH_TABLE_KEY_LACK;
}

/*
 *
 */
jni32_hash_table_res HIDDEN
jni32_del_hash_table_pair(const jni32_hash_table table, const key_type key)
{
	struct _jni32_hash_table_t *hash_table = table;
	size_t list_index;
	struct content_elem *list_iter;
	struct content_elem *del_elem;

	list_index = hash_table->hash_fptr(key);

	/* first element of the list*/
	if (hash_table->lists[list_index] != NULL) {
		if (key == hash_table->lists[list_index]->data.key) {
			del_elem = hash_table->lists[list_index];
			hash_table->lists[list_index] = hash_table->lists[list_index]->next;
			free(del_elem);
			return JNI32_HASH_TABLE_KEY_EXIST;
		}

		list_iter = hash_table->lists[list_index];

		while (list_iter->next != NULL) {
			if (key == list_iter->next->data.key) {
				del_elem = list_iter->next;
				list_iter->next = list_iter->next->next;
				free(del_elem);
				return JNI32_HASH_TABLE_KEY_EXIST;
			}
			list_iter = list_iter->next;
		}
	}

	return JNI32_HASH_TABLE_KEY_LACK;
}

