/* 
 * jni32_idmap.c
 *
 * 
 *
 */
 /*
 *
 * Copyright (C) 2005       Stanislav Polevic <stanislav.v.polevic@intel.com>
 * Copyright (C) 2005       Vasily Levchenko <vasily.v.levchenko@intel.com>
 * Copyright (C) 2005       Alexander Alexeev <alexander.n.alexeev@intel.com>
 * Copyright (C) 2005       Valery Ushakov <valery.e.ushakov@intel.com>
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
#include <pthread.h>
#include <jni.h>
#include <stdint.h>
#include <stdlib.h>

#include <mme_internal.h>
#include <jni32_hash_table.h>
#include <jni32_map.h>
#include <jni32_idmap.h>
#include <jni32_mth_parse.h>
#include <jni32_ref_tags.h>
#include <jni32_stack.h>


//XXX : add weak referenses


/*
 * FieldID defenitions
 */
#define FIELD_ID_HASH_FN_MODULE		701
#define FIELD_ID_MAP_INITIAL_SIZE	256
#define FIELD_ID_MAP_INCREMENT_SIZE	256
static size_t mod_fieldID_hash_func(const key_type value);
static void free_method_data(jni32_jmethodID method_data);
static void jni32_clear_methodIDs(void);
static size_t mod_methodID_hash_func(const key_type value);

struct fieldID_container { 
	struct jni32_map fid_map;
	jni32_hash_table fid_hash_tlb;
	pthread_mutex_t	cont_mutex;
};

struct fieldID_container HIDDEN fid_map;

/*
 * MethodID defenitions
 */
#define METHOD_ID_HASH_FN_MODULE	701
#define METHOD_ID_MAP_INITIAL_SIZE	256
#define METHOD_ID_MAP_INCREMENT_SIZE 256
size_t
mod_methodID_hash_func(const key_type value);

struct methodID_container {
	struct jni32_map mid_map;
	jni32_hash_table mid_hash_tlb;
	pthread_mutex_t	cont_mutex;
};

struct methodID_container HIDDEN mid_map;

/*
 * fieldID implementation
 */
static size_t
mod_fieldID_hash_func(const key_type value)
{
	return value % FIELD_ID_HASH_FN_MODULE;
};

uint32_t HIDDEN
marshal_fieldID(jfieldID fid)
{
	uint64_t key_fid64 = (uint64_t)fid;
	uint64_t value_fid32_ext;
	int32_t value_fid32;
	int32_t status;

	if (__PREDICT_FALSE(fid == NULL)) {
		return 0;
	}

	status = pthread_mutex_lock (&fid_map.cont_mutex);
	if (status != 0) return 0;;


	if (__PREDICT_FALSE(jni32_get_hash_table_value(fid_map.fid_hash_tlb,
													key_fid64, &value_fid32_ext)
						== JNI32_HASH_TABLE_KEY_EXIST)) {

		status = pthread_mutex_unlock (&fid_map.cont_mutex);
		if (status != 0) return 0;;

		return (SET_FILED_ID_TAG((uint32_t)value_fid32_ext));
	}

	jni32_map_add(&fid_map.fid_map, (void *)key_fid64, &value_fid32); 

	if (__PREDICT_FALSE(jni32_set_hash_table_value(fid_map.fid_hash_tlb,
													key_fid64, value_fid32)
						!= JNI32_HASH_TABLE_KEY_NEW)) {

		status = pthread_mutex_unlock (&fid_map.cont_mutex);
		if (status != 0) return 0;;

		return -1;
	}

	status = pthread_mutex_unlock (&fid_map.cont_mutex);
	if (status != 0) return 0;;

	return (SET_FILED_ID_TAG(value_fid32));
}

jfieldID HIDDEN
demarshal_fieldID(uint32_t *psp32)
{
	uint32_t fid32;
	jfieldID result;
	int32_t status;

	fid32 = btmix_pop_pointer(psp32);

	if (__PREDICT_FALSE(!IS_FILED_ID(fid32))) {
		return 0;
	}

	status = pthread_mutex_lock (&fid_map.cont_mutex);
	if (status != 0) return 0;;

	result = jni32_map_get(&fid_map.fid_map, RM_CERTAIN_TAG(fid32));

	status = pthread_mutex_unlock (&fid_map.cont_mutex);
	if (status != 0) return 0;;

	return result;
}

/*
 * methodID implementation
 */
static size_t
mod_methodID_hash_func(const key_type value)
{
	return value % METHOD_ID_HASH_FN_MODULE;
};

static void
free_method_data(jni32_jmethodID method_data)
{
	free((void *)method_data->method_info.name);
	free((void *)method_data);
	return;
}

uint32_t HIDDEN
marshal_methodID(jni32_jmethodID jni32_mid)
{
	if (__PREDICT_FALSE(jni32_mid == NULL)) {
		return 0;
	}
	return (SET_METHOD_ID_TAG(jni32_mid->mid32));
}

jni32_jmethodID HIDDEN
demarshal_methodID(uint32_t *psp32)
{
	uint32_t mid32;
	jni32_jmethodID result;
	int32_t status;

	mid32 = btmix_pop_pointer(psp32);
	if (__PREDICT_FALSE(!IS_METHOD_ID(mid32))) {
		/*if mid32 == NULL then it isn't method ID so returned NULL*/
		return NULL;
	}

	status = pthread_mutex_lock (&mid_map.cont_mutex);
	if (status != 0) return NULL;;

	result = jni32_map_get(&mid_map.mid_map, RM_CERTAIN_TAG(mid32));

	status = pthread_mutex_unlock (&mid_map.cont_mutex);
	if (status != 0) return NULL;;

	return result;
}

jni32_jmethodID HIDDEN
create_method_entry(jmethodID mid, const char *name, const char *sig)
{
	uint64_t key_mid64 = (uint64_t)mid;
	uint64_t value_mid32_ext;
	int32_t value_mid32;
	jni32_jmethodID method_data;
	int32_t status;

	status = pthread_mutex_lock (&mid_map.cont_mutex);
	if (status != 0) return NULL;;

	if (__PREDICT_FALSE(jni32_get_hash_table_value(mid_map.mid_hash_tlb,
													(uint64_t) mid,
													&value_mid32_ext)
						== JNI32_HASH_TABLE_KEY_EXIST)) {

		status = pthread_mutex_unlock (&mid_map.cont_mutex);
		if (status != 0) return NULL;;

		return ((jni32_jmethodID)mid_map.mid_map.data[value_mid32_ext].value);
	}

	method_data = malloc(sizeof(struct jni32_jmethodID_tag));

	if (method_data == NULL) {
	        return NULL;
	}

	method_data->mid64 = mid;

#if !defined(DMALLOC)
	method_data->method_info.name = strdup(name);
#else
	method_data->method_info.name = malloc(strlen(name) + 1);

	if (method_data->method_info.name == NULL) {
	        free_method_data(method_data);
		return NULL;
	}

	strcpy((char *)method_data->method_info.name, name);
#endif


	if (__PREDICT_FALSE(parse_sig(sig, &method_data->method_info) != 0)) {
		free_method_data(method_data);
		return NULL;
	}

	if (__PREDICT_FALSE(jni32_map_add(&mid_map.mid_map,
										(void *)method_data,
										&value_mid32) != 0)) {
		return NULL;
	}

	method_data->mid32 = value_mid32;

	if (__PREDICT_FALSE(JNI32_HASH_TABLE_KEY_NEW !=
						jni32_set_hash_table_value(mid_map.mid_hash_tlb,
													key_mid64, value_mid32))) {
		assert(!"method id error");
		return NULL;
	}

	status = pthread_mutex_unlock (&mid_map.cont_mutex);
	if (status != 0) return NULL;;

	return method_data;
}

static void
jni32_clear_methodIDs(void)
{
	uint32_t mid32 = 0;
	struct jni32_jmethodID_tag *mid64;
	mid64 = jni32_map_get(&mid_map.mid_map, mid32++);

	while (mid64) {
		if (mid64->method_info.argc != 0) { 
			free((void *)mid64->method_info.sig);
		}
		free((void *)mid64->method_info.name);
		free(mid64);
		mid64 = jni32_map_get(&mid_map.mid_map, mid32++);
	}
}

int32_t HIDDEN
init_id_maps(void)
{
	int32_t status;

	fid_map.fid_hash_tlb = jni32_hash_table_create(FIELD_ID_HASH_FN_MODULE, mod_fieldID_hash_func);

	if (__PREDICT_FALSE(fid_map.fid_hash_tlb == NULL)) {
		status = -1;
		goto out;
	}


	status = pthread_mutex_init (&fid_map.cont_mutex, NULL);
	if (status != 0) goto out;

	if (__PREDICT_FALSE(jni32_map_create(&fid_map.fid_map, FIELD_ID_MAP_INITIAL_SIZE, FIELD_ID_MAP_INCREMENT_SIZE) != 0)) {
		status = -1;
		goto out;
	}

	mid_map.mid_hash_tlb = jni32_hash_table_create(METHOD_ID_HASH_FN_MODULE, mod_methodID_hash_func);

	if (__PREDICT_FALSE(mid_map.mid_hash_tlb == NULL)) {
		status = -1;
		goto out;
	}

	status = pthread_mutex_init (&mid_map.cont_mutex, NULL);
	if (status != 0) goto out;

	if (__PREDICT_FALSE( jni32_map_create(&mid_map.mid_map, METHOD_ID_MAP_INITIAL_SIZE, METHOD_ID_MAP_INCREMENT_SIZE) != 0)) {
		status = -1;
		goto out;
	}

	status = 0;
out:
	return status;
}

int32_t HIDDEN
uninit_id_maps(void) 
{
	int32_t status;
	status = pthread_mutex_lock (&fid_map.cont_mutex);
	if (status != 0) goto out;

	jni32_clear_methodIDs();
	jni32_hash_table_destroy(fid_map.fid_hash_tlb);
	jni32_map_destroy(&fid_map.fid_map);

	status = pthread_mutex_unlock (&fid_map.cont_mutex);
	if (status != 0) goto out;

	status = pthread_mutex_destroy (&fid_map.cont_mutex);
	if (status != 0) goto out;

	status = pthread_mutex_lock (&mid_map.cont_mutex);
	if (status != 0) goto out;

	jni32_hash_table_destroy(mid_map.mid_hash_tlb);
	jni32_map_destroy(&mid_map.mid_map);

	status = pthread_mutex_unlock (&mid_map.cont_mutex);
	if (status != 0) goto out;

	status = pthread_mutex_destroy (&mid_map.cont_mutex);
	if (status != 0) goto out;

	status = 0;
out:
	return status;
}
