/* 
 * jni32_jnienv.c
 *
 * 
 *
 */
 /*
 *
 * Copyright (C) 2005       Stanislav Polevic <stanislav.v.polevic@intel.com>
 * Copyright (C) 2005       Vasily Levchenko <vasily.v.levchenko@intel.com>
 * Copyright (C) 2005       Alexander Alexeev <alexander.n.alexeev@intel.com>
 * Copyright (C) 2005       Valery Ushakov <valery.e.ushakov@intel.com>
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
#include <stdint.h>
#include <jni.h>

#include <jni32_jnienv.h>
#include <jni32_javavm.h>
#include <jni32_upcall.h>
#include <mme_internal.h>
#include <jni32.h>
#include <jni32_trampolines.h>
#include <jni32_refmap.h>
#include <jni32_ptrmap.h>
#include <jni32_upcall.h>
#include <jni32_diag.h>
#include <auxapp.h>
#include <btmix.h>


MME_THREAD_VAR HIDDEN struct JNIEnv32to64 * t_env32to64 = NULL;

/*****************************************************************************/
/*****************************************************************************/
static inline int32_t
jni32_jnienv_set_marshalled(struct JNIEnv32to64 *env)
{
	t_env32to64 = env;
	return 0;
}


/*****************************************************************************/
/*
 * Create IA32 JNIEnv with pointer to the function table of IA32
 * stubs, and with per-thread data, like real 64 bit JNIEnv.
 *
 * We will do this the first time ia32 function is called on the
 * current thread.	At this time we will need to init ia32 stack for
 * that thread and we can allocate this structure at the bottom of the
 * ia32 stack.
 */
/*****************************************************************************/
uint32_t HIDDEN
jni32_jnienv_marshal(ptr32_t * esp, JNIEnv *env, bool is_on_stack)
{
	ptr32_t new_esp;
	struct JNIEnv32to64 *env32to64;

	env32to64 = jni32_jnienv_get_marshalled();
	
	if (env32to64 != NULL) {
		/* psw: env is per-thread object */
		assert(env == env32to64->env64);
		return TRUNCATE_PTR(env32to64);
	}

	if (is_on_stack) {
		new_esp = *esp - sizeof(struct JNIEnv32to64);
		new_esp = ROUNDDOWN (new_esp, __alignof__ (struct JNIEnv32to64));
		*esp = new_esp;
		env32to64 = I2P (*esp);
		env32to64->is_malloc = false;
	} else {
		ptr32_t malloc_addr;
		malloc_addr = malloc32 (sizeof (*env32to64) * 2);
		if (malloc_addr == 0) {
			DBPRINT (true, "malloc32 failed\n");
			MME_THROW_EXCEPTION ("malloc32 failed");
			return 0;
		}
		env32to64 = I2P (ROUNDUP (malloc_addr,  __alignof__ (struct JNIEnv32to64)));
		env32to64->is_malloc = true;
		env32to64->malloc_addr = malloc_addr;
	}
	assert (IS_32BIT_VALUE(env32to64));

	/*
	* Stack pointer to the IA32 JNIEnv interface function table. As
	* it's the first member of the structure, (*env)->f() will do the
	* right thing in the IA32 world.
	*/
	env32to64->fns32 = g_ia32_runtime.jnienv32_fns;
	env32to64->env64 = env;

	DBPRINT (1, "fns32=%x env64=%p is_malloc=%d malloc_addr=%x\n",
			env32to64->fns32, env32to64->env64, env32to64->is_malloc, env32to64->malloc_addr);

	int32_t status;
	status = jni32_jnienv_set_marshalled(env32to64);
	if (status != 0) {
		return 0;
	}


	
	return TRUNCATE_PTR(env32to64);
}



int32_t HIDDEN
jni32_jnienv_destroy ()
{
	int32_t retval;
	struct JNIEnv32to64 *env;

	env = jni32_jnienv_get_marshalled ();

	if(env == NULL) {
		return 0;
	}

	DBPRINT (1, "fns32=%x env64=%p is_malloc=%d malloc_addr=%x\n",
		env->fns32, env->env64, env->is_malloc, env->malloc_addr);

	if (env->is_malloc) {
		free32 (env->malloc_addr);
	}

	retval = jni32_jnienv_set_marshalled (NULL);
	if (retval != 0) {
		DBPRINT (true, "jni32_jnienv_set_marshalled failed\n");
		MME_THROW_EXCEPTION ("jni32_jnienv_set_marshalled failed");
	}
	return retval;
}




/* define array of trampolines */
#define JNIENVLAST(_idx)	\
	jni32_trampoline_t HIDDEN jni32_jnienv_trampolines[_idx]

#include <jnifns.h>

= {

#define JNIENVFN(_idx, _type, _name, _args)	\
	[_idx] = (void *)JNIENV_IA64_TRAMPOLINE(_name),

#include <jnifns.h>
}; /* jni32_trampolines definition */




int64_t HIDDEN
jni32_jnienv_upcall(ia32el_context32 *ec32)
{
	int64_t result;

	result = jni32_upcall(ec32, jni32_jnienv_trampolines, MAX_JNIENV_IDX);

	return result;
}


/*
 * Trampolines implementation
 */

DEFINE_JNIENV_IA64_TRAMPOLINE(GetVersion)
{
	JNIEnv *env;
	jint result;
	ptr32_t sp32;
	struct JNIEnv32to64 *env32to64;

	sp32 = btmix_get_upcall_argp(ec32);
	assert(sp32 != 0);

	/* first argument is envp */
	env = jni32_jnienv_demarshal(&sp32);

	if (env == NULL) {
	        env32to64 = jni32_jnienv_get_marshalled();
		if (env32to64 == NULL) {
		        __SET_RETVAL_PRIMITIVE(jint, JNI_ERR);
			return 0;
		} else {
		        env = env32to64->env64;
			result = (*env)->GetVersion(NULL);
		}
	} else {
	        result = (*env)->GetVersion(env);
	}

	__SET_RETVAL_PRIMITIVE(jint, result);

	return 0;
}

DEFINE_JNIENV_IA64_TRAMPOLINE(MonitorEnter)
{
	JNIEnv *env;
	jobject obj;
	jint result;
	ptr32_t sp32;

	sp32 = btmix_get_upcall_argp(ec32);
	assert(sp32 != 0);

	/* first argument is envp */
	env = jni32_jnienv_demarshal(&sp32);
	assert(env != NULL);

	obj = jni32_demarshal_ref(&sp32);

	result = (*env)->MonitorEnter(env, obj);

	__SET_RETVAL_PRIMITIVE(jint, result);

	return 0;
}

DEFINE_JNIENV_IA64_TRAMPOLINE(MonitorExit)
{
	JNIEnv *env;
	jobject obj;
	jint result;
	ptr32_t sp32;

	sp32 = btmix_get_upcall_argp(ec32);
	assert(sp32 != 0);

	/* first argument is envp */
	env = jni32_jnienv_demarshal(&sp32);
	assert(env != NULL);

	obj = jni32_demarshal_ref(&sp32);

	result = (*env)->MonitorExit(env, obj);

	__SET_RETVAL_PRIMITIVE(jint, result);

	return 0;
}

DEFINE_JNIENV_IA64_TRAMPOLINE(GetJavaVM)
{
	JNIEnv *env_p, *env;
	ptr32_t sp32;
	uint32_t *vm32;
	JavaVM *vm;
	jint result;
	struct JNIEnv32to64 *env32to64;

	sp32 = btmix_get_upcall_argp(ec32);
	assert(sp32 != 0);

	env_p = jni32_jnienv_demarshal(&sp32);

	if (env_p == NULL) {
	        env32to64 = jni32_jnienv_get_marshalled();
		if (env32to64 == NULL) {
		        __SET_RETVAL_PRIMITIVE(jint, JNI_ERR);
			return 0;
		} else {
		        env_p = env32to64->env64;
			env = NULL;
		}
	} else {
	        env = env_p;
	}

	vm32 = demarshal_ptr(&sp32);

	if (vm32 == NULL) {
	        result = (*env_p)->GetJavaVM(env, NULL);
		__SET_RETVAL_PRIMITIVE(jint, result);
		return 0;
	}

	result = (*env_p)->GetJavaVM(env, &vm);
	if (__PREDICT_FALSE(result == JNI_OK)) {
		/* fill IA32_VM struct */
		*vm32 = TRUNCATE_PTR(jni32_javavm_marshal(vm));
	}

	__SET_RETVAL_PRIMITIVE(jint, result);

	return 0;
}



