/* 
 * jni32_mth_parse.c
 *
 * 
 *
 */
 /*
 *
 * Copyright (C) 2005       Stanislav Polevic <stanislav.v.polevic@intel.com>
 * Copyright (C) 2005       Vasily Levchenko <vasily.v.levchenko@intel.com>
 * Copyright (C) 2005       Alexander Alexeev <alexander.n.alexeev@intel.com>
 * Copyright (C) 2005       Valery Ushakov <valery.e.ushakov@intel.com>
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <jni.h>


#include <mme_internal.h>
#include <jni32_mth_parse.h>
#include <jni32_diag.h>

static int32_t isConstructor(JNIEnv *env, jobject tested_obj);
static size_t __get_sig_size(const char *sig);
static size_t __parse_sig_worker(const char *sig, const char start_token, const char end_token, char *out_sig);
static void replace_dot(char* str) UNUSED;
static int32_t get_type_for_string(const char* str);
static jsize add_class_name(JNIEnv *env, jobject class_obj, char* buf);
static jobject get_args_array(JNIEnv *env, jobject method_obj);
static jobject get_argument_name(JNIEnv *env, jobject arg_array_obj, jsize num) UNUSED;
static jsize get_result_type(JNIEnv *env, jobject method_obj, char* sig_buf);

static size_t
__get_sig_size(const char *sig)
{
	size_t sig_size;
	uint32_t i;
	int32_t obj_mode = 0;
	size_t result = 0;


	sig_size = strlen(sig);
	for (i = 0; i < sig_size; ++i) {
		if (obj_mode == 0) {
			switch (sig[i]) {
			case JNI_REF_CHAR:
				++result;
				obj_mode = 1;
			break;
			case JNI_ARRAY_CHAR:
				++result;
				++i;
				if (sig[i] == JNI_REF_CHAR) {
					obj_mode = 1;
				}
			break;
			case JNI_SIG_START:
				result = 0;
			break;
			case JNI_SIG_END:
				return result;
			break;
			default:
				result = obj_mode == 0 ? result + 1 : result;
			break;
			}
		}
		else if (sig[i] == JNI_REF_END) {
			obj_mode = 0;
		}
	}

	assert(0);
	return -1;
}

/*
 * Runs in two modes. 
 * If res is NULL - counts a size of new sig 
 * and returns new sig otherwise.
 */ 
static size_t
__parse_sig_worker(const char *sig, const char start_token,
					const char end_token, char *out_sig)
{
	size_t sig_size;
	char *sig_ptr;
	uint32_t i;
	uint32_t obj_mode = 0;
	size_t out_sig_size = 0;


	sig_size = strlen(sig);
	sig_ptr = strchr(sig, start_token);

	for (i = 0; i < sig_size && sig_ptr[i] != end_token; ++i) {
		if (sig_ptr[i] == start_token) {
			continue;
		}
		if (sig_ptr[i] == end_token) {
			
			return -1;
		}

		if (obj_mode == 0) {
			switch (sig_ptr[i]) {
			case JNI_BOOL_CHAR:
			case JNI_BYTE_CHAR:
			case JNI_CHAR_CHAR:
			case JNI_SHORT_CHAR:
			case JNI_INT_CHAR:
			case JNI_LONG_CHAR:
			case JNI_FLOAT_CHAR:
			case JNI_DOUBLE_CHAR:
				out_sig[out_sig_size] = sig_ptr[i];
				++out_sig_size;
			break;
			case JNI_REF_CHAR:
				obj_mode = 1;
				out_sig[out_sig_size] = sig_ptr[i];
				++out_sig_size;
				break;
			case JNI_ARRAY_CHAR:
				out_sig[out_sig_size] = JNI_REF_CHAR;
				++out_sig_size;
				++i;
				if (sig[i] == JNI_REF_CHAR) {
					obj_mode = 1;
				}
				break;
			case JNI_VOID_CHAR:
				out_sig[out_sig_size] = sig_ptr[i];
				++out_sig_size;
			break;
			default:
				
				assert(!"Invalid signature char");
				return -1;
			}
		} else if (sig[i] == JNI_REF_END) {
			obj_mode = 0;
		}
	}

	return out_sig_size;
}


/*
 * Parses signature and fills specified string with type tags.
 * Skips return type.
 * Result string is a bunch of type tags, one per character.
 */
int32_t HIDDEN
parse_sig(const char *sig, struct jni32_method_info *method)
{
	int32_t status;
	char *out_sig = NULL;
	char out_ret[] = {'\0', '\0'};
	size_t out_sig_size = 0;


	status = __parse_sig_worker(sig, JNI_SIG_END, '\0', out_ret);
	if (status == -1) {
		return -1;
	}

	out_sig_size = __get_sig_size(sig);
	if ((int64_t)out_sig_size == -1) {
		
		return -1;
	}
	else if (out_sig_size == 0) {
		/* We have no arguments at all */
		out_sig = "";
	} 
	else {
		out_sig = malloc(out_sig_size + 1);
		if (out_sig == NULL) {
			MME_THROW_EXCEPTION("malloc failed");
			return -1;
		}
		out_sig[out_sig_size] = '\0';
		status = __parse_sig_worker(sig, JNI_SIG_START, JNI_SIG_END, out_sig);
		if (status == -1) {
			return -1;
		}
	}

	method->sig	= out_sig;
	method->argc = out_sig_size;
	method->ret	= out_ret[0];

	
	return 0;
}


/*
 * Maping array (type number -> type letter)
 */
static const char* const m_type_symbols[] = {
	JNI_BOOL_SYM,
	JNI_BYTE_SYM,
	JNI_CHAR_SYM,
	JNI_SHORT_SYM,
	JNI_INT_SYM,
	JNI_LONG_SYM,
	JNI_FLOAT_SYM,
	JNI_DOUBLE_SYM,
	JNI_VOID_SYM
};

/*
 * Change class name to class path
 */
static void 
replace_dot(char* str)
{
	assert(str != NULL);

	size_t str_len = strlen(str);
	size_t i;
	for (i = 0; i < str_len; ++i) {
		if (str[i] == '.') str[i] = '/';
	}
}

/*
 *	Define type number by type name
 */
static int32_t 
get_type_for_string(const char* str)
{
	assert(str != NULL);


	if (!strcmp(str, JNI_BOOL_SIGN)) {
		return JNI_BOOL; 
	}
	else if (!strcmp(str, JNI_BYTE_SIGN))	{
		return JNI_BYTE; 
	}
	else if (!strcmp(str, JNI_CHAR_SIGN))	{
		return JNI_CHAR; 
	}
	else if (!strcmp(str, JNI_SHORT_SIGN))	{
		return JNI_SHORT; 
	}
	else if (!strcmp(str, JNI_INT_SIGN))	{
		return JNI_INT; 
	}
	else if (!strcmp(str, JNI_LONG_SIGN))	{
		return JNI_LONG; 
	}
	else if (!strcmp(str, JNI_FLOAT_SIGN))	{
		return JNI_FLOAT; 
	}
	else if (!strcmp(str, JNI_DOUBLE_SIGN)) {
		return JNI_DOUBLE; 
	}
	else if (!strcmp(str, JNI_VOID_SIGN)) {
		return JNI_VOID; 
	}
	else if (!strncmp(str, JNI_ARRAY_REF_SIGN, 2)) {
		return JNI_REF_ARRAY; 
	}
	else if (!strncmp(str, JNI_ARRAY_PRIM_SIGN, 1)) {
		return JNI_PRIM_ARRAY; 
	}
	else {
		return JNI_REF; 
	}

	assert(0);
	return -1;
}

/*
 * Return name of method from method_obj 
 */
// XXX: try to find the same proc and take it out in a separate file
jsize HIDDEN
get_method_name(JNIEnv *env, jobject method_obj, char* name_buf)
{
	jsize ret_size = 0;
	jclass method_class = 0;
	jmethodID methodid_getname = 0;
	jobject method_name_obj = 0;
	const char *name = 0;
	const char *init_str = "<init>";

	if (isConstructor(env, method_obj)) {
		ret_size = strlen(init_str) + 1;
		if (NULL != name_buf) {
			strcpy(name_buf, init_str);
			name_buf[ret_size - 1] = '\0';
		}
	} else {	
		method_class = (*env)->GetObjectClass(env, method_obj);
		if (__PREDICT_FALSE(method_class == NULL)) {
			return 0;
		}
		methodid_getname = (*env)->GetMethodID(env, method_class,
												"getName", "()Ljava/lang/String;");
		if ((methodid_getname == NULL)) {
			return 0;
		}

		method_name_obj = (*env)->CallObjectMethod(env, method_obj,
													methodid_getname);
		if ((method_name_obj == NULL)) {
			return 0;
		}

		name = (*env)->GetStringUTFChars(env, method_name_obj, NULL); 

		if (__PREDICT_FALSE(name == NULL)) {
			return 0;
		}

		ret_size = strlen(name) + 1;

		if (NULL != name_buf) {
			strcpy(name_buf, name);
			name_buf[ret_size - 1] = '\0';
		}

		(*env)->ReleaseStringUTFChars(env, method_name_obj, name);
	}

	return ret_size;

}

/*
 *	add class name of the class_obj content to the signature string
 */
static jsize
add_class_name(JNIEnv *env, jobject class_obj, char* buf)
{
	jclass class_class = 0;
	jmethodID methodid_getname = 0;
	jstring string_obj = 0;
	const char *type_name = 0;
	int32_t type_num = 0;
	jsize class_name_len = 0;

	class_class = (*env)->GetObjectClass(env, class_obj);
	if (__PREDICT_FALSE(class_class == NULL)) {
		return 0;
	}

	methodid_getname = (*env)->GetMethodID(env, class_class,
											"getName", "()Ljava/lang/String;");
	if (__PREDICT_FALSE(methodid_getname == NULL)) {
		return 0;
	}

	string_obj = (*env)->CallObjectMethod(env, class_obj, methodid_getname);
	if (__PREDICT_FALSE(string_obj == NULL)) {
		return 0;
	}

	type_name = (*env)->GetStringUTFChars(env, string_obj, NULL);
	if (__PREDICT_FALSE(type_name == NULL)) {
		return 0;
	}

	type_num = get_type_for_string(type_name);

	switch (type_num) {
	case JNI_BOOL:
	case JNI_BYTE:
	case JNI_CHAR:
	case JNI_SHORT:
	case JNI_INT:
	case JNI_LONG:
	case JNI_FLOAT:
	case JNI_DOUBLE:
		if (0 != buf) {
			strcat(buf, m_type_symbols[type_num]); 
		}
		class_name_len = class_name_len + 1; 
		break;
	case JNI_REF:
		class_name_len = class_name_len + strlen(type_name) + 2;
		if (0 != buf) {
			strcat(buf, "L");
			strcat(buf, type_name);
			strcat(buf, ";");
		}
		break;
	case JNI_REF_ARRAY: 
	case JNI_PRIM_ARRAY: 
		class_name_len = class_name_len + strlen(type_name);
		if (0 != buf) {
			strcat(buf, type_name); 
		}
		break;
	default: 
		assert(!"incorrect type number");
		break;
	}

	(*env)->ReleaseStringUTFChars(env, string_obj, type_name);
	return class_name_len;
}

/*
 *	return array object with argument's type of method_obj
 */
static jobject 
get_args_array(JNIEnv *env, jobject method_obj)
{
	jclass method_class;
	jmethodID methodid_getParameterTypes;
	jobject arg_array_obj;

	method_class = (*env)->GetObjectClass(env, method_obj);
	if (__PREDICT_FALSE(method_class == NULL)) {
		return 0;
	}

	methodid_getParameterTypes = (*env)->GetMethodID(env, method_class,
													"getParameterTypes",
													"()[Ljava/lang/Class;");

	if (__PREDICT_FALSE(methodid_getParameterTypes == NULL)) {
		return 0;
	}

	arg_array_obj = (*env)->CallObjectMethod(env, method_obj,
											methodid_getParameterTypes);
	if (__PREDICT_FALSE(arg_array_obj == NULL)) {
		return 0;
	}

	return arg_array_obj;
}

/*
 *	return string object with name of type from arguments's array
 */
static jobject
get_argument_name(JNIEnv *env, jobject arg_array_obj, jsize num)
{

		jobject argument_type;

		argument_type = (*env)->GetObjectArrayElement(env, arg_array_obj, num);
		if (__PREDICT_FALSE(argument_type == NULL)) {
			return NULL;
		}

		return NULL;// XXX: define what is realy returned
}

/*
 *	return string with result type of the method
 */
static jsize
get_result_type(JNIEnv *env, jobject method_obj, char* sig_buf)
{

	jclass method_class;
	jmethodID methodid_getReturnType;
	jobject ret_class_obj;
	jsize ret_type_len;

	if (isConstructor(env, method_obj)) {
		ret_type_len = 1;
		if (0 != sig_buf) {
			strcat(sig_buf, "V"); 
		}
	} else {	
		method_class = (*env)->GetObjectClass(env, method_obj);
		if (__PREDICT_FALSE(method_class == NULL)) {
			return 0;
		}


		methodid_getReturnType = (*env)->GetMethodID(env, method_class,
													"getReturnType",
													"()Ljava/lang/Class;");

		if (__PREDICT_FALSE(methodid_getReturnType == NULL)) {
			return 0;
		}

		ret_class_obj = (*env)->CallObjectMethod(env, method_obj,
												methodid_getReturnType);

		if (__PREDICT_FALSE(ret_class_obj == NULL)) {
			return 0;
		}

		ret_type_len = add_class_name(env, ret_class_obj, sig_buf);

	}
	return ret_type_len;
}


/*
 *	return consistent method's signature
 */
jsize HIDDEN
get_method_sig(JNIEnv *env, jobject method_obj, char* sig_buf)
{

	jobject arg_array_obj;
	jsize array_size;
	jsize sig_len = 0;
	jobject argument_class_obj;
	int32_t i = 0;

	arg_array_obj = get_args_array(env, method_obj);
	if (__PREDICT_FALSE(arg_array_obj == NULL)) {
		return 0;
	}

	array_size = (*env)->GetArrayLength(env, arg_array_obj); 
	if (__PREDICT_FALSE(-1 == array_size)) {
		return 0;
	}

	sig_len = sig_len + 1;
	if (sig_buf) {
		sig_buf = strcat(sig_buf, "("); 
	}


	for (i = 0; i < array_size; ++i) {
		argument_class_obj =
			(*env)->GetObjectArrayElement(env, arg_array_obj, i);

		if (argument_class_obj == NULL) {
			return 0;
		}

		sig_len = sig_len + add_class_name(env, argument_class_obj, sig_buf);
	}

	sig_len = sig_len + 1;
	if (sig_buf) {
		sig_buf = strcat(sig_buf, ")"); 
	}

	sig_len = sig_len + get_result_type(env, method_obj, sig_buf);

	sig_len = sig_len + 1;
	if (sig_buf) {
		sig_buf[sig_len - 1] = '\0'; 
	}
	return sig_len;
}

static jmethodID m_array_getComponentTypeID = NULL;
static jmethodID m_class_getNameID = NULL;

int32_t HIDDEN
get_array_component_jtype(JNIEnv *env, jobject array)
{
	jclass array_clazz, type_clazz, clazz_clazz;
	jobject componentTypeObj;
	jstring jname;
	int32_t jtype;
	const char *jname_chars;


	array_clazz = (*env)->GetObjectClass(env, array);
	assert(array_clazz != NULL);

	clazz_clazz = (*env)->GetObjectClass(env, array_clazz);
	assert(clazz_clazz != NULL);
	if (__PREDICT_FALSE(m_array_getComponentTypeID == NULL)) {
		m_array_getComponentTypeID = (*env)->GetMethodID(env, clazz_clazz,
							"getComponentType", "()Ljava/lang/Class;");
		assert(m_array_getComponentTypeID != NULL);
	}

	componentTypeObj = (*env)->CallObjectMethod(env, array_clazz,
						m_array_getComponentTypeID);
	type_clazz = (*env)->GetObjectClass(env, componentTypeObj);
	assert(type_clazz != NULL);
	if (__PREDICT_FALSE(m_class_getNameID == NULL)) {
		m_class_getNameID = (*env)->GetMethodID(env, clazz_clazz,
					"getName", "()Ljava/lang/String;");
		assert(m_class_getNameID != NULL);
	}

	jname = (*env)->CallObjectMethod(env, componentTypeObj, m_class_getNameID);
	assert(jname != NULL);

	jname_chars = (*env)->GetStringUTFChars(env, jname, NULL);
	assert(jname_chars != NULL);

	jtype = get_type_for_string(jname_chars);

	(*env)->ReleaseStringUTFChars(env, jname, jname_chars);

	return jtype;
}

size_t HIDDEN
get_jtype_size(int32_t jtype)
{
	size_t size;

	switch (jtype) {
	case JNI_BOOL:
		size = sizeof(jboolean);
	break;
	case JNI_BYTE:
		size = sizeof(jbyte);
	break;
	case JNI_CHAR:
		size = sizeof(jchar);
	break;
	case JNI_SHORT:
		size = sizeof(jshort);
	break;
	case JNI_INT:
		size = sizeof(jint);
	break;
	case JNI_LONG:
		size = sizeof(jlong);
	break;
	case JNI_FLOAT:
		size = sizeof(jfloat);
	break;
	case JNI_DOUBLE:
		size = sizeof(jdouble);
	break;
	case JNI_REF:
		size = sizeof(jobject);
	break;
	default:
		assert(!"Invalid java type");
		MME_THROW_EXCEPTION("Invalid java type");
		size = -1;
	break;
	}

	return size;
}

static int32_t
isConstructor(JNIEnv *env, jobject tested_obj)
{
	jclass unknown_class = (*env)->GetObjectClass(env, tested_obj);
	if (__PREDICT_FALSE(unknown_class == NULL)) {
		return 0;
	}

	jclass constr_class = (*env)->FindClass(env, "java/lang/reflect/Constructor");
	if (__PREDICT_FALSE(constr_class == NULL)) {
		return 0;
	}

	return (*env)->IsSameObject(env, unknown_class, constr_class);
}

int32_t HIDDEN 
isMethod(JNIEnv *env, jobject tested_obj)
{
	jclass unknown_class = (*env)->GetObjectClass(env, tested_obj);
	if (__PREDICT_FALSE(unknown_class == NULL)) {
		return 0;
	}

	jclass constr_class = (*env)->FindClass(env, "java/lang/reflect/Constructor");
	if (__PREDICT_FALSE(constr_class == NULL)) {
		return 0;
	}

	jclass method_class = (*env)->FindClass(env, "java/lang/reflect/Method");
	if (__PREDICT_FALSE(method_class == NULL)) {
		return 0;
	}

	return (((*env)->IsSameObject(env, unknown_class, constr_class)) ||
		((*env)->IsSameObject(env, unknown_class, method_class)));

}
