/* 
 * jni32_ptrmap.c
 *
 * 
 *
 */
 /*
 *
 * Copyright (C) 2005       Stanislav Polevic <stanislav.v.polevic@intel.com>
 * Copyright (C) 2005       Vasily Levchenko <vasily.v.levchenko@intel.com>
 * Copyright (C) 2005       Alexander Alexeev <alexander.n.alexeev@intel.com>
 * Copyright (C) 2005       Valery Ushakov <valery.e.ushakov@intel.com>
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
#include <stdio.h>
#include <stdint.h>
#include <pthread.h>

#include <jni32_ptrmap.h>
#include <jni32_hash_table.h>
/*
 * "IA64"->"IA32" pointer specific.
 */

// static declaration for ptrmap

#define HASH_FN_MODULE 701

static inline size_t
mod_hash_func(const key_type value) {
	return value % HASH_FN_MODULE;
}


struct jni32_ptrmap{
	pthread_mutex_t ptrmap_mutex;
	jni32_hash_table ptrs_hash_tlb;
};

static struct jni32_ptrmap m_ptrs_map = {PTHREAD_MUTEX_INITIALIZER, 0};

int32_t HIDDEN
init_ptrmap(void)
{
	int32_t status;

	

	m_ptrs_map.ptrs_hash_tlb = jni32_hash_table_create(HASH_FN_MODULE, mod_hash_func);

	if (m_ptrs_map.ptrs_hash_tlb == NULL) {
		DBPRINT (true, "jni32_hash_table_create failed\n");
		status = pthread_mutex_unlock (&m_ptrs_map.ptrmap_mutex);
		DBPRINT (status != 0, "pthread_mutex_unlock failed\n");
		return -1;
	}

	status = pthread_mutex_unlock (&m_ptrs_map.ptrmap_mutex);
	if (status != 0) {
		DBPRINT (status != 0, "pthread_mutex_unlock failed\n");
		return -1;
	}

	return 0;
}

int32_t HIDDEN
uninit_ptrmap(void) 
{
	int32_t status;

	status = pthread_mutex_lock (&m_ptrs_map.ptrmap_mutex);
	if (status != 0) {
		DBPRINT (status != 0, "pthread_mutex_lock failed\n");
		return -1;
	}

	jni32_hash_table_destroy(m_ptrs_map.ptrs_hash_tlb);
	m_ptrs_map.ptrs_hash_tlb = NULL;

	status = pthread_mutex_unlock (&m_ptrs_map.ptrmap_mutex);
	if (status != 0) {
		DBPRINT (status != 0, "pthread_mutex_unlock failed\n");
		return -1;
	}

	return 0;
}

HIDDEN void *
jni32_add_ptrmap(void *ptr64, size_t size) 
{
	int32_t status;

	jni32_hash_table_res hash_tbl_res;
	void *ptr32;

	if (ptr64 == NULL) {
		return NULL;
	}

	ptr32 = I2P(malloc32(size));

	if (ptr32 == NULL) {
		DBPRINT (true, "malloc32 failed\n");
		return NULL;
	}

	status = pthread_mutex_lock (&m_ptrs_map.ptrmap_mutex);
	if (status != 0) {
		DBPRINT (true, "pthread_mutex_lock failed\n");
		return NULL;
	}

	hash_tbl_res = jni32_set_hash_table_value(m_ptrs_map.ptrs_hash_tlb,
			(uint64_t)ptr32, (uint64_t)ptr64);

	if (__PREDICT_FALSE(hash_tbl_res == JNI32_HASH_TABLE_ERROR ||
				hash_tbl_res == JNI32_HASH_TABLE_KEY_EXIST)) {

		status = pthread_mutex_unlock (&m_ptrs_map.ptrmap_mutex);
		DBPRINT (status != 0, "pthread_mutex_unlock failed\n");
		return NULL;
	}

	status = pthread_mutex_unlock (&m_ptrs_map.ptrmap_mutex);
	if (status != 0) {
		DBPRINT (true, "pthread_mutex_unlock failed\n");
		return NULL;
	}

	return ptr32;
}

HIDDEN void *
jni32_get_ptrmap(void *ptr32) 
{
	union {
		void *ptr64;
		uint64_t ptr;
	} pc;
	jni32_hash_table_res hash;

	int32_t status;

	if (ptr32 == NULL) {
		return NULL;
	}

	status = pthread_mutex_lock (&m_ptrs_map.ptrmap_mutex);
	if (status != 0) {
		DBPRINT (true, "pthread_mutex_lock failed\n");
		return NULL;
	}

	hash = jni32_get_hash_table_value(m_ptrs_map.ptrs_hash_tlb, (uint64_t)ptr32, (uint64_t*)&pc.ptr);
	if (hash == JNI32_HASH_TABLE_KEY_LACK) {
		DBPRINT (true, "JNI32_HASH_TABLE_KEY_LACK\n");
		status = pthread_mutex_unlock (&m_ptrs_map.ptrmap_mutex);
		DBPRINT (status != 0, "pthread_mutex_unlock failed\n");
		return NULL; 
	}

	status = pthread_mutex_unlock (&m_ptrs_map.ptrmap_mutex);
	if (status != 0) {
		DBPRINT (true, "pthread_mutex_unlock failed\n");
		return NULL;
	}

	return pc.ptr64;
}

int32_t HIDDEN
jni32_del_ptrmap(void *ptr32, int32_t free_ptr32) 
{
	int32_t status;
	int32_t retval = -1;

	if (ptr32 == NULL) {
		goto out;
	}

	status = pthread_mutex_lock (&m_ptrs_map.ptrmap_mutex);
	if (status != 0) {
		DBPRINT (true, "pthread_mutex_lock failed\n");
		goto out;
	}

	if (jni32_del_hash_table_pair(m_ptrs_map.ptrs_hash_tlb, (uint64_t)ptr32) == JNI32_HASH_TABLE_KEY_LACK) {
		assert(!"del ptr from ptrmap error");
		goto out;
	}

	status = pthread_mutex_unlock (&m_ptrs_map.ptrmap_mutex);
	if (status != 0) {
		DBPRINT (true, "pthread_mutex_unlock failed\n");
		goto out;
	}

	if (free_ptr32 != 0) {
		assert (IS_32BIT_VALUE (ptr32));
		free32(TRUNCATE_PTR(ptr32));
	}

	retval = 0;
out:
	return 0;
}

