/* 
 * jni32_refmap.c
 *
 * 
 *
 */
 /*
 *
 * Copyright (C) 2005       Stanislav Polevic <stanislav.v.polevic@intel.com>
 * Copyright (C) 2005       Vasily Levchenko <vasily.v.levchenko@intel.com>
 * Copyright (C) 2005       Alexander Alexeev <alexander.n.alexeev@intel.com>
 * Copyright (C) 2005       Valery Ushakov <valery.e.ushakov@intel.com>
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
#include <stdio.h>
#include <pthread.h>
#include <stdlib.h>

#include <mme_internal.h>
#include <jni32_map.h>
#include <jni32_refmap.h>
#include <jni32_ref_tags.h>
#include <btmix.h>
#include <jni32_stack.h>
#include <jni32_diag.h>

#define LOCAL_REF_MAP_INITIAL_SIZE	256
#define LOCAL_REF_MAP_INCREMENT_SIZE	16

#define GLOBAL_REF_INITIAL_SIZE		256
#define GLOBAL_REF_INCREMENT_SIZE	256

#define WEAK_REF_INITIAL_SIZE		256
#define WEAK_REF_INCREMENT_SIZE		256





struct local_refmap;	/* forward */

struct local_refmap_data {
	void *ref64;
	void *aux;
};

struct local_frame {
	int32_t frame_start;
	int32_t frame_end;
	int32_t free;

	struct local_refmap *lmap;
	struct local_frame *prev;
};

struct local_refmap {
	struct jni32_map refmap;
	struct local_frame *top_frame;
};


struct global_refmap {
	struct jni32_map refmap;
	pthread_mutex_t refmap_mutex;
};


struct weak_refmap {
	struct jni32_map refmap;
	pthread_mutex_t refmap_mutex;
};

static struct weak_refmap m_weak_refmap_data;
static struct global_refmap m_global_refmap_data;
static MME_THREAD_VAR struct local_refmap * tm_local_refmap = NULL;


/* local funcs prototypes */
static void *jni32_demarshal_local_ref(struct local_refmap *lmap, uint32_t ref32);
static void *jni32_demarshal_global_ref(uint32_t ref32);
static void *jni32_demarshal_weak_ref(uint32_t ref32);
static struct local_refmap * jni32_get_current_local_refmap(void);
static int32_t jni32_local_refmap_create(struct local_refmap **lmap);
static void jni32_local_refmap_destroy(struct local_refmap **lmap);

/*****************************************************************************/
/*****************************************************************************/
HIDDEN void *
jni32_get_ref(uint32_t ref32)
{
	void *result;
	struct local_refmap *lmap;
	uint32_t tag_type;

	tag_type = GET_CERTAIN_TAG(ref32);
	DBPRINT (0, "ref32=%x tag_type=%x\n", ref32, tag_type);
	switch (tag_type) {
	case LOCAL_REF_TAG:
		lmap = jni32_get_current_local_refmap();
		result = jni32_demarshal_local_ref(lmap, ref32);
		break;
	case GLOBAL_REF_TAG:
		result = jni32_demarshal_global_ref(ref32);
		break;
	case WEAK_REF_TAG:
		result = jni32_demarshal_weak_ref(ref32);
		break;
	case NULL_REF_TAG:
		result = I2P(ref32);
		break;
	default:
		assert( ! "unknown tag" );
		MME_THROW_EXCEPTION("unknown tag");
		result = NULL;
	}

	
	DBPRINT (0, "result=%p\n", result);
	return result;
}

/*****************************************************************************/
/*****************************************************************************/
HIDDEN void *
jni32_demarshal_ref(ptr32_t * psp32)
{
	void *result;
	uint32_t ref32 = btmix_pop_jobject (psp32);

	result = jni32_get_ref (ref32);
	return result;
}

/*****************************************************************************/
/*****************************************************************************/
HIDDEN void *
jni32_delete_ref(uint32_t ref32)
{
	void *result;

	
	switch (GET_CERTAIN_TAG(ref32)) {
	case LOCAL_REF_TAG:
		result = jni32_local_ref_del(ref32);
	break;
	case GLOBAL_REF_TAG:
		result = jni32_global_ref_del(ref32);
	break;
	case WEAK_REF_TAG:
		result = jni32_weak_ref_del(ref32);
	break;
	case NULL_REF_TAG:
		result = I2P(ref32);
	break;
	default:
		assert(0);
		MME_THROW_EXCEPTION("unknown tag type");
		result = NULL;
	}

	
	return result;
}

/*****************************************************************************/
/*****************************************************************************/
int32_t HIDDEN
jni32_init_refmaps(void)
{
	int32_t status;

	status = jni32_map_create(&m_global_refmap_data.refmap,
			GLOBAL_REF_INITIAL_SIZE, GLOBAL_REF_INCREMENT_SIZE);
	if (status != 0) {
		goto out;
	}

	status = pthread_mutex_init (&m_global_refmap_data.refmap_mutex, NULL);
	if (status != 0) {
		goto out;
	}

	status = jni32_map_create(&m_weak_refmap_data.refmap,
			WEAK_REF_INITIAL_SIZE, WEAK_REF_INCREMENT_SIZE);
	if (status != 0) {
		goto out;
	}

	status = pthread_mutex_init (&m_weak_refmap_data.refmap_mutex, NULL);
	if (status != 0) {
		goto out;
	}

	status = 0;

out:

	return status;
}

/*****************************************************************************/
/*****************************************************************************/
int32_t HIDDEN
jni32_uninit_refmaps(void)
{
	int32_t status;

	status = pthread_mutex_lock (&m_global_refmap_data.refmap_mutex);
	if (status != 0) {
		goto out;
	}

	jni32_map_destroy(&m_global_refmap_data.refmap);

	status = pthread_mutex_unlock (&m_global_refmap_data.refmap_mutex);
	if (status != 0) {
		goto out;
	}

	status = pthread_mutex_destroy (&m_global_refmap_data.refmap_mutex);
	if (status != 0) {
		goto out;
	}

	status = pthread_mutex_lock (&m_weak_refmap_data.refmap_mutex);
	if (status != 0) {
		goto out;
	}

	jni32_map_destroy(&m_weak_refmap_data.refmap);

	status = pthread_mutex_unlock (&m_weak_refmap_data.refmap_mutex);
	if (status != 0) {
		goto out;
	}

	status = pthread_mutex_destroy (&m_weak_refmap_data.refmap_mutex);
	if (status != 0) {
		goto out;
	}



	status = 0;
out:

	return status;
}

/*
 * Local Frame
 */
/*****************************************************************************/
/*****************************************************************************/
int32_t HIDDEN
jni32_init_local_refmap(void)
{
	int32_t status;

	assert(tm_local_refmap == NULL);
	if (tm_local_refmap != NULL) {
		MME_THROW_EXCEPTION("unexpected call to init-map");
		return -1;
	}

	status = jni32_local_refmap_create(&tm_local_refmap);
	if (status != 0) {
		MME_THROW_EXCEPTION("refmap-create failed");
		return -1;
	}
	assert(tm_local_refmap != NULL);
	return 0;
}

/*****************************************************************************/
/*****************************************************************************/
bool HIDDEN
jni32_is_local_refmap_inited(void)
{
	/* tm_local_refmap is initialized by jni32_init_local_refmap() and
	 * set to NULL in jni32_uninit_local_refmap()
	 */
	if (tm_local_refmap != NULL) {
		return true;
	}
	return false;
}

/*****************************************************************************/
/*****************************************************************************/
int32_t
jni32_uninit_local_refmap(void)
{
	if (tm_local_refmap != NULL) {
		jni32_local_refmap_destroy(&tm_local_refmap);
	}

	return 0;
}

/*****************************************************************************/
/*****************************************************************************/
static struct local_refmap *
jni32_get_current_local_refmap(void)
{
	assert(tm_local_refmap != NULL);
	return tm_local_refmap;
}

/*****************************************************************************/
/*****************************************************************************/
static int32_t
jni32_local_refmap_create(struct local_refmap **lmap)
{
	int32_t status;
	assert(lmap != NULL);
	*lmap = malloc(sizeof(struct local_refmap));
	if (*lmap == NULL) {
		MME_THROW_EXCEPTION("malloc failed");
		return -1;
	}

	status = jni32_map_create(&(*lmap)->refmap, LOCAL_REF_MAP_INITIAL_SIZE,
			LOCAL_REF_MAP_INCREMENT_SIZE);
	if (status != 0) {
		MME_THROW_EXCEPTION("map_create failed");
		return -1;
	}

	(*lmap)->top_frame = NULL;
	
	return 0;
}

/*****************************************************************************/
/*****************************************************************************/
static void
jni32_local_refmap_destroy(struct local_refmap **lmap)
{
	struct local_frame *cur_frame;
	struct local_frame *del_frame;

	assert(*lmap != NULL);
	jni32_map_destroy(&(*lmap)->refmap);

	cur_frame = (*lmap)->top_frame;
	while (cur_frame != NULL) {
		del_frame = cur_frame;
		cur_frame = cur_frame->prev;
		free(del_frame);
	}

	free(*lmap);
	*lmap = NULL;
	
}


/*****************************************************************************/
/*****************************************************************************/
static int32_t
__jni32_push_local_ref_frame_intern(struct local_refmap *lmap, int32_t push_size)
{
	struct local_frame *new_frame, *prev_frame;
	int32_t need_more;
	int32_t free_tail;

	assert(lmap != NULL);
	new_frame = malloc(sizeof(struct local_frame));
	if (new_frame == NULL) {
		MME_THROW_EXCEPTION("malloc failed");
		return -1;
	}

	prev_frame = lmap->top_frame;

	new_frame->lmap = lmap;
	new_frame->prev = prev_frame;

	if (prev_frame == NULL) {
		new_frame->frame_start = 0;
	} else {
		new_frame->frame_start = prev_frame->frame_end;
	}

	/* make sure backing store has enough room */
	need_more = new_frame->frame_start + push_size - lmap->refmap.size;

	if (need_more > 0) {
		if (need_more < LOCAL_REF_MAP_INCREMENT_SIZE) {
			free_tail = jni32_map_grow(&lmap->refmap, LOCAL_REF_MAP_INCREMENT_SIZE);
		} else {
			free_tail = jni32_map_grow(&lmap->refmap, need_more);
		}

		if (free_tail < 0) {
			return -1;
		}
	}

	new_frame->free = new_frame->frame_start;
	new_frame->frame_end = new_frame->frame_start + push_size;

	struct jni32_map init_temp_map = {lmap->refmap.data, new_frame->frame_end, -1, 0};
	jni32_map_freelist_add(&init_temp_map, new_frame->frame_start);

	lmap->top_frame = new_frame;

	return 0;
}

/*****************************************************************************/
/*****************************************************************************/
int32_t HIDDEN
jni32_push_local_ref_frame(int32_t push_size)
{
	int32_t result;

	result =__jni32_push_local_ref_frame_intern (jni32_get_current_local_refmap(), push_size);

	return result;
}

/*****************************************************************************/
/*****************************************************************************/
int32_t HIDDEN
jni32_predef_size_push_local_ref_frame(void)
{
	int32_t result;
	
	result =__jni32_push_local_ref_frame_intern(
				jni32_get_current_local_refmap(),
				LOCAL_REF_MAP_INCREMENT_SIZE );
	return result;
}

/*****************************************************************************/
/*****************************************************************************/
static int32_t
__jni32_pop_local_ref_frame_intern(struct local_refmap *lmap) 
{
	struct local_frame *del_frame;

	assert(lmap != NULL);
	del_frame = lmap->top_frame;
	assert(del_frame != NULL);
	int32_t ref_ter;
	for (ref_ter = del_frame->frame_start; ref_ter < del_frame->frame_end; ++ref_ter) {
		if (lmap->refmap.data[ref_ter].value != NULL) {
			free(lmap->refmap.data[ref_ter].value);
		}
		lmap->refmap.data[ref_ter].value = NULL;
	}
	assert(del_frame != NULL);
	if (del_frame == NULL) {
		MME_THROW_EXCEPTION("frame not found");
		return -1;
	}
	lmap->top_frame = del_frame->prev;
	free (del_frame);
	
	return 0;
}

/*****************************************************************************/
/*****************************************************************************/
int32_t HIDDEN
jni32_pop_local_ref_frame(void) 
{
	int32_t result;
	result = __jni32_pop_local_ref_frame_intern( jni32_get_current_local_refmap());
	return result;
}

/*****************************************************************************/
/*****************************************************************************/
static uint32_t
__jni32_marshal_local_ref_intern(struct local_refmap *lmap, void *ref64) 
{
	struct local_refmap_data *new_ref_data;
	struct local_frame *top_frame;
	int32_t ref32;
	uint32_t retval;

	DBPRINT (0, "lmap=%p ref64=%p top_frame=%p\n", lmap, ref64, lmap->top_frame);

	assert(lmap != NULL);
	assert(lmap->top_frame != NULL);
	top_frame = lmap->top_frame;

	if (top_frame->free == -1) {
		if (jni32_map_grow(&lmap->refmap, LOCAL_REF_MAP_INCREMENT_SIZE) < 0) {
			MME_THROW_EXCEPTION("map_grow failed");
			retval = (uint32_t)-1;
			goto out;
		}
		top_frame->free = top_frame->frame_end;
		top_frame->frame_end = top_frame->frame_end + LOCAL_REF_MAP_INCREMENT_SIZE;

		struct jni32_map init_temp_map = {lmap->refmap.data, top_frame->frame_end, -1, 0};
		jni32_map_freelist_add(&init_temp_map, top_frame->free);
	}

	assert(top_frame->free >= top_frame->frame_start);
	assert(top_frame->free < top_frame->frame_end);

	/* get free slot from the head of the free list */
	new_ref_data = malloc(sizeof(struct local_refmap_data));
	assert(new_ref_data != NULL);
	if (new_ref_data == NULL) {
		MME_THROW_EXCEPTION("malloc failed");
		retval = (uint32_t)-1;
		goto out;
	} 

	ref32 = top_frame->free;
	assert(lmap->refmap.data[ref32].value == NULL);

	top_frame->free = lmap->refmap.data[ref32].next_free;

	new_ref_data->aux = lmap->top_frame;
	new_ref_data->ref64 = ref64;
	lmap->refmap.data[ref32].value = (void *)new_ref_data;

	retval = SET_LOCAL_REF_TAG(ref32);
out:
	DBPRINT (0, "top_frame=%p\n", lmap->top_frame);
	DBPRINT (0, "retval=%x\n", retval);
	return retval;
}

/*****************************************************************************/
/*****************************************************************************/
uint32_t HIDDEN
jni32_marshal_local_ref(void *ref64) 
{
	int32_t result;
	struct local_refmap * locref;
	
	if (ref64 == NULL) {
		result = 0;
		goto out;
	}
	locref = jni32_get_current_local_refmap();
	result = __jni32_marshal_local_ref_intern(locref, ref64);
out:
	DBPRINT (0, "ref64=%p ref32=%x\n", ref64, result);
	return result;
}



/*****************************************************************************/
/*****************************************************************************/
static void *
jni32_demarshal_local_ref(struct local_refmap *lmap, uint32_t ref32) 
{
	uint32_t untaged_ref32;
	untaged_ref32 = RM_CERTAIN_TAG(ref32);
	void *result;

	assert(lmap != NULL);
	if (GET_CERTAIN_TAG (ref32) != LOCAL_REF_TAG) {
		result = NULL;
		goto out;
	}

	DBPRINT (0, "top_frame=%p\n", lmap->top_frame);
	/* Is ref in refmap range? */
	DBPRINT (0, "frame_start=%d frame_end=%d free=%d\n",
			lmap->top_frame->frame_start,
			lmap->top_frame->frame_end,
			lmap->top_frame->free);
	assert((int32_t)untaged_ref32 < lmap->top_frame->frame_end);

	/* Is ref value is NULL? */
	if ((lmap->refmap.data[untaged_ref32].value == NULL)) {
	        result = NULL;
		goto out;
	}

	/* Is ref already freed? */
	if ((((struct local_refmap_data*)lmap->
	      refmap.data[untaged_ref32].value)->aux == NULL)) {
	        result = NULL;
		goto out;
	}

	result = ((struct local_refmap_data*)lmap->
			refmap.data[untaged_ref32].value)->ref64;
out:
	DBPRINT (0, "ref32=%x ref64=%p\n", ref32, result);
	return result;
}

/*****************************************************************************/
/*****************************************************************************/
/*
 * Delete ref32 from a local map.
 * Note, that the ref32 is not neccessarily from the current frame!
 */
static void *
__jni32_local_ref_del_intern(struct local_refmap *lmap, uint32_t ref32) 
{
	void *ref64;
	struct local_frame *cur_frame;
	uint32_t untaged_ref32;
	untaged_ref32 = RM_CERTAIN_TAG(ref32);

	assert(lmap != NULL);
	if (GET_CERTAIN_TAG (ref32) != LOCAL_REF_TAG) {
		return NULL;
	}

	/* Is ref in refmap range? */
	assert((int32_t)untaged_ref32 < lmap->top_frame->frame_end);
	/* Is ref already freed? */
	assert(((struct local_refmap_data*)lmap->
			refmap.data[untaged_ref32].value)->aux != NULL);

	/* Is ref value is NULL? */
	assert(lmap->refmap.data[untaged_ref32].value != NULL);

	ref64 = ((struct local_refmap_data*)lmap->
			refmap.data[untaged_ref32].value)->ref64;

	/* prepend to the free list of its frame */
	cur_frame = ((struct local_refmap_data*)lmap->
				refmap.data[untaged_ref32].value)->aux;
	free(lmap->refmap.data[untaged_ref32].value);
	lmap->refmap.data[untaged_ref32].value = NULL;
	lmap->refmap.data[untaged_ref32].next_free = cur_frame->free;
	cur_frame->free = untaged_ref32;

	
	return ref64;
}

/*****************************************************************************/
/*****************************************************************************/
HIDDEN void *
jni32_local_ref_del(uint32_t ref32) 
{
	void *result;
	result = __jni32_local_ref_del_intern(jni32_get_current_local_refmap(), ref32);
	return result;
}

/*
 * Global Refs
 */
/*****************************************************************************/
/*****************************************************************************/
uint32_t HIDDEN
jni32_marshal_global_ref(void *ref64) 
{
	int32_t status;
	int32_t result_value32;

	
	if (ref64 == NULL) {
		return 0;
	}

	status = pthread_mutex_lock (&m_global_refmap_data.refmap_mutex);
	if (status != 0) {
		return -1;
	}

	status = jni32_map_add(&m_global_refmap_data.refmap, ref64, &result_value32);
	if (status != 0) {
		MME_THROW_EXCEPTION("map_add failed");
		pthread_mutex_unlock (&m_global_refmap_data.refmap_mutex);
		return -1;
	}

	status = pthread_mutex_unlock (&m_global_refmap_data.refmap_mutex);
	if (status != 0) {
		return -1;
	}
	return SET_GLOBAL_REF_TAG(result_value32);
}

/*****************************************************************************/
/*****************************************************************************/
static void *
jni32_demarshal_global_ref(uint32_t ref32) 
{
	int32_t status;
	void *result;

	
	if (GET_CERTAIN_TAG (ref32) != GLOBAL_REF_TAG) {
		return NULL;
	}

	status = pthread_mutex_lock (&m_global_refmap_data.refmap_mutex);
	if (status != 0) {
		return NULL;
	}

	result = jni32_map_get(&m_global_refmap_data.refmap, RM_CERTAIN_TAG(ref32));

	status = pthread_mutex_unlock (&m_global_refmap_data.refmap_mutex);
	if (status != 0) {
		return NULL;
	}

	return result;
}

/*****************************************************************************/
/*****************************************************************************/
HIDDEN void *
jni32_global_ref_del(uint32_t ref32) 
{
	int32_t status;
	void *result;

	
	if (GET_CERTAIN_TAG (ref32) != GLOBAL_REF_TAG) {
		return NULL;
	}

	status = pthread_mutex_lock (&m_global_refmap_data.refmap_mutex);
	if (status != 0) {
		return NULL;
	}

	result = jni32_map_del(&m_global_refmap_data.refmap, RM_CERTAIN_TAG(ref32));

	status = pthread_mutex_unlock (&m_global_refmap_data.refmap_mutex);
	if (status != 0) {
		return NULL;
	}

	
	
	return result;
}

/*
 * Weak Refs
 */

/*****************************************************************************/
/*****************************************************************************/
uint32_t HIDDEN
jni32_marshal_weak_ref(void *ref64) 
{
	int32_t status;
	int32_t result_value32;

	
	if (ref64 == NULL) {
		return 0;
	}

	status = pthread_mutex_lock (&m_weak_refmap_data.refmap_mutex);
	if (status != 0) {
		return -1;
	}

	status = jni32_map_add(&m_weak_refmap_data.refmap, ref64, &result_value32);
	if (status != 0) {
		MME_THROW_EXCEPTION("map_add failed");
		pthread_mutex_unlock (&m_weak_refmap_data.refmap_mutex);
		return -1;
	}

	status = pthread_mutex_unlock (&m_weak_refmap_data.refmap_mutex);
	if (status != 0) {
		return -1;
	}
	
	return SET_WEAK_REF_TAG(result_value32);
}

/*****************************************************************************/
/*****************************************************************************/
static void *
jni32_demarshal_weak_ref(uint32_t ref32) 
{
	int32_t status;
	void *result;

	
	if (GET_CERTAIN_TAG (ref32) != WEAK_REF_TAG) {
		return NULL;
	}

	status = pthread_mutex_lock (&m_weak_refmap_data.refmap_mutex);
	if (status != 0) {
		return NULL;
	}

	result = jni32_map_get(&m_weak_refmap_data.refmap, RM_CERTAIN_TAG(ref32));

	status = pthread_mutex_unlock (&m_weak_refmap_data.refmap_mutex);
	if (status != 0) {
		return NULL;
	}

	return result;
}

/*****************************************************************************/
/*****************************************************************************/
HIDDEN void *
jni32_weak_ref_del(uint32_t ref32) 
{
	int32_t status;
	void *result;

	
	if (GET_CERTAIN_TAG (ref32) != WEAK_REF_TAG) {
		return NULL;
	}

	status = pthread_mutex_lock (&m_weak_refmap_data.refmap_mutex);
	if (status != 0) {
		return NULL;
	}

	result = jni32_map_del(&m_weak_refmap_data.refmap, RM_CERTAIN_TAG(ref32));

	status = pthread_mutex_unlock (&m_weak_refmap_data.refmap_mutex);
	if (status != 0) {
		return NULL;
	}
	
	return result;
}

