/* 
 * mme_downcall.c
 *
 * 
 *
 */
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
#include <stdarg.h>
#include <string.h>
#include <assert.h>
#include <jni.h>

#include <mme_jvm_api.h>
#include <ia32el_mme_api.h>
#include <mme_internal.h>
#include <jni32_refmap.h>
#include <jni32_jnienv.h>
#include <jni32_javavm.h>

/* local function prototypes */
static ptr32_t prepare_ia32_stack (uint64_t return_type, ptr32_t esp, char * param_list, va_list params);
static int32_t get_downcall_retval (int32_t return_type, uint64_t * retval);
static ptr32_t get_and_marshal_jnienv (ptr32_t * esp, JavaVM * vm64);
#if ! FIXED_STUBS
static ptr32_t prepare_return_stub (ptr32_t esp, int64_t return_type);
#endif


/*****************************************************************************/
/*****************************************************************************/
int32_t HIDDEN do_downcallV (int64_t return_type, uint64_t * return_value,
		void *addr, char *param_list, ...) 
{
	int32_t retval;
	va_list params;

	va_start (params, param_list);
	retval = do_downcall (false, return_type, return_value, addr, param_list, params);
	va_end (params);
	return retval;
}

int32_t HIDDEN do_downcall (bool is_jni, int64_t return_type, uint64_t * return_value,
		void *addr, char *param_list, va_list params) 
{
	ia32el_context32 ec32;
	ia32el_retval ia32el_ret;
	int32_t retval = -1;
	uint64_t jni_retval;


	DBPRINT (0, "is_jni=%d return_type=%ld, return_value=%p addr=%p param_list='%s'\n",
			is_jni, return_type, return_value, addr, param_list);
	ia32el_ret = g_ia32el_exports.ia32el_ia32_context_get (&ec32);
	if (ia32el_ret != IA32EL_SUCCESS) {
		DBPRINT (true, "ia32el_ia32_context_get failed\n");
		goto out;
	}

	/* XXX Check whether this thread is ready at the IA-32 level. If not then
	 * Finish initialization of child thread */

	/* Check if this is a call to malloc32 / free32 for NIO support */
	if (is_nio_call(param_list, params, return_value)) {
		return 0;
	}

	if (is_jni) {
		/* Push a new list to the stack of local reference lists */
		retval = jni32_predef_size_push_local_ref_frame ();
		if (retval != 0) {
			DBPRINT (true, "jni32_predef_size_push_local_ref_frame failed\n");
			goto out;
		}
	}

	/* put the arguments on the stack */
	ec32.esp = prepare_ia32_stack (return_type, ec32.esp, param_list, params);
	if (ec32.esp == 0) {
		DBPRINT (true, "prepare_ia32_stack failed\n");
		goto out;
	}

	/* init the fp state. the init values are taken from ia32x/btlib.h.
	 * the index in the ext-regs is taken from ia32exec_os_api.h
	 * (btg api header file) */
	ec32.fpu_state.control_word = /* FPCW_INIT_VAL */ 0x37f;
	ec32.fpu_state.tag_word = /* FPTW_INIT_VAL */ 0xffff;
	ec32.eflags = /* EFLAGS_INIT_VAL */ 0x202;
	*(unsigned int *) (void *) &ec32.xmm_and_bt_ext[/*MXCSR_INDEX_IN_EXT_REGS*/24] =
		/* MXCSR_INIT_VAL */ 0x1f80;

	/* Run the given function using btlib_mme_run */
	ec32.eip = (uintptr_t)addr;
	DBPRINT (0, "before ia32el_run - eip=%x esp=%x\n", ec32.eip, ec32.esp);
	ia32el_ret = g_ia32el_exports.ia32el_run (&ec32);
	if (ia32el_ret != IA32EL_SUCCESS) {
		DBPRINT (true, "ia32el_run failed\n");
		goto out;
	}

	/* get the return value */
	retval = get_downcall_retval (return_type, &jni_retval);
	if (retval != 0) {
		DBPRINT (true, "get_downcall_retval failed\n");
		goto out;
	}

	if (is_jni) {
		/* Pop the topmost list from the stack of local reference lists */
		retval = jni32_pop_local_ref_frame ();
		if (retval != 0) {
			DBPRINT (true, "jni32_pop_local_ref_frame failed\n");
			goto out;
		}
	}

out:
	if (retval == 0) {
		if (return_type != MME_JAVA_DATA_TYPE_VOID) {
			*return_value = jni_retval;
		}
	}

	return retval;
}
/*****************************************************************************/
/*****************************************************************************/
static int32_t get_downcall_retval (int32_t return_type, uint64_t * retval)
{
	int32_t ret = -1;
	ia32el_retval ia32el_ret;
	ptr32_t ia32_ref;
	ia32el_context32 ec32;

	/* Get the IA-32 context. the return value can be found according to
	 * IA-32 ABI : 
	 *	32 bit integer, pointer - EAX
	 *	64 bit integer - EDX(hi):EAX(lo)
	 *	floating point - ST(0)
	 */
	ia32el_ret = g_ia32el_exports.ia32el_ia32_context_get (&ec32);
	if (ia32el_ret != IA32EL_SUCCESS) {
		DBPRINT (true, "ia32el_ia32_context_get failed\n");
		goto out;
	}
	switch (return_type) {
		/* 8 bytes */
		case MME_JAVA_DATA_TYPE_LONG:
			*retval = ((uint64_t)ec32.edx << 32) | ec32.eax;
			break;
		/* 4 bytes */
		case MME_JAVA_DATA_TYPE_INT:
			*retval = ec32.eax;
			break;
		/* 2 bytes */
		case MME_JAVA_DATA_TYPE_SHORT:
		case MME_JAVA_DATA_TYPE_CHAR:
			*retval = ec32.eax;
			break;
		/* 1 byte */
		case MME_JAVA_DATA_TYPE_BYTE:
		case MME_JAVA_DATA_TYPE_BOOLEAN:
			*retval = ec32.eax;
			break;
		case MME_JAVA_DATA_TYPE_VOID:
			*retval = 0; /* dummy value */
			break;
		/* floating point value */
		case MME_JAVA_DATA_TYPE_DOUBLE:
			copy_from_ia32_stack (retval,
					I2P(ec32.esp - sizeof(double)),
					sizeof(double));
			break;
		case MME_JAVA_DATA_TYPE_FLOAT:
			copy_from_ia32_stack (retval,
					I2P(ec32.esp - sizeof(float)),
					sizeof(float));
			break;
		/* references */
		case MME_JAVA_DATA_TYPE_OBJECT:
			ia32_ref = ec32.eax;
			*retval = (uint64_t)jni32_get_ref (ia32_ref);
			break;
		default:
			DBPRINT (true, "unknown return type:%d\n", return_type);
			goto out;
	}
	DBPRINT (1, "retval=%lx\n", *retval);
	ret = 0;
out:
	return ret;
}

/*****************************************************************************/
/* Prepare a stack frame for the function according to the given parameter
 * list. The frame pointer will point to IA-32 code stub that traps to btlib
 * with BTLIB_TRAP_RETURN_FROM_RUN */
/*****************************************************************************/
static ptr32_t prepare_ia32_stack (uint64_t return_type, ptr32_t esp, char * param_list, va_list params)
{
	uint64_t param_list_len;
	char * p;
	uint32_t stack_size = 0;
	ptr32_t framep;
	ptr32_t stub_addr;
	void * argp;
	JNIEnv * env64;
	JavaVM * vm64;
	ptr32_t vm32 = 0;
	struct JavaVM32to64 * vms;
	jobject * obj64;
	ptr32_t env32 = 0, obj32 = 0;
	ptr32_t retval = 0;

	DBPRINT (0, "esp=%x param_list=%s\n", esp, param_list);
	/* get the total size of parameters */
	param_list_len = strnlen (param_list, MAX_PARAM_STR_LEN);
	if (param_list_len == MAX_PARAM_STR_LEN) {
		DBPRINT (true, "param list too long\n");
		goto out;
	}
	p = param_list;
	while (p < param_list + param_list_len) {
		switch (*p) {
			/* 8 bytes */
			case 'J':/*  Long */
			case 'D':/*  Double */
				stack_size += 8;
				break;
			/* 4 bytes */
			case 'V':/*  JAVAVM* pointer */
				assert (p == param_list);
				vm64 = va_arg (params, void *);
				vms = jni32_javavm_marshal (vm64);
				if (vms== 0) {
					DBPRINT (0, "jnimix_init_javavm failed\n");
					goto out;
				}
				assert (IS_32BIT_VALUE (vms));
				vm32 = TRUNCATE_PTR (vms);
				stack_size += 4;
				/* jnienv struct must be marshaled too at this point.
				 * it will not be possible while inside the OnLoad
				 * function, because the marshaling stub is located
				 * at current esp which is used by the emulated native
				 * code
				 */
				/*  !!! jnienv marshaling may change esp !!! */
				env32 = get_and_marshal_jnienv (&esp, vm64);
				if (env32 == 0) {
					DBPRINT (0, "get_and_marshal_jnienv failed\n");
					goto out;
				}
				break;
			case 'E':/*  JNIEnv* pointer */
				assert (p == param_list);
				env64 = va_arg (params, void *);
				/*  !!! jnienv marshaling may change esp !!! */
				env32 = jni32_jnienv_marshal (&esp, env64, true);
				if (env32 == 0) {
					DBPRINT (0, "jnimix_init_jnienv failed\n");
					goto out;
				}
				stack_size += 4;
				break;
			case 'R':/*  Reserved pointer */
			case 'L':/*  Object */
			case 'I':/*  Int */
			case 'F':/*  Float */
			case 'S':/*  Short */
			case 'C':/*  Char */
			case 'Z':/*  Boolean */
			case 'B':/*  Byte */
				stack_size += 4;
				break;
			default:
				DBPRINT (true, "unknown param type '%c'\n", *p);
				goto out;
		}
		p ++;
	}

#if FIXED_STUBS
	/* push the return address on the stack */
	stub_addr = TRUNCATE_PTR (get_stub_addr (IA32EL_TRAP_RETURN_FROM_RUN));
#else
	stub_addr = prepare_return_stub (esp, return_type);
	stack_size += ROUNDUP (esp - stub_addr, 4); /* place for the stub size */
#endif // FIXED_STUBS
	stack_size += sizeof (ptr32_t); /* place for return addr */
	framep = esp - stack_size;  /* XXX - check if framep is valid memory */
	assert (IS_32BIT_VALUE (framep));
	argp = I2P (framep);
	DBPRINT (0, "stub_addr=%x argp=%p stack_size=%d\n", stub_addr, argp, stack_size);

	copy_to_ia32_stack (argp, &stub_addr, sizeof (ptr32_t));
	argp += sizeof (ptr32_t);

	/* put the arguments on the stack */
	p = param_list;
	while (p < param_list + param_list_len) {
		switch (*p) {
			union {
				uint64_t u;
				double f;
			} arg64;
			union {
				uint32_t u;
				float f;
			} arg32;
			/* 8 bytes */
			case 'J':/*  Long */
			case 'D':/*  Double */
				arg64.u = va_arg (params, uint64_t);
#if MMEDEBUG
				if (*p == 'D') {
					DBPRINT (0, "%p:(%c)%f<%lx>\n", argp, *p, arg64.f, arg64.u);
				} else {
					DBPRINT (0, "%p:(%c)%ld<%lx>\n", argp, *p, arg64.u, arg64.u);
				}
#endif
				copy_to_ia32_stack (argp, &arg64.u, sizeof (uint64_t));
				argp += sizeof (uint64_t);
				break;
			/* 4 bytes */
			case 'L':/*  Object */
			case 'R':/*  Reserved pointer */
				obj64 = va_arg (params, void *);
				DBPRINT (0, "obj64=%p\n", obj64);
				obj32 = jni32_marshal_local_ref (obj64);
				if (obj32 == 0 && obj64 != NULL) {
					DBPRINT (true, "jnimix_init_jobject failed\n");
				}
				DBPRINT (0, "%p:(%c)%x\n", argp, *p, obj32);
				copy_to_ia32_stack (argp, &obj32, sizeof (obj32));
				argp += sizeof (obj32);
				break;
			case 'I':/*  Int */
			case 'S':/*  Short */
			case 'C':/*  Char */
			case 'Z':/*  Boolean */
			case 'B':/*  Byte */
				arg32.u = va_arg (params, uint32_t);
				DBPRINT (0, "%p:(%c)%d<%x>\n", argp, *p, arg32.u, arg32.u);
				copy_to_ia32_stack (argp, &arg32.u, sizeof (uint32_t));
				argp += sizeof (uint32_t);
				break;
			case 'F':/*  Float */
				/* float is promoted to double by va_arg */
				arg64.f = va_arg (params, double);
				DBPRINT (0, "%p:(%c)%E<%016lx>\n", argp, *p, arg64.f, arg64.u);
				arg32.f = (float)arg64.f;
				DBPRINT (0, "%E<%016x>\n", arg32.f, arg32.u);
				copy_to_ia32_stack (argp, &arg32.f, sizeof (float));
				argp += sizeof (float);
				break;
			case 'E':/*  JNIEnv* pointer */
				assert (env32 != 0);
				DBPRINT (0, "%p:(%c)%x\n", argp, *p, env32);
				copy_to_ia32_stack (argp, &env32, sizeof (env32));
				argp += sizeof (env32);
				break;
			case 'V':/*  JavaVM* pointer */
				assert (vm32 != 0);
				DBPRINT (0, "%p:(%c)%x\n", argp, *p, vm32);
				copy_to_ia32_stack (argp, &vm32, sizeof (vm32));
				argp += sizeof (vm32);
				break;
			default:
				DBPRINT (true, "unknown param type '%c'\n", *p);
				goto out;
		}
		p ++;
	}

	retval = framep;
out:
	DBPRINT (0, "retval=%x\n", retval);
	return retval;
}

/*****************************************************************************/
/* marshal the 64 bit env from 64 bit jvm  */
/*****************************************************************************/
static ptr32_t get_and_marshal_jnienv (ptr32_t * esp, JavaVM * vm64)
{
	int32_t ret;
	ptr32_t env32 = 0;
	union {
		JNIEnv * env64;
		void * ptr;
	} pc;


	ret = (*vm64)->GetEnv (vm64, &pc.ptr, JNI_VERSION_1_2);
	if (ret != 0) {
		DBPRINT (0, "GetEnv failed\n");
		goto out;
	}
	env32 = jni32_jnienv_marshal (esp, pc.env64, true);
	if (env32 == 0) {
		DBPRINT (0, "jnimix_init_jnienv failed\n");
		goto out;
	}
out:
	return env32;
}


#if ! FIXED_STUBS
/*****************************************************************************/
/* writes dynamically a stub of 32 bit code that restores
 * orig value of esp and traps to ia32el
 * input:	esp = orig value of esp
 * output:	address of the 32 bit stub
 *       ------------------------------------------------------
 * NOTE: CHANGES TO THIS FUNCTION MUST BE REFLECTED IN FUNCTION
 *	mme_dump.c:is_return_stub()
 *       ------------------------------------------------------
 */
/*****************************************************************************/
static ptr32_t prepare_return_stub (ptr32_t esp, int64_t return_type)
{
	union {
		char c[4];
		ptr32_t p;
	} pc;
	char return_stub[32], * rsp;
	int32_t stub_size;
	ptr32_t stub_addr;
	uint32_t retval_size = 0;



	/* mov esp, pc.c */
	pc.p = esp;
	rsp = return_stub;
	*rsp++ = 0xbc;
	*rsp++ = pc.c[0];
	*rsp++ = pc.c[1];
	*rsp++ = pc.c[2];
	*rsp++ = pc.c[3];

	/* fp values should be stored by 32 bit code in 
	 * order to use ia32el's definitions of precision etc. */
	if (return_type == MME_JAVA_DATA_TYPE_DOUBLE) {
		retval_size = 8;
		/* fstpl  -8(%esp) */
		*rsp ++ = 0xdd;
		*rsp ++ = 0x5c;
		*rsp ++ = 0x24;
		*rsp ++ = 0xf8;
	} else if (return_type == MME_JAVA_DATA_TYPE_FLOAT)  {
		retval_size = 4;
		/* fstps  -4(%esp) */
		*rsp ++ = 0xd9;
		*rsp ++ = 0x5c;
		*rsp ++ = 0x24;
		*rsp ++ = 0xfc;
	}

	/* jmpe RETURN_FROM_RUN */
	pc.p = IA32EL_TRAP_RETURN_FROM_RUN;
	*rsp++ = 0x0f;
	*rsp++ = 0xb8;
	*rsp++ = pc.c[0];
	*rsp++ = pc.c[1];
	*rsp++ = pc.c[2];
	*rsp++ = pc.c[3];

	stub_size = rsp - return_stub;
	assert (stub_size > 0);
	assert ((uint64_t)stub_size < sizeof (return_stub));
	stub_addr = ROUNDDOWN ((esp - stub_size - retval_size), 8);
	DBPRINT (0, "stub_addr=%x stub_size=%x esp=%x\n", stub_addr, stub_size, esp);
	memcpy (I2P(stub_addr), return_stub, stub_size);
	return stub_addr;
}






#endif // ! FIXED_STUBS

