/* 
 * mme_dump.c
 *
 * 
 *
 */
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <elf.h>
#include <link.h>
#include <linux/limits.h>

#include <ia32el_mme_api.h>
#include <auxapp.h>
#include <mme_internal.h>
#include <mme_jvm_api.h>
#include <jni32_refmap.h>

#define FDUMPF(fp,format,args...) do {						\
	fprintf (fp, format, ## args);\
	fprintf (stderr, format, ## args);\
}while (0)

#define IA32_PAGE_SIZE       0x1000
#define IA32_STACK_DUMP_SIZE 2048
#define IA32_PAGE_ALIGN(p)   ((p) & (~(IA32_PAGE_SIZE)+1))
#define DUMP_COLUMNS         6

typedef struct _binary_desc_tag {
	char full_path[PATH_MAX];
	uint32_t base_address;
	uint32_t bdh;
} binary_desc;

typedef struct _phdr_iter_data_tag {
	binary_desc ** bd;
	uint32_t * mn;
} phdr_iter_data;


static int32_t read_maps (binary_desc **, uint32_t *);
static int32_t find_bdh_and_print_symbol (binary_desc *, uint32_t, uint32_t, ptr32_t, uint32_t);
static int32_t iterate_phdr_callback (struct dl_phdr_info64 * info, size_t size UNUSED, void * data);

/*****************************************************************************/
/*****************************************************************************/
static void
dump_ia32_stack(FILE *fp, ia32el_context32 *ec32)
{
	uint32_t esp, i;
	uint32_t start;
	int32_t  bytes_to_dump;
	int64_t perm;

	esp = ec32->esp;
	start = IA32_PAGE_ALIGN(esp);
	bytes_to_dump = IA32_PAGE_SIZE;

	FDUMPF (fp, "IA-32 Stack:\n");
	FDUMPF (fp, "(* marks the word pointed to the stack pointer)\n");

	if (g_ia32el_exports.ia32el_query_memory_permission((void *)I2P(start), &perm) == IA32EL_FAILURE) {
		DBPRINT(1, "query ia32 memory permission failed\n");
	}

	if (!(perm & 0x1)) {
		FDUMPF (fp, "--inaccessible stack page--\n");
		return;
	}

	while(bytes_to_dump > 0) {
		FDUMPF (fp, "%08x" ": ", start);
		for (i = 0; i < DUMP_COLUMNS; i++) {
			char marker;
			if (esp >= start && esp < start + 4) {
				marker = '*';
			} else {
				marker = ' ';
			}
			FDUMPF (fp, "%08x" "%c ", *((uint32_t *)I2P(start)), marker);
			start += 4;
			bytes_to_dump -= 4;
		}
		FDUMPF (fp, "\n");
	}
}

/*****************************************************************************/
/*****************************************************************************/
static void
dump_ia32_code (FILE* fp, ia32el_context32 *ec32)
{
	uint32_t eip, i;
	uint32_t start;
	int32_t  bytes_to_dump;
	int64_t perm;

	eip = ec32->eip;
	start = IA32_PAGE_ALIGN(eip);
	bytes_to_dump = IA32_PAGE_SIZE;

	FDUMPF (fp, "IA-32 Code:\n");
	FDUMPF (fp, "(* marks the word pointed to the instruction pointer)\n");

	if (g_ia32el_exports.ia32el_query_memory_permission((void *)I2P(start), &perm) == IA32EL_FAILURE) {
		DBPRINT(1, "query ia32 memory permission failed\n");
	}

	if (!(perm & 0x1)) {
		FDUMPF (fp, "--inaccessible code page--\n");
		return;
	}

	while(bytes_to_dump > 0) {
		FDUMPF (fp, "%08x" ": ", start);
		for (i = 0; i < DUMP_COLUMNS; i++) {
			char marker;
			if (eip >= start && eip < start + 4) {
				marker = '*';
			} else {
				marker = ' ';
			}
			FDUMPF (fp, "%08x" "%c ", *((uint32_t *)I2P(start)), marker);
			start += 4;
			bytes_to_dump -= 4;
		}
		FDUMPF (fp, "\n");
	}
}

/*****************************************************************************/
/* there are 3 types of return stubs-see mme_downcall.c:prepare_return_stub():
 * 1:
 *    mov esp, xxx
 *    fstpl -8(esp)
 *    jmpe IA32EL_TRAP_RETURN_FROM_RUN
 * 2:
 *    mov esp, xxx
 *    fstps -4(esp)
 *    jmpe IA32EL_TRAP_RETURN_FROM_RUN
 * 3:
 *    mov esp, xxx
 *    jmpe IA32EL_TRAP_RETURN_FROM_RUN
 */
/*****************************************************************************/
static inline bool
is_return_stub (uint32_t eip)
{
	uint8_t * c;
	bool ret = false;

	DBPRINT (1, "eip=%x\n", eip);
	c = I2P (eip);

	if (*c != 0xbc) /* first byte in 'mov esp, xxx' instruction */
		goto out;
	/* go pass the 'mov esp, xxx' instruction */
	c += 5;
	if (*c == 0xdd) {
		/* fstpl */
		if ( !
				((*(c + 1) == 0x5c) &&
				(*(c + 2) == 0x24) &&
				(*(c + 3) == 0xf8))
				)
			goto out;
		else
			c += 4;
	} else if (*c == 0xd9) {
		/* fstps */
		if ( !
				((*(c + 1) == 0x5c) &&
				(*(c + 2) == 0x24) &&
				(*(c + 3) == 0xfc))
				)
			goto out;
		else
			c += 4;
	}
	/* jmpe */
	if ((*c != 0x0f) || (*(c + 1) != 0xb8))
		goto out;

	ret = true;
out:
	return ret;
}

/*****************************************************************************/
/*****************************************************************************/
static void
dump_ia32_stack_trace(FILE *fp, uint32_t eip, uint32_t ebp)
{
	/* trace ia32 stack until return stub (jmpe gate) */
	/* it seems we have to stop at the first nesting downcall level... */

	int32_t retval;
	uint64_t downcall_retval;
	int64_t perm;
	ptr32_t printout = 0;
	const uint32_t sizeof_printout = 256;
	binary_desc * bd = NULL;
	uint32_t mn = 0;

	DBPRINT (1, "eip=%x ebp=%x\n", eip, ebp);
	/* allocate a string in the 32 bit memory space */
	printout = malloc32 (sizeof_printout);
	if (printout == 0) {
		DBPRINT (true, "malloc32 failed");
		goto out;
	}
	memset (I2P(printout), 0, sizeof_printout);

	/* get the list of binaries */
	if (read_maps (&bd, &mn)) {
		DBPRINT (true, "read_maps failed\n");
		goto out;
	}
	DBPRINT (1, "mn=%d\n", mn);

	/* Skip the frame which is inaccessible */
	for (;;) {
		if (g_ia32el_exports.ia32el_query_memory_permission((void *)I2P(eip), &perm) ==
				IA32EL_FAILURE) {
			DBPRINT(1, "query ia32 memory permission failed\n");
			goto out;
		}

		if (perm & 0x1) {
			retval = find_bdh_and_print_symbol
				(bd, mn, eip, printout, sizeof_printout - 1);
			if (retval) {
				sprintf (I2P(printout), " -no symbol:ia32");
			}
		} else {
			sprintf (I2P(printout), " -no symbol:ia32");
		}
		FDUMPF (fp, "%s\n", (char*)I2P(printout));

		/* get the next frame */
		eip = *((uint32_t *)I2P(ebp + 4));
		ebp = *((uint32_t *)I2P(ebp));
		DBPRINT (1, "eip=%x ebp=%x\n", eip, ebp);

		if (is_return_stub (eip)) {
			break;
		}
	}

out:
	if (bd) {
		uint32_t j;
		for (j = 0; j < mn; j ++) {
			DBPRINT (1, "bd[%d] full_path=<%s> bdh=%x\n", j, bd[j].full_path, bd[j].bdh);
			if (bd[j].bdh) {
				retval = do_downcallV (MME_JAVA_DATA_TYPE_INT,
						&downcall_retval,
						I2P(g_ia32_runtime.binary_desc_finish),
						"I", bd[j].bdh);
				if (retval != 0) {
					DBPRINT (true, "calling binary_desc_finish failed\n");
				}
				if (downcall_retval != 0) {
					DBPRINT (true, "auxapp.binary_desc_finish failed\n");
				}
			}
		}
		assert (IS_32BIT_VALUE (bd));
		free32 ((ptr32_t)(uintptr_t)bd);
	}
	if (printout) {
		free32 (printout);
	}
}


/*****************************************************************************/
/*****************************************************************************/
int32_t
dump_ia32_context (FILE *fp, ia32el_context32 *ec32)
{
	FDUMPF (fp, "IA-32 Registers:\n");
	FDUMPF (fp, "\teax = %08x\tecx = %08x\tedx = %08x\tebx = %08x\n",
		ec32->eax,
		ec32->ecx,
		ec32->edx,
		ec32->ebx);
	FDUMPF (fp, "\tesp = %08x\tebp = %08x\tesi = %08x\tedi = %08x\n",
		ec32->esp,
		ec32->ebp,
		ec32->esi,
		ec32->edi);
	FDUMPF (fp, "\t es = %08x\t cs = %08x\t ss = %08x\t ds = %08x\n",
		ec32->es,
		ec32->cs,
		ec32->ss,
		ec32->ds);
	FDUMPF (fp, "\t fs = %08x\t gs = %08x\n",
		ec32->fs,
		ec32->gs);
	FDUMPF (fp, "\teip = %08x\teflags = %08x\n",
		ec32->eip,
		ec32->eflags);

	FDUMPF (fp, "\n");
	dump_ia32_stack(fp, ec32);
	FDUMPF (fp, "\n");
	FDUMPF (fp, "\n");
	dump_ia32_code(fp, ec32);
	FDUMPF (fp, "\n");
	FDUMPF (fp, "\n");
	FDUMPF (fp, "Thread Stack Trace:\n");

	dump_ia32_stack_trace(fp, ec32->eip, ec32->ebp);
	return 0;
}

/*****************************************************************************/
/* read the list of loaded 32 bit modules into the given list of binary
 * descriptors */
/*****************************************************************************/
static int32_t
read_maps (binary_desc ** bd, uint32_t * module_num)
{
	phdr_iter_data data;
	int32_t ret = -1, retval;
	binary_desc * bd64 = NULL;
	uint32_t size;

	data.bd = &bd64;
	data.mn = module_num;
	/* find the list of modules */
	DBPRINT (0, "bd=%p *bd=%p\n", data.bd, *(data.bd));
	retval = do_ld_iterate_phdr (iterate_phdr_callback, &data);
	if (retval != 0) {
		DBPRINT (true, "internal_ld_iterate_phdr failed\n");
		goto out;
	}

	/* copy the list to memory accessible to auxapp */
	size = sizeof (binary_desc) * *module_num;
	*bd = I2P (malloc32 (size));
	if (*bd == NULL) {
		DBPRINT (true, "malloc32 failed\n");
		goto out;
	}
	memcpy (*bd, bd64, size);
	/* the 64 bit list is not needed anymore */
	free (bd64);

	ret = 0;
out:
	DBPRINT (1, "ret=%d module_num=%d\n", ret, *module_num);
	return ret;
}
/*****************************************************************************/
/* this function gets called for each and every 32 bit module. add an
 * element to the list of binary-descriptors for each such modules that
 * has executable segment */
/*****************************************************************************/
static int32_t
iterate_phdr_callback (struct dl_phdr_info64 * info, size_t size UNUSED, void * data)
{
	phdr_iter_data * data64 = data;
	uint32_t j, new_bd_size, new_bd_bytes;
	binary_desc * new_bd;
	bool found = false;
	int32_t retval = -1;


	DBPRINT (0, "\n");
	assert (size == sizeof (*info));
	assert (info->dlpi_name && info->dlpi_name[0]);

	DBPRINT (0, "name=<%s> address=%lx\n", info->dlpi_name, info->dlpi_addr);

	for (j = 0; j < info->dlpi_phnum; j ++) {
		DBPRINT (0, "phdr[%d] type=%x flags=%x\n",
				j, info->dlpi_phdr[j].p_type, info->dlpi_phdr[j].p_flags);
		if ((info->dlpi_phdr[j].p_type == PT_LOAD) &&
				(info->dlpi_phdr[j].p_flags & PF_X)) {
			if ( ! found ) {
				/* this is the first time we encounter an executable
				 * segment for this module. add an element to the list
				 * of binary-descriptors */

				/* enlarge the list */
				new_bd_size = ++(*data64->mn);
				new_bd_bytes = new_bd_size * sizeof (binary_desc);
				*(data64->bd) = realloc (*(data64->bd), new_bd_bytes);
				if (*(data64->bd) == NULL) {
					DBPRINT (true, "realloc failed\n");
					goto out;
				}
				/* init the new element */
				new_bd = *(data64->bd) + new_bd_size - 1;
				strncpy (new_bd->full_path, info->dlpi_name,
						sizeof (new_bd->full_path) - 1);
				new_bd->full_path[sizeof (new_bd->full_path) - 1] = '\0';
				new_bd->base_address = info->dlpi_addr;

				assert (info->dlpi_phdr[j].p_offset == 0);
				assert (info->dlpi_phdr[j].p_vaddr == 0);

				found = true;
			} else {
				DBPRINT (true, "module <%s> has more than one executable segment\n",
						info->dlpi_name);
				assert ( ! "should not get here" );
				goto out;
			} /* ! found */
		} /* is executable segment */
	} /* loop over all phdr's */
	retval = 0;
out:
	DBPRINT (0, "retval=%d\n", retval);
	return retval;
}

/*****************************************************************************/
/* finds the module where the given address belongs and pass to auxapp to
 * get the description string */
/*****************************************************************************/
static int32_t
find_bdh_and_print_symbol (binary_desc *bdh_list, uint32_t bdh_num,
		uint32_t addr, ptr32_t output, uint32_t len)
{
	int32_t j = bdh_num - 1;
	uint64_t downcall_retval;
	int32_t retval = -1, ret;

	DBPRINT (1, "addr=%x\n", addr);
	while (j >= 0) {
		if (addr > bdh_list[j].base_address)
			break;
		j --;
	}
	if (j < 0) {
		DBPRINT (true, "%s: could not find module\n", __FUNCTION__);
		goto out;
	}
	DBPRINT (1, "bdh_list[%d] full_path=%p<%s>\n",
			j, bdh_list[j].full_path, bdh_list[j].full_path);
	if (bdh_list[j].bdh == 0) {
		ret = do_downcallV (MME_JAVA_DATA_TYPE_INT,
				&downcall_retval,
				I2P(g_ia32_runtime.binary_desc_init),
				"II", bdh_list[j].full_path,
				bdh_list[j].base_address);
		if (ret != 0) {
			DBPRINT (true, "calling binary_desc_init failed\n");
			goto out;
		}
		assert (IS_32BIT_VALUE (downcall_retval));
		bdh_list[j].bdh = (uint32_t) downcall_retval;

		if (bdh_list[j].bdh == 0) {
			DBPRINT (true, "%s: binary_desc_init failed\n", __FUNCTION__);
			goto out;
		}
	}

	ret = do_downcallV (MME_JAVA_DATA_TYPE_INT,
			&downcall_retval,
			I2P(g_ia32_runtime.get_symbol_info_str),
			"IIII", bdh_list[j].bdh, addr, output, len);
	if (ret != 0) {
		DBPRINT (true, "calling get_symbol_info_str failed\n");
		goto out;
	}

	retval = 0;

out:
	return retval;
}
