/* 
 * mme_internal.c
 *
 * 
 *
 */
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
#include <stdio.h>
#include <dlfcn.h>
#include <string.h>
#include <setjmp.h>
#include <pthread.h>
#include <stdarg.h>
#include <assert.h>
#include <stdlib.h>
#include <elf.h>
#include <link.h>
#include <sys/mman.h>
#include <sys/stat.h>

#include <linux/limits.h>
#include <ia32el_mme_api.h>
#include <mme_internal.h>
#include <mme_jvm_api.h>
#include <auxapp.h>
#include <jni32.h>
#include <jni32_refmap.h>
#include <jni32_stack.h>


#if MMEDEBUG
#define MME_LOG_FILE "./log.mme"
FILE * g_log_file = NULL;
#endif
#define AUX_APP "auxapp"

static char * m_search_path[] = {
	"/usr/local/bin",
	"/usr/lib",
	"/lib",
	"/usr/lib/ia32el",
	""
};

extern char **environ;

/* used for signal handling*/
extern int64_t g_tls_offset;

/* mme api function table  - mme passes this table to ia32el at initialization time */
static struct _mme_api_table_tag m_ia32el_api = {
	&mme_upcall,
	&mme_thread_at_clone,
	&mme_run_jvm_handler,
};

/* table of pointers to functions exported by ia32el */
struct _mme_ia32el_exports_tag HIDDEN g_ia32el_exports;

/* table of pointers to 32 bit functions that will be used internally by mme */
struct ia32_runtime_fptr_table HIDDEN g_ia32_runtime;

/* stubs */
HIDDEN void * g_stubs_area = NULL;

/* handle of ia32el shared-lib */
static void * m_ia32el = NULL;

/* pc at the point of exception in one of the 32 bit modules */
static uint32_t m_exception_pc = 0;

/* function prototypes */
#if FIXED_STUBS
static void * prepare_stub (uint32_t);
static int32_t free_stub_area (void);
#endif // FIXED_STUBS
static int32_t search_filename (char * name, char * fullpath, uint64_t pathlen);

static int32_t (*m_phdr_iter_callback) (struct dl_phdr_info64 *, size_t, void *);
/*****************************************************************************/
/*****************************************************************************/
int32_t HIDDEN
internal_mme_init (uint64_t flags, int64_t tls_offset)
{
	static bool is_initialized = false;
	bool signal_handlers_on = false;
	int32_t ret, retval = -1;
	char mme_verstr[IA32EL_MME_VER_STR_LEN];
	char filename[PATH_MAX];
	char * auxapp_argv[] = {NULL, NULL};
	ia32el_retval ia32el_ret;
	ptr32_t table_p;
	static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

	/* XXX do we really need this mutex here - or are we guaranteed not to
	 * have two threads trying to init mme at the same time */
	if (pthread_mutex_lock (&mutex)) {
		DBPRINT (true, "mutex_lock_failed\n");
		return -1;
	}
	/* init only once */
	if ( is_initialized ) {
		goto out;
	}
	
#if MMEDEBUG
#if USE_LOG_FILE
	/* open log file */
	g_log_file = fopen (MME_LOG_FILE, "w");
	if (g_log_file == NULL)
#endif
		g_log_file = stderr;
	setvbuf (g_log_file, NULL, _IONBF, 0ul);
#endif

/*	Store the given parameters in MTL's library memory	*/
	g_flags = flags;
	
/*	Load ia32el using dlopen	*/
	ret = search_filename (IA32EL_NAME, filename, sizeof (filename) - 1);
	if (ret != 0) {
		DBPRINT (true, "could not find %s\n", IA32EL_NAME);
		goto out;
	}
	m_ia32el = dlopen (filename, RTLD_LAZY);
	if (m_ia32el == NULL) {
		DBPRINT (true, "could not open %s: %s\n", IA32EL_NAME, dlerror());
		goto out;
	}
#define MAP_IA32EL_SYM(sym)									\
	g_ia32el_exports.ia32el_##sym = dlsym (m_ia32el, "ia32el_mme_" #sym);			\
	if (g_ia32el_exports.ia32el_##sym == NULL) {						\
		DBPRINT (true, "could not find symbol ia32el_mme_" #sym ":%s\n", dlerror());	\
		goto out;									\
	}

/*	Check the MME version of ia32el by calling mme_verstr. It must be
 *	equal to the version of MTL.	*/
	MAP_IA32EL_SYM (version);
	ia32el_ret = g_ia32el_exports.ia32el_version (mme_verstr);
	if (ia32el_ret != IA32EL_SUCCESS) {
		DBPRINT (true, "could not get ia32el version\n");
		goto out;
	}

	if (strncmp (mme_verstr,
				IA32EL_MME_VER_MAJOR "." IA32EL_MME_VER_MINOR,
				sizeof (mme_verstr))) {
		DBPRINT (true, "ia32el version mismatch:%s\n", mme_verstr);
		goto out;
	}
/*	Initialize ia32el by calling ia32el_mme_init, passing the API table as a parameter */
	MAP_IA32EL_SYM (init);
	ia32el_ret = g_ia32el_exports.ia32el_init (g_flags, &m_ia32el_api, tls_offset);
        g_tls_offset = tls_offset;
	if (ia32el_ret != IA32EL_SUCCESS) {
		DBPRINT (true, "ia32el init failed\n");
		goto out;
	}
/* 	find all the rest of the ia32el export functions */
	MAP_IA32EL_SYM (load_and_exec);
	MAP_IA32EL_SYM (ia32_context_get);
	MAP_IA32EL_SYM (memory_map);
	MAP_IA32EL_SYM (memory_unmap);
	MAP_IA32EL_SYM (run);
	MAP_IA32EL_SYM (signal_handler);
	MAP_IA32EL_SYM (thread_term);
	MAP_IA32EL_SYM (complete_child_init);
	MAP_IA32EL_SYM (complete_parent_thread_create);
	MAP_IA32EL_SYM (query_memory_permission);
	
/*	If signal handling policy is SIG_HANDLE then register signal handler for
 *	all signals which are used for notifying about hardware exceptions.	*/

	if (g_flags & MME_MTL_SIG_HANDLE) {
		ret = mme_register_signal_handlers ();
		if (ret != 0) {
			goto out;
		}
	}
	signal_handlers_on = true;

/*	Use the auxiliary application to find the addresses of all the symbols
 *	in the IA-32 runtime table	*/

	/*	Load the auxiliary application by calling ia32el_load_and_exec with
	 *	the appropriate parameters	*/
		ret = search_filename (AUX_APP, filename, sizeof (filename) - 1);
		if (ret != 0) {
			DBPRINT (true, "could not find %s\n", AUX_APP);
			goto out;
		}
		auxapp_argv[0] = filename;
		ia32el_ret = g_ia32el_exports.ia32el_load_and_exec (1, auxapp_argv, environ);
		if (ia32el_ret != IA32EL_SUCCESS) {
			DBPRINT (true, "ia32el load_and_exec failed\n");
			goto out;
	} else {
		/* ia32el returned */
		/* address of the function pointer table is located at esp */
		ia32el_context32 ec32;
		ia32el_ret = g_ia32el_exports.ia32el_ia32_context_get (&ec32);
		if (ia32el_ret != IA32EL_SUCCESS) {
			goto out;
		}
		memcpy (&table_p, I2P(ec32.esp), sizeof (ptr32_t));
		memcpy (&g_ia32_runtime, I2P(table_p), sizeof (g_ia32_runtime));
#if MMEDEBUG
		{
			char buff[4096];
			sprintf (buff, "g_ia32_runtime=%p\n", &g_ia32_runtime);
			sprintf (buff + strlen (buff), "\tdlopen=%x\n"
					"\tdlsym=%x\n"
					"\tdlerror=%x\n"
					"\tdlclose=%x\n"
					"\tpthread_create=%x\n"
					"\tpthread_attr_setstack=%x\n"
					"\tpthread_attr_setstacksize=%x\n"
					"\tpthread_attr_setdetachstate=%x\n"
					"\tpthread_attr_init=%x\n"
					"\tpthread_exit=%x\n"
					"\tmalloc=%x\n"
					"\trealloc=%x\n"
					"\tfree=%x\n"
					"\tjnienv32_fns=%x\n"
					"\tjavavm32_fns=%x\n"
					"\tdummy_thread_func=%x\n",
					g_ia32_runtime.dlopen,
					g_ia32_runtime.dlsym,
					g_ia32_runtime.dlerror,
					g_ia32_runtime.dlclose,
					g_ia32_runtime.pthread_create,
					g_ia32_runtime.pthread_attr_setstack,
					g_ia32_runtime.pthread_attr_setstacksize,
					g_ia32_runtime.pthread_attr_setdetachstate,
					g_ia32_runtime.pthread_attr_init,
					g_ia32_runtime.pthread_exit,
					g_ia32_runtime.malloc,
					g_ia32_runtime.realloc,
					g_ia32_runtime.free,
					g_ia32_runtime.jnienv32_fns,
					g_ia32_runtime.javavm32_fns,
					g_ia32_runtime.dummy_thread_func);
			DBPRINT (1, "%s", buff);
		}
#endif

	}

#if FIXED_STUBS
	/* prepare the stubs */
	g_stubs_area = g_ia32el_exports.ia32el_memory_map (0, STUBS_AREA_SIZE, PROT_READ|PROT_EXEC);
	if ((int64_t)g_stubs_area < 0) {
		DBPRINT (true, "ia32el_memory_map failed:%ld\n", (int64_t)g_stubs_area);
		goto out;
	}

	/* we cheat here on ia32el, changing the permissions of the memory
	 * it allocated for us. it should not matter much, since we're only
	 * in the init phase. we return to orig prot right away, before any
	 * 32 bit code is run */
	int32_t syscallret;
	syscallret = mprotect (g_stubs_area, STUBS_AREA_SIZE, PROT_READ | PROT_WRITE);
	if (syscallret < 0) {
		DBCODE (true, perror ("mprotect1 failed"));
		goto out;
	}
	(void)prepare_stub (IA32EL_TRAP_RETURN_FROM_RUN);
	syscallret = mprotect (g_stubs_area, STUBS_AREA_SIZE, PROT_READ | PROT_EXEC);
	if (syscallret < 0) {
		DBCODE (true, perror ("mprotect2 failed"));
		goto out;
	}
#endif // FIXED_STUBS
	/* finish successfuly */
	is_initialized = true;
	retval = 0;

out:
	if (pthread_mutex_unlock (&mutex)) {
		DBPRINT (true, "mutex_unlock_failed\n");
		retval = -1;
	}
	if (retval != 0) {
		if (m_ia32el != NULL) {
			dlclose (m_ia32el);
		}
		if (signal_handlers_on) {
			mme_unregister_signal_handlers ();
		}
	}
	return retval;

}

void HIDDEN
free32 (ptr32_t ptr32)
{
	int32_t retval;
	retval = do_downcallV (MME_JAVA_DATA_TYPE_VOID,
			NULL, I2P(g_ia32_runtime.free),
			"I", ptr32);
	if (retval != 0) {
		DBPRINT (1, "\tmme-retval=%d\n", retval);
	}
}

ptr32_t HIDDEN
realloc32 (ptr32_t ptr, uint32_t size)
{
	int32_t retval;
	uint64_t downcall_retval;
	ptr32_t ptr32 = 0;

	retval = do_downcallV (MME_JAVA_DATA_TYPE_INT,
			&downcall_retval, I2P(g_ia32_runtime.realloc),
			"II", ptr, size);
	if (retval != 0) {
		DBPRINT (1, "mme-retval=%d\n", retval);
		goto out;
	}
	assert (IS_32BIT_VALUE (downcall_retval));
	ptr32 = (ptr32_t)downcall_retval;
out:
	return ptr32;
}

ptr32_t HIDDEN
malloc32 (uint32_t size)
{
	int32_t retval;
	uint64_t downcall_retval;
	ptr32_t ptr32 = 0;

	retval = do_downcallV (MME_JAVA_DATA_TYPE_INT,
			&downcall_retval, I2P(g_ia32_runtime.malloc),
			"I", size);
	if (retval != 0) {
		DBPRINT (1, "mme-retval=%d\n", retval);
		goto out;
	}
	assert (IS_32BIT_VALUE (downcall_retval));
	ptr32 = (ptr32_t)downcall_retval;
out:
	return ptr32;
}

#if FIXED_STUBS
/*****************************************************************************/
/*****************************************************************************/
static void * prepare_stub (uint32_t stub_type)
{
	char * stub = NULL;
	union {
		uint32_t i;
		char c[4];
	} i2c;

	stub = get_stub_addr (stub_type);

	/* write the stub - "jmpe $stubtype" */
	i2c.i = stub_type;
	stub[0] = 0x0f;
	stub[1] = 0xb8;
	stub[2] = i2c.c[0];
	stub[3] = i2c.c[1];
	stub[4] = i2c.c[2];
	stub[5] = i2c.c[3];

	return stub;
}

/*****************************************************************************/
/*****************************************************************************/
static int32_t free_stub_area ()
{
	ia32el_retval ret;
	ret = g_ia32el_exports.ia32el_memory_unmap (g_stubs_area, STUBS_AREA_SIZE);
	return (ret == IA32EL_SUCCESS) ? 0 : -1;
}
#endif // FIXED_STUBS

/*****************************************************************************/
/*****************************************************************************/
int32_t HIDDEN internal_mme_unload ()
{
	int32_t retval = 0;

	retval |= jni32_term ();

#if FIXED_STUBS
	retval |= free_stub_area ();
#endif
	
	return retval;
}
/*****************************************************************************/
/*****************************************************************************/
static int32_t search_filename (char * name, char * fullpath, uint64_t pathlen)
{
	int32_t j, fail, ret;
	struct stat st;

	fail = 1;
	for (j = 0; m_search_path[j]; j++) {
		snprintf (fullpath, pathlen, "%s/%s", m_search_path[j], name);
		ret = stat (fullpath, &st);
		if (ret == 0) {
			fail = 0;
			break;
		}
	}

	return fail;
}

/*****************************************************************************/
/*****************************************************************************/
int32_t HIDDEN
internal_ld_iterate_phdr (phdr_iter_callback callback, void * data)
{
	int32_t retval = -1;
        mme_sem_data *semd64 = NULL;


	semd64 = (mme_sem_data *)data;
	*semd64->is_ia32_native_exception = 1;

	/* when we are here we should be still on exceptioned ia32 context */

	/* XXX after implementing the 32 bit stack-trace, this is not correct
	 * anymore. untill a better fix is available, we will assume that the
	 * pc at the point of exception is the one that was recorded when
	 * mme_dump_foriegn_exception was called XXX */
#if 0
        ia32el_context32 ec32;
        ia32el_retval ia32el_ret;
	ia32el_ret = g_ia32el_exports.ia32el_ia32_context_get (&ec32);
	if (ia32el_ret != IA32EL_SUCCESS) {
		DBPRINT (true, "ia32el_ia32_context_get failed\n");
		goto out;
	}
	*semd64->eip32 = (void *)I2P(ec32.eip);
#else 
	*semd64->eip32 = (void *)I2P(m_exception_pc);
#endif

	retval = do_ld_iterate_phdr (callback, data);

//out:

	return retval;
}
/*****************************************************************************/
/*****************************************************************************/
int32_t HIDDEN
do_ld_iterate_phdr (phdr_iter_callback callback, void * data)
{
	static pthread_mutex_t phdr_iter_mutex = PTHREAD_MUTEX_INITIALIZER;
	uint32_t data32 = 0;
	uint64_t downcall_retval;
	int32_t retval;
	bool local_frame_inited = false;


	DBPRINT (true, "0\n");
	if (pthread_mutex_lock (&phdr_iter_mutex)) {
		DBPRINT (true, "mutex_lock_failed\n");
		return -1;
	}

	assert (m_phdr_iter_callback == NULL);
	m_phdr_iter_callback = callback;
	/* convert to 32 bit value and store */

	if (!jni32_is_local_refmap_inited()) {
		retval = jni32_init_local_refmap ();
		if (retval < 0) {
			DBPRINT (true, "jni32_init_local_refmap failed\n");
			goto out;
		}
		local_frame_inited = true;
	}

	retval = jni32_predef_size_push_local_ref_frame ();
	if (retval != 0) {
		DBPRINT (true, "jni32_predef_size_push_local_ref_frame failed\n");
		goto out1;
	}

	data32 = jni32_marshal_local_ref (data);
	if (data32 == 0) {
		DBPRINT (true, "jni32_marshal_local_ref failed\n");
		retval = -1;
		goto out;
	}
	/* call the 32 bit dl_iterate_phdr */
	retval = do_downcallV (MME_JAVA_DATA_TYPE_INT,
			&downcall_retval, I2P(g_ia32_runtime.dl_iterate_phdr),
			"II", g_ia32_runtime.phdr_iter_callback, data32);
	if (retval != 0) {
		DBPRINT (true, "do_downcall failed\n");
		goto out;
	}
out:
	if (data32 != 0) {
		(void)jni32_local_ref_del (data32);
	}

        retval = jni32_pop_local_ref_frame ();
        if (retval != 0) {
                DBPRINT (true, "jni32_pop_local_ref_frame failed\n");
        }

	if (local_frame_inited) {
		retval = jni32_uninit_local_refmap ();
		if (retval != 0) {
			DBPRINT (true, "jni32_uninit_local_refmap failed\n");
		}
	}

out1:
	m_phdr_iter_callback = NULL;
	if (pthread_mutex_unlock (&phdr_iter_mutex)) {
		DBPRINT (true, "pthread_mutex_unlock failed\n");
	}

	return retval;
}


/*****************************************************************************/
/*****************************************************************************/
struct dl_phdr_info32 {
	ptr32_t dlpi_addr;
	ptr32_t dlpi_name;
	ptr32_t dlpi_phdr;
	Elf32_Half dlpi_phnum;
};

int32_t HIDDEN process_phdr_iter_callback (ia32el_context32 * ec32)
{
	uint32_t ap32, size32;
	ptr32_t info32p;
	struct dl_phdr_info32 * info32;
	void * data64;
	struct dl_phdr_info64 info64;
	size_t size64;
	int32_t retval;
	int32_t i;

	/* get the parameters to the 32 bit callback function:
	 * esp + 16: data
	 * esp + 12: size
	 * esp + 8 : info
	 * esp + 4 : return address
	 * esp     : 0x0
	 */
	ap32 = btmix_get_upcall_argp (ec32);
	info32p = btmix_pop_pointer (&ap32);
	size32 = btmix_pop_uint (&ap32);
	data64 = jni32_demarshal_ref (&ap32);
	if (data64 == NULL) {
		DBPRINT (true, "jni32_demarshal_ref failed\n");
		retval = -1;
		goto out;
	}

	/* convert to 64 bit values */
	info32 = I2P (info32p);
	info64.dlpi_name = (char *)(uintptr_t)info32->dlpi_name;

	if (info64.dlpi_name == NULL || strlen(info64.dlpi_name) == 0) {
		/* continue with the iterator, we need a valid ia32 module name */
                /* and we need to skip the exec path name */
		btmix_set_retval_int (ec32, 0);
		return 0;
	}
 
	info64.dlpi_addr = (Elf64_Addr)info32->dlpi_addr;
	info64.dlpi_phnum = (Elf64_Half)info32->dlpi_phnum;
	info64.dlpi_phdr = (Elf64_Phdr *)malloc(sizeof(Elf64_Phdr) * (info64.dlpi_phnum));
	Elf32_Phdr *phdr32 = (Elf32_Phdr *)I2P(info32->dlpi_phdr);
	for (i = 0; i < info64.dlpi_phnum; i++) {
		info64.dlpi_phdr[i].p_type   = (Elf64_Word)(phdr32[i].p_type);
		info64.dlpi_phdr[i].p_flags  = (Elf64_Word)(phdr32[i].p_flags);
		info64.dlpi_phdr[i].p_offset = (Elf64_Off)(phdr32[i].p_offset);
		info64.dlpi_phdr[i].p_vaddr  = (Elf64_Addr)(phdr32[i].p_vaddr);
		info64.dlpi_phdr[i].p_paddr  = (Elf64_Addr)(phdr32[i].p_paddr);
		info64.dlpi_phdr[i].p_filesz = (Elf64_Xword)(phdr32[i].p_filesz);
		info64.dlpi_phdr[i].p_memsz  = (Elf64_Xword)(phdr32[i].p_memsz);
		info64.dlpi_phdr[i].p_align  = (Elf64_Xword)(phdr32[i].p_align);
	}
	size64 = sizeof (info64);

	/* call the jvm's callback function */
	retval = m_phdr_iter_callback (&info64, size64, data64);

	free((void *)info64.dlpi_phdr);

	/* set the 32 bit return value */
	btmix_set_retval_int (ec32, retval);
out:
	DBPRINT (0, "retval=%d\n", retval);
	return retval;
}

/*****************************************************************************/
/*****************************************************************************/
/* call JVM upcall from this function, JVM should handle it */
void mme_run_jvm_handler (void * siginfo)
{
	g_java_callbacks.run_jvm_handler(siginfo);
}

/*****************************************************************************/
/*****************************************************************************/
int32_t internal_dump_exception_context (FILE *fp)
{
	ia32el_retval ia32el_ret;
	int32_t retval;
	ia32el_context32 ec32;

	ia32el_ret = g_ia32el_exports.ia32el_ia32_context_get (&ec32);

	if (ia32el_ret != IA32EL_SUCCESS) {
		DBPRINT (true, "ia32el_ia32_context_get failed\n");
		retval = -1;
		goto out;
	}
	m_exception_pc = ec32.eip;

	retval = dump_ia32_context(fp, &ec32);

out:
	return retval;
}

/*****************************************************************************/
/*****************************************************************************/
int32_t HIDDEN is_nio_call(char *param_list, va_list params, uint64_t *return_value)
{
	char *p;
	int32_t  retval = 0;
	uint64_t env;
	uint64_t size;
	uint64_t m;

	p = param_list;
	if (strcmp(p, "EM") && strcmp(p, "EK")) {
		retval = 0;
		goto out;  /* not an NIO call */
	}

	while (*p != '\0') {
		switch (*p) {
		case 'E':
			/* nothing needs to do with JNIEnv */
			env = va_arg(params, uint64_t);
			break;
		case 'M':
			size = va_arg (params, uint64_t);
			*return_value = (uint64_t)malloc32(size);
			retval = 1;
			break;
		case 'K':
			m = va_arg (params, uint64_t);
			free32(m);
			retval = 1;
			break;
		case 'P':
			m = va_arg (params, uint64_t);
			size = va_arg (params, uint64_t);
			*return_value = (uint64_t)realloc32(m, size);
			retval = 1;
			break;
		default:
			retval = 0;
			break;
		}
		p++;
	}
out:
	return retval;
}

