/* 
 * mme_jvm_api.c
 *
 * 
 *
 */
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
#include <stdlib.h>
#include <string.h>
#include <linux/limits.h>

#include <mme_jvm_api.h>
#include <ia32el_mme_api.h>
#include <mme_internal.h>
#include <auxapp.h>
#include <jni32.h>
#include <jni32_diag.h>

/* mode flags - jvm passes these flags to mme at initialization time */
uint64_t HIDDEN g_flags;

/* 32 bit memory area - for converting 64 bit strings to 32 bit strings */
static MME_THREAD_VAR ptr32_t tm_conversion_area = 0;


/*****************************************************************************/
/*****************************************************************************/
mme_retval mme_version (char version[MME_MTL_VER_STR_LEN])
{
	/* version string */
	static const char m_verstr[MME_MTL_VER_STR_LEN] = MTL_VER_MAJOR "." MTL_VER_MINOR;
	strcpy (version, m_verstr);
	return MME_SUCCESS;
}

/*****************************************************************************/
/*****************************************************************************/
mme_retval mme_complete_thread_init (mme_thread_init_flag flag, void *clone_params)
{
	int32_t ret;

	ret = after_thread_create (flag, clone_params);
	
	return (ret == 0) ? MME_SUCCESS : MME_FAILURE;
}
/*****************************************************************************/
/*****************************************************************************/
mme_retval mme_init (uint64_t flags, struct _mme_java_callback_tag * callbacks, int64_t tls_offset)
{
	int32_t retval;
	mme_retval mme_ret = MME_FAILURE;

	mme_ret = internal_mme_init (flags, tls_offset);
	if (mme_ret != MME_SUCCESS) {
		goto out;
	}

	retval = jni32_init (callbacks);
	if (retval < 0) {
		goto out;
	}

	retval = init_thread (false);
	if (retval < 0) {
		goto out;
	}

out:
	DBPRINT (1, "mme_ret=%d\n", mme_ret);
	return mme_ret;
}


/*****************************************************************************/
/*****************************************************************************/
uint64_t mme_downcall (int64_t return_type, void *addr, char *param_list, ...) 
{
	va_list params;
	int64_t retval;

	va_start (params, param_list);
	retval = mme_downcallV (return_type, addr, param_list, params);
	va_end (params);
	
	return retval;
}

uint64_t mme_downcallV (int64_t return_type, void* addr, char* param_list, va_list params)
{
	
	int32_t retval;
	uint64_t jni_retval;

	DBPRINT (1, "return_type=%ld addr=%p param_list=%s\n", return_type, addr, param_list);
	/* do the work */
	retval = do_downcall (true, return_type, &jni_retval, addr, param_list, params);
	/* raise exception in case of failure */
	if (retval) {
		DBPRINT (true, "do_downcall failed\n");
		MME_THROW_EXCEPTION ("internal error in do_downcall");
	}

	if (return_type == MME_JAVA_DATA_TYPE_FLOAT) {
		__asm__ __volatile__ ("ldfs f8=%0;;\n":: "m"(jni_retval));
	} else if (return_type == MME_JAVA_DATA_TYPE_DOUBLE) {
		__asm__ __volatile__ ("ldfd f8=%0;;\n":: "m"(jni_retval));
	}
#if MMEDEBUG
	if (return_type == MME_JAVA_DATA_TYPE_FLOAT) {
		float f;
		__asm__ __volatile__ ("stfs %0=f8;;\n": "=m"(f));
		DBPRINT (1, "fp-retval=%E\n", f);
		/* f8 is overwritten by the DBPRINT */
		__asm__ __volatile__ ("ldfs f8=%0;;\n":: "m"(f));
	} else if (return_type == MME_JAVA_DATA_TYPE_DOUBLE) {
		double f;
		__asm__ __volatile__ ("stfd %0=f8;;\n": "=m"(f));
		DBPRINT (1, "fp-retval=%E\n", f);
		/* f8 is overwritten by the DBPRINT */
		__asm__ __volatile__ ("ldfd f8=%0;;\n":: "m"(f));
	} else {
		DBPRINT (1, "int-retval=%lx<%ld><%lu>\n", jni_retval, jni_retval, jni_retval);
	}
#endif

	return jni_retval;
}
/*****************************************************************************/
/*****************************************************************************/
#define get_tm_conversion_area_size() (uint32_t)PATH_MAX
static void * get_tm_conversion_area ()
{
	void * areap = NULL;

	if (tm_conversion_area == 0) {
		tm_conversion_area = malloc32 (get_tm_conversion_area_size());
		if (tm_conversion_area == 0) {
			DBPRINT (true, "malloc failed\n");
			goto out;
		}
		DBPRINT (1, "tm_conversion_area=%x\n", tm_conversion_area);
	}
	areap = I2P (tm_conversion_area);
out:
	return areap;
}
/*****************************************************************************/
/*****************************************************************************/
void * mme_dlopen (const char * filename, int32_t flag) 
{
	mme_retval retval;
	void * lib = NULL;
	char * filename32;
	uint64_t downcall_retval;

	/* convert the filename to 32 bit pointer */
	assert (PATH_MAX <= get_tm_conversion_area_size());
	filename32 = get_tm_conversion_area ();
	filename32[PATH_MAX - 1] = '\0';
	strncpy (filename32, filename, PATH_MAX - 1);

	/* call the ia32 dlopen */
	DBPRINT (1, "dlopen: '%s'\n", filename32);
	retval = do_downcallV (MME_JAVA_DATA_TYPE_INT, &downcall_retval,
			I2P(g_ia32_runtime.dlopen), "II", filename32, flag);
	if (retval != MME_SUCCESS) {
		DBPRINT (1, "\tmme-retval=%d\n", retval);
		goto out;
	}
	assert (IS_32BIT_VALUE (downcall_retval));
	lib = I2P (downcall_retval);
	DBPRINT (1, "\tdowncall_retval=%lx lib=%p\n", downcall_retval, lib);
out:
	DBPRINT (1, "lib=%p\n", lib);
	return lib;
}
/*****************************************************************************/
/*****************************************************************************/
char * mme_dlerror (void) 
{
	mme_retval retval;
	union {
		void * errmsg;
		uint64_t ptr;
	} pc;

	retval = do_downcallV (MME_JAVA_DATA_TYPE_INT, (uint64_t*)&pc.ptr,
			I2P(g_ia32_runtime.dlerror), "");
	DBPRINT (1, "retval=%d\n", retval);
	DBPRINT (retval == MME_SUCCESS, "errmsg=%p\n", pc.errmsg);
	return (retval == MME_FAILURE) ? NULL : pc.errmsg;
}
/*****************************************************************************/
/*****************************************************************************/
void * mme_dlsym (void * handle, const char * symbol) 
{
	mme_retval retval;
	void * sym = NULL;
	char * symbol32;
	uint64_t downcall_retval;
	ptr32_t handle32;

	/* convert the symbol to 32 bit pointer */
	assert (PATH_MAX <= get_tm_conversion_area_size());
	symbol32 = get_tm_conversion_area ();
	symbol32[PATH_MAX - 1] = '\0';
	strncpy (symbol32, symbol, PATH_MAX - 1);

	/* call the ia32 dlsym */
	DBPRINT (1, "dlsym: handle=%p symbol='%s'\n", handle, symbol);
	assert (IS_32BIT_VALUE (handle));
	handle32 = TRUNCATE_PTR (handle);
	retval = do_downcallV (MME_JAVA_DATA_TYPE_INT, &downcall_retval,
			I2P(g_ia32_runtime.dlsym), "II", handle32, symbol32);
	if (retval != MME_SUCCESS) {
		DBPRINT (1, "\tmme-retval=%d\n", retval);
		goto out;
	}
	assert (IS_32BIT_VALUE (downcall_retval));
	sym = I2P (downcall_retval);
	DBPRINT (0, "\tdowncall_retval=%lx sym=%p\n", downcall_retval, sym);
out:
	DBPRINT (1, "sym=%p\n", sym);
	return sym;
}
/*****************************************************************************/
/*****************************************************************************/
int32_t mme_dlclose (void * handle) 
{
	mme_retval retval;
	ptr32_t handle32;
	uint64_t ret;

	handle32 = TRUNCATE_PTR (handle);
	retval = do_downcallV (MME_JAVA_DATA_TYPE_INT, &ret,
			I2P(g_ia32_runtime.dlclose), "I", handle);
	DBPRINT (1, "retval=%d\n", retval);
	DBPRINT (retval == MME_SUCCESS, "ret=%d\n", (int32_t)ret);
	return (retval == MME_FAILURE) ? -1 : (int32_t)ret;
}
/*****************************************************************************/
/*****************************************************************************/
mme_sighand_retval mme_signal_handler (int32_t signo, struct siginfo * si, struct sigcontext * sc) 
{
	ia32el_sighand_result ret;
	mme_sighand_retval    retval;

	ret = internal_mme_signal_handler (signo, si, sc);

	switch (ret) {
	case MME_IA32_NATIVE_EXCEPTION:
		retval = MME_SIGNAL_IA32_EXCEPTION;
		break;
	case MME_IA32_SIGNAL_HANDLED:
		retval = MME_SIGNAL_HANDLED;
		break;
	case MME_IA32_SIGNAL_NOT_HANDLED:
		retval = MME_SIGNAL_NOT_HANDLED;
		break;
	default:
		assert(0); /* not possible */
		retval = MME_SIGNAL_NOT_HANDLED;
		break;
	}

	return retval;
}
/*****************************************************************************/
/*****************************************************************************/
mme_retval mme_change_sighand (int32_t signo, struct sigaction * sc)
{
	bool ret;
	assert ((g_flags & MME_MTL_SIG_HANDLE) == 1);
	ret = change_sighand (signo, sc);
	return ret ? MME_SUCCESS : MME_FAILURE;
}
/*****************************************************************************/
/*****************************************************************************/
mme_retval mme_thread_init (void ** clone_params_p)
{
	int32_t retval;

	DBPRINT (1, "\n");
	retval = before_thread_create (clone_params_p);
	if (retval < 0) {
		return MME_FAILURE;
	}

	return MME_SUCCESS;
}
/*****************************************************************************/
/*****************************************************************************/
mme_retval mme_thread_term (void)
{
	int32_t retval;
	retval = internal_thread_term ();

	return (retval == 0) ? MME_SUCCESS : MME_FAILURE;
}
/*****************************************************************************/
/*****************************************************************************/
mme_retval mme_dl_iterate_phdr (phdr_iter_callback jvm_callback, void * data)
{
	int32_t retval;

	retval = internal_ld_iterate_phdr (jvm_callback, data);

	return (retval == 0) ? MME_SUCCESS : MME_FAILURE;
}
/*****************************************************************************/
/*****************************************************************************/
mme_retval mme_dump_foreign_context (FILE *fp)
{
	int32_t retval;
	retval = internal_dump_exception_context (fp);

	return (retval == 0) ? MME_SUCCESS : MME_FAILURE;
}
/*****************************************************************************/
/*****************************************************************************/
mme_retval mme_unload (int32_t reason UNUSED)
{
	mme_retval retval = MME_SUCCESS;
	int32_t internal_retval;

	DBPRINT (1, "\n");

	/* free the conversion area used in this file -
	 * must be done before the internal cleanup */
	if (tm_conversion_area) {
		free32 (tm_conversion_area);
	}

	/* do some internal cleanup */
	internal_retval = internal_mme_unload ();
	if (internal_retval < 0) {
		DBPRINT (true, "internal_mme_unload failed\n");
		retval = MME_FAILURE;
	}

	return retval;
}


