/* 
 * mme_thread.c
 *
 * 
 *
 */
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
#include <stdint.h>
#include <pthread.h>
#include <jni.h>
#include <stdlib.h>
#include <setjmp.h>

#include <jni32_refmap.h>
#include <jni32_jnienv.h>
#include <jni32_diag.h>
#include <mme_internal.h>
#include <ia32el_mme_api.h>
#include <auxapp.h>

/* child data - guarded by m_child_init_mutex */
static pthread_mutex_t m_child_init_mutex = PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP;
sigjmp_buf g_thread_func_jmpbuf;

/* parent data - guarded by m_parent_init_mutex */
static pthread_mutex_t m_parent_init_mutex = PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP;
static int32_t m_thread_in_init = 0;
static ptr32_t m_tid32 = 0;
static ia32el_context32 m_parent_ia32_ctx_at_clone;
static void * m_clone_params = NULL;

#define MME_DEFAULT_IA32_STACK_SIZE (128 * 1024) /* according to JVM setting */

/*****************************************************************************/
/* initialize a thread in the mme lib. don't get confused by the name - 
 * this is not what mme_thread_init does. */
/*****************************************************************************/
int32_t HIDDEN
init_thread (bool is_new_thread)
{
	int32_t retval;

	DBPRINT (1, "is_new_thread=%d\n", is_new_thread);
	/* init thread's map of local references */
	retval = jni32_init_local_refmap ();
	if (retval < 0) {
		DBPRINT (true, "jni32_init_local_refmap failed\n");
		goto out;
	}

	if (is_new_thread) {
		/* the case when this function is called from attach_thread */
		retval = jni32_predef_size_push_local_ref_frame ();
		if (retval != 0) {
			DBPRINT (true, "jni32_predef_size_push_local_ref_frame failed\n");
			goto out;
		}
	}

	DBPRINT (1, "retval=%d\n", retval);
	retval = 0;
out:
	return retval;
}



/*****************************************************************************/
/*****************************************************************************/
int32_t HIDDEN
before_thread_create (void ** clone_params_p)
{
	int32_t retval;
	uint64_t downcall_retval;
	int32_t ptc_retval;
	ptr32_t attr32;

	DBPRINT (1, "\n");

	/* only one thread can do this at a time */
	retval = pthread_mutex_lock (&m_parent_init_mutex);
	if (retval != 0) {
		DBPRINT (true, "pthread_mutex_lock failed\n");
		return MME_FAILURE;
	}
	m_thread_in_init = gettid ();

	/* get memory place holder for the 32 bit tid */
	assert (m_tid32 == 0);
	/* this memory will be freed at complete_thread_init */
	m_tid32 = malloc32 (sizeof (int32_t));
	if (m_tid32 == 0) {
		DBPRINT (true, "malloc failed\n");
		retval = -1;
		goto out;
	}

	attr32 = malloc32 (sizeof (pthread_attr32_t));
	if (attr32 == 0) {
		DBPRINT (true, "malloc failed\n");
		retval = -1;
		goto out;
	}
	/* call 32 bit pthread_attr_init and then pthread_attr_setstacksize */
	retval = do_downcallV (MME_JAVA_DATA_TYPE_INT, &downcall_retval,
			I2P (g_ia32_runtime.pthread_attr_init),
			"I", attr32);
	if (retval != 0) {
		DBPRINT(true, "pthread_att_init32 failed\n");
		free32 (attr32);
		goto out;
	}

	retval = do_downcallV (MME_JAVA_DATA_TYPE_INT, &downcall_retval,
			I2P (g_ia32_runtime.pthread_attr_setdetachstate),
			"II", attr32, PTHREAD_CREATE_DETACHED_32);
	if (retval != 0) {
		DBPRINT(true, "pthread_attr_setdetachstate failed\n");
		free32 (attr32);
		goto out;
	}

	retval = do_downcallV (MME_JAVA_DATA_TYPE_INT, &downcall_retval,
			I2P (g_ia32_runtime.pthread_attr_setstacksize),
			"II", attr32, MME_DEFAULT_IA32_STACK_SIZE);
	if (retval != 0) {
		DBPRINT(true, "pthread_attr_setstacksize failed\n");
		free32 (attr32);
		goto out;
	}

	/* call 32 bit pthread_create */
	retval = do_downcallV (MME_JAVA_DATA_TYPE_INT, &downcall_retval,
			I2P (g_ia32_runtime.pthread_create),
			"IIII", m_tid32, attr32, g_ia32_runtime.dummy_thread_func, NULL);
	if (retval != 0) {
		DBPRINT (true, "ptread_create32 failed\n");
		goto out;
	}
	ptc_retval = (int32_t)downcall_retval;
	free32(attr32);

	/* returning to JVM. at this point:
	 * 1) the parent thread finished initialization
	 *    stage at the 32 bit level and is right pass the point of clone.
	 * 2) ia32el has the parameters it needs to pass to the child thread ready
	 * the rest of the work will continue in mme_complete_thread_init, and that's
	 * where the mutex will be released, too */
	*clone_params_p = m_clone_params;
	retval = 0;

out:

	if (retval != 0) {
		retval = pthread_mutex_unlock (&m_parent_init_mutex);
		if (retval != 0) {
			DBPRINT (true, "pthread_mutex_unlock failed\n");
		}
		return -1;
	}

	DBPRINT (1, "retval=%d, clone_params=%p\n", retval, *clone_params_p);

	return 0;
}

/*****************************************************************************/
/*****************************************************************************/
int64_t HIDDEN
mme_thread_at_clone (void * params)
{
	ia32el_retval ia32el_ret;
	int64_t retval;
	int32_t tid;

	DBPRINT (1, "params=%p m_thread_in_init=%d\n", params, m_thread_in_init);
	tid = gettid ();
	if (tid == m_thread_in_init) {
		/* if the current thread is the one that initiated this clone then
		 * keep the context, keep the params-address and return 1 */
		ia32el_ret = g_ia32el_exports.ia32el_ia32_context_get (&m_parent_ia32_ctx_at_clone);
		if (ia32el_ret != IA32EL_SUCCESS) {
			/* return -1 on failure */
			DBPRINT (true, "ia32_context_get failed\n");
			retval = -1;
			goto out;
		}
		m_clone_params = params;
		retval = 1;
	} else {
		/* if the current thread is not in the thread-initialization 
		 * procedure than just return 0 */
		retval = 0;
	}
out:
	DBPRINT (1, "retval=%ld\n", retval);
	return retval;
}


/*****************************************************************************/
/*****************************************************************************/
int32_t HIDDEN after_thread_create (mme_thread_init_flag flag, void *clone_params)
{
	ia32el_retval ia32el_ret;
	volatile int32_t retval = 0;
	int32_t status, ret;
	ia32el_context32 ctx32;
	int64_t child_pid;

	DBPRINT (1, "flag=%d clone_params=%p\n", flag, clone_params);

	switch (flag) {
		case MME_THREAD_INIT_FAIL:
			ia32el_ret = g_ia32el_exports.
				ia32el_complete_parent_thread_create (clone_params, -1, &child_pid);
			if (ia32el_ret != IA32EL_SUCCESS) {
				DBPRINT (true, "ia32el_complete_parent_thread_create failed\n");
				retval = 1;
				break;
			}
			ia32el_ret = g_ia32el_exports.ia32el_run (&m_parent_ia32_ctx_at_clone);
			if (ia32el_ret != IA32EL_SUCCESS) {
				DBPRINT (true, "ia32el_run failed in parent\n");
				retval = 1;
			}
			break;
	case MME_THREAD_INIT_CHILD:
			retval = pthread_mutex_lock (&m_child_init_mutex);
			if (retval != 0) {
				DBPRINT (true, "pthread_mutex_lock failed\n");
				retval = -1;
				break;
			}
			status = sigsetjmp (g_thread_func_jmpbuf, 1);
			/* the child's thread func will trap to btlib, which will
			 * call mme_upcall with code THREAD_FUNC, which will make
			 * it longjmp back here */
			if (status == 0) {
				ia32el_ret = g_ia32el_exports.
					ia32el_complete_child_init (clone_params);
				if (ia32el_ret != IA32EL_SUCCESS) {
					DBPRINT (true, "ia32el_complete_child_init failed\n");
					retval = 1;
				}
			} else {
				retval = pthread_mutex_unlock (&m_child_init_mutex);
				if (retval != 0) {
					DBPRINT (true, "pthread_mutex_unlock failed\n");
					retval = -1;
					break;
				}
				/* init thread's map of local references */
				retval = jni32_init_local_refmap ();
				if (retval < 0) {
					DBPRINT (true, "jni32_init_local_refmap failed\n");
				}
			}
			break;
		case MME_THREAD_INIT_PARENT:
			ia32el_ret = g_ia32el_exports.
				ia32el_complete_parent_thread_create (clone_params, 1, &child_pid);
			if (ia32el_ret != IA32EL_SUCCESS) {
				DBPRINT (true, "ia32el_complete_parent_thread_create failed\n");
				retval = 1;
				break;
			}
			/* this is a JVM initiated thread and we should make eax to be
			 * child pid which supposed to be returned by ia32 sys_clone()
			 */
			assert(IS_32BIT_VALUE(child_pid));
			assert(child_pid > 0);
			m_parent_ia32_ctx_at_clone.eax = TRUNCATE_PTR(child_pid);

			ia32el_ret = g_ia32el_exports.ia32el_run (&m_parent_ia32_ctx_at_clone);
			if (ia32el_ret != IA32EL_SUCCESS) {
				DBPRINT (true, "ia32el_run failed in parent\n");
				retval = 1;
				break;
			}
			/* pthread_create 32 returned. check its return value
			 * and return error indication in case it failed */
			ia32el_ret = g_ia32el_exports.ia32el_ia32_context_get (&ctx32);
			if (ia32el_ret != IA32EL_SUCCESS) {
				DBPRINT (true, "ia32el_ia32_context_get failed\n");
				retval = 1;
				break;
			}
			DBPRINT (1, "eax=%x<%d>\n", ctx32.eax, ctx32.eax);
			if (ctx32.eax != 0) {
				DBPRINT (true, "32 bit pthread_create failed\n");
				retval = -1;
			}
			break;
		default:
			assert ( ! "unknown flag" );
			retval = 1;
	}

	if (flag != MME_THREAD_INIT_CHILD) {
		/* resources are reclaimed only in parent's context */
		free32 (m_tid32);
		m_tid32 = 0;
		m_thread_in_init = 0;
		ret = pthread_mutex_unlock (&m_parent_init_mutex);
		if (ret != 0) {
			DBPRINT (true, "pthread_mutex_unlock failed\n");
			retval = -1;
		}
	}

	DBPRINT (1, "retval=%d\n", retval);
	return retval;
}

int32_t HIDDEN
internal_thread_term ()
{
	int32_t retval;
	uint64_t downcall_retval;
	ia32el_retval ia32el_ret;

	retval = jni32_uninit_local_refmap ();
	if (retval != 0) {
		DBPRINT (true, "jni32_uninit_local_refmap failed\n");
		MME_THROW_EXCEPTION ("jni32_uninit_local_refmap failed");
	}
	retval = jni32_jnienv_destroy ();
	if (retval != 0) {
		DBPRINT (true, "jni32_jnienv_destroy failed\n");
		MME_THROW_EXCEPTION ("jni32_jnienv_destroy failed");
	}

	retval = do_downcallV (MME_JAVA_DATA_TYPE_VOID, &downcall_retval,
			I2P (g_ia32_runtime.pthread_exit),
			"I", NULL);
	if (retval != 0) {
		DBPRINT(true, "pthread_exit32 failed\n");
	}

	ia32el_ret = g_ia32el_exports.ia32el_thread_term ();
	if (ia32el_ret != IA32EL_SUCCESS) {
		DBPRINT (true, "ia32el_thread_term failed\n");
		MME_THROW_EXCEPTION ("ia32el_thread_term failed");
	}

	return retval;
}

