/*
 * ia32_machine.c
 *
 * wrapper functions for syscalls that deal with machine. 
 * common to both internal and external wrappers.
 * 
*/
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 * Copyright (C) 2005       Kevin Tian <kevin.tian@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include <ia32x_utils.h>

#include "ia32ss_common.h"
#include "user32.h"

int __sysctl(struct __sysctl_args *args);

long HIDDEN
sys32_sysinfo (__u32 info)
{
	struct sysinfo info64;
	struct sysinfo32 info32;
	int retval;

	if ((retval = SYSCALL(sysinfo, (&info64))) < 0)
		return retval;

	/* Check to see if any memory value is too large, or scale down if needed */
	if ((info64.totalram >> 32) || (info64.totalswap >> 32)) {
		int bitcount = 0;
		__u32 page_size = SYSCALL (getpagesize, ());

		while (info64.mem_unit < page_size) {
			info64.mem_unit <<= 1;
			bitcount++;
		}

		info64.totalram >>= bitcount;
		info64.freeram >>= bitcount;
		info64.sharedram >>= bitcount;
		info64.bufferram >>= bitcount;
		info64.totalswap >>= bitcount;
		info64.freeswap >>= bitcount;
		info64.totalhigh >>= bitcount;
		info64.freehigh >>= bitcount;
	}

	info32.uptime = (__s32)info64.uptime;
	info32.loads[0] = (__u32)info64.loads[0];
	info32.loads[1] = (__u32)info64.loads[1];
	info32.loads[2] = (__u32)info64.loads[2];
	info32.totalram = (__u32)info64.totalram;
	info32.freeram = (__u32)info64.freeram;
	info32.sharedram = (__u32)info64.sharedram;
	info32.bufferram = (__u32)info64.bufferram;
	info32.totalswap = (__u32)info64.totalswap;
	info32.freeswap = (__u32)info64.freeswap;
	info32.procs = info64.procs;
	info32.totalhigh = (__u32)info64.totalhigh;
	info32.freehigh = (__u32)info64.freehigh;
	info32.mem_unit = info64.mem_unit;

	if (g_safe_memcpy(I2P(info), &info32, sizeof(info32)) < 0)
		retval = -EFAULT;

	return retval;
}

long HIDDEN 
sys32_newuname (__u32 name)
{
	struct utsname uname;
	int retval;

	retval = SYSCALL (newuname, (&uname));

	if (!retval) {
		bt_memcpy(uname.machine, "i686\0\0\0", 8);
		if (g_safe_memcpy(I2P(name), &uname, sizeof(struct utsname)) < 0)
			retval = -EFAULT;
	}
	return retval;	
}

long HIDDEN
sys32_sched_getparam(__u32 pid, struct sched_param *param)
{
	long retval;
	struct sched_param buf;

	if (!param) {
		retval = -EINVAL;
		goto out;
	}
	retval = SYSCALL (sched_getparam, (pid, &buf));
	if (!IS_SYS_ERROR(retval)) {
		if (g_safe_memcpy(param, &buf, sizeof(struct sched_param)) < 0) {
			retval = -EFAULT;
		}
	}

out:
	return retval;
}

long HIDDEN
sys32_capget(cap_user_header_t head, cap_user_data_t data) 
{
	long retval;
	struct __user_cap_header_struct local_head;
	struct __user_cap_data_struct local_data;

	assert (sizeof (local_head) == sizeof (*head));
	assert (sizeof (local_data) == sizeof (*data));

	if (g_safe_memcpy(&local_head, head, sizeof(struct __user_cap_header_struct)) < 0) {
		return -EFAULT;
	}

	DBPRINT (0, "local version=0x%x pid=%d\n", local_head.version, local_head.pid);
	DBPRINT (0, "head  %p version=0x%x pid=%d\n", head, head->version, head->pid);
	if (local_head.version != _LINUX_CAPABILITY_VERSION) {
		assert(sizeof(local_head.version) == sizeof(__u32));
		assert(sizeof(head->version) == sizeof(__u32));
	
		local_head.version = _LINUX_CAPABILITY_VERSION;
		if (g_safe_memcpy(&(head->version), &(local_head.version), sizeof(local_head.version)) < 0) {
			return -EFAULT;
		}

		return -EINVAL;
	}

	retval = SYSCALL(capget, (&local_head, &local_data)); 
	if (!IS_SYS_ERROR(retval)) {
		if (g_safe_memcpy(data, &local_data, sizeof(struct __user_cap_data_struct)) < 0) {
			return -EFAULT;
		}
	}

	return retval;
}

long HIDDEN
sys32_capset(cap_user_header_t head, const cap_user_data_t data) 
{
	long retval;
	struct __user_cap_header_struct local_head;

	assert (sizeof (local_head) == sizeof (*head));

	if (g_safe_memcpy(&local_head, head, sizeof(struct __user_cap_header_struct)) < 0) {
		return -EFAULT;
	}
	DBPRINT (0, "local version=0x%x pid=%d\n", local_head.version, local_head.pid);
	DBPRINT (0, "head  %p version=0x%x pid=%d\n", head, head->version, head->pid);
	if (local_head.version != _LINUX_CAPABILITY_VERSION) {
		assert(sizeof(local_head.version) == sizeof(__u32));
		assert(sizeof(head->version) == sizeof(__u32));
	
		local_head.version = _LINUX_CAPABILITY_VERSION;
		if (g_safe_memcpy(&(head->version), &(local_head.version), sizeof(local_head.version)) < 0) {
			return -EFAULT;
		}
		return -EINVAL;
	}

	retval = SYSCALL(capset, (&local_head, data)); 

	return retval;
}

long HIDDEN 
sys32_getcpu(__u32 cpup, __u32 nodep, __u32 unused UNUSED)
{
	unsigned c, n;
	long retval;

	/* the third argument is not in use*/
	retval = SYSCALL(getcpu, (&c, &n, NULL));

	if (!IS_SYS_ERROR(retval)){
		if (g_safe_memcpy(I2P(cpup), &c, sizeof(unsigned)) < 0 || 
			g_safe_memcpy(I2P(nodep), &n, sizeof(unsigned)) < 0) {
			
			return -EFAULT;
		}
	}

	return retval;
}


