/*
 * user32.c
 *
 * misc functions needed by the external wrapper
 *
 */
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include <ia32x_utils.h>

#include "user32.h"
#include "ia32toia64.h"
#include "list.h"

#if DEBUG
long HIDDEN g_initialized = 0;
#endif // DEBUG

__thread long long_syscall_flag  __attribute__ ((tls_model ("global-dynamic"))) = 0l;
__thread long thread_personality  __attribute__ ((tls_model ("global-dynamic"))) = 0l;

/*****************************************************************************/
/* sync objects */
/*****************************************************************************/

HIDDEN mutex_handle g_ctxmap_sync = 0;

/*****************************************************************************/
/* let the library do the before/after fork things */
/*****************************************************************************/
long iti_before_fork ()
{
	int ret;
	/* no two locks are allowed to be taken together, 
	 * so the order is not important. */
	ret = lock_mregion_manager(MREGION_MANAGER_IA32);
	if (IS_SYS_ERROR (ret)) {
		goto out;
	}
	ret = mutex_lock (g_ctxmap_sync);
out:
	DBPRINT (IS_SYS_ERROR(ret), "mutx_lock failed:%d\n", ret);
	return ret;
}

long iti_after_fork (long is_parent)
{
	int ret;

	if (is_parent) {
		ret = mutex_unlock (g_ctxmap_sync);
		assert (ret == 0);
	} else {
		ret = mutex_init (g_ctxmap_sync);	
		assert (ret == 0);
	}
	init_mregion_manager_after_fork(MREGION_MANAGER_IA32, is_parent);

	return ret;
}


/*****************************************************************************/
/* get memory permissions for the given address
 * input:  addr - address to query
 *         prot - will hold the access permissions
 *         start - will hold the base address of the region
 *         len - will hold the size of the region
 * output: 1 if the addr is within a mapped memory area, 0 otherwise
 */
/*****************************************************************************/
long iti_get_memory_access_permissions (void *addr, unsigned *prot, 
                                        void **start, unsigned long *len)
{
	long ret_val;

	lock_mregion_manager(MREGION_MANAGER_IA32);
	ret_val = get_mregion_prot (MREGION_MANAGER_IA32, 
				   (unsigned long)addr, prot,
				   (unsigned long *)start, len, NULL);
	unlock_mregion_manager(MREGION_MANAGER_IA32);
	return (ret_val);
}
/*****************************************************************************/
/* initialization for the external wrapper */
/*****************************************************************************/

long iti_init (int page_size, long (*p_safe_memcpy)(void *, void *, unsigned long))
{
	long retval = 0;
	extern void before_break(void);

#if DEBUG
	extern struct {long a, b;} ia32_2_ia64[];
	long j;

	/* ensure that CONVERS_TAB_MAX_INDEX is defined correctly */
	for (j = 0; ia32_2_ia64[j].a != -1; j ++);
	assert (j - 1 == CONVERS_TAB_MAX_INDEX);
	
#endif // DEBUG
	/* init the sync objects */
	g_ctxmap_sync = mutex_create ();
	if (g_ctxmap_sync == NULL) {
		retval = -ENOMEM;
		goto out;
	}
	retval = mutex_init (g_ctxmap_sync);
	if (retval != 0) {
		goto out;
	}



	/* init the memory module */
	retval = init_mem_module (page_size, p_safe_memcpy);

	DBCODE (TRUE, g_initialized = 1);
out:

	return retval;
}

void iti_get_ia32toia64_version (char *version_str)
{
	bt_snprintf(version_str, MAX_UNIT_VER_STR_LEN, "%s", BUILD_NO);
	return;
}

/*****************************************************************************/
/*****************************************************************************/
long iti_long_syscall_interface (long **long_syscall_flag_p, 
                                 unsigned long *before_break_inst_p)
{
	extern void L_bbi (void);
	unsigned long * plabel = (unsigned long*)L_bbi;

	if (long_syscall_flag_p != NULL) {
		*long_syscall_flag_p = &long_syscall_flag;
	}	
	if (before_break_inst_p != NULL) {
		*before_break_inst_p = PLABEL_FUNC(plabel);
	}
	
	DBPRINT (0, "long_syscall_flag_p=%p before_break_inst_p=%lx\n", 
		 long_syscall_flag_p ? *long_syscall_flag_p : NULL, 
		 before_break_inst_p ? *before_break_inst_p : 0ul);
	return 0;
}

/*****************************************************************************/
/*****************************************************************************/
long iti_personality_interface (long **personality_p)
{
	if (personality_p != NULL) {
		*personality_p = &thread_personality;
	}
	
	return 0;
}

/*****************************************************************************/
/* this function is used when creating core files                            */
/* it enumerates over all mapped memory regions and write relevant info      */
/* to a buffer for later use in core_file                                    */
/* input - buffer for writing                                                */
/*         maximum memory regions that can be written                        */
/* output - number of memory regions that were written, or -1 if number      */
/*          of memory regions exceeds the maximum allowed                    */
/*****************************************************************************/

int iti_enumerate_memory_regions (struct memory_region *mem_regions_buffer,
                                  int max_memory_regions)
{
	list_handle region;
	int ret = 0;
	int err = 0;

	err = lock_mregion_manager(MREGION_MANAGER_IA32);
	assert (!IS_SYS_ERROR (err)); 
	
	region = get_first_mregion (MREGION_MANAGER_IA32);
	while (ret < max_memory_regions) {
		if (region == 0) {
			break;
		}

		get_mregion_prot (MREGION_MANAGER_IA32,
				 get_element_key(region), 
				 &mem_regions_buffer[ret].prot, 
				 &mem_regions_buffer[ret].address, 
				 &mem_regions_buffer[ret].len,
				 &mem_regions_buffer[ret].flags);
				 
		ret++;
		region = get_element_next (region);
	}

	if (ret == max_memory_regions) {
		err = unlock_mregion_manager(MREGION_MANAGER_IA32);
		assert (!IS_SYS_ERROR (err));
		return -1;
	}

	err = unlock_mregion_manager(MREGION_MANAGER_IA32);
	assert (!IS_SYS_ERROR (err));

	return ret;
}

/*****************************************************************************/
/* this function is used to retrive the closest (first) memory region        */
/* that is above the given address                                           */
/* input - address                                                           */
/*         pointer to a memory region                                        */
/* output - 0 if no appropriate memory region was found, 1 otherwise.        */
/*          if a memory region was found, the input pointer will be set to it*/
/*****************************************************************************/

long iti_closest_mem_region (unsigned long address, struct memory_region *buf)
{
	list_handle region;
	int err = 0;
	long ret = 0;
	unsigned long start;

	DBPRINT (0, "address=%ld\n", address);
	
	err = lock_mregion_manager(MREGION_MANAGER_IA32);
	assert (!IS_SYS_ERROR (err)); 
	
	region = get_first_mregion (MREGION_MANAGER_IA32);
	while (1) {
		if (region == 0) {
			break;
		}

		start = get_element_key (region);
		if (start > address) {
			ret = get_mregion_prot (
					MREGION_MANAGER_IA32,
					start,
					&buf->prot,
					&buf->address,
					&buf->len,
					&buf->flags);
			break;
		}
		region = get_element_next (region);
	}

	err = unlock_mregion_manager(MREGION_MANAGER_IA32);
	assert (!IS_SYS_ERROR (err));

	return ret;
}
