/*
 * user32_files.c
 *
 * wrapper functions for syscalls that deal with files.
 * used only by external wrapper
 *
 */
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 * Copyright (C) 2005       Kevin Tian <kevin.tian@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include <ia32x_utils.h>

#include "user32.h"

long HIDDEN
sys32_select (int nfds, __u32 readfds, __u32 writefds,
	      __u32 exceptfds, __u32 timeout)
{
	struct timeval32 to32;
	struct timeval to64;
	fd_set rfds, wfds, efds;
	
	int vt, vr, vw, ve, retval;

	DBPRINT (0, "nfds=%d readfds=%p writefds=%p exceptfds=%p timeout=%p\n", 
		 nfds, I2P(readfds), I2P(writefds), I2P(exceptfds), I2P(timeout));

	vt = IS_VALID_PTR(I2P(timeout));
	if (vt) {
		if (g_safe_memcpy((void *)&to32, I2P(timeout), sizeof(to32)) < 0) {
			return -EFAULT;
		}
		to64.tv_sec = to32.tv_sec;
		to64.tv_usec = to32.tv_usec;
	}
	vr = IS_VALID_PTR(I2P(readfds));
	if (vr && g_safe_memcpy(&rfds, I2P(readfds), sizeof(fd_set)) < 0) {
		return -EFAULT;
	}
	vw = IS_VALID_PTR(I2P(writefds));
	if (vw && g_safe_memcpy(&wfds, I2P(writefds), sizeof(fd_set)) < 0) {
		return -EFAULT;
	}
	ve = IS_VALID_PTR(I2P(exceptfds));
	if (ve && g_safe_memcpy(&efds, I2P(exceptfds), sizeof(fd_set)) < 0) {
		return -EFAULT;
	}


	retval = LONG_SYSCALL5 (select, nfds, 
			vr ? &rfds : NULL, 
			vw ? &wfds : NULL, 
			ve ? &efds : NULL,
			VALID_PTR(I2P(timeout), &to64));
	
	if (retval >= 0) {
		if (vt) {
			to32.tv_sec = to64.tv_sec;
			to32.tv_usec = to64.tv_usec;

			if (g_safe_memcpy(I2P(timeout), (void *)&to32, sizeof(to32)) < 0){
				return -EFAULT;
			}
		}

		if ( (vr && g_safe_memcpy(I2P(readfds), &rfds, sizeof(fd_set)) < 0) ||
		     (vw && g_safe_memcpy(I2P(writefds), &wfds, sizeof(fd_set)) < 0) ||
		     (ve && g_safe_memcpy(I2P(exceptfds), &efds, sizeof(fd_set)) < 0)) {
			return -EFAULT;
		}
	}

	return retval;
}

long HIDDEN
sys32_pselect (int nfds, __u32 readfds, __u32 writefds,
	      __u32 exceptfds, __u32 timeout)
{
	struct timespec32 ts32;
	struct timeval tv64;
	fd_set rfds, wfds, efds;
	
	int vt, vr, vw, ve, retval;

	DBPRINT (0, "nfds=%d readfds=%p writefds=%p exceptfds=%p timeout=%p\n", 
		 nfds, I2P(readfds), I2P(writefds), I2P(exceptfds), I2P(timeout));

	vt = IS_VALID_PTR(I2P(timeout));
	if (vt) {
		if (g_safe_memcpy((void *)&ts32, I2P(timeout), sizeof(ts32)) < 0) {
			return -EFAULT;
		}
		tv64.tv_sec = ts32.tv_sec;
		tv64.tv_usec = ts32.tv_nsec / 1000 + 1; /* time is inprecise here */
	}
	vr = IS_VALID_PTR(I2P(readfds));
	if (vr && g_safe_memcpy(&rfds, I2P(readfds), sizeof(fd_set)) < 0) {
		return -EFAULT;
	}
	vw = IS_VALID_PTR(I2P(writefds));
	if (vw && g_safe_memcpy(&wfds, I2P(writefds), sizeof(fd_set)) < 0) {
		return -EFAULT;
	}
	ve = IS_VALID_PTR(I2P(exceptfds));
	if (ve && g_safe_memcpy(&efds, I2P(exceptfds), sizeof(fd_set)) < 0) {
		return -EFAULT;
	}

	retval = LONG_SYSCALL5 (select, nfds, readfds, writefds, exceptfds, VALID_PTR(I2P(timeout), &tv64));
	
	if (retval >= 0) {
		if (vt) {
			ts32.tv_sec = tv64.tv_sec;
			ts32.tv_nsec = tv64.tv_usec * 1000;

			if (g_safe_memcpy(I2P(timeout), (void *)&ts32, sizeof(ts32)) < 0){
				return -EFAULT;
			}
		}

		if ( (vr && g_safe_memcpy(I2P(readfds), &rfds, sizeof(fd_set)) < 0) ||
		     (vw && g_safe_memcpy(I2P(writefds), &wfds, sizeof(fd_set)) < 0) ||
		     (ve && g_safe_memcpy(I2P(exceptfds), &efds, sizeof(fd_set)) < 0)) {
			return -EFAULT;
		}
	}

	return retval;
}

#define OPEN_MAX 256
long HIDDEN
sys32_poll(__u32 fds, __u32 nfds, __u32 timeout)
{
	struct pollfd buf[OPEN_MAX];
	long retval;

	DBPRINT(0, "fds=%lx, nfds=%d\n", fds, nfds);
	if (nfds > OPEN_MAX) {
		return -EINVAL;
	}
	if (g_safe_memcpy(buf, I2P(fds), sizeof(struct pollfd) * nfds) < 0) {
		return -EFAULT;
	}
	retval  = LONG_SYSCALL3(poll, buf, nfds, timeout);

	if (!IS_SYS_ERROR(retval) && g_safe_memcpy(I2P(fds), buf, sizeof(struct pollfd) * nfds) < 0) {
		return -EFAULT;
	}

	return retval;
}

/* ppoll is implemented by poll, the function is called by btlib_syscall_with_sigmask */
long HIDDEN
sys32_ppoll (__u32 fds, __u32 nfds, __u32 timeout)
{
	struct timespec32 ts32;
	int to= -1;
	long v, retval;

	DBPRINT (0, "fds=%p, nfds=%d timeout=%p\n", I2P(fds), nfds, I2P(timeout));

	v = IS_VALID_PTR(I2P(timeout));
	if (v) {
		if (g_safe_memcpy((void *)&ts32, I2P(timeout), sizeof(ts32)) < 0) {
			return -EFAULT;
		}
		to = ts32.tv_sec + ts32.tv_nsec / 1000000000; /* time is inprecise here */
	}

	retval = sys32_poll(fds, nfds, to);

	return retval;
}

long HIDDEN
sys32_llseek(int fd, __u32 offset_high, __u32 offset_low,
	     __u32 result, __u32 origin)
{
	__s64 offset;
	__s64 retval;

	offset = ((__s64)offset_high << 32) | offset_low;
	DBPRINT (0, "fd=%d offset=%lx origin=%d\n", fd, offset, origin);

	retval = SYSCALL (lseek, (fd, offset, origin));
	if (retval < 0) {
		return (int)retval;
	}

	/* It's loff_t(64bit) type */
	if (g_safe_memcpy (I2P(result), &retval, sizeof(retval)))
		return -EFAULT;
	return 0;
}

long HIDDEN
sys32_sendfile64(int out_fd, int in_fd, __u32 offset, __u32 count)
{
	__off64_t buf;
	__off64_t *ptr = (__off64_t*)I2P(offset);
	__off64_t *offset_ptr = ptr;
	long retval;

	if (IS_VALID_PTR(ptr)) {
		if (g_safe_memcpy(&buf, ptr, sizeof(__off64_t)) < 0) {
			return -EFAULT;
		}
		offset_ptr = &buf;
	}
	/* offset is allowed to be NULL, which means sending from current offset recorded in fd */
	retval = SYSCALL (sendfile, (out_fd, in_fd, offset_ptr, count));

	if (!IS_SYS_ERROR(retval) && IS_VALID_PTR(ptr)) {
		if (g_safe_memcpy(ptr, &buf, sizeof(__off64_t)) < 0) {
			return -EFAULT;
		}
	}

	return retval;
}
