/* 
 * user32_process.c
 *
 * wrapper functions for process related syscalls.
 * used by the external wrapper only
 *
 */
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include <ia32x_utils.h>
#include "user32.h"
#if NOKERNELALTROOT
#include "user32_pathwalk.h"
#endif

#ifndef S_IFMT
#define S_IFMT	0170000
#define S_IFREG	0100000
#endif

static int get_nargs (__u32 args);

#ifdef NOKERNELALTROOT

/*****************************************************************************/
/* direct invocation of scripts with 32 bit interps. returns 1 in case the 
 * script has no 32 bit interp, and negative error value in case of failure.
 * in case the script has 32 bit interp the function never returns.*/
/*****************************************************************************/
static long script_execve (char *interp_filename, char *arg,
		char *script_filename, int script_fd, char **argv, char **envp)
{
	/* Should use VLA here too.  */
	char interp_name[sizeof "/emul/ia32-linux" + 128];
	char *args[4];
	long ret = 1, r;
	size_t len;

	bt_memcpy (interp_name, "/emul/ia32-linux", sizeof ("/emul/ia32-linux") - 1);
	/* Caller guarantees interp_filename is at most 127 chars long.  */
	len = bt_strlen (interp_filename);
	if (len == 0 || len >= 128) {
		DBPRINT (1, "interp_name too long: %ld\n", len);
		goto out;
	}
	bt_memcpy (interp_name + sizeof ("/emul/ia32-linux") - 1, interp_filename, len + 1);
	DBPRINT (1, "interp_name='%s'\n", interp_name);
	/* check the interp's exec permissions */
	ret = check_exec_permissions_by_name (interp_name);
	if (ret != 1) {
		DBPRINT (1, "check_exec_permissions for interp failed:%ld\n", ret);
		/* if the interp was not found in emulation directory then let the
		 * system look for it elsewhere */
		if (ret == -ENOENT) {
			ret = 1;
		}
		goto out;
	}
	/* check the script's exec permissions */
	ret = check_exec_permissions_by_fd (script_fd);
	r = SYSCALL (close, (script_fd));
	DBPRINT (IS_SYS_ERROR (r), "close script failed:%ld\n", r); 
	if (ret != 1) {
		DBPRINT (1, "check_exec_permissions for script failed:%ld\n", ret);
		goto out;
	}
	/* copy the interp name to argv of the new exec */
	if (argv == NULL || *argv == NULL) {
		int i = 0;
		args[i++] = interp_filename;
		if (arg)
			args[i++] = arg;
		args[i++] = script_filename;
		args[i++] = NULL;
		argv = args;
	} else {
		*argv = script_filename;
		if (arg)
			*--argv = arg;
		*--argv = interp_filename;
	}
	/* issue the exec */
	return SYSCALL (execve, (interp_name, argv, envp));
out:
	(void)/*might be already closed*/SYSCALL (close, (script_fd));
	return ret;
}

/*****************************************************************************/
/* special treatment for scripts - this function checks if the given binary
 * is a script that has 32 bit interp. in such a case it tries to exec it
 * directly. returns 1 if the given file is not a script that has 32 bit interp
 * and negative error value in case of failure. in the case the given script is
 * a script with 32 bit interp the function does not return on success.*/
/*****************************************************************************/
static long execve_altroot_script (char *filename, char **argv, char **envp)
{
	long ret = 1;
	char buf[128]; /* Kernel's BINPRM_BUF_SIZE */
	size_t len;
	int fd = pws_open (filename, O_RDONLY, 0);

	if (fd < 0)
		return 1;

	len = SYSCALL (pread, (fd, buf, sizeof (buf) - 1, (off_t)0));
	if (len > 2 && buf[0] == '#' && buf[1] == '!') {
		char *p, *interp = buf + 2, *arg = NULL;

		buf[len] = '\0';
		while (*interp == ' ' || *interp == '\t')
			interp++;
		if (*interp == '\0') {
			ret = -ENOEXEC;
			goto out;
		}
		p = interp;
		while (*p && *p != ' ' && *p != '\t' && *p != '\n')
			p++;
		if (*p != '\0' && *p != '\n') {
			*p++ = '\0';
			while (*p == ' ' || *p == '\t')
				p++;
			if (*p != '\0' && *p != '\n') {
				char *argend;

				arg = p++;
				/* Kernel passes at most one argument.
				   If there are multiple words, it
				   strips trailing whitespace and passes
				   everything between start of 1st word
				   and end of last word as first
				   argument.  */
				for (argend = NULL; *p != '\0' && *p != '\n'; p++)
					if (*p != ' ' && *p != '\t')
						argend = NULL;
					else if (argend == NULL)
						argend = p;
				if (argend == NULL)
					argend = p;
				*argend = '\0';
			}
		} else
			*p = '\0';
		return script_execve (interp, arg, filename, fd, argv, envp);
	}
out:
	(void)SYSCALL (close, (fd));
	return ret;
}
#endif

/*****************************************************************************/
/* wrapper for 32 bit sys_execve:
 * 1) convert the 4 bytes pointer arrays to 8 bytes pointer arrays
 * 2) special handling for scripts with 32 bit interps */
/*****************************************************************************/
long HIDDEN sys32_execve (__u32 filename, __u32 argv, __u32 envp)
{
	char      **ia64_argv_envp, **ia64_argv, **ia64_envp;
	long      ret;
	int       argc, envc;  /* number of list */
	int       i;
	unsigned  addr32;
	unsigned  *p;

	DBPRINT (0, "filename=<%s> argv=%p envp=%p\n", 
		 (I2P(filename) != NULL) ? (char *)I2P(filename) : "(nil)", 
		 I2P(argv), I2P(envp));

	if ((argc = get_nargs (argv)) < 0) {
		return argc;
	}

	if ((envc = get_nargs (envp)) < 0) {
		return envc;
	}
	
	if ((unsigned)argc > (MAX_ARG_PAGES * IA32_PAGE_SIZE) / sizeof (__u32) - 1) {
		DBPRINT (FALSE, "Too many arguments\n");
		return -E2BIG;
	}

	if ((unsigned)envc > (MAX_ARG_PAGES * IA32_PAGE_SIZE) / sizeof (__u32) - 1) {
		DBPRINT (FALSE, "Too many environs\n");
		return -E2BIG;
	}

	/* allocate memory for the argument/enviroment list,
	   leave 2 space for NULL, 2 for altroot script.  */
	ia64_argv_envp = bt_malloc ((argc + envc + 4) * sizeof (void *));
	if (ia64_argv_envp == NULL) {
		DBPRINT (FALSE, "bt_malloc failed\n");
		return -ENOMEM;
	}
	ia64_argv = ia64_argv_envp + 2;
	ia64_envp = ia64_argv + argc + 1;

	/* copy argument/environment list */
	p = I2P (argv);
	if (p != NULL) {
		for (i = 0; i <= argc; i++, p++) {
			if (g_safe_memcpy (&addr32, p, sizeof (addr32)) < 0) {
				ret = -EFAULT;
				goto out;
			}
			ia64_argv[i] = I2P (addr32);
#if DEBUG & 0
			if (ia64_argv[i]) {
				DBPRINT (TRUE, "\targv[%d]='%s'\n", i, ia64_argv[i]);
			}
#endif
		}
	}
	p = I2P (envp);
	if (p != NULL) {
		for (i = 0; i <= envc; i++, p++) {
			if (g_safe_memcpy (&addr32, p, sizeof (addr32)) < 0) {
				ret = -EFAULT;
				goto out;
			}
			ia64_envp[i] = I2P (addr32);
#if DEBUG & 0
			if (ia64_envp[i]) {
				DBPRINT (TRUE, "\tenvp[%d]='%s'\n", i, ia64_envp[i]);
			}
#endif
		}
	}

#if NOKERNELALTROOT
	/* First see if we aren't trying to execute a script.
	   If so and the interpreter is available in altroot, execute
	   using that interpreter instead.  */
	struct stat s;
	ret = pws_stat ((char *)I2P(filename), &s);
	if (ret >= 0 && (s.st_mode & S_IFMT) == S_IFREG && (s.st_mode & 0111)) {
		ret = execve_altroot_script ((char *)I2P(filename),
					     VALID_PTR(I2P(argv), ia64_argv),
					     VALID_PTR(I2P(envp), ia64_envp));
	} else
		ret = 1;
	if (ret == 1)
		/* use the pathwalk sycall wrapper in user32_pathwalk.c */
		ret = pws_execve ((char *)I2P(filename), 
				  VALID_PTR(I2P(argv), ia64_argv),
				  VALID_PTR(I2P(envp), ia64_envp));
#else
	ret = SYSCALL (execve, ((char *)I2P(filename), 
				VALID_PTR(I2P(argv), ia64_argv),
				VALID_PTR(I2P(envp), ia64_envp)));
#endif

out:
	/* if we got here then the syscall failed */
	bt_free (ia64_argv_envp);

	DBPRINT (0, "ret=%ld\n", ret);
	return ret;
}

/*****************************************************************************/
/* wrapper for 32 bit sys_getpriority */
/*****************************************************************************/
long HIDDEN sys32_getpriority (int who, int which)
{
	long err;
	long ret;

	/* getpriority returns a value in the range -20 to 20 */
	ret = SYSCALL (getpriority, (who, which));
	__asm__ __volatile__ ("mov %0=r10;;\n" : "=r"(err)); 
	/* in IA-32 - the syscall returns the priority with offset 20 */
	if (err == 0) {
		ret = 20 - ret;
	}

	return ret;
}

/*****************************************************************************/
/* wrapper for 32 bit sys_nice */
/*****************************************************************************/
long HIDDEN sys32_nice(int increment)
{
	long newprio;
	long oldprio;
	long retval;
	long err;
	int tid = SYSCALL (gettid, ());

	oldprio = SYSCALL (getpriority, (PRIO_PROCESS, tid));
	__asm__ __volatile__ ("mov %0=r10;;\n" : "=r"(err)); 
	if (err < 0) {
		return oldprio;
	}

	newprio = oldprio + increment;
	newprio = MAX (-20, newprio);
	newprio = MIN (19, newprio);

	retval = SYSCALL (setpriority, (PRIO_PROCESS, tid, newprio));
	
	return retval;
}

/*****************************************************************************/
/* calculate number of arguments/enviroment variables */
/*****************************************************************************/
static int get_nargs (__u32 args)
{
	int      n = 0;
	unsigned arg;		/* pointer to argument */
	unsigned *p = I2P(args);
	
	if (p == NULL)
		return 0;

	for (;;) {
		if (g_safe_memcpy (&arg, p, sizeof (arg)) < 0) {
			return -EFAULT;	
		}
		if (arg == 0) {
			break;
		}
		n++;
		p++;
	}
	return n;
}

