/*
 * btlib_files.c
 *
 * wrapper functions for syscalls that deal with files. 
 * common to both internal and external wrappers.
 *
 */
 /*
 *
 * Copyright (C) 2005       Xiaolan Huang <xiaolan.huang@intel.com>
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 * Copyright (C) 2005       Gordon Jin <gordon.jin@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include <ia32x_utils.h>
#include <ia32toia64.h>
#include "btlib.h"

/* moved from ia32_files.c */
inline int HIDDEN iovec_conv(struct iovec32 *iov32, struct iovec *iov);


inline int HIDDEN
iovec_conv(struct iovec32 *iov32, struct iovec *iov)
{
	struct iovec32 iov_tmp;

	if (safe_memcpy (&iov_tmp, iov32, sizeof (iov_tmp)) < 0) {
		return -EFAULT;
	}
	DBPRINT (0, "base = %p len=%x\n", U2P(iov_tmp.iov_base), iov_tmp.iov_len);
	iov->iov_base = U2P(iov_tmp.iov_base);
	iov->iov_len = iov_tmp.iov_len;
	return 0;
}


long HIDDEN
btlib_writev (BTGENERIC_IA32_CONTEXT * bc32)
{
	int fd, count;
	struct iovec *iov;
	struct iovec32 * vector;
	int ret;

	fd = (int)IA32_SYSCALL_PARAM1(bc32);
	vector = (struct iovec32*)U2P(IA32_SYSCALL_PARAM2(bc32));
	count = (int)IA32_SYSCALL_PARAM3(bc32);

	DBPRINT (0, "fd=%d vector=%p count=%d\n", fd, vector, count);
	
	if ((int)count < 0) {
		return -EINVAL;
	}
	
	if (count > UIO_MAXIOV) {
		return -EINVAL;
	}
	
	iov = (struct iovec*)bt_malloc(count * sizeof(struct iovec));
	if (iov || count == 0) {
		int i;
		__s32 len = 0, len_orig;
		      
		for (i = 0; i < count; i++) {
			ret = iovec_conv(&vector[i], &iov[i]);
			if (ret < 0) {
				bt_free (iov);
				return ret;
			}

			/* if the sum of iov_len overflows __s32, EINVAL should be returned */
			len_orig = len;
			len += iov[i].iov_len;
			if (len < len_orig) {
				bt_free (iov);
				return -EINVAL;
			}
		}
		ret = LONG_SYSCALL3 (writev, fd, iov, count);
		bt_free(iov);
		return ret;
	}
	return -ENOMEM;
}

long 
btlib_readv (BTGENERIC_IA32_CONTEXT * bc32)
{
	int fd, count;
	struct iovec *iov;
	struct iovec32 * vector;
	int ret;

	fd = (int)IA32_SYSCALL_PARAM1(bc32);
	vector = (struct iovec32*)U2P(IA32_SYSCALL_PARAM2(bc32));
	count = (int)IA32_SYSCALL_PARAM3(bc32);

	DBPRINT (0, "fd=%d vector=%p count=%d\n", fd, vector, count);
	if ((int)count < 0) {
		return -EINVAL;
	}

	if (count > UIO_MAXIOV) {
		return -EINVAL;
	}
	
	iov = (struct iovec *)bt_malloc(count * sizeof(struct iovec));
	if (iov || count == 0) {
		int i;
		__s32 len = 0, len_orig;
		      
		for (i = 0; i < count; i++) {
			ret = iovec_conv(&(vector)[i], &iov[i]);
			if (ret < 0) {
				bt_free (iov);
				return ret;
			}

			/* if the sum of iov_len overflows __s32, EINVAL should be returned */
			len_orig = len;
			len += iov[i].iov_len;
			if (len < len_orig) {
				bt_free (iov);
				return -EINVAL;
			}
			if (g_has_smc) btlib_disable_smc_opt(iov[i].iov_base, iov[i].iov_len);
		}
	
		ret = LONG_SYSCALL3 (readv, fd, iov, count);
		bt_free(iov);
		return ret;
	}
	return -ENOMEM;
}

long
btlib_vmsplice(BTGENERIC_IA32_CONTEXT * bc32)
{
	int fd, count;
	struct iovec *iov;
	struct iovec32 * vector;
	long ret;

	fd = (int)IA32_SYSCALL_PARAM1(bc32);
	vector = (struct iovec32*)U2P(IA32_SYSCALL_PARAM2(bc32));
	count = (int)IA32_SYSCALL_PARAM3(bc32);

	if (count < 0) {
		return -EINVAL;
	}

	if (count > UIO_MAXIOV) {
		return -EINVAL;
	}

	iov = (struct iovec *)bt_malloc(count * sizeof(struct iovec));
	if (iov || count == 0) {
		int i;
		__s32 len = 0, len_orig;

		for (i = 0; i < count; i++) {
			ret = iovec_conv(&(vector)[i], &iov[i]);
			if (ret < 0) {
				bt_free (iov);
				return ret;
			}

			/* if the sum of iov_len overflows __s32, EINVAL should be returned */
			len_orig = len;
			len += iov[i].iov_len;	
			if (len < len_orig) {
				bt_free (iov);
				return -EINVAL;
			}
			if (g_has_smc) btlib_disable_smc_opt(iov[i].iov_base, iov[i].iov_len);
		}

		ret = SYSCALL (vmsplice, (fd, iov, (unsigned long)count, IA32_SYSCALL_PARAM4(bc32)));
		bt_free(iov);
		return ret;
	}
	return -ENOMEM;
}
