/* 
 * btlib_memory.c
 *
 * 
 *
 */
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include <ia32x_utils.h>

#include <ia32toia64.h>

#include "btlib.h"
#include "btlib_sync.h"

#define MAX_PROC_MOUNTS_SIZE (16 * 1024)
#define MOUNT_LEN            (1024)

/* function prototypes */
static long set_btrans_permissions (unsigned long prot);
static long mmap32 (unsigned int addr, unsigned int len, unsigned int prot, 
                    unsigned int flags, unsigned int fd, unsigned int pgoff);
static unsigned set_linux_permissions (unsigned prot);
static long munmap32 (unsigned addr, unsigned len);
static int is_on_hugetlb_fs (int fd);
static unsigned long check_hugetlbfs (void);
static void generate_ia32_proc_maps_file();
inline int ultostrx(unsigned long num, char* buff);
inline long do_native_read(int fd, char* buff, int size);

/* fd_map list, used for hugetlb judgement */
struct fd_status_map m_huge_fds;
unsigned long m_hugetlbfs_exists = 0;

#define IS_IA32_ADDRESS(addr) ((((unsigned long)(addr)) & (0xFFFFFFFF)) == ((unsigned long)addr))

/*protected by ia32 memory lock*/
struct m_proc_maps_file_desc m_ia32_proc_maps = {NULL, 0, 0};
struct m_proc_maps_file_desc m_ia64_proc_maps = {NULL, 0, 0};

/*****************************************************************************/
/* initialize the memory module. must be called before any
 * exported function is called
 */
/*****************************************************************************/
void
btlib_init_memory_module ()
{
	DBPRINT (btl_prt_mem, "\n");

	/* initialize the m_huge_fds */
	m_hugetlbfs_exists = check_hugetlbfs ();

	m_huge_fds.start_fd = 0;
	m_huge_fds.used_fd = 0;
	bt_memset (m_huge_fds.fd_map, FD_UNKNOWN, (size_t)FD_MAP_SIZE);
	m_huge_fds.next = NULL;

	m_ia32_proc_maps.buff = (char*) SYSCALL(mmap, (BTLIB_MEMORY_BASE, (unsigned long)PAGE_SIZE, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANON, -1, 0l));
	assert(m_ia32_proc_maps.buff >= (char*)(BTLIB_MEMORY_BASE));
	m_ia32_proc_maps.max_size = PAGE_SIZE;
	m_ia32_proc_maps.cur_size = 0;

	m_ia64_proc_maps.buff = (char*) SYSCALL(mmap, (BTLIB_MEMORY_BASE, (unsigned long)PAGE_SIZE, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANON, -1, 0l));
	assert(m_ia64_proc_maps.buff >= (char*)(BTLIB_MEMORY_BASE));
	m_ia64_proc_maps.max_size = PAGE_SIZE;
	m_ia64_proc_maps.cur_size = 0;
}

inline int ultostrx(unsigned long num, char* buff)
{
	int i;
	int n;
	assert(sizeof (buff) >= 8);

	for (i=0; i < 8; i++) {
		n = num % 16;

//		DBPRINT(1, "i=%d num=0x%lx, char=%d\n", i, num, n);

		buff[7-i] = n > 9 ? n - 10 + 'a' : n + '0';
		num = num >> 4;
	}

	DBPRINT(0, "num %x to %c%c%c%c %c%c%c%c\n", num, buff[0], buff[1], buff[2], buff[3], buff[4], buff[5], buff[6], buff[7] );

	return 8;
}

static void generate_ia32_proc_maps_file()
{
	int n = 0;
#define CURCHAR 		(m_ia64_proc_maps.buff[n + i])
#define WRITE_BUFF_START	(&(m_ia32_proc_maps.buff[m_ia32_proc_maps.cur_size]))
#define WRITE_BUFF_PUSH(c)	\
{ \
	m_ia32_proc_maps.buff[m_ia32_proc_maps.cur_size] = (char)(c); \
	m_ia32_proc_maps.cur_size++; \
}

	/**/
	while (n < m_ia64_proc_maps.cur_size) {
		unsigned long start64, end64, offset64;
		unsigned long start32, end32, offset32;
		unsigned long reg_start, reg_size;
		int len;
		int prot64 = 0, prot32;
		int i = 0;
		/*read one line*/

//		DBPRINT (1, "n=%d i=%d\n%s\n", n, i, &CURCHAR);

		/*start address*/
		start64 = bt_axtoul(&CURCHAR, &len);
		i += len; 

		while (CURCHAR == ' ') i++;
		assert (CURCHAR == '-');
		i++;
		while (CURCHAR == ' ') i++;

		/*end address*/
		end64 = bt_axtoul(&CURCHAR, &len);
		i += len; 
		while (CURCHAR == ' ') i++;

		/*check memory address*/
		if (!IS_IA32_ADDRESS(start64)) {
			/*no need to go on*/
			break;
		}

		/*memory prot*/
		if (CURCHAR == 'r' || CURCHAR == 'R') {
			prot64 |= PROT_READ;
		}
		i++;
		if (CURCHAR == 'w' || CURCHAR == 'W') {
			prot64 |= PROT_WRITE;
		}
		i++;
		if (CURCHAR == 'x' || CURCHAR == 'X') {
			prot64 |= PROT_EXEC;
		}
		i++;
		if (CURCHAR == 's' || CURCHAR == 'S') {
			prot64 |= PROT_SEM;
		}
		i++;
		while (CURCHAR == ' ') i++;
		
		/* offset */
		offset64 = bt_axtoul(&CURCHAR, &len);
		i += len; 
		
		/* set n to the start position of the string to be copied */
		n += i;
		i = 0;

		//DBPRINT (1, "%lx - %lx prot=%x offset=%lx\n", start64, end64, prot64, offset64);
		/* find the end of the line*/
		while (CURCHAR != '\n')i++;
		assert(n+i < m_ia64_proc_maps.max_size);
		//replace \n with \0
		CURCHAR='\0'; 
		
		offset32 = offset64;
		start32 = start64;
		/* check the memory region, split the line if necessary */
		do {
			int try = PAGE_SIZE / IA32_PAGE_SIZE - 1;
			long ret;

			//DBPRINT (1, "start32=%lx end32=%lx\n", start32, end32);
			ret = 0;
			ret = get_mregion_prot (MREGION_MANAGER_IA32, start32, (unsigned *)(&prot32), &reg_start, &reg_size, NULL); 
			while (try > 0 && ret == 0 && start32 + IA32_PAGE_SIZE < end64) {
				start32 += IA32_PAGE_SIZE;
				offset32 += IA32_PAGE_SIZE;
				try--;
				ret = get_mregion_prot (MREGION_MANAGER_IA32, start32, (unsigned *)(&prot32), &reg_start, &reg_size, NULL); 
				//DBPRINT(1, "get mergion prot start32=%lx -> regstart =%lx regsize=%lx\n", start32, reg_start, reg_size);
			}
			if (start32 >= end64 || ret == 0) {
				//DBPRINT (1, "the ia64 memoryregion is not mapped for ia32 actually\n");
				break; //next line of ia64 maps file
			}
			//DBPRINT (1, "offset32=%lx\n", offset32);
			assert(reg_start <= start32);
			assert(start32 >= start64 && start32 < end64);

			DBPRINT(0, "end64 = %lx, reg_start=%lx reg_size=%lx\n", end64, reg_start, reg_size);
			end32 = (end64 <= reg_start + reg_size) ? end64 : reg_start + reg_size;

			/**
			 *  write to map32 
			 */
			//DBPRINT (1, "to write %lx-%lx prot=%x\n", start32, end32, prot32);
			//DBPRINT (1, "ia32 cursize=%x maxsizd=%x\n", m_ia32_proc_maps.cur_size, m_ia32_proc_maps.max_size);

			/*extend the buff for ia32 if necessary, assume that one line will not takes 1024 bytes*/
			if (m_ia32_proc_maps.max_size - m_ia32_proc_maps.cur_size < 1024) {
				unsigned long address;

				address = SYSCALL(mremap, ((unsigned long)(m_ia32_proc_maps.buff), (unsigned long)(m_ia32_proc_maps.max_size), 
					(unsigned long)(m_ia32_proc_maps.max_size * 2), MREMAP_MAYMOVE, (unsigned long)(m_ia32_proc_maps.buff)));
				assert (!IS_SYS_ERROR(address));
				m_ia32_proc_maps.buff = (char*)(address);
				m_ia32_proc_maps.max_size = m_ia32_proc_maps.max_size * 2;
				
			}

			/* region address */
			m_ia32_proc_maps.cur_size += ultostrx(start32, WRITE_BUFF_START);
			WRITE_BUFF_PUSH('-');
			m_ia32_proc_maps.cur_size += ultostrx(end32, WRITE_BUFF_START);
			WRITE_BUFF_PUSH(' ');

			/* prot */
			WRITE_BUFF_PUSH(((prot32 & PROT_READ) == PROT_READ) ? 'r' : '-');
			WRITE_BUFF_PUSH(((prot32 & PROT_WRITE) == PROT_WRITE) ? 'w' : '-');
			WRITE_BUFF_PUSH(((prot32 & PROT_EXEC) == PROT_EXEC) ? 'x' : '-');
			WRITE_BUFF_PUSH(((prot64 & PROT_SEM) == PROT_SEM) ? 's' : 'p'); /*mem region doesn't save this info*/
			WRITE_BUFF_PUSH(' ');

			/* offset */
			m_ia32_proc_maps.cur_size += ultostrx(offset32, WRITE_BUFF_START);

			/*write rest*/
			assert (i > 0);
			ret = (unsigned long)bt_memcpy(WRITE_BUFF_START, &(m_ia64_proc_maps.buff[n]), (unsigned long)(i));
			assert ((char*)ret == WRITE_BUFF_START);
			m_ia32_proc_maps.cur_size += i;
			WRITE_BUFF_PUSH('\n');

			/* prepare for next splitted line to write*/
			offset32 += end32 - start32;
			start32 = end32;

		}while (end64 > reg_start + reg_size);

		while (CURCHAR == ' ' || CURCHAR == '\0') i++;
//		DBPRINT (1, "current char=%d=%c\n", CURCHAR, CURCHAR);
		n += i;
		i = 0;
	} 

	m_ia32_proc_maps.buff[m_ia32_proc_maps.cur_size] = 0;
	DBPRINT (0, "ia32 maps file size=%d strsize=%d \n%s", m_ia32_proc_maps.cur_size, bt_strlen(m_ia32_proc_maps.buff), m_ia32_proc_maps.buff);
	assert(m_ia32_proc_maps.cur_size > 0);
}

inline long
do_native_read(int fd, char* buff, int size)
{
	off_t orig_offset; 
	ssize_t retlen; 

	do { 
		orig_offset = SYSCALL (lseek, (fd, (off_t)0, SEEK_SET)); 
		assert (!IS_SYS_ERROR(orig_offset)); 
		retlen = SYSCALL(read, (fd, buff, (size_t)size)); 
	} while(retlen == -EINTR); 

	return (long)retlen; 
}

/* the function should be protected by memory lock*/
struct m_proc_maps_file_desc *
btlib_get_proc_maps_content(int fd)
{
	assert (!(m_ia32_proc_maps.cur_size == 0 && !is_memory_maps_changed(MREGION_MANAGER_IA32)));
	if (is_memory_maps_changed(MREGION_MANAGER_IA32)) {
		int ret;

		m_ia32_proc_maps.cur_size = 0;
		/* read the content of ia64 maps file */
		ret = do_native_read(fd, m_ia64_proc_maps.buff, m_ia64_proc_maps.max_size);
		assert (ret >= 0);

		while (ret == m_ia64_proc_maps.max_size) {
			unsigned long address;

			/* the buff is too small for the file*/
			/* extend the buff*/
			address = SYSCALL(mremap, ((unsigned long)(m_ia64_proc_maps.buff), (unsigned long)(m_ia64_proc_maps.max_size), 
				(unsigned long)(m_ia64_proc_maps.max_size * 2), MREMAP_MAYMOVE, (unsigned long)(m_ia64_proc_maps.buff)));
			assert (!IS_SYS_ERROR(address));
			m_ia64_proc_maps.buff = (char*)(address);
			m_ia64_proc_maps.max_size = m_ia64_proc_maps.max_size * 2;

			/* read the content again */
			ret = do_native_read(fd, m_ia64_proc_maps.buff, m_ia64_proc_maps.max_size);
			assert (ret >= 0);

		}

		m_ia64_proc_maps.cur_size = ret;
//		DBCODE(1, m_ia64_proc_maps.buff[m_ia64_proc_maps.cur_size] = 0);

		//DBPRINT (1, "m_ia64_proc_maps buff size =%d, cur size=%d\n", m_ia64_proc_maps.max_size, m_ia64_proc_maps.cur_size);

		generate_ia32_proc_maps_file();
		reset_memory_maps(MREGION_MANAGER_IA32);
	}

	return &m_ia32_proc_maps;
}


/* 
   check if there si a hugetlb fs in the system 
   return value:
        0: hugetlb fs not exists
        1: hugetlb exists
*/
static unsigned long
check_hugetlbfs (void)
{
	/* currently we always conservativley return 1 to indicates hugetlb FS existing */
	return 1;
}

/*****************************************************************************/
/* convert from linux permissions to BTGeneric permissions */
/*****************************************************************************/
static long
set_btrans_permissions (unsigned long prot)
{
	assert ((MEM_READ == PROT_READ) &&
		(MEM_WRITE == PROT_WRITE) && (MEM_EXECUTE == PROT_EXEC));
	return prot;
}

/*****************************************************************************/
/* convert from BTGeneric permissions to linux permissions */
/*****************************************************************************/
static unsigned
set_linux_permissions (unsigned prot)
{
	assert ((MEM_READ == PROT_READ) &&
		(MEM_WRITE == PROT_WRITE) && (MEM_EXECUTE == PROT_EXEC));
	return prot;
}

/*****************************************************************************/
/* get the permissions of the memory at a given address 
 * returns the permssions of the page where the given address falls, or 
 * a negative error code on failure.
 */
/*****************************************************************************/
long
btlib_memory_query_permissions (void *address, void **region_start,
                                unsigned long *region_size)
{
	unsigned perm = 0;
	long ret;

	/* first try to query the wrapper lib */
	ret = iti_get_memory_access_permissions (address, &perm, region_start,
						 region_size);
	if (ret == 0) {

		DBPRINT (btl_prt_mem, "address not mapped\n");

		/* the address was not found in the wrapper lib */

		/* linux maps the first page as ro */
		if (PAGESTART ((unsigned long) address, btl_page_size) != 0ul) {
			perm = PROT_NONE;
		}

		if (region_start != NULL) {
			*region_start = (void *) PAGESTART ((unsigned long) address, 
							    btl_page_size);
		}
		if (region_size != NULL) {
			*region_size = btl_page_size;
		}
	}

	DBPRINT (btl_prt_mem, "perm=%x\n", perm);

	return perm;
}

/*****************************************************************************/
/* do the memory mapping for IA-32 addresses */
/*****************************************************************************/
static long
mmap32 (unsigned int addr, unsigned int len, unsigned int prot, 
        unsigned int flags, unsigned int fd, unsigned int pgoff)
{
	unsigned long retval;

	/* be careful with the following debug-print -
	 * we cannot be sure that mmap_args is a valid pointer 
	 */
	DBPRINT (0, "addr=%x len=%x prot=%x flags=%x fd=%d pgoff=%x\n",
		 addr, len, prot, flags, fd, pgoff);

	/* return -EINVAL when fd is on hugetlbfs */
	if (m_hugetlbfs_exists && !(flags & MAP_ANONYMOUS)) {
		int status = get_fd_hugetlb_status ((int)fd);

		if (status == FD_UNKNOWN) {
			/* get the fd hugetls status */
			status = is_on_hugetlb_fs ((int)fd);
			if (status != FD_UNKNOWN) {
				/* set the fd hugetls status in g_huge_fds[] */
				set_fd_hugetlb_status ((int)fd, status);
			}
		}							
			
		if (status == FD_HUGE) {
			return (unsigned long)-EINVAL;
		}
	}

	/* the memory may be allocated already, and this map will change the permission of it*/
	if ((flags & MAP_FIXED) != 0) {
		long perm = 0;
		int no_of_pages;

		perm = set_btrans_permissions ((unsigned long)prot);
		no_of_pages = (len - 1) / IA32_PAGE_SIZE + 1;

		/* notify ia32exec.bin of the change */
		DBPRINT (btl_prt_mem, "FIXED map no_of_pages=%d prot=%lx\n", no_of_pages, prot);
		BTGENERIC_NOTIFY_CHANGE_PERMISSION_REQUEST (U2P(addr), no_of_pages, perm);
	}
	
	/* do the mapping */
	retval = ILP32_WRAPPER6 (IA32_SYSCALL_MMAP2, addr, len, prot, flags, fd, pgoff);

	DBPRINT (btl_prt_mem, "map32 retval=%lx\n", retval);

	if (!IS_SYS_ERROR (retval)) {
		/* make sure we don't return a high address to the application */
		if (retval >= 0x100000000) {
			DBPRINT (btl_prt_mem, "mmap32:un-mapping\n");
			retval = SYSCALL (munmap, (retval, (unsigned long)len));
			sys_err_chk (retval, "munmap");
			return (unsigned long)-ENOMEM;
		}
		/* if the mapped region is executable then notify btgeneric */ 
		else if ((prot & PROT_EXEC) == PROT_EXEC) {
			long rettmp;
			struct stat statbuf;
			char name[PATH_MAX];
			int namelen = 0;

#if ! RELEASE
			//disable dll reuse on debug/lite version
			if (btl_no_dll_reuse  
#if DEBUG
			   && (!btl_prt_dll)
#endif
			   ) {
				BTGENERIC_FLUSH_IA32_INSTRUCTION_CACHE (
                                        (void *) retval, len, BT_FLUSH_ALLOC);
			}
			else
#endif
			if (((int)fd) > 0) {
				void *ret;

				rettmp = sys_fstat((int)fd, &statbuf);
				assert(!rettmp);

				ret = bt_memcpy (&(name[namelen]), (char*)(&statbuf.st_dev), 
						 sizeof(statbuf.st_dev));
				assert (ret == &(name[namelen]));
				namelen += sizeof(statbuf.st_dev);
				 	
				ret = bt_memcpy (&(name[namelen]), (char*)(&statbuf.st_ino), 
						 sizeof(statbuf.st_ino));
				assert (ret == &(name[namelen]));
				namelen += sizeof(statbuf.st_ino);

				
				/* report dll load */
#if DEBUG			
				// print the info of DLL before report on DEBUG  version 
				if (btl_prt_dll) {
					char buf[256];
					int i;

					#define GET_BYTE(num, i) 	( (num >> 4*i) & 0xF ) 
					#define BYTE_TO_CHAR(num, i) 	( GET_BYTE(num, i) >0x9 ? \
									  GET_BYTE(num, i) + 'a' -10: \
									  GET_BYTE(num, i) + '0' \
									)
					for (i = 0; i < namelen; i++) {
						buf[2 * i] = BYTE_TO_CHAR(name[i], 1);
						buf[2 * i + 1] = BYTE_TO_CHAR(name[i], 0);
					}
					buf[2 * i] = '\0';
				}

				//disable dll for debug version
				if (btl_no_dll_reuse) {
					BTGENERIC_FLUSH_IA32_INSTRUCTION_CACHE (
	                                        (void *) retval, len, BT_FLUSH_ALLOC);
				} else {
					BTGENERIC_REPORT_LOAD ((void *) retval, len, name, namelen);
				}
				
#else				//report DLL load anyway on RELASE/LITE version
				BTGENERIC_REPORT_LOAD ((void *) retval, len, name, namelen);
#endif

			} else if ((flags & MAP_FIXED) == 0){
				DBPRINT (btl_prt_mem, 
					 "report to flush instruction cache %lx %u\n", retval, len);
				BTGENERIC_FLUSH_IA32_INSTRUCTION_CACHE (
					(void *) retval, len, BT_FLUSH_ALLOC);
			}
			
		}


#if SHMEM_ORDER_SOLUTION
		if (flags & MAP_SHARED) {
			apply_shared_memory_ordering_solution ();
		}
#endif
	}

	return retval;
}

/*****************************************************************************/
/* do the memory un-mapping for IA-32 addresses */
/*****************************************************************************/
static long
munmap32 (unsigned addr, unsigned len)
{
	long retval;

	DBPRINT (btl_prt_mem, "addr=0x%x len=0x%x=%d\n", addr, len, len);
	DBPRINT (btl_prt_dll, "\n DLL_UNLOAD 0x%p 0x%x\n", (void*)(U2P(addr)), len);

	if ((unsigned long)addr + (unsigned long) len >= IA32_MAX_USER_ADDR ) {
		return -EINVAL;
	}

	/* notify btgeneric */
#if ! RELEASE
	if (btl_no_dll_reuse) {
		BTGENERIC_FLUSH_IA32_INSTRUCTION_CACHE (U2P (addr), len, BT_FLUSH_FREE);
	} else
#endif
	{
		BTGENERIC_REPORT_UNLOAD( U2P(addr) , len, "NULL");
	}

	/* do the unmapping */
	retval = ILP32_WRAPPER2 (IA32_SYSCALL_MUNMAP, (unsigned long) addr,
				 (unsigned long) len);
	DBPRINT (btl_prt_mem, "retval=%ld\n", retval);
	return retval;
}

/*****************************************************************************/
/* do the memory re-mapping for IA-32 addresses */
/*****************************************************************************/
long
btlib_ia32_mremap (const BTGENERIC_IA32_CONTEXT *bc32)
{
	void *retval;
	unsigned addr, old_len, new_len;
	unsigned flags, new_addr;
	int perm;

	DBPRINT (btl_prt_mem, "\n");

	addr = IA32_SYSCALL_PARAM1 (bc32);
	old_len = IA32_SYSCALL_PARAM2 (bc32);
	new_len = IA32_SYSCALL_PARAM3 (bc32);
	flags = IA32_SYSCALL_PARAM4 (bc32);
	new_addr = IA32_SYSCALL_PARAM5 (bc32);

	perm = btlib_memory_query_permissions (U2P (addr), NULL, NULL);
	if (perm < 0) {
		retval = (void *) (long) perm;
		goto out;
	}
	if (perm & MEM_EXECUTE) {
		BTGENERIC_FLUSH_IA32_INSTRUCTION_CACHE (U2P (addr), old_len,
							BT_FLUSH_FREE);
	}

	DBPRINT (btl_prt_mem, "addr=%x old_len=%d perm=%x\n", addr, old_len, perm);
	DBPRINT (btl_prt_mem, "new_len=%d flags=%x new_addr=%x\n", new_len, flags, 
		 new_addr);

	/* do the remapping */
	retval = (void *) ILP32_WRAPPER5 (IA32_SYSCALL_MREMAP, 
					  (unsigned long) addr,
					  (unsigned long) old_len,
					  (unsigned long) new_len,
					  (unsigned long) flags,
					  (unsigned long) new_addr);

	DBPRINT (btl_prt_mem, "retval=%p\n", retval);

	if (perm & PROT_EXEC) {
		if (IS_SYS_ERROR ((long) retval)) {
			/* undo the report_unload */
			BTGENERIC_FLUSH_IA32_INSTRUCTION_CACHE (U2P (addr),
								old_len,
								BT_FLUSH_ALLOC);
		} else {
			assert ((uintptr_t) retval < 0x100000000ul);
			BTGENERIC_FLUSH_IA32_INSTRUCTION_CACHE ((void *) retval,
								new_len,
								BT_FLUSH_ALLOC);
		}
	}
      out:
	return (long) retval;
}

/*****************************************************************************/
/* implement the IA-32 mmap2 syscall */
/*****************************************************************************/
long
btlib_ia32_mmap2 (const BTGENERIC_IA32_CONTEXT * bc32)
{
	long retval = 0;

	DBPRINT (btl_prt_mem, "\n");

	/* put the arguments in a struct fitting to mmap */
	retval = mmap32 (IA32_SYSCALL_PARAM1 (bc32),
			 IA32_SYSCALL_PARAM2 (bc32),
			 IA32_SYSCALL_PARAM3 (bc32),
			 IA32_SYSCALL_PARAM4 (bc32),
			 IA32_SYSCALL_PARAM5 (bc32),
			 IA32_SYSCALL_PARAM6 (bc32));

	/* do the mapping */
	DBPRINT (0, "mmap32=%lx\n", retval);
	return retval;
}

/*****************************************************************************/
/* implement the IA-32 mmap syscall */
/*****************************************************************************/
long
btlib_ia32_mmap (const BTGENERIC_IA32_CONTEXT * bc32)
{
	legacy_mmap_args_t args;
	long retval = 0;

	retval = safe_memcpy (&args, U2P(IA32_SYSCALL_PARAM1(bc32)), sizeof (args));
	if (retval == 0) {
		unsigned pgoff;
		if (IA32_PAGE_OFF(args.offset)) {
			return -EINVAL;
		}
		pgoff = args.offset >> IA32_PAGE_SHIFT;
	
		/* do the mapping */
		retval = mmap32 (args.addr, args.len, args.prot, args.flags,
				 args.fd, pgoff);
	}
	return retval;
}

/*****************************************************************************/
/* implement the IA-32 munmap syscall */
/*****************************************************************************/
long
btlib_ia32_munmap (const BTGENERIC_IA32_CONTEXT *bc32)
{
	long retval;

	DBPRINT (btl_prt_mem, "\n");
	retval = munmap32 (IA32_SYSCALL_PARAM1 (bc32), IA32_SYSCALL_PARAM2 (bc32));
	DBPRINT (btl_prt_mem, "retval=%ld\n", retval);
	return retval;
}

/*****************************************************************************/
/* implement the IA-32 mprotect syscall */
/*****************************************************************************/
long
btlib_ia32_mprotect (BTGENERIC_IA32_CONTEXT *bc32)
{
	long perm = 0;
	int no_of_pages;
	long retval;
	unsigned long size, addr, prot;

	addr = (unsigned long) IA32_SYSCALL_PARAM1 (bc32);
	size = (unsigned long) IA32_SYSCALL_PARAM2 (bc32);
	prot = (unsigned long) IA32_SYSCALL_PARAM3 (bc32);

	perm = set_btrans_permissions (prot);
	no_of_pages = (size - 1) / IA32_PAGE_SIZE + 1;

	/* notify ia32exec.bin of the change */
	DBPRINT (btl_prt_mem, "addr=%lx size=%lu no_of_pages=%d prot=%lx\n",
		 addr, size, no_of_pages, prot);
	BTGENERIC_NOTIFY_CHANGE_PERMISSION_REQUEST (addr, no_of_pages, perm);

	/* do the syscall */
	retval = ILP32_WRAPPER3 (IA32_SYSCALL_MPROTECT, addr, size, prot);
	return retval;
}

/*****************************************************************************/
/* allocate memory for IA-32 Execution Layer 
 * all anonymous mappings above 4G must go through this function - no direct 
 * calls to mmap are allowed.
 */
/*****************************************************************************/
void *
btlib_mem_alloc (unsigned long start, size_t length, unsigned prot)
{
	long retval, mm_retval;
	int linux_prot;
	int flags = MAP_PRIVATE | MAP_ANONYMOUS;
	unsigned long free_memory_addr;


	linux_prot = set_linux_permissions (prot);

	lock_mregion_manager(MREGION_MANAGER_BTGENERIC);

	/* find free memory */
	free_memory_addr = find_free_mregion (MREGION_MANAGER_BTGENERIC, start, length, (unsigned)flags);
	/* no enough memory */
	if ((long)free_memory_addr == -ENOMEM) {
		DBPRINT (1, "no enough memory\n");
		retval = free_memory_addr;
		goto out;
	}
	assert (IS_BTGENERIC_MEM_ADDR (free_memory_addr)); 

	/* syscall to allocat the memory */
	retval = (long) SYSCALL (mmap, (free_memory_addr, length, linux_prot, flags, -1, 0ul));
	if (IS_SYS_ERROR(retval)) {
		DBPRINT(1, "failed to mmap addr=0x%lx len=0x%lx, prot=0x%x, flags=0x%x\n", 
			free_memory_addr, length, linux_prot, flags);
		goto out;
	}

	DBPRINT((unsigned long)retval!=free_memory_addr, 
		"addr=0x%lx len=0x%lx, prot=0x%x, flags=0x%x ret=%lx\n",
		free_memory_addr, length, linux_prot, flags, retval);


	/* free the illegal memory and try to allocate a legal one    
	 * in multi-thread programe, some thread may allocat memory by mmap directly
	 * The memory may be for share memory or temporary usage
	 * So that memory may be marked free in memory manage system,
	 * but acutally allocated already 
         * Following code will help to find free memory in such scenario.
         */
	free_memory_addr = BTGENERIC_MEM_BASE;
	while ((!IS_BTGENERIC_MEM_ADDR ((unsigned long)retval)) && 
	       (free_memory_addr < BTGENERIC_MAX_ADDRESS)) {
		
		long tmp_retval;

		/* free the illegal memory*/
		tmp_retval = SYSCALL (munmap, ((unsigned long)retval, length));
	        if (IS_SYS_ERROR (tmp_retval)) {
			DBPRINT(1, "failed to munmap memory addr=0x%lx len=0x%lx\n", retval, length);
			goto out;
		}

		/* try to find new freememory*/
		free_memory_addr = find_free_mregion(MREGION_MANAGER_BTGENERIC, free_memory_addr, length, (unsigned)flags);
		assert (IS_BTGENERIC_MEM_ADDR (free_memory_addr));

		/* allocate memory */
		retval = (long) SYSCALL (mmap, (free_memory_addr, length, linux_prot, flags, -1, 0ul));
		if (IS_SYS_ERROR (retval)) {
			DBPRINT(1, "failed to mmap addr=0x%lx len=0x%lx, prot=0x%x, flags=0x%x\n", 
				free_memory_addr, length, linux_prot, flags);
			goto out;
		}
		
		/* next memory address */
		free_memory_addr += PAGE_SIZE;
	}
	if (!IS_BTGENERIC_MEM_ADDR ((unsigned long)retval)) {
		long tmp_retval;
		tmp_retval = SYSCALL (munmap, ((unsigned long)retval, length));
		assert(!IS_SYS_ERROR(tmp_retval));
		retval = -ENOMEM;
		DBPRINT(1, "failed to find free memory after scan the 4G memory\n");
		goto out;
	}


	/* update mm system */
	mm_retval = insert_mregion (MREGION_MANAGER_BTGENERIC, (unsigned long)retval, length, 
				    (unsigned)linux_prot, (unsigned)flags);
	assert(mm_retval == 0);

out:
	unlock_mregion_manager(MREGION_MANAGER_BTGENERIC);
	DBPRINT (0, "retval = 0x%lx\n", retval);
	assert (IS_BTGENERIC_MEM_ADDR((unsigned long)retval) || 
		retval == -ENOMEM || IS_SYS_ERROR(retval));
	return (void *) retval;
}

/*****************************************************************************/
/* free a mapped memory area */
/*****************************************************************************/
long
btlib_mem_free (void *start, size_t length)
{
	long retval, mm_retval;

	assert((unsigned long)start==PAGE_ALIGN((unsigned long)start));

	/* syscall */
	DBPRINT (btl_prt_mem, "start=%p length=%lu\n", start, length);
	retval = SYSCALL (munmap, ((unsigned long)start, length));

	/* update mm system */
	if (!IS_SYS_ERROR(retval)) {
		lock_mregion_manager(MREGION_MANAGER_BTGENERIC);
		mm_retval = remove_mregion (MREGION_MANAGER_BTGENERIC, (unsigned long)start, (unsigned long)length);
		assert(mm_retval==0);
		unlock_mregion_manager(MREGION_MANAGER_BTGENERIC);
	}
	
	return retval;
}

/*****************************************************************************/
/* change permissions for ia32exec.bin needs */
/*****************************************************************************/
long
btlib_memory_change_permissions (unsigned long start, unsigned long size,
                                 unsigned prot)
{
	unsigned linux_prot = 0;
	long retval, mm_retval;

	linux_prot = set_linux_permissions (prot);

	/* syscall */
	retval = SYSCALL (mprotect, (start, (size_t) size, linux_prot));

	/* update mm system */
	if (!IS_SYS_ERROR(retval)) {
		lock_mregion_manager(MREGION_MANAGER_BTGENERIC);
		mm_retval = set_mregion_prot (MREGION_MANAGER_BTGENERIC, start, size, prot);
	        assert(mm_retval==0);
		unlock_mregion_manager(MREGION_MANAGER_BTGENERIC);
	}

	/* TODO: need to return the currrent permissions on failure ? */
	return retval;
}

/*****************************************************************************/
/* set the fd hugetlb status in m_huge_fds
 * 
 * input:	fd:	the fd we want to set the status
 * 		status:	the new status of this fd
 * 		
 * output: 	0 :	success
 *		-1:	fail
 */
/*****************************************************************************/
int 
set_fd_hugetlb_status(int fd, int status) 
{
	struct fd_status_map * p, * q;
	
	p = &m_huge_fds;
	assert (p != NULL);
	while (p) {
		if (GROUP_OF_FDITEM (fd) == p->start_fd) {
			/* we found the group of this fd */
			if ((status == FD_UNKNOWN) && 
			    (p->fd_map[INDEX_OF_FDITEM (fd)] != FD_UNKNOWN)) {
				/* now, we will only set status to FD_UNKNOWN when close()
				 * try to reuse the group
				 */
				assert (p->used_fd > 0);
				p->used_fd--;
				if ((p->used_fd == 0) && (p != &m_huge_fds)) {
					/* never reuser m_huge_fds */
					assert (UNUSED_FD_GROUP_FLAG < 0);
					p->start_fd = UNUSED_FD_GROUP_FLAG;
				}
			}

			if ((status != FD_UNKNOWN) && 
			    (p->fd_map[INDEX_OF_FDITEM (fd)] == FD_UNKNOWN)) {
				p->used_fd++;
			}

			p->fd_map[INDEX_OF_FDITEM (fd)] = status;
			return 0;
		}
		
		p = p->next;
	}

	/* we can not find the fd */
	if (status == FD_UNKNOWN) {
		/* we only setup FD_UNKNOWN in close(), no need to allocate new item */
		return 0;
	}

	/* try to reuse one free item at first */
	q = p = &m_huge_fds;
	while (p && (p->start_fd != UNUSED_FD_GROUP_FLAG)) {
		q = p;
		p = p->next;
	}
	
	if (p) {
		/* we found one free item, re-use it! */
		assert (p->start_fd == UNUSED_FD_GROUP_FLAG);
		
		p->start_fd = GROUP_OF_FDITEM (fd);
		p->fd_map[INDEX_OF_FDITEM (fd)] = status;
		p->used_fd = 1;
	} else {
		/* we cound not find one free item, allocate one */
		p = (struct fd_status_map *) bt_malloc (sizeof (struct fd_status_map));
		if (p == NULL) {
			/* can not require more memory */
			return -1;
		}

		q->next = p;
		
		p->start_fd = GROUP_OF_FDITEM (fd);
		bt_memset (m_huge_fds.fd_map, FD_UNKNOWN, (size_t)FD_MAP_SIZE);
		p->fd_map[INDEX_OF_FDITEM (fd)] = status;
		p->used_fd = 1;
		p->next = NULL;
	}

	return 0;
}

/*****************************************************************************/
/* get the fd hugetlb status in m_huge_fds
 * 
 * input:	fd:	the fd we want to get the status
 * 		
 * output: 	fd hugetlb status
 */
/*****************************************************************************/
int 
get_fd_hugetlb_status (int fd) 
{
	struct fd_status_map * p;
	
	p = &m_huge_fds;
	assert (p != NULL);
	while (p) {
		if (GROUP_OF_FDITEM (fd) == p->start_fd) {
			return p->fd_map[INDEX_OF_FDITEM (fd)];
		}
		
		p = p->next;
	}

	return FD_UNKNOWN;
}

/*****************************************************************************/
/* judge if the fd is on one  hugetlb via /proc/mounts
 * when there are errors, always return FD_UNKNOWN
 * 
 * input:	fd:	the fd we want to judge the status
 * 		
 * output: 	fd status
 */
/*****************************************************************************/
static int 
is_on_hugetlb_fs (int fd) 
{
	struct stat stat_buff;
	char mount_point[MOUNT_LEN];
	char mount_type[MOUNT_LEN];
	char *p, *q, *temp;
	char *mounts = NULL;

	int  maps_file;
	int  retval = FD_UNKNOWN, ret;
	int  size;
	dev_t device;

	assert (m_hugetlbfs_exists);

	/* get the device No. of fd */
	ret = SYSCALL (fstat, (fd, &stat_buff));
	if (IS_SYS_ERROR (ret)) {
		DBPRINT (1, "failed to stat fd %d\n", fd);
		return retval;
	}
	device = stat_buff.st_dev;

	/* open mounts maps file */
	maps_file = SYSCALL (open, (PROC_MOUNT_NAME, O_RDONLY, 0));
	if (IS_SYS_ERROR (maps_file)) {
		DBPRINT (1, "failed to open %s!\n", PROC_MOUNT_NAME);
		return retval;
	}

	mounts = (char *)bt_malloc ((size_t)MAX_PROC_MOUNTS_SIZE);
	if (mounts == NULL) {
		DBPRINT (1, "failed to allocate memory!\n");
		goto out;
	}
	bt_memset (mounts, 0, (size_t)MAX_PROC_MOUNTS_SIZE);

	/* go through the mounts maps file to see if there are hugetlbfs
 	 * 	the structure of /proc/mounts is like:
	 *		source_dir mount_point mount_type ....
	 */
	size = SYSCALL (read, (maps_file, mounts, (size_t) MAX_PROC_MOUNTS_SIZE));
	if (IS_SYS_ERROR (size)) {
		DBPRINT (1, "failed to read %s!\n", PROC_MOUNT_NAME);
		goto out;
	}
	/* tentatively set the buffer size to 64K, should be enough for 
	   all real world cases */
	assert (size <= MAX_PROC_MOUNTS_SIZE); 

	retval = FD_NOTHUGE;

	/* fast path, there is no huge tls filesystem in the system */
	if ((temp = bt_strstr (mounts, HUGETLBFS_NAME)) == NULL) {
		goto out;
	}

	/* locate the 1st line containing HUGETLBFS_NAME */
	while ((*temp != '\n') && (temp != &mounts[0])) {
		temp--;
	}
	if (temp != &mounts[0]) {
		temp++;
	}

	/* handle one line in one iteration - p is the begin of a line, 
	   q is the end of a line */
	p = temp; 
	while (p != &mounts[size]) {
		char *point, *type;

		q = bt_strstr (p, "\n");
		*q = 0;

		DBPRINT (0, "handling: %s\n", p);
		
		/* find a line with HUGETLBFS_NAME */
		temp = bt_strstr (p, HUGETLBFS_NAME);
		if (temp == NULL) {
			p = q + 1;
			continue;
		}

		/* find mount point */
		point = bt_strstr (p, " ");
		point++;

		/* find mount type */
		type = bt_strstr (point, " ");	
		*type = 0;
		type++;	

		assert (bt_strlen (point) <= MOUNT_LEN);
		bt_strcpy (&mount_point[0], point);

		temp = bt_strstr (type, " ");
		*temp = 0;

		assert (bt_strlen (type) <= MOUNT_LEN);
		bt_strcpy (&mount_type[0], type);

		if (bt_strncmp (&mount_type[0], HUGETLBFS_NAME, (size_t)MOUNT_LEN) == 0) {
			/* find one hugetlbfs */
			/* get the device No. of mount_point */
			DBPRINT (0, "find hugetlbfs mount point: %s, mount type: %s\n", 
				 mount_point, mount_type);
			ret = SYSCALL (stat, (mount_point, &stat_buff));
			if ((!IS_SYS_ERROR (ret)) && (stat_buff.st_dev == device)) {
				retval = FD_HUGE;
				break;	
			}
		}
		p = q + 1;
	}
out:
	bt_free (mounts);
	SYSCALL (close, (maps_file));
	return retval;
}

int btlib_check_region_prot(unsigned long address, unsigned long size, int prot)
{
	unsigned long start = PAGE_START(address), end = LAST_PAGE_IN_REGION(address, size);
	unsigned long addr;
	int valid = 1, perm, exist;
	extern int get_page_prot_e (unsigned long start_addr, int *is_exist);

	DBPRINT ((address >> 32) != 0, "address=%lx size=%lx, prot=%x\n", address, size, prot);
	/*not legal ia32 memory*/
	if (((address >> 32) != 0 ) || ((size >> 32) != 0) || (((address + size - 1) >> 32) != 0)) {
		return 0;
	}

	assert (((unsigned long)prot & ((unsigned long)(~(PROT_WRITE | PROT_READ | PROT_EXEC)))) == 0);
	DBPRINT(0, "%x[%x] =>  %x-%x\n", address, size, start, end);
	for (addr = start; addr <= end;  addr += PAGE_SIZE) {
		perm = get_page_prot_e(addr, &exist); 
		DBPRINT (0, "addr=%x is_exist=%d perm=%x require %x\n", addr, exist, perm, prot);
		if (!exist || ((perm & prot) != prot)) {
			valid = 0;
			break;
		}
	}

	return valid;
}
