/*  
 * btlib_networking.c
 *  
 * this file contains the code for emulation of the IA-32 ipc syscall. 
 * 
*/
 /*
 *
 * This file has been derived from GNU libc. Please see COPYING.LIBC
 * for more information
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include <ia32x_utils.h>
#include <ia32toia64.h>
#include "btlib.h"

long HIDDEN sendmsg32 (int fd, __u32 msg32, unsigned flags);
long HIDDEN recvmsg32 (int fd, __u32 msg32, unsigned int flags);
static int msghdr32_to_msghdr(struct msghdr * msg, struct msghdr32 * msg32);
static int cmsghdr32_to_cmsghdr(struct msghdr * msg);
static int cmsg32_to_cmsg(struct msghdr * msg, struct msghdr32 * msg32);
static int movaddr_disable_smc_opt(void* addr, int* len);

static int movaddr_disable_smc_opt(void* addr, int* len)
{
	int l;

	if (safe_memcpy(&l, len, sizeof(int)) < 0) {
		return -EFAULT;
	}

	if (l < 0) {
		return 1;
	}

	btlib_disable_smc_opt(len, sizeof(int));
	btlib_disable_smc_opt(addr, l);
	return 0;
}

struct cmsghdr32_list {
	struct cmsghdr32 * item;
	struct cmsghdr32_list * next;	
};

#define SOCKETCALL_ARGUMENTS_SIZE(i) (i * sizeof(__u32))
static const unsigned long socketcall_arguments_size[18]={
	0,                            /* dummy */
	SOCKETCALL_ARGUMENTS_SIZE(3),  /* sys_socket(2) */
	SOCKETCALL_ARGUMENTS_SIZE(3),  /* sys_bind(2) */
	SOCKETCALL_ARGUMENTS_SIZE(3),  /* sys_connect(2) */
	SOCKETCALL_ARGUMENTS_SIZE(2),  /* sys_listen(2) */
	SOCKETCALL_ARGUMENTS_SIZE(3),  /* sys_accept(2) */
    	SOCKETCALL_ARGUMENTS_SIZE(3),  /* sys_getsockname(2) */
	SOCKETCALL_ARGUMENTS_SIZE(3),  /* sys_getpeername(2) */
	SOCKETCALL_ARGUMENTS_SIZE(4),  /* sys_socketpair(2) */
	SOCKETCALL_ARGUMENTS_SIZE(4),  /* sys_send(2) */
	SOCKETCALL_ARGUMENTS_SIZE(4),  /* sys_recv(2) */
	SOCKETCALL_ARGUMENTS_SIZE(6),  /* sys_sendto(2)	*/
    	SOCKETCALL_ARGUMENTS_SIZE(6),  /* sys_recvfrom(2) */
	SOCKETCALL_ARGUMENTS_SIZE(2),  /* sys_shutdown(2) */
	SOCKETCALL_ARGUMENTS_SIZE(5),  /* sys_setsockopt(2) */
	SOCKETCALL_ARGUMENTS_SIZE(5),  /* sys_getsockopt(2) */
	SOCKETCALL_ARGUMENTS_SIZE(3),  /* sys_sendmsg(2) */
	SOCKETCALL_ARGUMENTS_SIZE(3)}; /* sys_recvmsg(2) */

long btlib_socketcall (BTGENERIC_IA32_CONTEXT * bc32)
{
	int retval;
	__u32 syscall_parameters[6];
	int call;
	__u32 args; 

	call = (int)IA32_SYSCALL_PARAM1(bc32);
	args = (__u32)IA32_SYSCALL_PARAM2(bc32);

	if (call < 1 || call > SYS_SOCKETCALL_MAXCALL) {
		return -EINVAL;
	}
	DBPRINT(0, "call=%d args=%d size=%d\n", call, args, socketcall_arguments_size[call]);
	if (safe_memcpy(syscall_parameters, U2P(args), 
			  socketcall_arguments_size[call]) < 0) {
		return -EFAULT;
	}

	switch(call) {
	case SYS_SOCKET:
		retval = SYSCALL(socket,
			         ((int)syscall_parameters[0], 
			          (int)syscall_parameters[1], 
			          (int)syscall_parameters[2]));
		break;
	case SYS_BIND:
		retval = SYSCALL(bind,
			         ((int)syscall_parameters[0], 
			          (struct sockaddr *)UL(syscall_parameters[1]), 
			          (int)syscall_parameters[2]));
		break;
	case SYS_CONNECT:
		retval = SYSCALL(connect,
				 ((int)syscall_parameters[0], 
				  (struct sockaddr *)UL(syscall_parameters[1]), 
				  (int)syscall_parameters[2]));
		break;
	case SYS_ACCEPT:
		if (g_has_smc) movaddr_disable_smc_opt(U2P(syscall_parameters[1]), U2P(syscall_parameters[2]));
		retval = SYSCALL(accept,
			         ((int)syscall_parameters[0], 
				  (struct sockaddr *)UL(syscall_parameters[1]), 
			          (int *)UL(syscall_parameters[2])));
		break;
	case SYS_GETSOCKNAME:
		if (g_has_smc) movaddr_disable_smc_opt(U2P(syscall_parameters[1]), U2P(syscall_parameters[2]));
		retval = SYSCALL(getsockname,
				 ((int)syscall_parameters[0], 
				  (struct sockaddr *)UL(syscall_parameters[1]), 
				  (int *)UL(syscall_parameters[2])));
			break;
	case SYS_GETPEERNAME:
		if (g_has_smc) movaddr_disable_smc_opt(U2P(syscall_parameters[1]), U2P(syscall_parameters[2]));
		retval = SYSCALL(getpeername,
				 ((int)syscall_parameters[0], 
				  (struct sockaddr *)UL(syscall_parameters[1]), 
				  (int *)UL(syscall_parameters[2])));
		break;
	case SYS_SOCKETPAIR:
	{
		int sv[2];
		retval = SYSCALL(socketpair,
				 ((int)syscall_parameters[0], 
				  (int)syscall_parameters[1], 
				  (int)syscall_parameters[2], 
				  sv));
		if (!IS_SYS_ERROR(retval) && safe_memcpy(U2P(syscall_parameters[3]), sv, sizeof(int) * 2) < 0) {
			/*close the socket*/
			retval = SYSCALL(close, (sv[0]));
			assert (retval == 0);
			retval = SYSCALL(close, (sv[1]));
			assert (retval == 0);
			retval = -EFAULT;
		}
		
	}
		break;
	case SYS_LISTEN:
		retval = LONG_SYSCALL2(listen,
				       (int)syscall_parameters[0], 
				       (int)syscall_parameters[1]);
		break;
	case SYS_SEND:
		retval = LONG_SYSCALL4(send, 
				       (int)syscall_parameters[0], 
				       (void *)UL(syscall_parameters[1]), 
				       (int)syscall_parameters[2], 
				       (int)syscall_parameters[3]);
		break;
	case SYS_SENDTO:
		retval = LONG_SYSCALL6(sendto, 
				       (int)syscall_parameters[0], 
				       (void *)UL(syscall_parameters[1]), 
				       (int)syscall_parameters[2], 
				       (int)syscall_parameters[3], 
				       (struct sockaddr*)UL(syscall_parameters[4]), 
				       syscall_parameters[5]);
		break;
	case SYS_RECV:
		if (g_has_smc) btlib_disable_smc_opt(syscall_parameters[1], syscall_parameters[2]);
		retval = LONG_SYSCALL4(recv, 
				       (int)syscall_parameters[0], 
				       (void *)UL(syscall_parameters[1]), 
				       (int)syscall_parameters[2], 
				       (int)syscall_parameters[3]);
		break;
	case SYS_RECVFROM:
		if (g_has_smc) {
			btlib_disable_smc_opt(syscall_parameters[1], syscall_parameters[2]);
			movaddr_disable_smc_opt(U2P(syscall_parameters[4]), U2P(syscall_parameters[5]));
		}
		retval = LONG_SYSCALL6(recvfrom, 
				       (int)syscall_parameters[0], 
				       (void*)UL(syscall_parameters[1]), 
				       (int)syscall_parameters[2], 
				       (int)syscall_parameters[3], 
				       (struct sockaddr*)UL(syscall_parameters[4]), 
				       (void*)UL(syscall_parameters[5]));
		break;
	case SYS_SHUTDOWN:
		retval = SYSCALL(shutdown,
				 ((int)syscall_parameters[0],
				  (int)syscall_parameters[1]));
		break;
	case SYS_SETSOCKOPT:
		retval = SYSCALL(setsockopt,
				 ((int)syscall_parameters[0], 
				  (int)syscall_parameters[1], 
				  (int)syscall_parameters[2], 
				  (char *)UL(syscall_parameters[3]), 
				  (int)syscall_parameters[4]));
		break;
	case SYS_GETSOCKOPT:
		if (g_has_smc) movaddr_disable_smc_opt(U2P(syscall_parameters[4]), U2P(syscall_parameters[5]));
		retval = SYSCALL(getsockopt,
				 ((int)syscall_parameters[0], 
				  (int)syscall_parameters[1], 
				  (int)syscall_parameters[2], 
				  (void*)UL(syscall_parameters[3]), 
				  (void*)UL(syscall_parameters[4])));
			break;
	case SYS_SENDMSG:
		retval = sendmsg32((int)syscall_parameters[0], 
				       (__u32) UL(syscall_parameters[1]), 
				       (unsigned)syscall_parameters[2]);
		break;
	case SYS_RECVMSG:
		retval = recvmsg32((int)syscall_parameters[0], 
			   	       (__u32) UL(syscall_parameters[1]), 
				       (unsigned)syscall_parameters[2]);
		break;
	default:
		retval = -EINVAL;
		break;
	}

	DBPRINT(0,"retval=%lx", retval);
	return retval;
}


/* from user32_networking.c*/
extern inline int iovec_conv(struct iovec32 * iov32, struct iovec * iov);

/*****************************************************************************/
/* from /usr/include/bits/socket.h */
/*****************************************************************************/
inline struct cmsghdr *
__cmsg_nxthdr (struct msghdr *__mhdr, struct cmsghdr *__cmsg)
{
  if ((size_t) __cmsg->cmsg_len < sizeof (struct cmsghdr))
    /* The kernel header does this so there may be a reason.  */
    return 0;

  __cmsg = (struct cmsghdr *) ((unsigned long) __cmsg
			       + CMSG_ALIGN (__cmsg->cmsg_len));
  if ((unsigned char *) (__cmsg + 1) > ((unsigned char *) __mhdr->msg_control
					+ __mhdr->msg_controllen)
      || ((unsigned char *) __cmsg + CMSG_ALIGN (__cmsg->cmsg_len)
	  > ((unsigned char *) __mhdr->msg_control + __mhdr->msg_controllen)))
    /* No more entries.  */
    return 0;
  return __cmsg;
}

static inline struct cmsghdr32 * cmsg32_nxthdr (struct msghdr * mhdr, 
						struct cmsghdr32 * cmsg32, 
						int cmsg_len)
{
	cmsg32 = (struct cmsghdr32 *) UL((unsigned char *) cmsg32 +
		CMSG32_ALIGN (cmsg_len));
	if ((unsigned char *) (cmsg32 + 1) > ((unsigned char *) mhdr->msg_control
			+ mhdr->msg_controllen)
		|| ((unsigned char *) cmsg32 + CMSG32_ALIGN (cmsg_len) 
			> ((unsigned char *) mhdr->msg_control + mhdr->msg_controllen))) {
		/* No more entries.  */
		return NULL;
	}
	return cmsg32;
}


/* 
 * simulate sendmsg system call, due to 32-bit and 64-bit issue, 
 * we must handle: 
 *	1. io vector 
 *	2. control message header 
 */
long HIDDEN 
sendmsg32 (int fd, __u32 msg32, unsigned flags)
{
	struct msghdr msg;
	struct iovec * iov = NULL;
	int retval, ret;
	int free_msg_iov = 0;
	unsigned int i;
	struct iovec32 *my_iovec32;

	DBPRINT (0, "fd=%d msg32=%p flags=%x\n", fd, U2P(msg32), flags);

	bt_memset (&msg, 0, sizeof(msg));
	retval = msghdr32_to_msghdr (&msg, (struct msghdr32 *)UL(msg32));
	if (retval < 0) {
		goto out;
	}

	if (IS_VALID_PTR(msg.msg_iov) && (msg.msg_iovlen > 0)) {
		iov = (struct iovec *) bt_malloc (sizeof(struct iovec) * msg.msg_iovlen);
		if (iov == NULL) {
			retval = -ENOMEM;
			goto out;
		}
		free_msg_iov= 1;
	
		my_iovec32 = (struct iovec32 *)U2P((((struct msghdr32 *)(U2P(msg32)))->msg_iov));
	
		for (i = 0; i < msg.msg_iovlen; i++) {
			ret = iovec_conv(&my_iovec32[i], &iov[i]);
			if (ret < 0) {
				retval = -EFAULT;
				goto out;
			} 
		}
		msg.msg_iov = iov;
	}

	if (msg.msg_controllen > INT_MAX) {
		retval = -ENOBUFS;
		goto out;
	}

	if (IS_VALID_PTR(msg.msg_control)) {
		if (msg.msg_controllen > 0) {
			retval = cmsghdr32_to_cmsghdr (&msg);
			if (retval < 0)
				goto out;
		}
	}

	flags &= 0x7FFFFFFF; /* clear MSG_CMSG_COMPAT bit before calling the system call */
	msg.msg_flags = flags;
	retval = LONG_SYSCALL3 (sendmsg, fd, &msg, flags);

	if (msg.msg_controllen != 0) {
		bt_free(msg.msg_control);
	}

out:
	if (free_msg_iov) {
		bt_free (iov);
	}
	return retval;
}

/* 
 * simulate recvmsg system call, due to 32-bit and 64-bit issue, 
 * we must handle: 
 *	1. io vector 
 *	2. control message header 
 */
long HIDDEN 
recvmsg32 (int fd, __u32 msg32, unsigned int flags)
{
	struct msghdr msg;
	struct iovec * iov = NULL;
	int retval, ret;
	int free_msg_iov = 0, free_msg_control = 0;
	unsigned int cmsg_num, i;
	struct iovec32 *my_iovec32;

	bt_memset (&msg, 0, sizeof(msg));
	retval = msghdr32_to_msghdr (&msg, (struct msghdr32 *)UL(msg32));
	if (retval < 0) {
		goto out;
	}

	if (IS_VALID_PTR(msg.msg_iov) && (msg.msg_iovlen > 0)) {
		iov = (struct iovec *) bt_malloc (sizeof(struct iovec) * msg.msg_iovlen);
		if (iov == NULL) {
			retval = -ENOMEM;
			goto out;
		}
		free_msg_iov = 1;
	
		my_iovec32 = (struct iovec32 *)U2P((((struct msghdr32 *)(U2P(msg32)))->msg_iov));
	
		for (i = 0; i < msg.msg_iovlen; i++) {
			ret = iovec_conv(&my_iovec32[i], &iov[i]);
			if (ret < 0) {
				retval = -EFAULT;
				goto out;
			} 
			if (g_has_smc) btlib_disable_smc_opt(U2P(my_iovec32[i].iov_base), my_iovec32[i].iov_len);
		}
		msg.msg_iov = iov;
	}

	cmsg_num = msg.msg_controllen / (sizeof(struct cmsghdr32) + sizeof(int));
	msg.msg_controllen += cmsg_num * 8;

	if (IS_VALID_PTR(msg.msg_control) && (msg.msg_controllen > 0)) {
		msg.msg_control = bt_malloc(msg.msg_controllen);
		if (msg.msg_control == NULL) {
			retval = -ENOMEM;
			goto out;
		}
		free_msg_control = 1;
	}
	
	retval = LONG_SYSCALL3 (recvmsg, fd, &msg, flags);
	if (IS_SYS_ERROR (retval)) {
		goto out;
	}

	if (IS_VALID_PTR(msg.msg_control)) {
		ret = cmsg32_to_cmsg(&msg, (struct msghdr32 *)UL(msg32));
		if (ret < 0)
			retval = ret;
	}

	/* bt_free the localy allocated memory */
out:
	if (free_msg_iov) {
		bt_free (msg.msg_iov);
	}
	if (free_msg_control) {
		bt_free (msg.msg_control);
	}

	return retval;
}

/*****************************************************************************/
/* convert msghdr32 to msghdr
 * input: msg - destination, 64-bit msghdr 
 *		msg32 - source, 32-bit msghdr
 * output: 0 on success, negative error code otherwise
 */					
/*****************************************************************************/
static int msghdr32_to_msghdr (struct msghdr * msg, struct msghdr32 * msg32)
{
	struct msghdr32 lm32;

	if (safe_memcpy (&lm32, msg32, sizeof(lm32)) < 0) {
		return -EFAULT;
	}

	if (lm32.msg_iovlen > UIO_MAXIOV) {
		return -EMSGSIZE;
	}

	msg->msg_name		= U2P(lm32.msg_name);
	msg->msg_namelen	= lm32.msg_namelen;
	msg->msg_iovlen		= lm32.msg_iovlen;
	msg->msg_flags		= lm32.msg_flags;
	msg->msg_iov		= U2P(lm32.msg_iov);	
	msg->msg_control	= U2P(lm32.msg_control);
	msg->msg_controllen	= lm32.msg_controllen;

	return 0;
}

/*****************************************************************************/
/* 
 * convert 32-bit style cmsghdr to 64-bit style cmsghdr. memory to hold 64-bit
 * cmsghdr information is allocted in this function and must be freed by the 
 * caller.
 * input: msg->msg_control - 32-bit cmsghdr 
 * output: 0 on success, negative error code otherwise
 */					
/*****************************************************************************/
static int cmsghdr32_to_cmsghdr(struct msghdr * msg)
{
	struct cmsghdr * cmsg, * cmsg_bak = NULL;
	struct cmsghdr32 * cmsg32;
	struct cmsghdr32_list * cmsghdr32_list_head = NULL, * cmsghdr32_list_tail = NULL, * p;
	unsigned long	cmsglen = 0;
	unsigned long cmsg32len = 0;
	unsigned long	cmsg_now_offset = UL(CMSG32_FIRSTHDR (msg));
	long retval = 0;
	int	free_cmsg = 0;

	/* record cmsg32 */
	for (cmsg32 = CMSG32_FIRSTHDR (msg);
		cmsg32 != NULL;
		cmsg32 = cmsg32_nxthdr(msg, cmsg32, (int)cmsg32len)) {
		
		struct cmsghdr32 cmsg32_tmp;
		if (safe_memcpy (&cmsg32_tmp, cmsg32, sizeof (cmsg32_tmp)) < 0) {
			retval = -EFAULT;
			goto out;
		}
		cmsg32len = CMSG32_LEN(cmsg32_tmp);

		p = (struct cmsghdr32_list *)bt_malloc(sizeof(struct cmsghdr32_list));
		if (p == NULL) {
			retval = -ENOBUFS;
			goto out;
		}
		
		if(cmsghdr32_list_head == NULL) {
			cmsghdr32_list_head = cmsghdr32_list_tail = p;
		}
		else {
			cmsghdr32_list_tail->next = p;
			cmsghdr32_list_tail = p;
		}
		
		p->item = cmsg32;
		p->next = NULL;		
	}

	/* get cmsglen */
	for (cmsg32 = CMSG32_FIRSTHDR (msg);
		cmsg32 != NULL;
		cmsglen += cmsg32len + CMSG_DIFF) {

		struct cmsghdr32 cmsg32_tmp;
		if (safe_memcpy (&cmsg32_tmp, cmsg32, sizeof (cmsg32_tmp)) < 0) {
			retval = -EFAULT;
			goto out;
		}
		
		cmsg32len = CMSG32_LEN(cmsg32_tmp);
		cmsg_now_offset += cmsg32len;
		if ((CMSG32_ALIGN(cmsg32len) < CMSG32_ALIGN(sizeof(struct cmsghdr32)))
			|| (cmsg32 > (struct cmsghdr32 *)cmsg_now_offset)) {
			retval = -EINVAL;
			goto out;
		}
		
		cmsg32 = cmsg32_nxthdr(msg, cmsg32, (int)cmsg32len);
	}

	/* generate cmsg! */
	cmsg = (struct cmsghdr *)bt_malloc(cmsglen);
	if (cmsg == NULL) {
		retval = -ENOBUFS;
		goto out;
	}
	free_cmsg = 1;

	/* construct cmsg */
	p = cmsghdr32_list_head;
	cmsg_bak = cmsg;
	while (p) {
		assert(IS_VALID_PTR(p->item));
		cmsg->cmsg_len = p->item->cmsg_len + CMSG_DIFF;
		cmsg->cmsg_level = p->item->cmsg_level;
		cmsg->cmsg_type = p->item->cmsg_type;
		if (safe_memcpy (CMSG_DATA(cmsg), CMSG32_DATA(p->item), 
			(unsigned long)(p->item->cmsg_len  - CMSG32_ALIGN(sizeof(struct cmsghdr32)))) < 0) {
			retval = -EFAULT;
			goto out;
		}
		
		cmsg = (struct cmsghdr *)UL(((unsigned char *)cmsg + CMSG_ALIGN(cmsg->cmsg_len)));
		p = p->next;
	}

	msg->msg_controllen = cmsglen;
	msg->msg_control = cmsg_bak;
	free_cmsg = 0;

out:
	if (free_cmsg) {
		bt_free (cmsg_bak);
	}
	while (cmsghdr32_list_head) {
		p = cmsghdr32_list_head->next;
		bt_free (cmsghdr32_list_head);
		cmsghdr32_list_head = p;
	}
	return retval;
		
}
/* 
 * convert 64-bit control/ancillary data to 32-bit 
 * input: msg - 64-bit msghdr 
 *		msg32 - 32-bit msghdr
 * output: void
 */ 
static int cmsg32_to_cmsg(struct msghdr * msg, struct msghdr32 * msg32)
{
	struct cmsghdr * cmsg;
	struct cmsghdr32 * cmsg32, cmsg32_tmp;
	unsigned len = 0;
	unsigned int datalen;

	cmsg32 = (struct cmsghdr32 *)UL(msg32->msg_control);
	for (cmsg = CMSG_FIRSTHDR(msg);
		cmsg != NULL; 
		cmsg = CMSG_NXTHDR(msg, cmsg)) {

		cmsg32_tmp.cmsg_level = cmsg->cmsg_level;
		cmsg32_tmp.cmsg_type = cmsg->cmsg_type;
		datalen = cmsg->cmsg_len - CMSG_ALIGN(sizeof(*cmsg));
		cmsg32_tmp.cmsg_len = (CMSG32_ALIGN(sizeof(struct cmsghdr32))) + 
					datalen;

		if (safe_memcpy (cmsg32, &cmsg32_tmp, sizeof (cmsg32_tmp)) < 0) {
			return -EFAULT;
		}

		if (safe_memcpy (CMSG32_DATA(cmsg32), CMSG_DATA(cmsg), (unsigned long)datalen) < 0) {
			return -EFAULT;
		}
		
		len += cmsg32_tmp.cmsg_len;

		cmsg32 = (struct cmsghdr32 *)UL((unsigned char *)cmsg32 + CMSG32_ALIGN(cmsg32->cmsg_len));
	}
	msg32->msg_controllen = len;

	return 0;
}
