/*  
 * btlib_syscall.c
 *  
 *  
 * 
*/
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include "btlib.h"
#include <ia32x_utils.h>

#include <ia32toia64.h>

#include "btlib_segment.h"
#include "btlib_syscall.h"
#include "btlib_sync.h"
#include "btlib_stats.h"
#include "btlib_ptemu.h"
#include "btlib_sig_utils.h"
#include "btlib_pending_que.h"

extern long HIDDEN (*g_safe_memcpy)(void *, void *, unsigned long);
/* the IA-32 kernel space address */
long btl_kernel_page = 0;

/* the address of restore code for sigreturn */
long btl_restorer;
/* the address of rt_restore code for sigreturn */
long btl_rt_restorer;

struct special_file;
typedef int (*sf_handler) (BTGENERIC_IA32_CONTEXT *, struct special_file *);
/* special file handling */
struct special_file {
	int fd;
	sf_handler read_handler;
	sf_handler lseek_handler;
	unsigned long position;
};

struct special_filename {
	char filename[PATH_MAX];
	sf_handler open_handler;
	sf_handler read_handler;
	sf_handler lseek_handler;
};

struct special_file *m_special_files = NULL;
const int m_custom_handlers_size = 2;
int m_special_files_size = 0;

static int special_file_insert (int fd, const struct special_filename *spec);
static int special_file_remove (int fd);
static const struct special_filename *get_special_file_spec (char *filename);
static struct special_file *get_special_file_handlers (int fd);
static int btlib_open (BTGENERIC_IA32_CONTEXT *);
static int btlib_close (BTGENERIC_IA32_CONTEXT *);
static int btlib_read (BTGENERIC_IA32_CONTEXT *);
static int btlib_lseek (BTGENERIC_IA32_CONTEXT *);
static int proc_self_stat_read (BTGENERIC_IA32_CONTEXT * bc, struct special_file *sf);
static int proc_self_stat_lseek (BTGENERIC_IA32_CONTEXT * bc, struct special_file *sf);
static int proc_self_maps_read (BTGENERIC_IA32_CONTEXT * bc, struct special_file *sf);
static int proc_self_maps_lseek (BTGENERIC_IA32_CONTEXT * bc, struct special_file *sf);
static long btlib_sysctl (BTGENERIC_IA32_CONTEXT * bc32);
static long btlib_futex (BTGENERIC_IA32_CONTEXT * bc32);
static long btlib_splice(BTGENERIC_IA32_CONTEXT * bc32);

const struct special_filename m_custom_handlers[] = {
	{"/proc/self/stat", NULL, proc_self_stat_read, proc_self_stat_lseek},
	{"/proc/self/maps", NULL, proc_self_maps_read, proc_self_maps_lseek}, /*stat and map has the same lseek*/
};

#if DEBUG
const char *syscall_names[] = {
/*0*/ "nyi",
/*1*/ "exit",
/*2*/ "fork",
/*3*/ "read",
/*4*/ "write",
/*5*/ "open",
/*6*/ "close",
/*7*/ "waitpid",
/*8*/ "creat",
/*9*/ "link",
/*10*/ "unlink",
/*11*/ "execve",
/*12*/ "chdir",
/*13*/ "time",
/*14*/ "mknod",
/*15*/ "chmod",
/*16*/ "lchown",
/*17*/ "break",
/*18*/ "oldstat",
/*19*/ "lseek",
/*20*/ "getpid",
/*21*/ "mount",
/*22*/ "umount",
/*23*/ "setuid",
/*24*/ "getuid",
/*25*/ "stime",
/*26*/ "ptrace",
/*27*/ "alarm",
/*28*/ "oldfstat",
/*29*/ "pause",
/*30*/ "utime",
/*31*/ "stty",
/*32*/ "gtty",
/*33*/ "access",
/*34*/ "nice",
/*35*/ "ftime",
/*36*/ "sync",
/*37*/ "kill",
/*38*/ "rename",
/*39*/ "mkdir",
/*40*/ "rmdir",
/*41*/ "dup",
/*42*/ "pipe",
/*43*/ "times",
/*44*/ "prof",
/*45*/ "brk",
/*46*/ "setgid",
/*47*/ "getgid",
/*48*/ "signal",
/*49*/ "geteuid",
/*50*/ "getegid",
/*51*/ "acct",
/*52*/ "umount2",
/*53*/ "lock",
/*54*/ "ioctl",
/*55*/ "fcntl",
/*56*/ "mpx",
/*57*/ "setpgid",
/*58*/ "ulimit",
/*59*/ "oldolduname",
/*60*/ "umask",
/*61*/ "chroot",
/*62*/ "ustat",
/*63*/ "dup2",
/*64*/ "getppid",
/*65*/ "getpgrp",
/*66*/ "setsid",
/*67*/ "sigaction",
/*68*/ "sgetmask",
/*69*/ "ssetmask",
/*70*/ "setreuid",
/*71*/ "setregid",
/*72*/ "sigsuspend",
/*73*/ "sigpending",
/*74*/ "sethostname",
/*75*/ "setrlimit",
/*76*/ "getrlimit",
/*77*/ "getrusage",
/*78*/ "gettimeofday",
/*79*/ "settimeofday",
/*80*/ "getgroups",
/*81*/ "setgroups",
/*82*/ "select",
/*83*/ "symlink",
/*84*/ "oldlstat",
/*85*/ "readlink",
/*86*/ "uselib",
/*87*/ "swapon",
/*88*/ "reboot",
/*89*/ "readdir",
/*90*/ "mmap",
/*91*/ "munmap",
/*92*/ "truncate",
/*93*/ "ftruncate",
/*94*/ "fchmod",
/*95*/ "fchown",
/*96*/ "getpriority",
/*97*/ "setpriority",
/*98*/ "profil",
/*99*/ "statfs",
/*100*/ "fstatfs",
/*101*/ "ioperm",
/*102*/ "socketcall",
/*103*/ "syslog",
/*104*/ "setitimer",
/*105*/ "getitimer",
/*106*/ "stat",
/*107*/ "lstat",
/*108*/ "fstat",
/*109*/ "olduname",
/*110*/ "iopl",
/*111*/ "vhangup",
/*112*/ "idle",
/*113*/ "vm86old",
/*114*/ "wait4",
/*115*/ "swapoff",
/*116*/ "sysinfo",
/*117*/ "ipc",
/*118*/ "fsync",
/*119*/ "sigreturn",
/*120*/ "clone",
/*121*/ "setdomainname",
/*122*/ "uname",
/*123*/ "modify_ldt",
/*124*/ "adjtimex",
/*125*/ "mprotect",
/*126*/ "sigprocmask",
/*127*/ "create_module",
/*128*/ "init_module",
/*129*/ "delete_module",
/*130*/ "get_kernel_syms",
/*131*/ "quotactl",
/*132*/ "getpgid",
/*133*/ "fchdir",
/*134*/ "bdflush",
/*135*/ "sysfs",
/*136*/ "personality",
/*137*/ "afs_syscall",
/*138*/ "setfsuid",
/*139*/ "setfsgid",
/*140*/ "_llseek",
/*141*/ "getdents",
/*142*/ "_newselect",
/*143*/ "flock",
/*144*/ "msync",
/*145*/ "readv",
/*146*/ "writev",
/*147*/ "getsid",
/*148*/ "fdatasync",
/*149*/ "_sysctl",
/*150*/ "mlock",
/*151*/ "munlock",
/*152*/ "mlockall",
/*153*/ "munlockall",
/*154*/ "sched_setparam",
/*155*/ "sched_getparam",
/*156*/ "sched_setscheduler",
/*157*/ "sched_getscheduler",
/*158*/ "sched_yield",
/*159*/ "sched_get_priority_max",
/*160*/ "sched_get_priority_min",
/*161*/ "sched_rr_get_interval",
/*162*/ "nanosleep",
/*163*/ "mremap",
/*164*/ "setresuid",
/*165*/ "getresuid",
/*166*/ "vm86",
/*167*/ "query_module",
/*168*/ "poll",
/*169*/ "nfsservctl",
/*170*/ "setresgid",
/*171*/ "getresgid",
/*172*/ "prctl",
/*173*/ "rt_sigreturn",
/*174*/ "rt_sigaction",
/*175*/ "rt_sigprocmask",
/*176*/ "rt_sigpending",
/*177*/ "rt_sigtimedwait",
/*178*/ "rt_sigqueueinfo",
/*179*/ "rt_sigsuspend",
/*180*/ "pread",
/*181*/ "pwrite",
/*182*/ "chown",
/*183*/ "getcwd",
/*184*/ "capget",
/*185*/ "capset",
/*186*/ "sigaltstack",
/*187*/ "sendfile",
/*188*/ "getpmsg",
/*189*/ "putpmsg",
/*190*/ "vfork",
/*191*/ "ugetrlimit",
/*192*/ "mmap2",
/*193*/ "truncate64",
/*194*/ "ftruncate64",
/*195*/ "stat64",
/*196*/ "lstat64",
/*197*/ "fstat64",
/*198*/ "lchown32",
/*199*/ "getuid32",
/*200*/ "getgid32",
/*201*/ "geteuid32",
/*202*/ "getegid32",
/*203*/ "setreuid32",
/*204*/ "setregid32",
/*205*/ "getgroups32",
/*206*/ "setgroups32",
/*207*/ "fchown32",
/*208*/ "setresuid32",
/*209*/ "getresuid32",
/*210*/ "setresgid32",
/*211*/ "getresgid32",
/*212*/ "chown32",
/*213*/ "setuid32",
/*214*/ "setgid32",
/*215*/ "setfsuid32",
/*216*/ "setfsgid32",
/*217*/ "pivot_root",
/*218*/ "mincore",
/*219*/ "madvise",
/*220*/ "getdents64",
/*221*/ "fcntl64",
/*222*/ "NYI",
/*223*/ "security",
/*224*/ "sys_gettid",
/*225*/ "readahead",
/*226*/ "setxattr",
/*227*/ "lsetxattr",
/*228*/ "fsetxattr",
/*229*/ "getxattr",
/*230*/ "lgetxattr",
/*231*/ "fgetxattr",
/*232*/ "listxattr",
/*233*/ "llistxattr",
/*234*/ "flistxattr",
/*235*/ "removexattr",
/*236*/ "lremovexattr",
/*237*/ "fremovexattr",
/*238*/ "tkill",
/*239*/ "sendfile64",
/*240*/ "futex",
/*241*/ "sched_setaffinity",
/*242*/ "sched_getaffinity",
/*243*/ "set_thread_area",
/*244*/ "get_thread_area",
/*245*/ "io_setup",
/*246*/ "io_destroy",
/*247*/ "io_getevents",
/*248*/ "io_submit",
/*249*/ "io_cancel",
/*250*/ "fadvise64",
/*251*/ "free_hugepages",
/*252*/ "exit_group",
/*253*/ "lookup_dcookie",
/*254*/ "epoll_create",
/*255*/ "epoll_ctl",
/*256*/ "epoll_wait",
/*257*/ "remap_file_pages",
/*258*/ "set_tid_address",
/*259*/ "timer_create",
/*260*/ "timer_settime",
/*261*/ "timer_gettime",
/*262*/ "timer_getoverrun",
/*263*/ "timer_delete",
/*264*/ "clock_settime",
/*265*/ "clock_gettime",
/*266*/ "clock_getres",
/*267*/ "clock_nanosleep",
/*268*/ "NYI",
/*269*/ "fstatfs64",
/*270*/ "tgkill",
/*271*/ "utimes",
/*272*/ "fadvise64_64",
/*273*/ "NYI",
/*274*/ "mbind",
/*275*/ "get_mempolicy",
/*276*/ "set_mempolicy",
/*277*/ "mq_open",
/*278*/ "mq_unlink",
/*279*/ "mq_timedsend",
/*280*/ "mq_timedreceive",
/*281*/ "mq_notify",
/*282*/ "mq_getsetattr",
/*283*/ "NYI",
/*284*/ "waitid",
/*285*/ "setaltroot(not support)",
/*286*/ "add_key",
/*287*/ "request_key",
/*288*/ "keyctl",
/*289*/ "ioprio_set",
/*290*/ "ioprio_get",
/*291*/ "inotify_init",
/*292*/ "inotify_add_watch",
/*293*/ "inotify_rm_watch",
/*294*/ "migrate_pages(not support)",
/*295*/ "openat",
/*296*/ "mkdirat",
/*297*/ "mknodat",
/*298*/ "fchownat",
/*299*/ "futimesat",
/*300*/ "fstatat64",
/*301*/ "unlinkat",
/*302*/ "renameat",
/*303*/ "linkat",
/*304*/ "symlinkat",
/*305*/ "readlinkat",
/*306*/ "fchmodat",
/*307*/ "faccessat",
/*308*/ "pselect",
/*309*/ "ppoll",
/*310*/ "unshare",
/*311*/ "set_robust_list",
/*312*/ "get_robust_list",
/*313*/ "splice",
/*314*/ "sync_file_range",
/*315*/ "tee",
/*316*/ "vmsplice",
/*317*/ "move_pages",
/*318*/ "getcpu",
/*319*/ "epoll_pwait",
/*320*/ "utimensat",
/*321*/ "signalfd",
/*322*/ "timerfd_create",
/*323*/ "eventfd",
/*324*/ "fallocate",
/*325*/ "timerfd_settime",
/*326*/ "timerfd_gettime",
};

#define SYSCALL_NAMES_MAX_ID ((int)CONVERS_TAB_MAX_INDEX)
#endif	// DEBUG

/* function prototypes */
static long ia32_wait (BTGENERIC_IA32_CONTEXT * bc32);
static long ia32_waitid (BTGENERIC_IA32_CONTEXT *bc32);
long ia32_system_call (BTGENERIC_IA32_CONTEXT * bc32, void *ia64p);
static int need_deliver_pending (int syscall_id);
static long btlib_readlink (BTGENERIC_IA32_CONTEXT * bc32);
//static long btlib_getcwd (BTGENERIC_IA32_CONTEXT * bc32);
//static long btlib_mincore (BTGENERIC_IA32_CONTEXT * bc32);
extern char g_has_smc;
static long btlib_smc_wrapper (unsigned long start, unsigned long len, BTGENERIC_IA32_CONTEXT * bc32);
static long btlib_ngetxattr(BTGENERIC_IA32_CONTEXT* bc32);
static long btlib_nlistxattr(BTGENERIC_IA32_CONTEXT* bc32); 
static long btlib_personality (BTGENERIC_IA32_CONTEXT * bc32);
#if DEBUG
static void UNUSED print_syscall_before (int syscall_id);
static void UNUSED print_syscall_after (int syscall_id, long retval);
static void UNUSED print_socketcall_info (BTGENERIC_IA32_CONTEXT * bc32);
static void UNUSED print_getdents (BTGENERIC_IA32_CONTEXT * bc32);
static void UNUSED print_ipc_info (BTGENERIC_IA32_CONTEXT * bc32);
static void UNUSED print_fcntl_info (BTGENERIC_IA32_CONTEXT * bc32);
#endif

/* array with bit field entry for each syscall specifying:
LONG_SYSCALL_ATTRIB         bit#0 set
NEED_DELIVER_PENDING_ATTRIB bit#1 set
NEED_RESTART_ATTRIB         bit#2 set
*/
unsigned char syscall_attrib[MAX_IA32_SYSCALL+1];


static int
need_deliver_pending (int syscall_id)
{
	/* for sigaction/rt_sigaction/signal cases, we'd better check the 
	   sighandler also, but not checking doesn't do harm */
	if ((syscall_id == IA32_SYSCALL_SIGPROCMASK) ||
	    (syscall_id == IA32_SYSCALL_RT_SIGPROCMASK) ||
	    (syscall_id == IA32_SYSCALL_SIGACTION) ||
	    (syscall_id == IA32_SYSCALL_RT_SIGACTION) ||
	    (syscall_id == IA32_SYSCALL_SIGNAL)) {
		return 1;
	} else {
		return 0;
	}
}


/*****************************************************************************/
/* initialize the syscall_attrib array for each syscall
*/
/*****************************************************************************/
static void
init_syscall_attribs ()
{
	unsigned long i;

	for (i=0; i<=MAX_IA32_SYSCALL; i++) {
		syscall_attrib[i] = 0;
		if (is_long_syscall((int)i)) {
			syscall_attrib[i] |= LONG_SYSCALL_ATTRIB;
		}
		if (need_deliver_pending((int)i)) {
			syscall_attrib[i] |= NEED_DELIVER_PENDING_ATTRIB;
		}
		if (syscall_need_restart((int)i)) {
			syscall_attrib[i] |= NEED_RESTART_ATTRIB;
		}
	}
}

/*****************************************************************************/
/* call the wrapper syscall for IA-32 syscalls
 * input  : bc32 - the IA-32 context of the thread
 *          ia64p - btlib may pass pointers above 4G for some of the syscalls.
 * output : the return value of the syscall
 * remarks: in case of long syscalls, the syscall_id flag
 *          of the thread is set before calling the wrapper
 *          and cleared when returning from it.
 *
 *          ALL LONG SYSCALLS MUST GO THROUGH THIS FUNCTION
 */
/*****************************************************************************/
long
ia32_system_call (BTGENERIC_IA32_CONTEXT * bc32, void *ia64p)
{
	unsigned long param2;
	int syscall_id;
	long retval;
	sigset32_t mask, oldmask;

	assert (bc32 != NULL);

	syscall_id = cur_thread_data->syscall_id;
#if 0
if(syscall_id == 324) {
	fprintf(stdout, "sys_fallocate - fd=%d mode=%x arg3=%lx arg4=%lx, arg5=%lx arg6=%lx\n",
		IA32_SYSCALL_PARAM1(bc32),
		IA32_SYSCALL_PARAM2(bc32),
		IA32_SYSCALL_PARAM3(bc32),
		IA32_SYSCALL_PARAM4(bc32),
		IA32_SYSCALL_PARAM5(bc32),
		IA32_SYSCALL_PARAM6(bc32));
}
#endif
	/* wait syscalls may want to pass a status parameter which is
	   not IA-32 address */
	param2 = (((syscall_id == IA32_SYSCALL_WAITPID) ||
		   (syscall_id == IA32_SYSCALL_WAIT4)) && (ia64p != NULL)) ?
			(unsigned long) ia64p : (unsigned long) IA32_SYSCALL_PARAM2 (bc32);

	/*disable smc opt if necessary */
	if (g_has_smc == 1) {
		if (syscall_id == IA32_SYSCALL_PREAD || syscall_id == IA32_SYSCALL_LOOPUP_DCOOKIE) {
			btlib_disable_smc_opt(IA32_SYSCALL_PARAM2(bc32), IA32_SYSCALL_PARAM3(bc32));
		} else if (syscall_id == IA32_SYSCALL_MQ_TIMEDRECEIVE) {
			btlib_disable_smc_opt(IA32_SYSCALL_PARAM2(bc32), IA32_SYSCALL_PARAM3(bc32));
			btlib_disable_smc_opt(IA32_SYSCALL_PARAM4(bc32), sizeof(unsigned int));
		} else if (syscall_id == IA32_SYSCALL_TIMER_CREATE) {
			btlib_disable_smc_opt(IA32_SYSCALL_PARAM3(bc32), sizeof(timer_t));
		} else if (syscall_id == IA32_SYSCALL_SYSLOG) {
			int type = IA32_SYSCALL_PARAM1(bc32);
			if (type == 2 || type ==3 || type == 4) {
				btlib_disable_smc_opt(IA32_SYSCALL_PARAM2(bc32), IA32_SYSCALL_PARAM3(bc32));
			}
		} else if (syscall_id == IA32_SYSCALL_IO_GETEVENTS) {
			btlib_disable_smc_opt(IA32_SYSCALL_PARAM4(bc32), sizeof(struct io_event) * IA32_SYSCALL_PARAM3(bc32));
		} else if (syscall_id == IA32_SYSCALL_IO_CANCEL) {
			struct iocb buff, * ptr = (struct iocb*)U2P(IA32_SYSCALL_PARAM2(bc32)); 

			if (ptr != NULL && safe_memcpy(&buff, ptr, sizeof(struct iocb)) < 0) {
				return -EFAULT;
			}
			btlib_disable_smc_opt(buff.aio_buf, buff.aio_nbytes);
		} else if (syscall_id == IA32_SYSCALL_IO_SUBMIT) {
			struct iocb buff, * ptr;
			long i, nr;
			unsigned *iocbpp;
			__u32 addr32;

			nr = (long)IA32_SYSCALL_PARAM2(bc32); 
			iocbpp = (unsigned *)U2P(IA32_SYSCALL_PARAM3(bc32));
			if (nr < 0) {
				return -EINVAL;
			}
	
			// check the access right of iocbpp
			if (!access_ok((unsigned long)iocbpp, (unsigned long)(nr * sizeof(addr32)))){
				return -EFAULT;
			}
			for (i = 0; i < nr; i++) {
				if (safe_memcpy (&addr32, iocbpp, sizeof (addr32)) < 0) {
					return -EFAULT;
				} 
				ptr = (struct iocb *)U2P(addr32);
				if (safe_memcpy(&buff, ptr, sizeof(struct iocb)) < 0) {
					return -EFAULT;
				}
				btlib_disable_smc_opt(buff.aio_buf, buff.aio_nbytes) ;
				iocbpp++;
			}
		} else if (syscall_id == IA32_SYSCALL_MOVE_PAGES)  {
			btlib_disable_smc_opt(IA32_SYSCALL_PARAM5(bc32), sizeof(int) * IA32_SYSCALL_PARAM2(bc32));
		}

	}//end g_has_smc == 1

	/* for sigsuspend, signal handler uses signal mask specified by the parameter */
	if (syscall_id == IA32_SYSCALL_SIGSUSPEND) {
		if (safe_memcpy (&mask.sig[0], U2P (IA32_SYSCALL_PARAM3 (bc32)),
				 sizeof (mask.sig[0])) < 0) {
			return -EFAULT;
		}

		mask.sig[1] = 0;
	} else if (syscall_id == IA32_SYSCALL_RT_SIGSUSPEND) {
		if (IA32_SYSCALL_PARAM2 (bc32) != sizeof (sigset32_t)) {
			return -EINVAL;
		}

		if (safe_memcpy (&mask, U2P (IA32_SYSCALL_PARAM1 (bc32)),
				 sizeof (mask)) < 0) {
			return -EFAULT;
		}
	}

	/* for sigsuspend, check signal queue first with new signal mask */
	if ((syscall_id == IA32_SYSCALL_SIGSUSPEND) ||
	    (syscall_id == IA32_SYSCALL_RT_SIGSUSPEND)) {

		sigdelset32 (&mask, SIGKILL);
		sigdelset32 (&mask, SIGSTOP);

		oldmask = cur_thread_data->sigmask;
		cur_thread_data->sigmask = mask;

		if (!btlib_que_is_empty (cur_thread_data->pending_que) ||
		    !btlib_que_is_empty (btl_shared_que)) {
			long *lsf = NULL;

			iti_long_syscall_interface (&lsf, NULL);
			*lsf = NOT_IN_SYSCALL;
			cur_thread_data->bc32->eax = (unsigned) -EINTR;
			restart_btg (NULL, &oldmask);
			assert (!"should not get here");
		}
	}

      repeat_call:
	/* call the IA-32 syscall using the 64 bit wrapper */
	if (syscall_id == IA32_SYSCALL_SOCKETCALL) {
		retval = btlib_socketcall(cur_thread_data->bc32);
	} else if (syscall_id == IA32_SYSCALL_READV) {
		retval = btlib_readv(cur_thread_data->bc32);
	} else if (syscall_id == IA32_SYSCALL_WRITEV) {
		retval = btlib_writev(cur_thread_data->bc32);
	} else if (syscall_id == IA32_SYSCALL_SPLICE) {
                retval = btlib_splice(cur_thread_data->bc32);
	} else if (syscall_id == IA32_SYSCALL_VMSPLICE) {
		retval = btlib_vmsplice(cur_thread_data->bc32);
	} else if (syscall_id == IA32_SYSCALL_FUTEX) {
                retval = btlib_futex(cur_thread_data->bc32);
	} else {
		retval = ILP32_WRAPPER6 ((unsigned long) syscall_id,
				 (unsigned long) IA32_SYSCALL_PARAM1 (bc32),
				 param2,
				 (unsigned long) IA32_SYSCALL_PARAM3 (bc32),
				 (unsigned long) IA32_SYSCALL_PARAM4 (bc32),
				 (unsigned long) IA32_SYSCALL_PARAM5 (bc32),
				 (unsigned long) IA32_SYSCALL_PARAM6 (bc32));
	}
	if (retval == -EINTR) {
		int signo;

		assert (HAS_LONG_SYSCALL_ATTRIB (syscall_id));
		assert (cur_thread_data->tid == SYSCALL (gettid, ()));
		DBPRINT (btl_prt_syscall, "interrupted long syscall\n");

		/* send the signal */
		if (!btlib_que_is_real_empty (cur_thread_data->pending_que)) {
			/* for signal with permanent signal handler and blocked, 
			   the EINTR is ont a real one */
			signo = btlib_que_lastsig (cur_thread_data->pending_que);

			if (has_permanent_handler (signo) &&
			    sigismember32 (&cur_thread_data->sigmask, signo)) {
				goto repeat_call;
			}

			*cur_thread_data->long_syscall_flag = NOT_IN_SYSCALL;
			if ((syscall_id == IA32_SYSCALL_SIGSUSPEND) ||
			    (syscall_id == IA32_SYSCALL_RT_SIGSUSPEND)) {
				restart_btg (NULL, &oldmask);
			} else {
				restart_btg (NULL, NULL);
			}
			assert (!"should not get here");
		}
	}
	DBPRINT (0, "return %d\n", retval);
	return retval;
}

/*****************************************************************************/
/* implement the IA-32 wait4 and waitpid syscalls */
/*****************************************************************************/
static long
ia32_wait (BTGENERIC_IA32_CONTEXT * bc32)
{
	int child_pid = 0;
	int *orig_statusp = NULL;
	int pid, options;
	int status;

	/* get the syscall params */
	pid = (int) IA32_SYSCALL_PARAM1 (bc32);
	orig_statusp = U2P (IA32_SYSCALL_PARAM2 (bc32));
	options = (int) IA32_SYSCALL_PARAM3 (bc32);

 
#if PTEMU_ENABLED
      repeat_waiting:
#endif
	DBPRINT (btl_prt_syscall, "pid=%d status=%p options=%x\n",
		 pid, &status, options);

	child_pid = ia32_system_call ((BTGENERIC_IA32_CONTEXT *) bc32, &status);

	/* see whether the syscall completed */
	DBPRINT (btl_prt_syscall, "lsf=%ld\n",
		 *cur_thread_data->long_syscall_flag);
	if (*cur_thread_data->long_syscall_flag != SYSCALL_COMPLETED) {
		goto out;
	}

	DBPRINT (btl_prt_pt, "ia32_system_call=%d status=%d\n",
		 child_pid, status);
#if PTEMU_ENABLED
	if (child_pid > 0) {
		long retval;
		/* 
		 * do the work for the ptrace emulation. btlib_ptemu_wait
		 * returns with either FINISH_WAIT or DISCARD_WAIT. 
		 * in case of FINISH_WAIT there is nothing special to do. 
		 * in case of DISCARD_WAIT we need to pretend as if the wait
		 * syscall did not return any info. so - if the syscall is blocking
		 * we just repeat it. if it is non-blocking then we change the 
		 * return-value to 0.
		 */
		retval = btlib_ptemu_wait (child_pid, &status);
		DBPRINT (btl_prt_syscall, "btlib_ptemu_wait=%s\n",
			 (retval == DISCARD_WAIT) ? "DISCARD" : "FINISH");
		if (retval == DISCARD_WAIT) {
			DBPRINT (btl_prt_syscall, "WNOHANG=%d\n",
				 (options & WNOHANG) ? 1 : 0);

			if (options & WNOHANG) {
				child_pid = 0;
			} else {
				*cur_thread_data->long_syscall_flag = AFTER_CHECK_QUE;
				goto repeat_waiting;
			}
		}
	}
#endif

#if DEBUG
	if (btl_prt_thread | btl_prt_syscall) {
		if (child_pid > 0) {
			if (__WIFEXITED (status)) {
				DBPRINT (TRUE, "exit %d\n",
					 __WEXITSTATUS (status));
			} else if (__WIFSIGNALED (status)) {
				DBPRINT (TRUE, "kill signal %d\n",
					 __WTERMSIG (status));
			} else { 
				DBPRINT (TRUE, "stop signal %d\n", __WSTOPSIG (status)); 
			} 
		} 
	} 
#endif /* if we are sure that the thread terminated then cleanup its resources */
	if (child_pid > 0) {
		g_safe_memcpy(orig_statusp, &status, sizeof(status));

		if (!__WIFSTOPPED (status)) {
			btlib_cleanup_thread (child_pid);
		}
	}
      out:
	return child_pid;
}

/*****************************************************************************/
/* implement the IA-32 waitid syscall */
/*****************************************************************************/
static long
ia32_waitid (BTGENERIC_IA32_CONTEXT *bc32)
{
	int child_pid = 0;
	struct siginfo32 *orig_infop = NULL;
	struct siginfo32 orig_info;
	int res, pid, options, status;

	/* get the syscall params */
	pid = (int) IA32_SYSCALL_PARAM2 (bc32);
	orig_infop = U2P (IA32_SYSCALL_PARAM3 (bc32));
	options = (int) IA32_SYSCALL_PARAM4 (bc32);

	if (orig_infop != NULL)
		g_safe_memcpy(&orig_info, orig_infop, sizeof(orig_info));

#if PTEMU_ENABLED
repeat_waiting:
#endif
	DBPRINT (btl_prt_syscall, "pid=%d options=%x\n", pid, options);

	res = ia32_system_call ((BTGENERIC_IA32_CONTEXT *) bc32, NULL);

	/* see whether the syscall completed */
	DBPRINT (btl_prt_syscall, "lsf=%ld\n", *cur_thread_data->long_syscall_flag);
	if (*cur_thread_data->long_syscall_flag != SYSCALL_COMPLETED) {
		goto out;
	}

	/* if orig_infop is NULL and the call succeeded, then the child pid
	 * is the return value */
	if (orig_infop == NULL) {
		if (res > 0) {
			child_pid = res;
		} else {
			child_pid = 0;
			status = 0;
		}
	} else if (safe_memcpy(&child_pid, &(orig_infop->_sifields._sigchld._pid), sizeof(child_pid)) < 0 ||
		   safe_memcpy(&status, &(orig_infop->_sifields._sigchld._status), sizeof(status)) < 0){
		return -EFAULT;
	}

	DBPRINT (btl_prt_pt, "ia32_system_call=%d status=%d\n", 
		 res, status);


#if PTEMU_ENABLED

	if (child_pid > 0) {
		long retval;
		/* 
		 * do the work for the ptrace emulation. btlib_ptemu_wait
		 * returns with either FINISH_WAIT or DISCARD_WAIT. 
		 * in case of FINISH_WAIT there is nothing special to do. 
		 * in case of DISCARD_WAIT we need to pretend as if the wait
		 * syscall did not return any info. so - if the syscall is blocking
		 * we just repeat it. if it is non-blocking then we change the 
		 * return-value to 0.
		 */
		retval = btlib_ptemu_wait (child_pid, &status);
		DBPRINT (btl_prt_syscall, "btlib_ptemu_wait=%s\n",
			 (retval == DISCARD_WAIT) ? "DISCARD" : "FINISH");
		if (retval == DISCARD_WAIT) {
			DBPRINT (btl_prt_syscall, "WNOHANG=%d\n", (options & WNOHANG) ? 1 : 0);
	    
			if (options & WNOHANG) {
				res = child_pid = 0;
				if (orig_infop != NULL)
					g_safe_memcpy(orig_infop, &orig_info, sizeof(orig_info));
			} else {
				*cur_thread_data->long_syscall_flag = AFTER_CHECK_QUE;	
				goto repeat_waiting;
			}
	  	}
	}
#endif

#if DEBUG

	  if (btl_prt_thread | btl_prt_syscall) {
		if (child_pid > 0) {
			if (options & WEXITED) {
				DBPRINT (TRUE, "exit %d\n", status);
			} else if (options & WSTOPPED) {
				DBPRINT (TRUE, "stop signal %d\n", status);
			} else {
				DBPRINT (TRUE, "kill signal %d\n", status);
			}
		}
	  }
#endif
	/* if we are sure that the thread terminated then cleanup its resources */
	if ((options & WEXITED) && (child_pid > 0)) {
		BTGENERIC_THREAD_ABORTED (child_pid);
		btlib_cleanup_thread (child_pid);
	}

out:
	return res;
}
/*****************************************************************************/
/*
 * btlib_system_call
 * Remarks: most of the syscalls are delegated to the system  - 
 *          this function calls ia32_system_call, which calls
 *          ilp32_wrapper. but some of the syscall are handled
 *          internally by btlib. 
 *          as for now, all long syscalls are delegated to the system.
 *          if this will change - the function which handles the long
 *          syscall will have to take care of interrupted syscalls, the
 *          same way it is done in ia32_system_call.
 */
/*****************************************************************************/
void
btlib_system_call ()
{
	long retval = -ENOSYS;
	int syscall_id;
	int is_rt = 1;
#if PTEMU_ENABLED
	int ptrace_syscall = 0;
#endif
	syscall_id = IA32_SYSCALL_ID (cur_thread_data->bc32);
	cur_thread_data->in_nonblocking_syscall = !(HAS_LONG_SYSCALL_ATTRIB(syscall_id));
	DBPRINT (btl_prt_syscall, "id-%d(%s) called eip=%x esp=%x\n",
		 syscall_id,
		 (syscall_id <= SYSCALL_NAMES_MAX_ID) ?
		 syscall_names[syscall_id] : "???",
		 cur_thread_data->bc32->eip, cur_thread_data->bc32->esp);

	STATCODE (btl_stats, btlib_update_stat (BTL_STAT_SYSCALL, COUNT, 0ul));

#if DEBUG && 0
	print_syscall_before (syscall_id);
#endif

	/* 
	   make sure the que is empty before invoking
	   1 - long syscalls 
	   2 - system call that may affect the pending que, e.g., 
	   sigaction(SIG_IGN) should remove corresponding pending 
	   signal
	 */

	STATCODE (btl_stats && HAS_LONG_SYSCALL_ATTRIB(syscall_id),
		  btlib_update_stat (BTL_STAT_LSYSCALL, COUNT, 0ul));

	if (HAS_LONG_SYSCALL_ATTRIB(syscall_id) || HAS_NEED_DELIVER_PENDING_ATTRIB (syscall_id)) {
		DBPRINT (btl_prt_syscall, "long syscall\n");
		assert (*cur_thread_data->long_syscall_flag == NOT_IN_SYSCALL);

		*cur_thread_data->long_syscall_flag = CHECK_QUE;
		/* it's enough to check the private que here */
		if (!btlib_que_is_empty (cur_thread_data->pending_que)) {
			DBPRINT (btl_prt_syscall, "signal pending\n");
			*cur_thread_data->long_syscall_flag = NOT_IN_SYSCALL;
			longjmp (*cur_thread_data->btgeneric_run_context, DELIVER);
			assert (!"should not get here");
		}
		*cur_thread_data->long_syscall_flag = AFTER_CHECK_QUE;
	}

	/* advance eip pass the int 0x80 instruction */
	cur_thread_data->syscall_id = cur_thread_data->bc32->eax;
	cur_thread_data->bc32->eax = (unsigned) -ENOSYS;
	cur_thread_data->bc32->eip += 2;

#if PTEMU_ENABLED
	/* 
	 * if there is a ptrace_syscall request then suspend yourself 
	 * in IA-32 systems the os prepares things as follows - 
	 * 1) syscall_id is kept in orig_eax
	 * 2) eax holds -ENOSYS
	 * 3) the debugee is stopped with sigtrap
	 * 4) if the option TRACESYSGOOD is set then the signal number 
	 *    is ored with 0x80
	 *
	 * we try to emulate all this stuff here. the orig_eax is 
	 * kept in the tls. 
	 */
	if ((get_cur_thread_ptrace () & PTEMU_PTRACED) &&
	    (get_cur_thread_ptrace () & PTEMU_SYSCALL)) {

		ptrace_syscall = 1;

		/* put SIGTRAP in the tia_signal and suspend yourself
		   by sending PTEMU_SIG. this will be caught by 
		   btlib of the debugger */
		bt_memset (&cur_thread_data->tia_signal, 0,
			   sizeof (cur_thread_data->tia_signal));
		cur_thread_data->tia_signal.siginfo.si_signo = SIGTRAP;
		cur_thread_data->tia_signal.siginfo.si_signo |= IA32_SYSCALL_INT;

		DBPRINT (btl_prt_pt,
			 "PTEMU_SYSCALL syscall_id=%d si_signo=%d\n",
			 syscall_id,
			 cur_thread_data->tia_signal.siginfo.si_signo);

		retval = btlib_ptemu_sig0_selfkill (TRUE);
		sys_err_chk (retval, "kill");
	}
#endif	// PTEMU_ENABLED

	switch (syscall_id) {
	case IA32_SYSCALL_OPENAT:
	case IA32_SYSCALL_OPEN:
		retval = btlib_open (cur_thread_data->bc32);
		break;
	case IA32_SYSCALL_READ:
		retval = btlib_read (cur_thread_data->bc32);
		break;
	case IA32_SYSCALL_LSEEK:
	case IA32_SYSCALL_LLSEEK:
		retval = btlib_lseek (cur_thread_data->bc32);
		break;
	case IA32_SYSCALL_VFORK:
		/* TODO - add special handling for vfork */
		/* fall through */
	case IA32_SYSCALL_FORK:
		retval = btlib_fork (cur_thread_data->bc32);
		break;
	case IA32_SYSCALL_READLINK:
		retval = btlib_readlink (cur_thread_data->bc32);
		break;
	case IA32_SYSCALL_CLONE:
		retval = btlib_clone (cur_thread_data->bc32);
		break;
	case IA32_SYSCALL_MMAP:
		retval = btlib_ia32_mmap (cur_thread_data->bc32);
		break;
	case IA32_SYSCALL_MMAP2:
		retval = btlib_ia32_mmap2 (cur_thread_data->bc32);
		break;
	case IA32_SYSCALL_MUNMAP:
		retval = btlib_ia32_munmap (cur_thread_data->bc32);
		break;
	case IA32_SYSCALL_MREMAP:
		retval = btlib_ia32_mremap (cur_thread_data->bc32);
		break;
	case IA32_SYSCALL_EXIT_GROUP:
		btlib_group_exit (cur_thread_data->bc32);
		break;
	case IA32_SYSCALL_EXIT:
		btlib_thread_exit (cur_thread_data->bc32);
		break;
	case IA32_SYSCALL_SIGACTION:
		retval = btlib_sigaction (cur_thread_data->bc32);
		break;
	case IA32_SYSCALL_RT_SIGACTION:
		retval = btlib_rt_sigaction (cur_thread_data->bc32);
		break;
	case IA32_SYSCALL_SIGNAL:
		retval = btlib_signal (cur_thread_data->bc32);
		break;
	case IA32_SYSCALL_SIGALTSTACK:
		retval = btlib_sigaltstack (cur_thread_data->bc32);
		break;
	case IA32_SYSCALL_SIGRETURN:
		is_rt = 0;
		/* fall through */
	case IA32_SYSCALL_RT_SIGRETURN:
		btlib_sigreturn (cur_thread_data->bc32, is_rt);
		break;
	case IA32_SYSCALL_MPROTECT:
		retval = btlib_ia32_mprotect (cur_thread_data->bc32);
		break;
	case IA32_SYSCALL_WAITPID:
	case IA32_SYSCALL_WAIT4:
		retval = ia32_wait (cur_thread_data->bc32);
		break;
	case IA32_SYSCALL_WAITID:
	        retval = ia32_waitid (cur_thread_data->bc32);
		break;
	case IA32_SYSCALL_PERSONALITY:
		retval = btlib_personality (cur_thread_data->bc32);
		break;
	case IA32_SYSCALL_MODIFY_LDT:
		retval = btlib_modify_ldt (cur_thread_data->bc32);
		break;
	case IA32_SYSCALL_PTRACE:
#ifdef PTEMU_ENABLED
		retval = btlib_ptrace (cur_thread_data->bc32);
		break;
#endif
	case IA32_SYSCALL_SET_THREAD_AREA:
		retval = btlib_set_thread_area (cur_thread_data->bc32);
		break;
	case IA32_SYSCALL_GET_THREAD_AREA:
		retval = btlib_get_thread_area (cur_thread_data->bc32);
		break;
	case IA32_SYSCALL_SET_TID_ADDRESS:
		retval = btlib_set_tid_address (cur_thread_data->bc32);
		break;
	case IA32_SYSCALL_USELIB:
		assert (!"unimplemented syscall");
		retval = -ENOSYS;
		break;
#if ALL_THREAD_AFFINITY
	case IA32_SYSCALL_SCHED_SETAFFINITY:
		/* if IA-32 thread explicitly calls setaffinity()
		   we'll pass the call to the wrapper, and if it succeeds,
		   we record the new explicitly-set mask so it will be inherited
		   correctly in case of fork or clone. */
		if (btl_narrowed_affinity) {
			/* in case we applied affinity solution and one of the threads calls
			sched_setaffinity, we have to apply ordering solution*/
			DBPRINT(btl_prt_thread, 
				"MT app called sched_setaffinity, applying shared memory solution\n");
			apply_shared_memory_ordering_solution();
		}
		retval = ia32_system_call (cur_thread_data->bc32, NULL);
		if (!IS_SYS_ERROR (retval)) {
			/* param1 of the call is target_tid */
			btlib_record_explicit_affinity_request (
				(int)IA32_SYSCALL_PARAM1 (cur_thread_data-> bc32));
		}
		break;
#endif	/* ALL_THREAD_AFFINITY */

	case IA32_SYSCALL_IPC:
#if SHMEM_ORDER_SOLUTION
		/* ipc call of type (first param) "SHMAT", is a request for shared memory.
		   In this case we have to apply the shared-memory-process-affinity solution */
		if ((int) IA32_SYSCALL_PARAM1 (cur_thread_data->bc32) == SHMAT) {
			DBPRINT (btl_prt_mem || btl_prt_thread,
				 "ipc called with flag SHMAT, Applying shared memory solution\n");
			apply_shared_memory_ordering_solution ();
		}
#endif
		/* now call the default handler of sys-calls */
		retval = btlib_ipc(cur_thread_data->bc32);
		break;
	case IA32_SYSCALL_SIGPROCMASK:
		is_rt = 0;
		/* fall through */
	case IA32_SYSCALL_RT_SIGPROCMASK:
		retval = btlib_sigprocmask (cur_thread_data->bc32, is_rt);
		break;
	case IA32_SYSCALL_SIGPENDING:
		is_rt = 0;
		/* fall through */
	case IA32_SYSCALL_RT_SIGPENDING:
		retval = btlib_sigpending (cur_thread_data->bc32, is_rt);
		break;
	case IA32_SYSCALL_CLOSE:
		retval = btlib_close (cur_thread_data->bc32);
		if (!IS_SYS_ERROR (retval) && m_hugetlbfs_exists) {
			set_fd_hugetlb_status (
				(int)IA32_SYSCALL_PARAM1 (cur_thread_data->bc32), FD_UNKNOWN);
		}
		break;
	case IA32_SYSCALL_EXECVE:
		btlib_terminate_all_native_thread();
		retval = ia32_system_call (cur_thread_data->bc32, NULL);
		/* if execve failed, native thread will not start again, 
		 * it will not be removed from native thread list 
		 */
		break;
	case IA32_SYSCALL_PSELECT:
	case IA32_SYSCALL_PPOLL:
	case IA32_SYSCALL_EPOLL_PWAIT:
		retval = btlib_syscall_with_sigmask(cur_thread_data->bc32);
		break; 

	/* smc mechansim 
	 * check the pointer passed to kernel
	 */

	/* REDO syscalls*/
	case IA32_SYSCALL_GETCWD:
		retval = btlib_smc_wrapper(
			(unsigned long)IA32_SYSCALL_PARAM1(cur_thread_data->bc32), 
			(unsigned long)IA32_SYSCALL_PARAM2(cur_thread_data->bc32), 
			cur_thread_data->bc32);
		break;
	case IA32_SYSCALL_MINCORE:
	{
		#define VECTOR_SIZE(length)  (((length) + PAGE_SIZE - 1) / PAGE_SIZE)
		retval = btlib_smc_wrapper(
			(unsigned long)IA32_SYSCALL_PARAM3(cur_thread_data->bc32), 
			VECTOR_SIZE((unsigned long)(IA32_SYSCALL_PARAM2(cur_thread_data->bc32))), 
			cur_thread_data->bc32);
	}
		break;
	case IA32_SYSCALL_SCHED_GETAFFINITY:
		retval = btlib_smc_wrapper(
			(unsigned long)IA32_SYSCALL_PARAM2(cur_thread_data->bc32),
			(unsigned long)sizeof(struct sched_param),
			cur_thread_data->bc32);
		break;
	case IA32_SYSCALL_GETXATTR:
	case IA32_SYSCALL_LGETXATTR:
	case IA32_SYSCALL_FGETXATTR:
		retval = btlib_ngetxattr(cur_thread_data->bc32);
		break;
	case IA32_SYSCALL_LISTXATTR:
	case IA32_SYSCALL_LLISTXATTR:
	case IA32_SYSCALL_FLISTXATTR:
		retval = btlib_nlistxattr(cur_thread_data->bc32);
		break;
	/*DISABLE smc optimization on related page*/
	case IA32_SYSCALL_SYSCTL:
		retval = btlib_sysctl(cur_thread_data->bc32);
		break;
	default:
		retval = ia32_system_call (cur_thread_data->bc32, NULL);
		break;
	}

	DBPRINT (btl_prt_syscall, "id=%d(%s) - returned %lx(%ld)\n",
		 syscall_id,
		 (syscall_id <= SYSCALL_NAMES_MAX_ID) ?
		 syscall_names[syscall_id] : "???", retval, retval);

	DBPRINT (btl_prt_syscall && retval == -EFAULT, "SMC id=%d arg1=%x arg2=%x arg3=%x arg4=%x arg5=%x arg6=%x\n", 
		syscall_id,
		IA32_SYSCALL_PARAM1 (cur_thread_data->bc32),
		IA32_SYSCALL_PARAM2 (cur_thread_data->bc32),
		IA32_SYSCALL_PARAM3 (cur_thread_data->bc32),
		IA32_SYSCALL_PARAM4 (cur_thread_data->bc32),
		IA32_SYSCALL_PARAM5 (cur_thread_data->bc32),
		IA32_SYSCALL_PARAM6 (cur_thread_data->bc32));

	after_syscall_returns (syscall_id, retval, ptrace_syscall);

	restart_btg (NULL, NULL);

	assert (!"should not get here");
}

void after_syscall_returns (int syscall_id, long retval, int ptrace_syscall)
{
	/* check whether the long syscall was executed */
	if (HAS_LONG_SYSCALL_ATTRIB(syscall_id)) {
		long state = *cur_thread_data->long_syscall_flag;
		
		DBPRINT (btl_prt_sig, "id=%d(%s) - long syscall returned %lx(%ld), state = %ld\n",
			 syscall_id,
		 	 (syscall_id <= SYSCALL_NAMES_MAX_ID) ?
		 	 syscall_names[syscall_id] : "???", retval, retval, state);
		
		*cur_thread_data->long_syscall_flag = NOT_IN_SYSCALL;
		if ((state != SYSCALL_COMPLETED) && (state != AFTER_CHECK_QUE)) {
			/* the syscall was not executed. repeat it */
			DBPRINT (btl_prt_sig || btl_prt_syscall, "id=%d(%s) - long syscall returned %lx(%ld), state = %ld, repeat it\n",
			 syscall_id,
		 	 (syscall_id <= SYSCALL_NAMES_MAX_ID) ?
		 	 syscall_names[syscall_id] : "???", retval, retval, state);
			
			cur_thread_data->bc32->eip -= 2;
			cur_thread_data->bc32->eax =
			    cur_thread_data->syscall_id;
			cur_thread_data->syscall_id = IA32_SYSCALL_NONE;
			return;
		}
	}
#if DEBUG && 0
	print_syscall_after (syscall_id, retval);
#endif	// DEBUG

	/* store the return value in eax */
	if ( !cur_thread_data->is_eax_changed ){
		cur_thread_data->bc32->eax = retval;
	}else {
		cur_thread_data->is_eax_changed = 0;
	}

#if PTEMU_ENABLED
	/* 
	 * in case of fork() -
	 * if the child process is ptraced then it should be stopped with SIGSTOP
	 */
	if (((syscall_id == IA32_SYSCALL_FORK) || (syscall_id == IA32_SYSCALL_CLONE)) && 
	    (cur_thread_data->bc32->eax == 0)) {

		if (btl_is_ptraced != IS_PTRACED_NO) {
			DBPRINT (btl_prt_thread | btl_prt_pt, "child ptraced\n");

			DBPRINT (btl_prt_pt, "setting thread ptrace flag\n");
			assert (!g_is_mme_mode);
			*cur_thread_ptrace () = PTEMU_PTRACED;
			cur_thread_data->tia_signal.siginfo.si_signo = SIGSTOP;
			retval = btlib_ptemu_sig0_selfkill (TRUE);
			sys_err_chk (retval, "kill");
		}
	} else if (ptrace_syscall) {
		/* if there is a ptrace_syscall request then suspend yourself */
		DBPRINT (btl_prt_pt, "PTEMU_SYSCALL end syscall\n");
		/* 
		 * put SIGTRAP in the tia_signal and suspend yourself
		 * by sending PTEMU_SIG. this will be caught by 
		 * btlib of the debugger
		 */
		bt_memset (&cur_thread_data->tia_signal, 0,
			   sizeof (cur_thread_data->tia_signal));
		cur_thread_data->tia_signal.siginfo.si_signo = SIGTRAP;
		cur_thread_data->tia_signal.siginfo.si_signo |= IA32_SYSCALL_INT;
		retval = btlib_ptemu_sig0_selfkill (TRUE);
		sys_err_chk (retval, "kill");
	}
#endif
	cur_thread_data->syscall_id = IA32_SYSCALL_NONE;
	*cur_thread_data->long_syscall_flag = NOT_IN_SYSCALL;
}

struct sysctl_args32 {
	__u32 name;
	__s32 nlen;
	__u32 oldval;
	__u32 oldlenp;
	__u32 newval;
	__u32 newlen;
	__u32 __unused[4];
};

static int
get_timespec (struct timespec32 *ts, struct timespec *ts64)
{
	int err = 0;
	struct timespec32 ts32;

	if (safe_memcpy (&ts32, ts, sizeof (ts32)) < 0) {
		err = -EFAULT;
	}
	ts64->tv_sec = ts32.tv_sec;
	ts64->tv_nsec = ts32.tv_nsec;
	
	return err;
}

static long
btlib_splice(BTGENERIC_IA32_CONTEXT * bc32)
{
	long retval;
	loff_t * off_in, * off_out;
	loff_t in, out, *inptr, *outptr;

	off_in = (loff_t*) U2P(IA32_SYSCALL_PARAM2(cur_thread_data->bc32));
	off_out = (loff_t*) U2P(IA32_SYSCALL_PARAM4(cur_thread_data->bc32));

	inptr = (off_in == NULL)? NULL : &in;
	outptr = (off_out == NULL) ? NULL : &out;

	DBPRINT (0, " - fd_in=%d off_in=%p fd_out=%d off_out=%p len=%d flag=%x\n", 
		(int) IA32_SYSCALL_PARAM1 (bc32), off_in,
                         (int) IA32_SYSCALL_PARAM3 (bc32), off_out,
                         (size_t) IA32_SYSCALL_PARAM5 (bc32),
                         (unsigned int) IA32_SYSCALL_PARAM6 (bc32));

	if (off_in != NULL && safe_memcpy(&in, off_in, sizeof(loff_t)) < 0) {
		inptr = off_in;
		btlib_disable_smc_opt(off_in, sizeof(loff_t));
		DBPRINT(1, "off_in=%p failed to copy\n", off_in);
	}

	if (off_out != NULL && safe_memcpy(&out, off_out, sizeof(loff_t)) < 0) {
		outptr = off_out;
		btlib_disable_smc_opt(off_out, sizeof(loff_t));
		DBPRINT(1, "off_out=%p failed to copy\n", off_out);
	}
	DBPRINT (0, "inptr=%p outptr=%p\n", inptr, outptr);
	retval = SYSCALL (splice, (
                         (int) IA32_SYSCALL_PARAM1 (bc32), inptr, 
                         (int) IA32_SYSCALL_PARAM3 (bc32), outptr,
                         (size_t) IA32_SYSCALL_PARAM5 (bc32),
                         (unsigned int) IA32_SYSCALL_PARAM6 (bc32)));

	if (retval >= 0) {	
		if (inptr == &in && safe_memcpy(off_in, &in, sizeof(loff_t)) < 0) {
			return -EFAULT;
		}
		if (outptr == &out && safe_memcpy(off_out, &out, sizeof(loff_t)) < 0) {
			return -EFAULT;
		}
	}

	return retval;
}


static long 
btlib_futex (BTGENERIC_IA32_CONTEXT * bc32)
{
	__u32 futex, futex2, timeout;
	struct timespec ts64;
	struct timespec *pts64;
	int retval, op, val1, val2;
	int cmd;

	futex = IA32_SYSCALL_PARAM1(bc32);
	op = (int)IA32_SYSCALL_PARAM2(bc32);
	val1 = (int)IA32_SYSCALL_PARAM3(bc32);
	timeout = IA32_SYSCALL_PARAM4(bc32);
	futex2 = IA32_SYSCALL_PARAM5(bc32);
	val2 = (int)IA32_SYSCALL_PARAM6(bc32);

	cmd = op & FUTEX_CMD_MASK;

	pts64 = (struct timespec *)(long)timeout;

	btlib_disable_smc_opt(futex, sizeof(__u32));

	if (cmd == FUTEX_REQUEUE || cmd == FUTEX_CMP_REQUEUE) {
		btlib_disable_smc_opt(futex2, sizeof(__u32));
	}
	else if ((cmd == FUTEX_WAIT) && (U2P(timeout) != NULL)) {
		if (get_timespec ((struct timespec32 *)U2P(timeout), &ts64) < 0) {
			return -EFAULT;
		}
		pts64 = &ts64;	
	}

	retval = LONG_SYSCALL6 (futex, U2P(futex), op, val1, pts64, U2P(futex2), val2);
	
	return retval;
}

static long 
btlib_sysctl (BTGENERIC_IA32_CONTEXT * bc32)
{
	struct sysctl_args32 arg32;
	struct __sysctl_args arg64;
	__u64 olen64 = 0;
	int v, retval;

	if (safe_memcpy((void*)&arg32, U2P(IA32_SYSCALL_PARAM1(bc32)), sizeof(arg32)) < 0)
		return -EFAULT;


	arg64.oldval = U2P(arg32.oldval);
	v = IS_VALID_PTR(arg64.oldval);
	if (v) {
		__u32 olen32;
		if (safe_memcpy(&olen32, U2P(arg32.oldlenp), sizeof(__u32)) < 0)
			return -EFAULT;
		olen64 = (__u64)olen32;
		arg64.oldlenp = &olen64;

		if(g_has_smc) btlib_disable_smc_opt(arg64.oldval, olen32);
	} else
		arg64.oldlenp = NULL;
	
	arg64.name = U2P(arg32.name);
	arg64.nlen = arg32.nlen;
	arg64.newval = U2P(arg32.newval);
	arg64.newlen = (__u64)arg32.newlen;
	bt_memset(&arg64.__unused, 0, sizeof(arg64.__unused));

	retval = SYSCALL (_sysctl, (&arg64));

	if (!retval && v && safe_memcpy(U2P(arg32.oldlenp), &olen64, sizeof(__u32))) {
		return -EFAULT;
	}
	return retval;
}



#define GET_LEGAL_SIZE(addr, size) (access_ok(addr, size) ? size : (USER_SPACE_SIZE - addr))

/*****************************************************************************/
/* implement the readlink syscall - handle reads of /proc/self/exe and the like */
/*****************************************************************************/
static long
btlib_readlink (BTGENERIC_IA32_CONTEXT * bc32)
{
	char *ulinkp = U2P (IA32_SYSCALL_PARAM1 (bc32));
	char *uret = U2P (IA32_SYSCALL_PARAM2 (bc32));
	unsigned long usize = IA32_SYSCALL_PARAM3 (bc32);

	char link[PATH_MAX + 2];
	long ret;

	ret = safe_stringcopy (link, ulinkp, sizeof (link) - 1);
	if (IS_SYS_ERROR (ret)) {
		ret = -EFAULT;
		goto out;
	}
	if (ret > PATH_MAX) {
		ret = -ENAMETOOLONG;
		goto out;
	}
	link[ret] = '\0';
	DBPRINT (0, "link='%s'\n", link);

	/* we need to copy full-path of ia32 binary to the user given
	 * buffer if the link is something link /proc/self/exe */
	int need_to_copy = 0;
	const char *proc_self_exe = "/proc/self/exe";
	char proc_pid_exe[PATH_MAX];

	if (bt_strncmp (proc_self_exe, link, bt_strlen (proc_self_exe) + 1) == 0) {
		need_to_copy = 1;
	} else {
		ret = bt_snprintf (proc_pid_exe, sizeof (proc_pid_exe) - 1,
				   "/proc/%d/exe", SYSCALL (getpid, ()));
		assert (ret >= 0 && (unsigned long) ret < sizeof (proc_pid_exe) - 1);
		if (bt_strncmp (proc_pid_exe, link, bt_strlen (proc_pid_exe) + 1) == 0) {
			need_to_copy = 1;
		} else {
			ret = bt_snprintf (proc_pid_exe,
					   sizeof (proc_pid_exe) - 1,
					   "/proc/%d/exe", SYSCALL (gettid, ()));
			assert (ret >= 0 && (unsigned long) ret < sizeof (proc_pid_exe) - 1);
			if (bt_strncmp (proc_pid_exe, link, bt_strlen (proc_pid_exe) + 1) == 0) {
				need_to_copy = 1;
			}
		}
	}
	if (need_to_copy) {
		unsigned long act_size, t1;
		t1 = bt_strnlen (btl_ia32_bin_full_path,
				 sizeof (btl_ia32_bin_full_path) - 1);
		act_size = MIN (t1, usize);
		ret = safe_memcpy (uret, btl_ia32_bin_full_path, act_size);
		ret = IS_SYS_ERROR (ret) ? -EFAULT : (int) act_size;
	} else {
		int valid = 1;
		/* otherwise, just delegate to the kernel */
		while (valid) {
			ret = ILP32_WRAPPER3 (IA32_SYSCALL_READLINK, ulinkp, uret, usize);
			valid = 0;
			if (ret == -EFAULT && g_has_smc) {
				lock_mregion_manager(MREGION_MANAGER_IA32);
				valid = btlib_check_region_prot((unsigned long)uret, (unsigned long)usize, PROT_WRITE);
				unlock_mregion_manager(MREGION_MANAGER_IA32);

				if (valid) { 
					btlib_disable_smc_opt(uret, usize); 
				}
			}

		}
		
	}
      out:
	DBPRINT (0, "ret=%ld\n", ret);
	return ret;
}


/*****************************************************************************/
/* implement the syscalls - redo the system call if it fails due to SMC
 * mechanism  
 * be careful to use the function - 
 * 1. the syscall should not change the kernel status at all
 * 2. the syscall should not be a long syscall
 * 3. only one pointer are passed to kernel in this syscall, its for write only
 */
/*****************************************************************************/
static long 
btlib_smc_wrapper (unsigned long start, unsigned long len, BTGENERIC_IA32_CONTEXT * bc32)
{
	long retval;
	unsigned long syscall_id = (unsigned long) cur_thread_data->syscall_id;

	assert(access_ok(start, 1));
	assert(!HAS_LONG_SYSCALL_ATTRIB(syscall_id));
	retval = ILP32_WRAPPER6 (syscall_id,
		 (unsigned long) IA32_SYSCALL_PARAM1 (bc32),
		 (unsigned long) IA32_SYSCALL_PARAM2 (bc32),
		 (unsigned long) IA32_SYSCALL_PARAM3 (bc32),
		 (unsigned long) IA32_SYSCALL_PARAM4 (bc32),
		 (unsigned long) IA32_SYSCALL_PARAM5 (bc32),
		 (unsigned long) IA32_SYSCALL_PARAM6 (bc32));

	if (retval == -EFAULT && g_has_smc) {
		unsigned long size = GET_LEGAL_SIZE(start, len);

		DBPRINT(size != len, "addr = %lx size=%lx => size=%lx\n", start, len, size);
		btlib_disable_smc_opt(start, size);

		retval = ILP32_WRAPPER6 (syscall_id,
			 (unsigned long) IA32_SYSCALL_PARAM1 (bc32),
			 (unsigned long) IA32_SYSCALL_PARAM2 (bc32),
			 (unsigned long) IA32_SYSCALL_PARAM3 (bc32),
			 (unsigned long) IA32_SYSCALL_PARAM4 (bc32),
			 (unsigned long) IA32_SYSCALL_PARAM5 (bc32),
			 (unsigned long) IA32_SYSCALL_PARAM6 (bc32));
	}
	
	return retval;
}
#ifdef BUF_SIZE
#undef BUF_SIZE
#endif

#define BUF_SIZE 256
static long
btlib_ngetxattr(BTGENERIC_IA32_CONTEXT* bc32)
{
	char buf[BUF_SIZE];
	char* ptr = buf;
	unsigned long size = (unsigned long)IA32_SYSCALL_PARAM4(bc32); 
	long retval;

	assert((unsigned long) cur_thread_data->syscall_id >= IA32_SYSCALL_GETXATTR &&
		(unsigned long) cur_thread_data->syscall_id <= IA32_SYSCALL_FGETXATTR);
	if (size > BUF_SIZE) {
		ptr = bt_malloc(size);
		if (ptr == NULL)  
			return -ENOMEM;
	}

	retval = ILP32_WRAPPER4 (cur_thread_data->syscall_id,
			 (unsigned long) IA32_SYSCALL_PARAM1 (bc32),
			 (unsigned long) IA32_SYSCALL_PARAM2 (bc32),
			 (unsigned long) ptr,
			 (unsigned long) IA32_SYSCALL_PARAM4 (bc32));
	
	if (!IS_SYS_ERROR(retval)) {
		assert (retval >= 0);
		if (size != 0 && safe_memcpy(U2P(IA32_SYSCALL_PARAM3(bc32)), ptr, (unsigned long)retval) < 0 ) {
			retval = -EFAULT;
		}
	}

	if (ptr != buf)	 {
		bt_free(ptr);
	}

	return retval;
}
static long
btlib_nlistxattr(BTGENERIC_IA32_CONTEXT* bc32) 
{
	char buf[BUF_SIZE];
	char* ptr = buf;
	unsigned long size = (unsigned long)IA32_SYSCALL_PARAM3(bc32); 
	long retval;

	assert((unsigned long) cur_thread_data->syscall_id >= IA32_SYSCALL_LISTXATTR &&
		(unsigned long) cur_thread_data->syscall_id <= IA32_SYSCALL_FLISTXATTR);
	if (size > BUF_SIZE) {
		ptr = bt_malloc(size);
		if (ptr == NULL)  
			return -ENOMEM;
	}

	retval = ILP32_WRAPPER3 (cur_thread_data->syscall_id,
			 (unsigned long) IA32_SYSCALL_PARAM1 (bc32),
			 (unsigned long) ptr,
			 (unsigned long) IA32_SYSCALL_PARAM3 (bc32));
	
	if (!IS_SYS_ERROR(retval)) {
		assert (retval >= 0);
		if (size != 0 && safe_memcpy(U2P(IA32_SYSCALL_PARAM2(bc32)), ptr, (unsigned long)retval) < 0 ) {
			retval = -EFAULT;
		}
	}

	if (ptr != buf)	 {
		bt_free(ptr);
	}

	return retval;
}
/*****************************************************************************/
/* implement the personality syscall - don't allow changing the personality  */
/*****************************************************************************/
#define GET_CURRENT_PERSONALITY 0xffffffff

static long
btlib_personality (BTGENERIC_IA32_CONTEXT * bc32)
{
	unsigned person;
	unsigned person_bak;
	long ret;

	person = IA32_SYSCALL_PARAM1 (bc32);
	person_bak = person;

	/* do not put READ_IMPLIES_EXEC into kernel */
	if (person != GET_CURRENT_PERSONALITY) {
		person &= ~READ_IMPLIES_EXEC;
	}

	/* we do not permit the change of personality */
	if ((person != PER_LINUX) && (person != GET_CURRENT_PERSONALITY)) {
		DBPRINT (btl_prt_syscall, "cannot change personality\n");
		return -EINVAL;
	}
#if DEBUG
	person = SYSCALL (personality, ((unsigned long) GET_CURRENT_PERSONALITY));
	assert (person == PER_LINUX32);
#endif

	if (cur_thread_data->personality == NULL) {
		ret = iti_personality_interface (&cur_thread_data->personality);
		assert (ret == 0);
	}

	/* build up the return value */
	ret = PER_LINUX;
	if (RX_PERSON (*cur_thread_data->personality)) {
		ret |= READ_IMPLIES_EXEC;
	}

	/* reset personality */
	if (person_bak != GET_CURRENT_PERSONALITY) {
#if ! RELEASE
		if (btl_always_rie) {
			person_bak |= READ_IMPLIES_EXEC;
		}
#endif
		*cur_thread_data->personality = person_bak;
	}

	return ret;
}

/************************************************************************** 
* init btlib sigreturn code address *
**************************************************************************/
int
btlib_init_syscall_module (void)
{
	/* allocate one IA-32 kernel page */
	btl_kernel_page = iti_alloc_kernel_page (PROT_EXEC | PROT_READ | PROT_WRITE);
	if (btl_kernel_page < 0) {
		return -1;
	}

	btl_restorer = btl_kernel_page;
	btl_rt_restorer = btl_kernel_page + LEGACY_RESTORER_SIZE * 2;

	bt_memcpy ((void *) btl_restorer, RESTORER_CODE, LEGACY_RESTORER_SIZE);
	bt_memcpy ((void *) btl_rt_restorer, RT_RESTORER_CODE,
		   LEGACY_RESTORER_SIZE);

	init_syscall_attribs ();

	return 0;
}

/************************************************************************** 
* lookup a special-file-spec of special files *
**************************************************************************/
static const struct special_filename *
get_special_file_spec (char *filename)
{
	long j;
	char proc_pid[PATH_MAX];

	DBPRINT(btl_prt_syscall || btl_prt_mem, "%s\n", filename);
	for (j = 0; j < m_custom_handlers_size; j++) {
		if (bt_strcmp (filename, m_custom_handlers[j].filename) == 0) {
			break;
		} else if (bt_strncmp (m_custom_handlers[j].filename, "/proc/self",
			   bt_strlen ("/proc/self")) == 0) {
			bt_snprintf (proc_pid, (unsigned long) PATH_MAX,
				     "/proc/%d", SYSCALL (gettid, ()));
			bt_strncat (proc_pid + bt_strlen (proc_pid),
				    m_custom_handlers[j].filename +
				    bt_strlen ("/proc/self"),
				    PATH_MAX - bt_strlen (proc_pid) - 1);
			if (bt_strcmp (filename, proc_pid) == 0) {
				break;
			}
		} 
	}
	return (j == m_custom_handlers_size) ? NULL : &m_custom_handlers[j];
}

/************************************************************************** 
* lookup the special-file-handlers of special files *
**************************************************************************/
static struct special_file *
get_special_file_handlers (int fd)
{
	long j;
	struct special_file *ret = NULL;
	for (j = 0; j < m_special_files_size; j++) {
		DBPRINT (0, "j=%ld fd=%d\n", j, m_special_files[j].fd);
		if (fd == m_special_files[j].fd) {
			ret = &m_special_files[j];
			break;
		}
	}
	DBPRINT (0, "size=%ld fd=%d ret=%p\n", m_special_files_size, fd, ret);
	return ret;
}

/************************************************************************** 
* insert a new element to the list of special files *
**************************************************************************/
static int
special_file_insert (int fd, const struct special_filename *spec)
{
	long j;

	DBPRINT (btl_prt_syscall, "fd=%d\n", fd);

	btlib_acquire_lock (BTL_LOCK_IND_SPECIAL_FILE);
	if (m_special_files == NULL) {
		m_special_files = bt_calloc (1ul, sizeof (struct special_file));
		if (m_special_files == NULL) {
			assert (m_special_files_size == 0);
			DBPRINT (TRUE, "bt_calloc failed\n");
			btlib_abort_all ();
		}
		m_special_files_size = 1;
		m_special_files[0].fd = -1;
	}
	for (j = 0; j < m_special_files_size; j++) {
		if (m_special_files[j].fd == -1)
			break;
	}
	if (j == m_special_files_size) {
		/* need to enlarge the m_special_files */
		m_special_files =
		    bt_realloc (m_special_files,
				(++m_special_files_size) *
				sizeof (struct special_file));
		if (m_special_files == NULL) {
			DBPRINT (TRUE, "bt_calloc failed\n");
			btlib_abort_all ();
		}
	}
	/* j now holds the index of the new element */
	m_special_files[j].fd = fd;
	m_special_files[j].read_handler = spec->read_handler;
	m_special_files[j].lseek_handler = spec->lseek_handler;
	m_special_files[j].position = 0;
	btlib_release_lock (BTL_LOCK_IND_SPECIAL_FILE);

	return 0;
}

/************************************************************************** 
* remove a new element from the list of special files *
**************************************************************************/
static int
special_file_remove (int fd)
{
	long j;
	struct special_file *list = m_special_files;

	DBPRINT (btl_prt_syscall, "fd=%d\n", fd);
	for (j = 0; j < m_special_files_size; j++) {
		if (list[j].fd == fd)
			break;
	}
	if (j < m_special_files_size) {
		/* element found */
		list[j].fd = -1;
		return 0;
	}
	return -1;
}

/************************************************************************** 
* opens a file - looks in the list of files that need special
* handling, and sets special handlers in case it is needed.
* the return value of this function should be used as the return
* value of the open syscall *
**************************************************************************/
static int
btlib_open (BTGENERIC_IA32_CONTEXT * bc32)
{
	const struct special_filename *sfn;
	char filename[PATH_MAX];
	int ret, fd;
	char *filename32;

	/* filename should be in safe memory before we start working with it */
	filename32 = U2P (IA32_SYSCALL_PARAM1 (bc32));
	switch (cur_thread_data->syscall_id)
	{
	case IA32_SYSCALL_OPEN:
		filename32 = U2P (IA32_SYSCALL_PARAM1 (bc32));
		break;
	case IA32_SYSCALL_OPENAT:
		filename32 = U2P (IA32_SYSCALL_PARAM2 (bc32));
		break;
	default:
		DBPRINT(1, "syscall_id = %d\n", cur_thread_data->syscall_id);
		assert(!"illegal syscall id");
	}


	ret = safe_stringcopy (filename, filename32, sizeof (filename) - 1);
	if (IS_SYS_ERROR (ret)) {
		DBPRINT (0, "memcpy failed filename32=%p\n", filename32);
		fd = ret;
		goto out;
	}
	filename[ret] = 0;


	/* /proc/<pid>/auxv cannot be expressed as a simple file name,
	   so here is an explicit rule for it */
	if ((bt_strncmp (filename, "/proc", 5ul) == 0) &&
			(bt_strstr (filename, "/auxv") != 0)) {
		fd = -ENOENT;
		goto out;
	}

	/* get the special handlers spec for this file */
	sfn = get_special_file_spec (filename);
	if (sfn == NULL || sfn->open_handler == NULL) {
		/* no special handler for sys_open */
		fd = ia32_system_call (bc32, NULL);
	} else {
		DBPRINT (btl_prt_syscall, "special handler for sys_open\n");
		fd = (*sfn->open_handler) (bc32, NULL);
	}
	if (!IS_SYS_ERROR (fd)) {
		/* if there is a special handler for other syscalls then
		 * insert a new element to the handlers list */
		if (sfn != NULL && (sfn->read_handler != NULL || sfn->lseek_handler != NULL)) {
			DBPRINT (btl_prt_syscall, "special handler for sys_read or sys_lseek\n");
			special_file_insert (fd, sfn);
		}
	}
out:
	return fd;
}

/************************************************************************** 
* close a file - looks in the list of files that need special
* handling. if the file is in the list then close it. 
**************************************************************************/
static int
btlib_close (BTGENERIC_IA32_CONTEXT * bc32)
{
	int fd = IA32_SYSCALL_PARAM1 (bc32);
	int ret;

	special_file_remove (fd);	/* don't care about the result */
	ret = ia32_system_call (bc32, NULL);
	return ret;
}

/************************************************************************** 
* syscalls that may require special handling - look for special handler and
* if exisit then call it instead of calling the regular syscall-wrapper.
**************************************************************************/
static int
btlib_read (BTGENERIC_IA32_CONTEXT * bc32)
{
	int fd = IA32_SYSCALL_PARAM1 (bc32);
	struct special_file *sfn = get_special_file_handlers (fd);
	int ret;

	if (g_has_smc) btlib_disable_smc_opt(IA32_SYSCALL_PARAM2(bc32), IA32_SYSCALL_PARAM3(bc32)); 

	DBPRINT (0, "fd=%d sfn=%p read_handler=%p\n",
		 fd, sfn, sfn ? sfn->read_handler : NULL);
	ret = (sfn == NULL || sfn->read_handler == NULL) ?
		ia32_system_call (bc32, NULL) : (*sfn->read_handler) (bc32, sfn);
	return ret;
}

/*****************************************************************************/
/*****************************************************************************/
static int
btlib_lseek (BTGENERIC_IA32_CONTEXT * bc32)
{
	int fd = IA32_SYSCALL_PARAM1 (bc32);
	struct special_file *sfn = get_special_file_handlers (fd);
	int ret;
	ret = (sfn == NULL || sfn->lseek_handler == NULL) ?
		ia32_system_call (bc32, NULL) : (*sfn->lseek_handler) (bc32, sfn);
	return ret;
}

static int 
proc_self_maps_read (BTGENERIC_IA32_CONTEXT * bc32, struct special_file *sf)
{
	struct m_proc_maps_file_desc* maps_content_p; 
	int fd = IA32_SYSCALL_PARAM1(bc32);
	char* userbuff = (char*)U2P(IA32_SYSCALL_PARAM2(bc32));
	int count = IA32_SYSCALL_PARAM3(bc32);
	int n, ret;

	DBPRINT (btl_prt_syscall, "fd=%d count=%d current position=%d\n", sf->fd, count, sf->position);

	lock_mregion_manager(MREGION_MANAGER_IA32);

	maps_content_p = btlib_get_proc_maps_content(fd); /*get the content of /proc/self/maps on ia32 level*/
	DBPRINT (btl_prt_syscall, "content cursize=%d\n", maps_content_p->cur_size);

	if ((int)(sf->position) >= maps_content_p->cur_size) 
		return 0;

	n = ((sf->position + count) <= (unsigned long)maps_content_p->cur_size) ? count : (maps_content_p->cur_size - (int)(sf->position)); 
	ret = safe_stringcopy(userbuff, &(maps_content_p->buff[sf->position]), (unsigned long)n);
	sf->position += ret;
	DBPRINT (btl_prt_syscall, "new position = %d\n", sf->position);

	unlock_mregion_manager(MREGION_MANAGER_IA32);
	return ret;
}

static int
proc_self_maps_lseek (BTGENERIC_IA32_CONTEXT * bc32, struct special_file *sf)
{
	int syscall_id = cur_thread_data->syscall_id;
	int whence;
	long offset;
	int ret = 0;

	if (syscall_id == IA32_SYSCALL_LSEEK) {
		offset = (long) IA32_SYSCALL_PARAM2 (bc32);
		whence = IA32_SYSCALL_PARAM3 (bc32);
	} else {
		assert (syscall_id == IA32_SYSCALL_LLSEEK);
		unsigned lo = IA32_SYSCALL_PARAM3 (bc32);
		offset = IA32_SYSCALL_PARAM2 (bc32);
		offset <<= 32;
		offset |= lo;
		whence = IA32_SYSCALL_PARAM5 (bc32);
	}
	switch (whence) {
	case SEEK_SET:
		sf->position = offset;
		break;
	case SEEK_CUR:
		/* do nothing */
		sf->position += offset;
		break;
	case SEEK_END:
		sf->position = 0;
		break;
	default:
		ret = -EINVAL;
	}

	return sf->position;
}
/*****************************************************************************/
/* read handler for /proc/self/stat - replace the memory and signal related
 * fields */
/*****************************************************************************/
static int
proc_self_stat_read (BTGENERIC_IA32_CONTEXT * bc32, struct special_file *sf)
{
	char ia32_stat_fields[2048];
	int state = 1;
	char *start_code;
	char *wchan;
	extern btrans_binary_map_t g_application_info;
	long j = 0;
	long ret, cpyret;
	char static_buf[4 * 1024];
	char *bufp = static_buf;
	int is_malloc = 0;
	unsigned long size = sizeof (static_buf);
	sigset32_t pending, ignored, catch;
	char *ubuf;
	unsigned ubuf_size;
	off_t cur_offset;
    long lseek_ret;

    lseek_ret = SYSCALL (lseek, (sf->fd, 0UL, SEEK_CUR));
	if (IS_SYS_ERROR (lseek_ret)) {
		ret = lseek_ret;
		goto out;
	}
    cur_offset=lseek_ret;
	DBPRINT (btl_prt_syscall, "fd=%d\n", sf->fd);

	/* first read the file */
	do {
        lseek_ret = SYSCALL (lseek, (sf->fd, 0UL, SEEK_SET));
		if (IS_SYS_ERROR (lseek_ret)) {
			ret = lseek_ret;
			goto out;
		}
		ret = SYSCALL (read, (sf->fd, bufp, size));
		/* after read, the original file offset should be recovered.*/
		if (IS_SYS_ERROR (ret)) {
		        lseek_ret = SYSCALL (lseek, (sf->fd, cur_offset, SEEK_SET));
			    goto out;
		} else {
		    lseek_ret = SYSCALL (lseek, (sf->fd, cur_offset, SEEK_SET));
		    if (IS_SYS_ERROR (lseek_ret)) {
			    ret = lseek_ret;
			    goto out;
     		     }
		}
		if ((unsigned long) ret == size) {
			/* need to enlarge the buffer */
			size *= 2;
			bufp = is_malloc ?  bt_realloc (bufp, size) : bt_malloc (size);
			if (bufp == NULL) {
				ret = -ENOMEM;
				goto out;
			}
			is_malloc = 1;
		} else {
			break;
		}
	} while (1);

	DBPRINT (0, "stat='%s'\n", bufp);
	/* ret now holds the size of the string */

	/* rewrite the fields */
	/*************************************/

	/* goto 'start_code' field */
	while (state > 0 && state < 26) {
		while (j < ret) {
			if (bufp[j++] == ' ') {
				break;
			}
			assert (j < ret);
		}
		state++;
	}
	assert (state == 26);
	start_code = bufp + j;

	ubuf_size = IA32_SYSCALL_PARAM3 (bc32);
	if ((sf->position + ubuf_size) > (unsigned long) (start_code - bufp)) {

		DBPRINT (0, "need translation\n");

		/* goto pass 'sigcatch' field ('wchan' field) */
		while (state > 25 && state < 35) {
			while (j < ret) {
				if (bufp[j++] == ' ') {
					break;
				}
				assert (j < ret);
			}
			state++;
		}
		wchan = bufp + j;
		size = bt_strlen (wchan);

		/* get the signal related values */
		ret = get_sigpending (&pending, (int) sizeof (unsigned));
		if (IS_SYS_ERROR (ret)) {
			DBPRINT (TRUE, "get_sigpending failed:%ld\n", ret);
			goto out;
		}
		ret = get_sigignored (&ignored, (int) sizeof (unsigned));
		if (IS_SYS_ERROR (ret)) {
			DBPRINT (TRUE, "get_sigignored failed:%ld\n", ret);
			goto out;
		}
		ret = get_sigcatch (&catch, (int) sizeof (unsigned));
		if (IS_SYS_ERROR (ret)) {
			DBPRINT (TRUE, "get_sigcatch failed:%ld\n", ret);
			goto out;
		}

		/* prepare 32 bit fields */
		bt_snprintf (ia32_stat_fields, sizeof (ia32_stat_fields),
			     "%u %u %u %u %u %u %u %u %u ",
			     g_application_info.start_code,
			     g_application_info.end_code,
			     g_application_info.start_stack,
			     bc32->esp,
			     bc32->eip,
			     pending.sig[0],
			     cur_thread_data->sigmask,
			     ignored.sig[0], catch.sig[0]
		    );
		DBPRINT (0, "ia32_stat_fields=%s\n", ia32_stat_fields);

		/* copy to buffer */
		assert (start_code < wchan);
		if (bt_strlen (ia32_stat_fields) > (unsigned long) (wchan - start_code)) {
			assert (!"ia32_stat_fields string is too long");
			ret = -1;
			goto out;
		}
		bt_strcpy (start_code, ia32_stat_fields);
		DBPRINT (0, "len=%ld\n", bt_strlen (ia32_stat_fields));

		char *pp = start_code + bt_strlen (ia32_stat_fields);
		bt_memmove (pp, wchan, size);
		*(pp + size) = '\0';
		ret = bt_strlen (bufp);

		DBPRINT (0, "total len=%ld\n", ret);

		/* now put the contents into user's buffer */
		size = MIN ((ret - sf->position), ubuf_size);
	} else {
		size = ubuf_size;
	}
#if DEBUG & 1
	{
		unsigned k;
		DBPRINT (0, "chunk='");
		for (k = 0; k < size; k++) {
			fprintf (g_log_file, "%c", *(bufp + sf->position + k));
		}
		fprintf (g_log_file, "'\n");
	}
#endif
	ubuf = U2P (IA32_SYSCALL_PARAM2 (bc32));
	cpyret = safe_memcpy (ubuf, bufp + sf->position, size);
	sf->position += size;
	ret = IS_SYS_ERROR (cpyret) ? cpyret : (long) size;
      out:
	if (is_malloc) {
		bt_free (bufp);
	}
	DBPRINT (0, "ret=%ld\n", ret);
	return ret;
}

static int
proc_self_stat_lseek (BTGENERIC_IA32_CONTEXT * bc32, struct special_file *sf)
{
	int syscall_id = cur_thread_data->syscall_id;
	int whence;
	long offset;
	int ret = 0;

	if (syscall_id == IA32_SYSCALL_LSEEK) {
		offset = (long) IA32_SYSCALL_PARAM2 (bc32);
		whence = IA32_SYSCALL_PARAM3 (bc32);
	} else {
		assert (syscall_id == IA32_SYSCALL_LLSEEK);
		unsigned lo = IA32_SYSCALL_PARAM3 (bc32);
		offset = IA32_SYSCALL_PARAM2 (bc32);
		offset <<= 32;
		offset |= lo;
		whence = IA32_SYSCALL_PARAM5 (bc32);
	}
	switch (whence) {
	case SEEK_SET:
		sf->position = offset;
		break;
	case SEEK_CUR:
		/* do nothing */
		break;
	case SEEK_END:
		sf->position = 0;
		break;
	default:
		ret = -EINVAL;
	}

	return ret;
}

#if DEBUG
/************************************************************************** 
* print info of sockecall syscall *
**************************************************************************/

char *call_names[] = {
	"",
	"SYS_SOCKET",
	"SYS_BIND",
	"SYS_CONNECT",
	"SYS_LISTEN",
	"SYS_ACCEPT",
	"SYS_GETSOCKNAME",
	"SYS_GETPEERNAME",
	"SYS_SOCKETPAIR",
	"SYS_SEND",
	"SYS_RECV",
	"SYS_SENDTO",
	"SYS_RECVFROM",
	"SYS_SHUTDOWN",
	"SYS_SETSOCKOPT",
	"SYS_GETSOCKOPT",
	"SYS_SENDMSG",
	"SYS_RECVMSG"
};
static void
print_socketcall_info (BTGENERIC_IA32_CONTEXT * bc32)
{
	int call;
	unsigned *a;
	char details_str[16 * 1024];
	struct sockaddr32 *sap;
	struct msghdr32 *smh;
	long ret;

	call = IA32_SYSCALL_PARAM1 (bc32);
	a = U2P (IA32_SYSCALL_PARAM2 (bc32));
	assert ((call > 0) && (call <= SYS_RECVMSG) && (a != NULL));

	switch (call) {
	case SYS_SOCKET:
		ret = bt_snprintf (details_str, sizeof (details_str) - 1,
				   "family=%d type=%d protocal=%d", a[0], a[1], a[2]);
		break;
	case SYS_BIND:
	case SYS_CONNECT:
	case SYS_ACCEPT:
	case SYS_GETSOCKNAME:
	case SYS_GETPEERNAME:
		sap = U2P (a[1]);
		ret = bt_snprintf (details_str, sizeof (details_str) - 1,
				   "fd=%d addr.family=%u "
				   "addr.data='%x %x %x %x %x %x %x %x %x %x %x %x %x %x' "
				   "addrlen=%x",
				   a[0],
				   sap->sa_family,
				   sap->sa_data[0], sap->sa_data[1],
				   sap->sa_data[2], sap->sa_data[3],
				   sap->sa_data[4], sap->sa_data[5],
				   sap->sa_data[6], sap->sa_data[7],
				   sap->sa_data[8], sap->sa_data[9],
				   sap->sa_data[10], sap->sa_data[11],
				   sap->sa_data[12], sap->sa_data[13], a[2]);
		break;
	case SYS_LISTEN:
		ret = bt_snprintf (details_str, sizeof (details_str) - 1,
				   "fd=%d backlog=%d", a[0], a[1]);
		break;
	case SYS_SOCKETPAIR:
		ret = bt_snprintf (details_str, sizeof (details_str) - 1,
				   "family=%d type=%d protocal=%d vec=%x", a[0],
				   a[1], a[2], a[3]);
		break;
	case SYS_SEND:
	case SYS_RECV:
		ret = bt_snprintf (details_str, sizeof (details_str) - 1,
				   "fd=%d buf=%x len=%x flags=%x", a[0], a[1],
				   a[2], a[3]);
		break;
	case SYS_SENDTO:
	case SYS_RECVFROM:
		sap = U2P (a[4]);
		ret = bt_snprintf (details_str, sizeof (details_str) - 1,
				   "fd=%d buf=%x len=%x flags=%x addr.family=%u "
				   "addr.data='%x %x %x %x %x %x %x %x %x %x %x %x %x %x' "
				   "addrlen=%x", a[0], a[1], a[2], a[3],
				   sap->sa_family, sap->sa_data[0],
				   sap->sa_data[1], sap->sa_data[2],
				   sap->sa_data[3], sap->sa_data[4],
				   sap->sa_data[5], sap->sa_data[6],
				   sap->sa_data[7], sap->sa_data[8],
				   sap->sa_data[9], sap->sa_data[10],
				   sap->sa_data[11], sap->sa_data[12],
				   sap->sa_data[13], a[5]);
		break;
	case SYS_SHUTDOWN:
		ret = bt_snprintf (details_str, sizeof (details_str) - 1,
				   "fd=%d how=%d", a[0], a[1]);
		break;
	case SYS_SETSOCKOPT:
	case SYS_GETSOCKOPT:
		ret = bt_snprintf (details_str, sizeof (details_str) - 1,
				   "fd=%d level=%d optname=%d optval=%x optlen=%d",
				   a[0], a[1], a[2], a[3], a[4]);
		break;
	case SYS_SENDMSG:
	case SYS_RECVMSG:
		smh = U2P (a[1]);
		ret = bt_snprintf (details_str, sizeof (details_str) - 1,
				   "fd=%d "
				   "msg.name=%x msg.namelen=%x msg.iov=%x msg.iovlen=%x "
				   "msg.control=%x msg.controllen=%x msg.flags=%x "
				   "flags=%x", a[0], smh->msg_name,
				   smh->msg_namelen, smh->msg_iov,
				   smh->msg_iovlen, smh->msg_control,
				   smh->msg_controllen, smh->msg_flags, a[2]);
	}

	assert (ret >= 0 && (unsigned long) ret < sizeof (details_str) - 1);
	fprintf (g_log_file, "socketcall: call=%s %s\n", call_names[call],
		 details_str);
}

static void
print_getdents (BTGENERIC_IA32_CONTEXT * bc32)
{
	int fd = IA32_SYSCALL_PARAM1 (bc32);
	struct linux_dirent64 *UNUSED dirent = U2P (IA32_SYSCALL_PARAM2 (bc32));
	unsigned UNUSED count = IA32_SYSCALL_PARAM3 (bc32);

	fprintf (g_log_file, "getdents: fd=%d\n", fd);
}

char *ipc_calls[] = {
	"",
	"SEMOP",		/*1 */
	"SEMGET",		/*2 */
	"SEMCTL",		/*3 */
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"MSGSND",		/*11 */
	"MSGRCV",		/*12 */
	"MSGGET",		/*13 */
	"MSGCTL",		/*14 */
	"",
	"",
	"",
	"",
	"",
	"",
	"SHMAT",		/*21 */
	"SHMDT",		/*22 */
	"SHMGET",		/*23 */
	"SHMCTL"		/*24 */
};

char *sem_ctl_val[] = {
	"IPC_RMID",		/*0 */
	"IPC_SET",		/*1 */
	"IPC_STAT",		/*2 */
	"IPC_INFO",		/*3 */
	"",			/*4 */
	"",			/*5 */
	"",			/*6 */
	"",			/*7 */
	"",			/*8 */
	"",			/*9 */
	"",			/*10 */
	"GETPID",		/*11 */
	"GETVAL",		/*12 */
	"GETALL",		/*13 */
	"GETNCNT",		/*14 */
	"GETZCNT",		/*15 */
	"SETVAL",		/*16 */
	"SETALL",		/*17 */
	"SEM_STAT",		/*18 */
	"SEM_INFO"		/*19 */
};
char *msg_ctl_val[] = {
	"IPC_RMID",		/*0 */
	"IPC_SET",		/*1 */
	"IPC_STAT",		/*2 */
	"IPC_INFO",		/*3 */
	"",			/*4 */
	"",			/*5 */
	"",			/*6 */
	"",			/*7 */
	"",			/*8 */
	"",			/*9 */
	"",			/*10 */
	"MSG_STAT",		/*11 */
	"MSG_INFO"		/*12 */
};

char *shm_ctl_val[] = {
	"IPC_RMID",		/*0 */
	"IPC_SET",		/*1 */
	"IPC_STAT",		/*2 */
	"IPC_INFO",		/*3 */
	"",			/*4 */
	"",			/*5 */
	"",			/*6 */
	"",			/*7 */
	"",			/*8 */
	"",			/*9 */
	"",			/*10 */
	"SHM_LOCK",		/*11 */
	"SHM_UNLOCK",		/*12 */
	"SHM_STAT",		/*13 */
	"SHM_INFO"		/*14 */
};

struct msgbuf32 {
	__s32 mtype;
	char mtext[1];
};
#define IPC_64 0x0100

static void
print_ipc_info (BTGENERIC_IA32_CONTEXT * bc32)
{
	unsigned call, first, second, third, ptr, fifth, cmd, ipc64;
	char buffer[4096] = { 0 };
	char *p;
	unsigned j;
	unsigned long chars_left = sizeof (buffer) - 1;
	long len;

	call = IA32_SYSCALL_PARAM1 (bc32);
	first = IA32_SYSCALL_PARAM2 (bc32);
	second = IA32_SYSCALL_PARAM3 (bc32);
	third = IA32_SYSCALL_PARAM4 (bc32);
	ptr = IA32_SYSCALL_PARAM5 (bc32);
	fifth = IA32_SYSCALL_PARAM6 (bc32);

	len = bt_snprintf (buffer, chars_left,
			   "(%d) print_ipc_info: call=%u('%s'):",
			   cur_thread_data->tid, call, 
			   (call > 0 && call < 25) ? ipc_calls[call] : "");

	p = buffer + len;
	chars_left -= len;

	switch (call) {
	case SEMOP:
	case SEMTIMEDOP:
		len = bt_snprintf (p, chars_left, "semid=%d nops=%d ", first, second);
		p += len;
		chars_left -= len;
		if (call == SEMTIMEDOP) {
			int *t32 = (void *) (uintptr_t) fifth;
			len = bt_snprintf (p, chars_left, "timeout=%u,%u", *t32, *(t32 + 1));
			p += len;
			chars_left -= len;
		}
		len = bt_snprintf (p, chars_left, "\n");
		chars_left -= len;

		for (j = 0; j < second; j++) {
			struct sembuf *sb = (struct sembuf *) (uintptr_t) ptr + j;
			len = bt_snprintf (p, chars_left,
					   "\tsem_num=%d sem_op=%d sem_flg=%x\n",
					   sb->sem_num, sb->sem_op, sb->sem_flg);
			p += len;
			chars_left -= len;
		}
		break;
	case SEMGET:
		len = bt_snprintf (p, chars_left, "key=%d nsems=%d semflg=%x\n",
				   first, second, third);
		p += len;
		chars_left -= len;
		break;
	case SEMCTL:
		ipc64 = (third & IPC_64) ? 1 : 0;
		cmd = ipc64 ? (third ^ IPC_64) : third;
		len = bt_snprintf (p, chars_left,
				   "semid=%d semnum=%d cmd=%d('%s%s') val=%d(%x)\n",
				   first, second, third, ipc64 ? "IPC_64 | " : "",
				   (cmd < 20) ? sem_ctl_val[cmd] : "", ptr, ptr);
		p += len;
		chars_left -= len;
		break;
	case MSGSND:
		len = bt_snprintf (p, chars_left, "msqid=%d msgsz=%d msgflg=%x\n",
				   first, second, third);
		p += len;
		chars_left -= len;
		for (j = 0; j < second; j++) {
			struct msgbuf32 *mb = (void *) (uintptr_t) ptr + j;
			len = bt_snprintf (p, chars_left, "\tmtype=%u mtext=%x\n",
					   mb->mtype, mb->mtext[0]);
			p += len;
			chars_left -= len;
		}
		break;
	case MSGRCV:
		{
			unsigned version = call >> 16;
			unsigned msgtyp = fifth;
			if (!version) {
				struct ipc_kludge *uipck =
					(struct ipc_kludge *) (uintptr_t) ptr;
				assert (ptr != 0);

				msgtyp = uipck->msgtyp;
			}

			len = bt_snprintf (p, chars_left,
					   "msqid=%d msgsz=%d msgtyp=%d msgflg=%x\n",
					   first, second, msgtyp, third);
			p += len;
			chars_left -= len;
		}
		for (j = 0; j < second; j++) {
			struct msgbuf32 *mb = (void *) (uintptr_t) ptr + j;
			len = bt_snprintf (p, chars_left, "\tmtype=%u mtext=%x\n",
					   mb->mtype, mb->mtext[0]);
			p += len;
			chars_left -= len;
		}
		break;
	case MSGGET:
		len = bt_snprintf (p, chars_left, "key=%d msgflg=%x\n", first,
				   second);
		p += len;
		chars_left -= len;
		break;
	case MSGCTL:
		ipc64 = (second & IPC_64) ? 1 : 0;
		cmd = ipc64 ? (second ^ IPC_64) : second;
		len = bt_snprintf (p, chars_left, "msqid=%d cmd=%d('%s%s')\n",
				   first, second, ipc64 ? "IPC_64 | " : "",
				   (cmd <= 12) ? msg_ctl_val[cmd] : "");
		p += len;
		chars_left -= len;
		break;
	case SHMAT:
		len = bt_snprintf (p, chars_left,
				   "shmid=%d shmflag=%x shmaddr=%x\n", first,
				   second, ptr);
		p += len;
		chars_left -= len;
		break;
	case SHMDT:
		len = bt_snprintf (p, chars_left, "shmaddr=%x\n", ptr);
		p += len;
		chars_left -= len;
		break;
	case SHMGET:
		len = bt_snprintf (p, chars_left, "key=%d size=%x shmflg=%x\n",
				   first, second, third);
		p += len;
		chars_left -= len;
		break;
	case SHMCTL:
		ipc64 = (second & IPC_64) ? 1 : 0;
		cmd = ipc64 ? (second ^ IPC_64) : second;
		len = bt_snprintf (p, chars_left, "shmid=%d cmd=%d('%s%s')\n",
				   first, second, ipc64 ? "IPC_64 | " : "",
				   (cmd <= 14) ? shm_ctl_val[cmd] : "");
		p += len;
		chars_left -= len;
		break;
	default:
		len = bt_snprintf (p, chars_left, "unknown call\n");
	}

	fprintf (g_log_file, "%s", buffer);
}

static char *fcntl_cmd_str[] = {
	"F_DUPFD",
	"F_GETFD",
	"F_SETFD",
	"F_GETFL",
	"F_SETFL",
	"F_GETLK",
	"F_SETLK",
	"F_SETLKW",
	"F_SETOWN",
	"F_GETOWN",
	"F_SETSIG",
	"F_GETSIG"
};
static void
print_fcntl_info (BTGENERIC_IA32_CONTEXT * bc32)
{
	long ret;
	struct stat s;
	int syscall_id = IA32_SYSCALL_ID (bc32);
	int fd = IA32_SYSCALL_PARAM1 (bc32);
	unsigned cmd = IA32_SYSCALL_PARAM2 (bc32);
	unsigned arg = IA32_SYSCALL_PARAM3 (bc32);
	assert (syscall_id == 221 || syscall_id == 55);

	DBPRINT (TRUE, "syscall fcntl%s fd=%d, cmd=%d<%s>\n",
		 (syscall_id == 221) ? "64" : "", fd, cmd,
		 (cmd <= 11) ? fcntl_cmd_str[cmd] : "unknown");
	DBPRINT (cmd == F_SETFD, "FD_CLOEXEC=%d\n", ((arg & FD_CLOEXEC) != 0));

	ret = SYSCALL (fstat, (fd, &s));
	if (!IS_SYS_ERROR (ret)) {
		DBPRINT (TRUE, "dev=%ld ino=%ld\n", s.st_dev, s.st_ino);
	} else {
		DBPRINT (TRUE, "fstat=%ld\n", ret);
	}
}

static void
print_syscall_before (int syscall_id)
{
/*	if (syscall_id == 184) {
		cap_user_header_t head = (cap_user_header_t) U2P(IA32_SYSCALL_PARAM1 (cur_thread_data->bc32));
		DBPRINT (1, "capget arg1=%p arg2=%lx\n", 
			head, 
			(unsigned long)(IA32_SYSCALL_PARAM2 (cur_thread_data->bc32)));

		DBPRINT (head != NULL && head != (void*)(-1), "version = 0x%x, pid= 0x%x\n", head->version, head->pid);
	}
*/
	if (syscall_id == IA32_SYSCALL_WRITE) {
		unsigned count = IA32_SYSCALL_PARAM3 (cur_thread_data->bc32);
		char *str = U2P (IA32_SYSCALL_PARAM2 (cur_thread_data->bc32));
		char wbuf[2048];
		unsigned j;
		char c;
		int err=0;

		for (j = 0; (j < sizeof (wbuf) - 1) && (j < count); j++) {
			if (safe_memcpy(&c, str+j, 1ul) < 0) {
				err=1;
				break;
			}
			if ((*(str + j) < ' ') || (*(str + j) > '~')) {
				break;
			} else {
				wbuf[j] = *(str + j);
			}
		}
		wbuf[j] = 0;
		DBPRINT (1, "syscall write fd=%d string='%s' err=%d %d bytes in all\n",
			 IA32_SYSCALL_PARAM1 (cur_thread_data->bc32), wbuf, err, j);

	}

	if (syscall_id == 172) {
		DBPRINT (1, "prctl (opt=%d, arg2=%lx, arg3=%lx, arg4=%lx, arg5=%lx)\n",
			IA32_SYSCALL_PARAM1 (cur_thread_data->bc32),
			IA32_SYSCALL_PARAM2 (cur_thread_data->bc32),
			IA32_SYSCALL_PARAM3 (cur_thread_data->bc32),
			IA32_SYSCALL_PARAM4 (cur_thread_data->bc32),
			IA32_SYSCALL_PARAM5 (cur_thread_data->bc32));
	}

#if 0
	if (syscall_id == IA32_SYSCALL_OPEN) {
		DBPRINT (1,
			 "syscall open file-name=%p<%s> flags=%o mode=%o\n",
			 U2P (IA32_SYSCALL_PARAM1 (cur_thread_data->bc32)),
			 (char *)
			 U2P (IA32_SYSCALL_PARAM1 (cur_thread_data->bc32)),
			 (int) IA32_SYSCALL_PARAM2 (cur_thread_data->bc32),
			 (int) IA32_SYSCALL_PARAM3 (cur_thread_data->bc32));
	}
	if (syscall_id == IA32_SYSCALL_PSELECT) {
		sigset32_t* ptr;
		DBPRINT(0,
			"syscall pselect (nfs=%d, readfd=%p, writefd=%p, exceptfds=%p, timeout=%p, sigmask=%p)\n",
			(unsigned long)IA32_SYSCALL_PARAM1 (cur_thread_data->bc32),
			U2P (IA32_SYSCALL_PARAM2 (cur_thread_data->bc32)),
			U2P (IA32_SYSCALL_PARAM3 (cur_thread_data->bc32)),
			U2P (IA32_SYSCALL_PARAM4 (cur_thread_data->bc32)),
			U2P (IA32_SYSCALL_PARAM5 (cur_thread_data->bc32)),
			U2P (IA32_SYSCALL_PARAM6 (cur_thread_data->bc32))
			);

		ptr = U2P (IA32_SYSCALL_PARAM6 (cur_thread_data->bc32));
		DBPRINT(0, "sigmask %x/%x\n", ptr->sig[0], ptr->sig[1]);

	}
	if (syscall_id == IA32_SYSCALL_OPENAT) {
		DBPRINT (1,
			 "syscall open file-name=%p<%s> flags=%o mode=%o\n",
			 U2P (IA32_SYSCALL_PARAM2 (cur_thread_data->bc32)),
			 (char *)
			 U2P (IA32_SYSCALL_PARAM2 (cur_thread_data->bc32)),
			 (int) IA32_SYSCALL_PARAM3 (cur_thread_data->bc32),
			 (int) IA32_SYSCALL_PARAM4 (cur_thread_data->bc32));
	}

	if (syscall_id == IA32_SYSCALL_FSTATAT64 ) {
		DBPRINT (1,
			"fstatat64 ( fd = %d, name = %s, buf = %p, flag = 0x%x)\n",
			(int)IA32_SYSCALL_PARAM1(cur_thread_data->bc32),
			(char*) U2P (IA32_SYSCALL_PARAM2(cur_thread_data->bc32)),
			U2P (IA32_SYSCALL_PARAM3(cur_thread_data->bc32)), 
			(int)IA32_SYSCALL_PARAM4(cur_thread_data->bc32));
	}
	if (syscall_id == IA32_SYSCALL_STAT64 ){
		DBPRINT (1,
			"stat64 ( name = %s, buf = %p)\n",
			(char*) U2P (IA32_SYSCALL_PARAM1(cur_thread_data->bc32)),
			U2P (IA32_SYSCALL_PARAM2(cur_thread_data->bc32))); 

	}
	if (syscall_id == IA32_SYSCALL_KILL) {
		DBPRINT (0,
			 "syscall kill pid=%d signo=%d\n",
			 (int) IA32_SYSCALL_PARAM1 (cur_thread_data->bc32),
			 (int) IA32_SYSCALL_PARAM2 (cur_thread_data->bc32));
	}
	if (syscall_id == IA32_SYSCALL_TKILL) {
		DBPRINT (0,
			 "syscall tkill tid=%d signo=%d\n",
			 (int) IA32_SYSCALL_PARAM1 (cur_thread_data->bc32),
			 (int) IA32_SYSCALL_PARAM2 (cur_thread_data->bc32));
	}



	if (syscall_id == 221 /*IA32_SYSCALL_FCNTL64 */ ) {
		print_fcntl_info (cur_thread_data->bc32);
	}
	if (syscall_id == IA32_SYSCALL_EXECVE) {
		btlib_thread_data_t *kuku UNUSED = cur_thread_data;
		char *file = U2P (IA32_SYSCALL_PARAM1 (cur_thread_data->bc32));
		DBPRINT (0, "syscall execve file='%s'\n", file);
		/*
		   int j;
		   unsigned *argv = U2P (IA32_SYSCALL_PARAM2 (cur_thread_data->bc32));
		   for (j = 0; *argv != 0; j++, *argv++) {
		   DBPRINT (0, "\targv[%d]='%s'\n",
		   j, (char *) (uintptr_t) * argv);
		   }
		 */
	}
	if (syscall_id == IA32_SYSCALL_IPC) {
		print_ipc_info (cur_thread_data->bc32);
	}
	if (syscall_id == IA32_SYSCALL_STAT64) {
		fprintf (g_log_file,
			 "(%d) syscall stat64 file-name='%s'\n",
			 cur_thread_data->tid, (char *) (uintptr_t)
			 IA32_SYSCALL_PARAM1 (cur_thread_data->bc32));

	}
	if (syscall_id == IA32_SYSCALL_EXECVE) {
		btlib_thread_data_t *kuku UNUSED = cur_thread_data;
		char *file = U2P (IA32_SYSCALL_PARAM1 (cur_thread_data->bc32));
		unsigned *argv =
		    U2P (IA32_SYSCALL_PARAM2 (cur_thread_data->bc32));
		int j;
		fprintf (g_log_file, "(%d) syscall execve file='%s'\n",
			 cur_thread_data->tid, file);
		for (j = 0; *argv != 0; j++, *argv++) {
			fprintf (g_log_file, "\targv[%d]='%s'\n",
				 j, (char *) (uintptr_t) * argv);
		}
	}
	if (syscall_id == 258 /* set_tid_address */ ) {
		fprintf (g_log_file,
			 "(%d) syscall set_tid_address tidptr=%p\n",
			 cur_thread_data->tid,
			 U2P (IA32_SYSCALL_PARAM1 (cur_thread_data->bc32)));
	}
	if (syscall_id == IA32_SYSCALL_STAT64) {
		fprintf (g_log_file,
			 "(%d) syscall stat64 file-name='%s'\n",
			 cur_thread_data->tid, (char *) (uintptr_t)
			 IA32_SYSCALL_PARAM1 (cur_thread_data->bc32));

	}
	if (syscall_id == IA32_SYSCALL_ACCESS) {
		fprintf (g_log_file,
			 "(%d) syscall access file-name='%s' flags=%d mode=%d\n",
			 cur_thread_data->tid, (char *) (uintptr_t)
			 IA32_SYSCALL_PARAM1 (cur_thread_data->bc32),
			 (int) IA32_SYSCALL_PARAM2 (cur_thread_data->bc32),
			 (int) IA32_SYSCALL_PARAM3 (cur_thread_data->bc32));
	}
	if (syscall_id == IA32_SYSCALL_LSTAT64) {
		fprintf (g_log_file,
			 "(%d) syscall lstat file-name='%s'\n",
			 cur_thread_data->tid, (char *) (uintptr_t)
			 IA32_SYSCALL_PARAM1 (cur_thread_data->bc32));

	}
	if (syscall_id == IA32_SYSCALL_IPC) {
		print_ipc_info (cur_thread_data->bc32);
	}
	if (syscall_id == IA32_SYSCALL_LLSEEK) {
		fprintf (g_log_file,
			 "(%d) syscall llseek fd=%d off_hi=%x off_lo=%x orig=%d\n",
			 cur_thread_data->tid,
			 (int) IA32_SYSCALL_PARAM1 (cur_thread_data->bc32),
			 (int) IA32_SYSCALL_PARAM2 (cur_thread_data->bc32),
			 (int) IA32_SYSCALL_PARAM3 (cur_thread_data->bc32),
			 (int) IA32_SYSCALL_PARAM5 (cur_thread_data->bc32));
	}
	if (syscall_id == IA32_SYSCALL_WRITE) {
		unsigned count = IA32_SYSCALL_PARAM3 (cur_thread_data->bc32);
		char *str = U2P (IA32_SYSCALL_PARAM2 (cur_thread_data->bc32));
		char wbuf[2048];
		unsigned j;
		for (j = 0; (j < sizeof (wbuf) - 1) && (j < count); j++) {
			if ((*(str + j) < ' ') || (*(str + j) > '~')) {
				break;
			} else {
				wbuf[j] = *(str + j);
			}
		}
		wbuf[j] = 0;
		fprintf (g_log_file, "(%d) syscall write fd=%d string='%s'\n",
			 cur_thread_data->tid,
			 IA32_SYSCALL_PARAM1 (cur_thread_data->bc32), wbuf);

	}
	if (syscall_id == IA32_SYSCALL_SOCKETCALL) {
		print_socketcall_info (cur_thread_data->bc32);
	}
	if (syscall_id == IA32_SYSCALL_CLOSE) {
		fprintf (g_log_file, "(%d) syscall close fd=%d\n",
			 cur_thread_data->tid,
			 (int) IA32_SYSCALL_PARAM1 (cur_thread_data->bc32));
	}
	if (syscall_id == 10 /*unlink */ ) {
		fprintf (g_log_file,
			 "(%d) syscall unlink file-name='%s'\n",
			 cur_thread_data->tid, (char *)
			 U2P (IA32_SYSCALL_PARAM1 (cur_thread_data->bc32)));
	}
	if (syscall_id == IA32_SYSCALL_OPEN) {
		fprintf (g_log_file,
			 "(%d) syscall open file-name='%s' flags=%d mode=%d\n",
			 cur_thread_data->tid, (char *)
			 U2P (IA32_SYSCALL_PARAM1 (cur_thread_data->bc32)),
			 (int) IA32_SYSCALL_PARAM2 (cur_thread_data->bc32),
			 (int) IA32_SYSCALL_PARAM3 (cur_thread_data->bc32));
	}
	if (syscall_id == IA32_SYSCALL_EPOLL_CREATE) {
		fprintf (g_log_file,
			 "(%d) syscall epoll_create size=%d\n",
			 cur_thread_data->tid,
			 (int) IA32_SYSCALL_PARAM1 (cur_thread_data->bc32));
	}
	if (syscall_id == IA32_SYSCALL_EPOLL_CTL) {
		fprintf (g_log_file,
			 "(%d) syscall epoll_ctl epfd=%d op=%d fd=%d\n",
			 cur_thread_data->tid,
			 (int) IA32_SYSCALL_PARAM1 (cur_thread_data->bc32),
			 (int) IA32_SYSCALL_PARAM2 (cur_thread_data->bc32),
			 (int) IA32_SYSCALL_PARAM3 (cur_thread_data->bc32));
	}
	if (syscall_id == IA32_SYSCALL_EPOLL_WAIT) {
		fprintf (g_log_file,
			 "(%d) syscall epoll_wait epfd=%d maxevents=%d timeout=%d\n",
			 cur_thread_data->tid,
			 (int) IA32_SYSCALL_PARAM1 (cur_thread_data->bc32),
			 (int) IA32_SYSCALL_PARAM3 (cur_thread_data->bc32),
			 (int) IA32_SYSCALL_PARAM4 (cur_thread_data->bc32));
	}
#endif
}

static void
print_syscall_after (int syscall_id, long retval)
{
	if (syscall_id == IA32_SYSCALL_FSTATAT64){
		struct stat32 *s = (struct stat32*)U2P(IA32_SYSCALL_PARAM3(cur_thread_data->bc32));
		
		DBPRINT(1, "return value = %d stat=0x%p size=%d\n", retval, s, sizeof(*s));
		DBPRINT(1, "dev=0x%lx\t inode=0x%lx\t link=0x%lx\n", s->st_dev, s->st_ino, s->st_nlink);
		DBPRINT(1, "mode=0x%lx\t uid=%d\t gid=%d\n", s->st_mode, s->st_uid, s->st_gid);
	}
	if (syscall_id == IA32_SYSCALL_READ) {
		long count = retval;
		char *str = (char *) (uintptr_t) IA32_SYSCALL_PARAM2 (cur_thread_data->bc32);
		char rbuf[2048];
		unsigned j;

		for (j = 0; (j < sizeof (rbuf) - 1) && (j < count); j++) {
			if ((*(str + j) < ' ') || (*(str + j) > '~')) {
				break;
			} else {
				rbuf[j] = *(str + j);
			}
		}
		rbuf[j] = 0;
		DBPRINT (1, "syscall read fd=%d string='%s' retval=%ld\n",
			 IA32_SYSCALL_PARAM1 (cur_thread_data->bc32), rbuf, retval);
	}
	if (syscall_id == IA32_SYSCALL_GETCWD) {
		char *buf = (char *) (uintptr_t) IA32_SYSCALL_PARAM1 (cur_thread_data->bc32);
		unsigned len = IA32_SYSCALL_PARAM2 (cur_thread_data->bc32);

		if ((retval > 0) && (retval < len)) {
			fprintf (g_log_file, "(%d) syscall getcwd='%s'\n",
				 cur_thread_data->tid, buf);
		} else {
			fprintf (g_log_file,
				 "(%d) syscall getcwd: len=%u retval=%ld\n",
				 cur_thread_data->tid, len, retval);
		}
	}

	if (syscall_id == IA32_SYSCALL_GETDENTS64) {
		print_getdents (cur_thread_data->bc32);
	}

	if (syscall_id == IA32_SYSCALL_READ) {
		long count = retval;
		char *str = (char *) (uintptr_t) IA32_SYSCALL_PARAM2 (cur_thread_data->bc32);
		char rbuf[2048];
		unsigned j;
		for (j = 0; (j < sizeof (rbuf) - 1) && (j < count); j++) {
			if ((*(str + j) < ' ') || (*(str + j) > '~')) {
				break;
			} else {
				rbuf[j] = *(str + j);
			}
		}
		rbuf[j] = 0;
		fprintf (g_log_file, "(%d) syscall read fd=%d string='%s'\n",
			 cur_thread_data->tid,
			 IA32_SYSCALL_PARAM1 (cur_thread_data->bc32), rbuf);
	}
}

#endif	// DEBUG

