/*
 * btlib_thread.c
 *	
 * 
 * 
*/
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include "btlib.h"
#include <ia32x_utils.h>

#include <ia32toia64.h>

#include "btlib_sig_utils.h"
#include "btlib_signal.h"
#include "btlib_asm.h"
#include "btlib_segment.h"
#include "btlib_sync.h"
#include "btlib_ptemu.h"
#include "btlib_stats.h"
#include "../ia32x_utils/include/mutex_internal.h"

/* The cloning data struct */
typedef struct btlib_clone_data_t {
	BTGENERIC_IA32_CONTEXT start_context;
	void *clst;
	btlib_thread_data_t *blst;
	k_sigset_t sigmask;
	long personality;
	void * child_sem;
	int child_tid;
} btlib_clone_data_t;

typedef struct btlib_native_clone_data_t {
	struct btlib_native_clone_data_t * next; // Pointer to next struct list
	btlib_clone_data_t clone_data; // data used for cloning
	int thread_tid;   // Thread id of native thread
	int native_thread_type;  // type of native thread
	int thread_should_die;  // is set to TRUE if trying to kill the thread before it finished initalizing
	void * thread_exist_sem; // Existance semaphore. Held as long as the native thread exists.
} btlib_native_clone_data_t;

btlib_native_clone_data_t * native_thread_list = NULL; // list of native threads in process

void __btlib_exit (int err_code, void *stack, unsigned long len);
extern size_t __default_stack_size;
extern long pthread_create (pthread_t *, void *, void * (*) (void *), void *);

/* are we part of a thread group */
static int m_thread_group = 0;
/* thread list, protected by BTL_LOCK_IND_THREAD */
btlib_thread_data_t *m_used_thread_list = NULL;
static long m_kernel_support_nptl = 0;

/* variables used for the affinity and shared mem handling */
#if SHMEM_ORDER_SOLUTION || ALL_THREAD_AFFINITY
static cpu_set_t m_orig_process_affinity_mask = { {0} };
static int m_did_clone = 0;
static int m_did_shared_mem = 0;
static int m_grpid = -1;
int btl_narrowed_affinity = 0;
int btl_did_ordered_code = 0;
#define NUM_BLANKS_TIL_PGRP 4
#define NUM_BLANKS_TIL_UITME 8
#endif

static long prepare_child (void *arg);
static void before_thread_exit (void);
static void list_add (btlib_thread_data_t * elem, btlib_thread_data_t ** list);
static void list_del (btlib_thread_data_t * elem, btlib_thread_data_t ** list);
static void *allocate_thread_data (void **, btlib_thread_data_t **, void **);
static void clone_signals (unsigned, btlib_thread_data_t *);
static long do_fork (btlib_thread_data_t * parent RELEASE_UNUSED,
		     unsigned clone_flags,
		     unsigned ia32_tos,
		     int *parent_tidptr,
		     int *child_tidptr, BTGENERIC_IA32_CONTEXT * bc32);
static void set_sig_inherit (void);
static void cleanup_thread (btlib_thread_data_t * tbr, int free_thread_stack);
static void init_thread_data (btlib_thread_data_t * blst, void *clst, void *stack_lowest_address, unsigned long stack_size);

#if ALL_THREAD_AFFINITY
static int select_one_cpu (cpu_set_t * cur_mask, cpu_set_t * new_mask);
static int get_current_cpu (void);
static int get_random_cpu (cpu_set_t * cpu_mask);
#define DEFAULT_MHZ 1500L
// NUM_THREAD_SCALABILITY_CALLS_TO_CHECK_AFTER must be power of two
#define NUM_THREAD_SCALABILITY_CALLS_TO_CHECK_AFTER (0x100)
#define SCALABILITY_THRESHOLD 15
#define HIGH_SCALABILITY_THRESHOLD 8
#define FRACTION_OF_SECOND_BETWEEN_THREAD_STATE_CHECKS 0.5
#define NUM_FOR_VERY_MULTI_THREADED 2
unsigned long m_cycles_between_ts_checks;
unsigned long m_last_ts_check_timestamp = 0;
unsigned long BtlpNumThreads = 0;
unsigned long m_in_thread_scalability_analysis = 0;
unsigned int m_num_consec_multi_threaded = 0;
unsigned int m_num_consec_very_multi_threaded = 0;
unsigned int m_last_num_threads_in_paralled = 1;
#if THREAD_SCALABILIY_WITH_IDLE_TIME
static unsigned long get_cur_num_secs_idle (void);
static unsigned long m_sec_idle_at_istrt = 0;
#endif

#define CYCLES_IN_MILI_SECOND (btl_cycles_in_second / 1000)
/* half of the interval is multi-threaded => thread scalability */
#define INVERSE_FRACTION_OF_INTERVAL_FOR_THREAD_SCALABILITY 2
/* one quarter of the interval is idle time => worthwhile to switch
   to ordered code in order to remove affinity */
#define INVERSE_FRACTION_OF_IDLETIME_FOR_THREAD_SCALABILITY 4
#if DEBUG
static int is_only_one_cpu_allowed (cpu_set_t * cpu_mask);
unsigned long btl_start_time = 0;
#endif	/* DEBUG */
#endif	/* ALL_THREAD_AFFINITY */

#if DEBUG
void print_thread_list (char *);
#endif	// DEBUG

#if SHMEM_ORDER_SOLUTION || ALL_THREAD_AFFINITY
static long set_affinity_to_all_threads (cpu_set_t * new_affinity_mask);
#endif

/*****************************************************************************/
/* utility functions for CPU masks.	*/
/*****************************************************************************/
static inline int
__CPU_IS_EQUAL (cpu_set_t * set1, cpu_set_t * set2)
{
	unsigned int __i;
	for (__i = 0; __i < sizeof (cpu_set_t) / sizeof (__cpu_mask); ++__i)
		if (set1->__bits[__i] != set2->__bits[__i])
			return 0;
	return 1;
}
static inline int
__CPU_IS_ZERO (cpu_set_t * set)
{
	unsigned int __i;
	for (__i = 0; __i < sizeof (cpu_set_t) / sizeof (__cpu_mask); ++__i)
		if (set->__bits[__i] != 0)
			return 0;
	return 1;
}

#if DEBUG
static inline void
__CPU_PRINT (char *msg, cpu_set_t * set)
{
	char set_str[4096] = { 0 };
	char *setp = set_str;
	unsigned int cpu;
	unsigned long chars_left = sizeof (set_str) - 1;
	long len;

	for (cpu = 0; cpu < 8 * sizeof (*set) - 1; cpu++) {
		if (__CPU_ISSET (cpu, set)) {
			assert (chars_left > 0);
			len = bt_snprintf (setp, chars_left, "%u ", cpu);
			setp += len;
			chars_left -= len;
		}
	}
	fprintf (g_log_file, "%s: cpu_mask=[ %s]\n", msg, set_str);
}
#endif
/*****************************************************************************/

/*****************************************************************************/
/* 
 * initialize the thread module. must be called before any exported function 
 * is called
 */
/*****************************************************************************/
void
btlib_init_thread_module ()
{
	long retval;
	retval = SYSCALL (set_tid_address, (0));
	m_kernel_support_nptl = (retval == -ENOSYS) ? FALSE : TRUE;
	btlib_init_tls_module ();
	
}

/**************************************************************************/
/*
 * main_thread_init
 *
 * Purpose:	
 *      1. Allocate thread specific struct and initialize it.
 *      2. Allocate clst+glst memory for BTGeneric and call BTGeneric_Thread_Init.
 */
/**************************************************************************/
long
btlib_main_thread_init_tsd ()
{
	void *clst;
	void **clstp;
	unsigned long tls_size;

	tls_size = (unsigned long) btl_clst_glst_size + sizeof (btlib_thread_data_t),
	    /* allocate clst+glst */
	    clst = btlib_mem_alloc (INITIAL_BTLIB_ADDRESS,
				    tls_size,
				    PROT_READ | PROT_WRITE | PROT_EXEC);
	if (IS_SYS_ERROR ((long) clst)) {
		DBPRINT (TRUE, "mmap failed: %ld\n", (long) clst);
		return -ENOMEM;
	}
	assert (clst != NULL);
	assert ((uintptr_t) cur_thread_clst ==
		ROUNDUP ((uintptr_t) cur_thread_clst, sizeof (void *)));
	clstp = (void **) (__r13 + g_clstp_offset);
	*clstp = clst;

	DBPRINT (0, "cur_thread_data=%p clst=%p\n", cur_thread_data,
		 cur_thread_clst);

	return 0;
}

void
btlib_set_initial_ia32_exec_context (BTGENERIC_IA32_CONTEXT * bc32)
{
	/* set the start context */
	unsigned long p_glst;
	p_glst=bc32->p_glst;
	bt_memset (bc32, 0, sizeof (BTGENERIC_IA32_CONTEXT));
       bc32->p_glst=p_glst;
	bc32->fpu_state.control_word = FPCW_INIT_VAL;
	bc32->fpu_state.tag_word = FPTW_INIT_VAL;
	bc32->eflags = EFLAGS_INIT_VAL;
	*(unsigned int *) (void *) &bc32->
	    xmm_and_bt_ext[MXCSR_INDEX_IN_EXT_REGS] = MXCSR_INIT_VAL;

	/* same with the kernel */
	bc32->ds = __USER_DS;
	bc32->es = __USER_DS;
	bc32->ss = __USER_DS;
	bc32->cs = __USER_CS;

	SET_DESC_TABLE_REGISTERS (*bc32);
	bc32->composition_bits =
		CTX32_CRITICAL | CTX32_GP | CTX32_FLOATING_POINT | CTX32_XMM |
		CTX32_SEGMENT | CTX32_BT_EXTENDED;
}

/**************************************************************************/
/**************************************************************************/
long
btlib_main_thread_init (const btrans_binary_map_t * app_info, long personality)
{
	BTGENERIC_IA32_CONTEXT bc32;
	BT_STATUS_CODE btg_ret;
	long ret RELEASE_UNUSED;
	k_sigset_t mask64;
	sigset32_t mask32;

	/* cur_thread_data already assigned by btlib_main_thread_init_tsd */
	init_thread_data (cur_thread_data, cur_thread_clst, 0, 0ul);
	cur_thread_data->__r13 = (uintptr_t) __r13;

	btlib_que_init (&cur_thread_data->pending_que);

	/* store the tid */
	cur_thread_data->tid = SYSCALL (gettid, ());

	/* the first element in the list. 
	 * since adding to the list is always at the beginning, iterating
	 * the list can be done by starting with m_used_thread_list backwards
	 */
	m_used_thread_list = cur_thread_data;
	cur_thread_data->next = NULL;
	cur_thread_data->prev = NULL;

	/* create a signal-handelr table */
	cur_thread_data->sighandlers = btlib_create_sighand_table ();
	assert (cur_thread_data->sighandlers != NULL);

	/* set the inherited mask and disposition */
	set_sig_inherit ();

	ret = SYSCALL (rt_sigprocmask,
		       (SIG_SETMASK, NULL, &mask64, sizeof (k_sigset_t)));
	sys_err_chk (ret, "rt_sigprocmask");
	sigset64_2_sigset32 (&mask64, &mask32);
	set_sigmask (&mask32);

	/* get the address of the long syscall flag */
	ret = iti_long_syscall_interface (&cur_thread_data->long_syscall_flag, NULL);
	assert (ret == 0);
	DBPRINT (0, "long_syscall_flag=%p\n",
		 cur_thread_data->long_syscall_flag);

	/* initialize the thread */
	btg_ret = BTGENERIC_THREAD_INIT (cur_thread_clst, app_info->start_stack, 0ul);
	assert (btg_ret == BT_STATUS_SUCCESS);

	/* set the start context */
	btlib_set_initial_ia32_exec_context (&bc32);

	bc32.esp = (unsigned) app_info->start_stack;
	bc32.eip = (unsigned) app_info->entry_point;

	btg_ret = BTGENERIC_IA32_CONTEXT_SET (NULL, &bc32);
	assert (btg_ret == BT_STATUS_SUCCESS);

#if PTEMU_ENABLED
	/* see if the new app was invoked by a debugger */
	DBPRINT (0, "btl_is_ptraced=%ld\n", btl_is_ptraced);
	if (btl_is_ptraced != IS_PTRACED_NO) {
		DBPRINT (btl_prt_pt, "setting thread ptrace flag\n");
		assert (!g_is_mme_mode);
		*cur_thread_ptrace () = PTEMU_PTRACED;
	}
#endif	// PTEMU_ENABLED

	/* record personality */
#if ! RELEASE
	if (btl_always_rie) {
		personality |= READ_IMPLIES_EXEC;
	}
#endif
	ret = iti_personality_interface (&cur_thread_data->personality);
	assert (ret == 0);
	*cur_thread_data->personality = personality;

	return 0;
}

/******************************************************************************/
/*
 *  btlib_cleanup_thread()
 */
/******************************************************************************/
void
btlib_cleanup_thread (int tbr_pid)
{
	btlib_thread_data_t *tbr;
	int found = 0;

	DBPRINT (btl_prt_thread, "tbr_pid=%d\n", tbr_pid);

	assert (cur_thread_data->tid == SYSCALL (gettid, ()));
	assert (cur_thread_data->tid != tbr_pid);

	/* find the thread to be removed */
	btlib_acquire_lock (BTL_LOCK_IND_THREAD);
	for (tbr = m_used_thread_list; tbr != NULL; tbr = tbr->next) {
		if (tbr->tid == tbr_pid) {
			found = 1;
			break;
		}
	}
	btlib_release_lock (BTL_LOCK_IND_THREAD);

	/* the thread should be in the used-thread-list only if it
	   terminated abnormaly. otherwise - in btlib_thread_exit it would
	   have been removed from the list */

	if (!found) {
		return;
	}
	assert (tbr->is_exited == 0);

	/* thread exited abnormaly */
	DBPRINT (btl_prt_thread, "abnormal\n");

#if DEBUG 
	/* if it's holding a critical section then we are doomed */
	if (is_holding_critical_section (tbr)) {
		DBPRINT (TRUE, "holding cs - abort!!\n");
		btlib_abort_all ();
	}
#endif
	list_del (tbr, &m_used_thread_list);

	/* bt_free the thread data */
	cleanup_thread (tbr, TRUE);
}

/******************************************************************************/
/* bt_free resources held by the given thread
 * input : tbr, thread to be bt_free 
 *         free_thread_stack, whether to bt_free thread stack
 * output: none
 */
/******************************************************************************/
static void
cleanup_thread (btlib_thread_data_t * tbr, int free_thread_stack)
{
	btlib_try_free_sighand_table (tbr);
	btlib_que_free (&tbr->pending_que);	/* bt_free pending que */

	if (tbr->stackblock != 0) {
		btlib_deallocate_stack (tbr->stackblock, &tbr->tid, free_thread_stack);
	}
}

/******************************************************************************/
/******************************************************************************/
void *
clone_main (void *arg)
{
	prepare_child (arg);
	/* run the new thread */
	DBPRINT (btl_prt_thread, "run\n");
	cur_thread_data->longjmp_status = RUN;
	btrans_jump ();

	/* free BTLib thread stack, which was deallocated in __btlib_thread_exit */
	if (cur_thread_data->ia32_thread_exited == 1) {
		SYSCALL (munmap, ((unsigned long)cur_thread_data->stackblock, __default_stack_size));
	}

	/* this is just for the compiler to be happy */
	return NULL;
}


static long
prepare_child (void * arg)
{
	long retval;
	BT_STATUS_CODE btg_ret;
	btlib_clone_data_t *clone_params = (btlib_clone_data_t *) arg;
	unsigned RELEASE_UNUSED flags = IA32_SYSCALL_PARAM1 (&clone_params->start_context);
	void **clstp;

	/* init the thread desc */
	clstp = (void **) (__r13 + g_clstp_offset);
	*clstp = clone_params->clst;
	cur_thread_data->__r13 = (uintptr_t) __r13;

	/* store the new thread id */
	cur_thread_data->tid = SYSCALL (gettid, ());
	
	assert (cur_thread_data->in_critical_section == 0);
#if DEBUG
	assert (cur_thread_data->locks == 0);
#endif

	sigset64_2_sigset32 (&clone_params->sigmask, &cur_thread_data->sigmask);

	/* get the address of the long syscall flag */
	retval = iti_long_syscall_interface (&cur_thread_data->long_syscall_flag, NULL);
	assert (retval == 0);

	DBPRINT (0, "long_syscall_flag=%p\n", cur_thread_data->long_syscall_flag);

	/* get the address of the thread personality */
	retval = iti_personality_interface (&cur_thread_data->personality);
	assert (retval == 0);
	*cur_thread_data->personality = clone_params->personality;

	/* we are child now, eax should be 0 */
	clone_params->start_context.esp = IA32_SYSCALL_PARAM2 (&clone_params->start_context);
	clone_params->start_context.eax = 0;

	/* 
	 * add current thread to used threa list, should be done in child to 
	 * avoid cases that child exits before parent adds it to the list
	 */
	btlib_acquire_lock (BTL_LOCK_IND_THREAD);
	list_add (cur_thread_data, &m_used_thread_list);
	btl_num_threads++;

	btlib_release_lock (BTL_LOCK_IND_THREAD);

	/* add FS actual offset and GS actual offset to the calling sequence */
	retval = (long) BTGENERIC_THREAD_INIT (cur_thread_clst,
					       (__u32) clone_params->start_context.esp,
					       (__u64) 0);

	assert (retval == (long) BT_STATUS_SUCCESS);

	DBPRINT (0, "esp=%x eip=%x ebx=%x\n",
		 clone_params->start_context.esp, clone_params->start_context.eip,
	  	 clone_params->start_context.ebx);

	CHECK_DESC_TABLE_REGISTERS (clone_params->start_context);
	clone_params->start_context.composition_bits =
		CTX32_CRITICAL | CTX32_GP | CTX32_SEGMENT |
		CTX32_FLOATING_POINT | CTX32_XMM | CTX32_BT_EXTENDED;

	btg_ret = BTGENERIC_IA32_CONTEXT_SET (NULL, &clone_params->start_context);

	retval = btlib_check_segment (cur_thread_data->bc32);

	/* 
	 * if the base address of one of the segments changed then 
	 * it must be that clone was called with clone-settls
	 */
	assert ((retval == 0) || (flags & CLONE_SETTLS));

	assert (btg_ret == BT_STATUS_SUCCESS);

	retval = set_sigmask (&cur_thread_data->sigmask);
	sys_err_chk (retval, "set_sigmask");


	/* put tid in child/parent-tidptr and signal the parent */
	DBPRINT (0, "child_tidptr=%x child_sem=%p\n",
		 cur_thread_data->child_tidptr, clone_params->child_sem);
	if (cur_thread_data->parent_tidptr != 0) {
		*cur_thread_data->parent_tidptr = cur_thread_data->tid;
	}
	if (cur_thread_data->child_tidptr != 0) {
		*cur_thread_data->child_tidptr = cur_thread_data->tid;
	}
	if (clone_params->child_sem != NULL) {
		clone_params->child_tid = cur_thread_data->tid;
		assert (g_is_mme_mode);
		btlib_signal_semaphore (&clone_params->child_sem);
	} else {
		/* clone_params is not needed anymore */
		bt_free (clone_params);
	}

	/* a new thread needs to be stopped with sigstop */
	if (!g_is_mme_mode ) {
		if (get_cur_thread_ptrace ()) {
			assert (get_cur_thread_ptrace () |  PTEMU_PTRACED);
			bt_memset (&cur_thread_data->tia_signal, 0, 
				   sizeof (cur_thread_data->tia_signal));
			cur_thread_data->tia_signal.siginfo.si_signo = SIGSTOP;
			retval = btlib_ptemu_sig0_selfkill (TRUE);
			sys_err_chk (retval, "kill");
		}
	}

	return 0;
}

/**************************************************************************/
/* set the sighand table for the new thread
 * input : clone_flags - the flags for the clone syscall
 *         parent - the parent's thread data struct
 *         child - the child's thread data struct
 * output: none
 */
/**************************************************************************/
static void
clone_signals (unsigned clone_flags, btlib_thread_data_t * child)
{
	struct signal_handler_table *cur_sighand;

	assert (child != NULL);

	cur_sighand = cur_thread_data->sighandlers;

	btlib_acquire_lock (BTL_LOCK_IND_SIGHAND);
	if (clone_flags & CLONE_SIGHAND) {
		cur_sighand->ref_count++;
		child->sighandlers = cur_sighand;
	} else {
		child->sighandlers = btlib_create_sighand_table ();
		assert (child->sighandlers != NULL);

		bt_memcpy (child->sighandlers->handlers, cur_sighand->handlers,
			   sizeof (struct signal_handler_desc) * (_NSIG + 1));
	}
	btlib_release_lock (BTL_LOCK_IND_SIGHAND);

	GCC_BUILTIN_SYNC_SYNCHRONIZE ();
}

/**************************************************************************/
/* IA-32 set_tid_address syscall implementation */
/**************************************************************************/
long
btlib_set_tid_address (const BTGENERIC_IA32_CONTEXT * bc32)
{
	long retval;

	/* tidptr points to data page, disable smc protection on it*/
	btlib_disable_smc_opt((unsigned long)IA32_SYSCALL_PARAM1(bc32), sizeof(int));
	/* abort execution if set_tid_address is not supported */
	retval = SYSCALL (set_tid_address, (U2P (IA32_SYSCALL_PARAM1 (bc32))));
	sys_err_chk (retval, "set_tid_address");

	return retval;
}

#if ALL_THREAD_AFFINITY
/******************************************************************************
 btlib_record_explicit_affinity_request - records the affinity mask that was
 set due to the user's (IA-32 thread) explicit request. This mask will override
 the mask seleceted by the thread-affinity or shared-mem-process-affinity 
 solutions.This mask will be inherited in case of clone or fork.

 This function is called only after the set_affinity ia32-sys-call was passed
 to the wrapper and returned successfully.
*******************************************************************************/
void
btlib_record_explicit_affinity_request (int target_tid)
{
	long retval;

	/* Note: the user may request affinity mask that inlcudes more 
	   processors that are actually available. The kernel ANDs the 
	   available cpus with the user's requested cpus and sets the actaul 
	   affinity mask accordingly. This is the affinity mask
	   we'll record here and not the user's initial request */
	cpu_set_t current_affinity_mask;

	/* check if the affinity mask is for current thread,
	   if not, we do nothing about it for now... */
	if (target_tid == cur_thread_data->tid) {
		bt_memset (&current_affinity_mask, 0, sizeof (current_affinity_mask));
		retval = SYSCALL (sched_getaffinity,
			          (target_tid,
			          (unsigned int) sizeof (current_affinity_mask),
			          &current_affinity_mask.__bits[0]));
		sys_err_chk (retval, "sys_sched_getaffinity");
		DBPRINT (btl_prt_thread,
			 "tid=%d explicit affinity mask requested=%lx\n",
			 target_tid, current_affinity_mask.__bits[0]);
		cur_thread_data->explicit_affinity_mask = current_affinity_mask;

		if (m_did_clone == 0) {
			/* this is the main thread that asks for explicit 
			   affinity_mask (before any clones), this will be 
			   recorded as the process's orig mask */
			m_orig_process_affinity_mask = cur_thread_data->explicit_affinity_mask;
		}
	}
	return;
}

/**************************************************************************
get_current_cpu - attempts to read from /proc/self/stat and retrieve
 the current processor the thread is executing on.
Return Value:
 Success: the processoor num the thread is currently executin on
 Failure: -1
**************************************************************************/
#define STAT_FILE "/proc/self/stat"
#define STAT_STR_LENGTH 512ul
int
get_current_cpu ()
{
	int stat_file;
	char stat_str[STAT_STR_LENGTH];
	int current_processor = -1, retval;

	/* read from /proc/self/stat file, which is the processor we're executing on */
	stat_file = SYSCALL (open, (STAT_FILE, O_RDONLY, 0));
	if (IS_SYS_ERROR (stat_file)) {
		DBPRINT (btl_prt_thread, "Error: open failed: %d\n", stat_file);
		return -1;
	} else {
		retval = SYSCALL (read, (stat_file, stat_str, STAT_STR_LENGTH));
		if (IS_SYS_ERROR (retval)) {
			DBPRINT (btl_prt_thread,
				 "Error: failed to read from %s: %d\n",
				 STAT_FILE, retval);
			SYSCALL (close, (stat_file));
			return -1;
		} else {
			char *p = bt_strchr (stat_str, ')');
			int j = 1;
			while ((p != NULL) && (j < 37)) {
				p = bt_strchr (p + 1, ' ');
				j++;
			}
			if (j == 37) {
				p = bt_strchr (p + 1, ' ');
				if (p != NULL) {
					*p = '\0';
					current_processor = bt_atol (p + 1);
				}
			}

			DBPRINT (btl_prt_thread, "current processor - %d\n",
				 current_processor);
		}
		SYSCALL (close, (stat_file));
	}

	return current_processor;
}

/**************************************************************************
get_random_cpu - randomly selects a CPU number out of all the allowed CPUs
                 according to the input cpu_mask
Return Value:
 Success: the selected CPU number
 Failure: -1
**************************************************************************/
int
get_random_cpu (cpu_set_t * cpu_mask)
{
	int num_of_allowed_cpus, cpu_X_out_of, randnum;
	unsigned int cpu;

	/* how many cpus are allowed according to cpu mask */
	num_of_allowed_cpus = 0;
	for (cpu = 0; cpu < 8 * sizeof (*cpu_mask) - 1; cpu++) {
		if (__CPU_ISSET (cpu, cpu_mask)) {
			num_of_allowed_cpus++;
		}
	}

	/* now get a random number */
	{
		struct timeval tv;
		long ret = SYSCALL (gettimeofday, (&tv, NULL));
		if (IS_SYS_ERROR (ret)) {
			randnum = 0;
		} else {
			randnum = tv.tv_sec;
		}
	}

	/* so we'll use the X'th cpu out of these allowed */
	cpu_X_out_of = randnum % num_of_allowed_cpus;
	for (cpu = 0; cpu < 8 * sizeof (*cpu_mask) - 1; cpu++) {
		if (__CPU_ISSET (cpu, cpu_mask)) {
			if (cpu_X_out_of == 0) {
				DBPRINT (btl_prt_thread,
					 "randomly selected cpu %d\n", cpu);
				return cpu;
			}
		}
		cpu_X_out_of--;
	}

	/* if we get here, some error happened... */
	DBPRINT (btl_prt_thread, "Error: couldn't select a cpu\n");
	return -1;
}

/**************************************************************************
select_one_cpu - This function implements the heuristic for selecting only
 one cpu for all the process's threads to run on, out of all allowed cpus 
 according to the cpu_mask.
Return Value:
 Success: the new cpu-affinity-mask that contains only one cpu
 Failure: 0 (empty affinity-mask)
**************************************************************************/
static int
select_one_cpu (cpu_set_t * cur_mask, cpu_set_t * new_mask)
{
	int cpu;

	DBPRINT (btl_prt_thread, "allowed cpu_mask=%lx\n", cur_mask->__bits[0]);

	/* try getting current cpu we're executing on */
	cpu = get_current_cpu ();
	if (cpu < 0) {
		cpu = get_random_cpu (cur_mask);
	}
	DBPRINT (btl_prt_thread, "new_mask=%lx\n", new_mask->__bits[0]);
	if (cpu >= 0) {
		__CPU_SET (cpu, new_mask);
	}
	assert (is_only_one_cpu_allowed (new_mask));
	assert (cpu >= 0);
	return (cpu >= 0) ? 0 : -1;

}

#if DEBUG
/**************************************************************************
is_only_one_cpu_allowed - make sure that the input cpu_mask specifies only
one cpu.
Return Value:
 Success: 1
 Failure: 0
**************************************************************************/
int
is_only_one_cpu_allowed (cpu_set_t * cpu_mask)
{
	unsigned int cpus_allowed;
	unsigned long cpu;

	cpus_allowed = 0;

	DBCODE (0, __CPU_PRINT ("is_only_one_cpu_allowed", cpu_mask));
	for (cpu = 0; cpu < 8 * sizeof (*cpu_mask) - 1; cpu++) {
		if (__CPU_ISSET (cpu, cpu_mask)) {
			cpus_allowed++;
		}
	}

	if (cpus_allowed == 1)
		return 1;

	return 0;
}
#endif				/* DEBUG */
#endif				/* ALL_THREAD_AFFINITY */
/******************************************************************************/
/* 
 * implement the IA-32 clone syscall 
 *	param1 - clone flags
 *	param2 - child stack
 *	param3 - parent tid pointer
 *	param4 - new TLS area for new thread (child), user_desc pointer
 *	param5 - child tid pointer
 *
 * NOTES: some clone flags must be handled inside BTLib
 *	CLONE_SETTLS : 
 *	1. It requests that the very first userland instruction of the new thread 
 *         can access thread local data by thread pointer. 
 *	2. No limitation for the parent thread. 
 *	3. BTLib must handle this by pasinging its own TLS pointer to kernel 
 *         because IA-32 kernel implements TLS using GDT.
 *
 *	CLONE_PARENT_SETTID/CLONE_CHILD_SETTID : 
 *	The parent_tidptr must be filled with new thread's tid before both 
 *	parent and child are back to userland. There's no way for BTLib
 *	to guarantee that, so it must pass kernel tid pointer of translated 
 *	process.
 *
 *	CLONE_CHILD_CLEARTID :
 *	The kernel takes child_tidptr as a futex, record child_tidptr somewhere.
 *	When the thread exits, the kernel is responsible to clear it, and wake 
 *	up thread waiting on it. BTLib should passing kernel its own 
 *	child_tidptr because we want to use it for BTLib thread memory 
 *	management. And for child_tidptr of translated process, BTLib should 
 *	should simulate kernel behavior
 *
 *	xlin2: usage of CLONE_CHILD_CLEARTID changed because there is a race
 *             condition: the thread waiting on the futex can be waken up even
 *             if the child thread call exit_group
 */
/******************************************************************************/
long
btlib_clone (BTGENERIC_IA32_CONTEXT * bc32)
{
	void *stacktop = 0;
	unsigned ia32_stack = 0;
	btlib_clone_data_t *clone_params;
	//btlib_thread_desc   *ctd;
	btlib_thread_data_t *child;
	void *child_clst;
	btlib_thread_data_t *parent;
	unsigned clone_flags;
	long retval, r;
	void *new_r13;
	void *parent_tidptr = NULL, *child_tidptr = NULL;
	k_sigset_t mask, oldmask;
#if ALL_THREAD_AFFINITY
	cpu_set_t current_affinity_mask, new_affinity_mask;
	int tid;
#endif	/* ALL_THREAD_AFFINITY */

	/* Locate the thread data for the current thread. */
	parent = cur_thread_data;
	assert (parent->tid == SYSCALL (gettid, ()));

	/* get the syscall params */
	clone_flags = IA32_SYSCALL_PARAM1 (bc32);
	ia32_stack = IA32_SYSCALL_PARAM2 (bc32);
	DBPRINT (btl_prt_thread, "clone_flags=0x%x ia32_stack=0x%x\n",
		 clone_flags, ia32_stack);

	if (g_has_smc) btlib_disable_smc_opt(ia32_stack, PAGE_SIZE);
	/* don't handle shared pid */
	if (clone_flags & CLONE_PID) {
		DBPRINT (TRUE, "someone is requesting a shared pid!\n");
		return -EPERM;
	}
	/* if the memory is not shared then a special treatment is required */
	if (!(clone_flags & CLONE_VM)) {
#if DEBUG
		if (clone_flags & CLONE_SETTLS) {
			assert (!"Don't support!\n");
		}
#endif
		retval = do_fork (parent, clone_flags, ia32_stack,
				  U2P (IA32_SYSCALL_PARAM3 (bc32)),
				  U2P (IA32_SYSCALL_PARAM5 (bc32)), bc32);

		return retval;
	}
	/* verify that we run on an adequate kernel */
	if (IS_NPTL_FLAGS (clone_flags)) {
		if (!m_kernel_support_nptl) {
			DBPRINT (TRUE, "CLONE_THREAD is called on an old kernel\n");
			return -EINVAL;
		}
	}

	/* allocate new stack and thread data for the child thread. */
	stacktop = allocate_thread_data (&new_r13, &child, &child_clst);
	DBPRINT (btl_prt_thread, "stacktop=%p\n", stacktop);
	if (IS_SYS_ERROR (stacktop)) {
		return (long) stacktop;
	}

	/* allocate pending que for new thread */
	btlib_que_init (&child->pending_que);

	child->tid = 0;
	/* 
	 * param4 is a pointer to user_desc (TLS area of new thread) 
	 * if CLONE_SETTLS is specified 
	 */
	if (clone_flags & CLONE_SETTLS) {
		/* 
		 * set child's TLS here because if validating user_desc fails, 
		 * the 64 bit clone should not be called!
		 */
		int idx;
		user_desc u_info;

		retval = safe_memcpy (&u_info,
				      U2P (IA32_SYSCALL_PARAM4 (bc32)),
				      sizeof (u_info));
		if (retval != 0) {
			return -EFAULT;
		}

		if (is_udesc_empty (&u_info)) {
			return -EINVAL;
		}

		idx = u_info.entry_number;

		if (idx < GDT_USER_ENT_FIRST || idx > GDT_USER_ENT_LAST) {
			return -EINVAL;
		}
		set_desc (&child->tls_array[idx - GDT_USER_ENT_FIRST], &u_info);
	} else {
		/* child should have same tls array with parent if not specifying 
		   CLONE_SETTLS */
		bt_memcpy (child->tls_array, cur_thread_data->tls_array,
			   sizeof (desc_struct) * GDT_USER_ENT_NUM);
	}

	if (clone_flags & CLONE_THREAD) {
		m_thread_group = 1;
	}

	clone_signals (clone_flags, child);

	/* the clone-params struct will be released in clone_main */
	clone_params = bt_malloc (sizeof (*clone_params));
	if (clone_params == NULL) {
		DBPRINT (TRUE, "bt_malloc failed\n");
		cleanup_thread (child->stackblock, TRUE);
		return -ENOMEM;
	}

	clone_params->child_sem = NULL;
	clone_params->child_tid = 0;
	clone_params->start_context = *bc32;
	clone_params->blst = child;
	clone_params->clst = child_clst;
	sigset32_2_sigset64 (&cur_thread_data->sigmask, &clone_params->sigmask);

	if ( ! g_is_mme_mode ) {
		if (!(clone_flags & CLONE_UNTRACED)&& !(clone_flags & CLONE_DETACHED) ) {
			*thread_ptrace (new_r13) = get_cur_thread_ptrace ();
		}
	}

	if ((clone_flags & CLONE_PARENT_SETTID) ||
	    (clone_flags & CLONE_CHILD_SETTID)) {
		/* CLONE_VM must be specified upon here, so CLONE_PARENT_SETTID and
		   CLONE_CHILD_SETTID is exactly the same */
		parent_tidptr = U2P (IA32_SYSCALL_PARAM3 (bc32));
		if (g_has_smc) btlib_disable_smc_opt(parent_tidptr, sizeof(int));
	}

	if (clone_flags & CLONE_CHILD_CLEARTID) {
		child_tidptr = U2P (IA32_SYSCALL_PARAM5 (bc32));
		if (g_has_smc) btlib_disable_smc_opt(child_tidptr, sizeof(int));
	}

	DBPRINT (btl_prt_thread,
		 "stackblock=%p stacksize=%lx flags=%x parent_tidptr=%p, new_r13=%p child_tidptr=%p\n",
		 child->stackblock, child->user_stack_size, clone_flags,
		 parent_tidptr, new_r13, child_tidptr);

#if ALL_THREAD_AFFINITY
	/* affinity solution for inter-thread ordering hazard */
	if (m_did_clone == 0
#if !RELEASE
	    && btl_no_affinity == 0
#endif
	    ) {
		/* this is the first time clone is called */
		m_did_clone = 1;

		/* Before narrowing the thread's affinity mask, we must verify that
		   (1) shared-mem-process-affinity or shared-mem-ordering solution  or switch to
		   ordered code wasn't already applied
		   (2) explicit-mask is still 0, we don't want to override user's explicit request */
		if ((m_did_shared_mem == 0 && btl_did_ordered_code == 0) &&
		    __CPU_IS_ZERO (&cur_thread_data->explicit_affinity_mask)) {

			assert ( ! g_is_mme_mode );
			/* just before calling clone, we'll reduce the affinity mask of the 
			   thread to allow only one CPU. The affinity mask will be inherited by the
			   child thread */
			tid = SYSCALL (gettid, ());
			bt_memset (&current_affinity_mask, 0, sizeof (current_affinity_mask));
			bt_memset (&new_affinity_mask, 0, sizeof (new_affinity_mask));
			retval = SYSCALL (sched_getaffinity, (tid,
				          (unsigned int) sizeof (current_affinity_mask),
				          &current_affinity_mask.__bits[0]));
			sys_err_chk (retval, "sys_sched_getaffinity");

			/* we'll save the original affinity mask of this process, so in case
			   of fork we'll be able to use it */
			m_orig_process_affinity_mask = current_affinity_mask;

			/* the heuristic for selecting one CPU out of all the allowed CPUs
			   is contained in the "select_one_cpu" function */
			retval = select_one_cpu (&current_affinity_mask, &new_affinity_mask);
			assert (retval == 0);

			retval = SYSCALL (sched_setaffinity, (tid,
					   (unsigned int) sizeof (new_affinity_mask),
					   &new_affinity_mask.__bits[0]));
			sys_err_chk (retval, "sys_sched_setaffinity");
			btl_narrowed_affinity = 1;
		}
	}
#if DEBUG
	if (__CPU_IS_ZERO (&cur_thread_data->explicit_affinity_mask) && btl_narrowed_affinity) {
		tid = SYSCALL (gettid, ());
		bt_memset (&current_affinity_mask, 0, sizeof (current_affinity_mask));
		retval = SYSCALL (sched_getaffinity, (tid,
					(unsigned int) sizeof (current_affinity_mask),
					&current_affinity_mask.__bits[0]));
		sys_err_chk (retval, "sys_sched_getaffinity");
		assert (is_only_one_cpu_allowed (&current_affinity_mask));
	}
#endif	/* DEBUG */
#endif	/* ALL_THREAD_AFFINITY */

	/* BTLib always uses thread local data */
	clone_flags |= CLONE_SETTLS;

	if ( ! g_is_mme_mode ) {
		/* child's blst will be assigned only in clone_main and the 
		 * pending que cannot be accessed before that. child must not
		 * get any signal before the pending que is ready */
		sigfillset (&mask);
		r = SYSCALL (rt_sigprocmask, (SIG_SETMASK, &mask, &oldmask, sizeof (k_sigset_t)));
		sys_err_chk (r, "rt_sigprocmask(1)");

		/* pass personality to child */
		assert (cur_thread_data->personality != NULL);
		clone_params->personality = *cur_thread_data->personality;

		/* create the child thread */
		retval = __btlib_clone (clone_main,
					child->stackblock,
					(unsigned long) (child->user_stack_size),
					(unsigned long) clone_flags,
					(void *) clone_params,
					parent_tidptr, new_r13, child_tidptr);

		/* restore the sigmask */
		r = SYSCALL (rt_sigprocmask, (SIG_SETMASK, &oldmask, NULL, sizeof (k_sigset_t)));
		sys_err_chk (r, "rt_sigprocmask(2)");
	} else { 
		/* mme mode */
		DBPRINT (0, "child_tidptr=%x parent_tidptr=%x\n",
			 child_tidptr, parent_tidptr);
		child->child_tidptr = child_tidptr;
		child->parent_tidptr = parent_tidptr;
		/* init the child semaphore */
		retval = btlib_init_semaphore (&clone_params->child_sem, 0);
		if (retval < 0) {
			DBPRINT (TRUE, "btlib_init_semaphore failed\n");
			goto out;
		}
		/* check whether this thread creation was initiated by MME */
		DBPRINT (0, "mme clone point\n");
		long ret = btlib_mme_is_ia64_thread (clone_params);
		if (ret < 0) {
			/* failure */
			retval = -1;
		} else {
			if (ret == 0) {
				/* not initiated by MME. */

				/* create the thread */
				ret = pthread_create (&cur_thread_data->pthread_h,
						      NULL, clone_main, clone_params);
				DBPRINT (0, "pthread_create=%d\n", ret);
				/* we only get here in parent's context */
				retval = btlib_complete_mme_parent_thread_create (clone_params, ret);
			} else {
				/* initiated by MME */
				retval = 0; /* compiler complains about 'use of uninitialized variable' */
				/* initiated by MME. need to return to MME directly here,
				 * so that execution of 32 bit pthread_create will stop here */
				longjmp (*cur_thread_data->btgeneric_run_context, RETURN);
				assert ( ! "should not get here" );
			}
		}
	}

out:
	DBPRINT (retval && btl_prt_thread, "child pid=%ld<%lx>\n", retval, retval);
	if (IS_SYS_ERROR (retval)) {
		DBPRINT (TRUE, "clone failed=%ld\n", retval);
		/* undo allocate_thread_data */
		cleanup_thread (child, TRUE);
	}

	return retval;
}

/******************************************************************************/
/******************************************************************************/
long btlib_complete_mme_parent_thread_create (void * params, int64_t retval)
{
	btlib_clone_data_t * clone_params = params;
	long ret, child_pid = retval;

	DBPRINT (0, "params=%p retval=%ld\n", params, retval);
	if (retval >= 0) {
		/* the semaphore will be released by the child after it
		 * puts the correct values in child/parent _tidptr fields
		 * in its cur_thread_data, and the tid of the new thread
		 * in the clone_param structure */
		(void)btlib_wait_semaphore (&clone_params->child_sem, TRUE);
		assert (clone_params->child_tid != 0);
		child_pid = clone_params->child_tid;
	}
	ret = btlib_destroy_semaphore (&clone_params->child_sem);
	assert (ret == 0);
	/* clone_params is not needed anymore */
	bt_free (clone_params);
	return child_pid;
}

/******************************************************************************/
/* Main func of native thread (64 bit)
 */
/******************************************************************************/
static void *
native_clone_main (void *arg)
{
	long retval;
	void **clstp;
	btlib_native_clone_data_t *native_clone_params = (btlib_native_clone_data_t *) arg;

	DBPRINT (0, "\n****************** Inside new Native thread *****************\n");
	/* Grab the existance Semaphore */
	btlib_wait_semaphore((void *)(&(native_clone_params->thread_exist_sem)), 1);

	if (native_clone_params->thread_should_die == 0) {
		/* init the thread desc */
		clstp = (void **) (__r13 + g_clstp_offset);
		*clstp = native_clone_params->clone_data.clst;
		cur_thread_data->__r13 = (uintptr_t) __r13;

		assert (cur_thread_data->in_critical_section == 0);
		/* store the new thread id */
		cur_thread_data->tid = SYSCALL (gettid, ());

		retval = (long) BTGENERIC_THREAD_INIT (cur_thread_clst, (__u32) 1, (__u64) 0);

		assert (retval == (long) BT_STATUS_SUCCESS);

		/* Run the actual thread's code */
		BTGENERIC_NATIVE_THREAD_START ((__u64) native_clone_params->native_thread_type);

		/* After returning from BTGeneric (Thread is about to end) */
		BTGENERIC_THREAD_TERMINATED ();
	}

	/* Release the existance Semaphore */
	btlib_signal_semaphore((void *)(&(native_clone_params->thread_exist_sem)));

	return NULL;
}

/******************************************************************************/
/* clone a native (64 bit) thread
 */
/******************************************************************************/
void *
btlib_clone_native (int native_thread_type)
{
	void *stacktop = 0;
	btlib_native_clone_data_t *clone_params;
	void *child_clst;
	btlib_thread_data_t *child;
	btlib_thread_data_t *parent;
	unsigned clone_flags;
	long retval, r, sem;
	void *new_r13;
	k_sigset_t mask, oldmask;


	/* Locate the thread data for the current thread. */
	parent = cur_thread_data;
	assert (parent->tid == SYSCALL (gettid, ()));

	/* get the syscall params */
	clone_flags =
		CLONE_VM | CLONE_SIGHAND | CLONE_FS | CLONE_FILES | CLONE_SETTLS |
		CLONE_THREAD | CLONE_DETACHED | CLONE_CHILD_SETTID | CLONE_CHILD_CLEARTID;

	/* verify that we run on an adequate kernel */
	if (IS_NPTL_FLAGS (clone_flags)) {
		if (!m_kernel_support_nptl) {
			DBPRINT (TRUE, "CLONE_THREAD is called on an old kernel\n");
			return 0;
		}
	}
	/* allocate new stack and thread data for the child thread. */
	stacktop = allocate_thread_data (&new_r13, &child, &child_clst);
	if (IS_SYS_ERROR (stacktop)) {
		return 0;
	}

	/* allocate pending que for new thread */
	btlib_que_init (&child->pending_que);

	child->tid = 0;

	//clone_signals (clone_flags, child);

	/* Below is a tailored TT solution for CS: */
	//child->sighandlers->cs = btlib_create_critical_section();
	//btlib_init_critical_section(child->sighandlers->cs);
	clone_signals (0, child);

	/*
	 * the clone-params struct will be released in clone_main. it
	 * is very rare that more than one such a struct will be used
	 * simultanuously. therefore, there is no benefit in using the
	 * memory pools here, only an overhead
	 */
	clone_params = bt_malloc (sizeof (*clone_params));
	if (clone_params == NULL) {
		DBPRINT (TRUE, "bt_malloc failed\n");
		cleanup_thread (child->stackblock, TRUE);
		return 0;
	}

	/* Connect structure in native thread list */
	clone_params->next = native_thread_list;
	native_thread_list = clone_params;
	
	clone_params->clone_data.blst = child;
	clone_params->clone_data.clst = child_clst;
	clone_params->native_thread_type = native_thread_type;
	clone_params->thread_should_die = 0;
	clone_params->thread_tid = -1;
	sem = btlib_init_semaphore((void *)(&(clone_params->thread_exist_sem)), 1);
	if (sem < 0) {
		DBPRINT (TRUE, "Creating existance semaphore failed\n");
		cleanup_thread (child->stackblock, TRUE);
		return 0;
	}

	DBPRINT (btl_prt_thread,
		 "stackblock=%p stacksize=%lx flags=%x new_r13=%p \n",
		 child->stackblock, child->user_stack_size, clone_flags,
		 new_r13);

	if ( ! g_is_mme_mode ) {
		sigfillset (&mask);
		retval = SYSCALL (rt_sigprocmask,
				  (SIG_SETMASK, &mask, &oldmask, sizeof (k_sigset_t)));
		sys_err_chk (retval, "rt_sigprocmask(1)");
		clone_params->clone_data.sigmask = oldmask;

		retval = __btlib_clone (native_clone_main,
					child->stackblock,
					(unsigned long) (child->user_stack_size),
					(unsigned long) clone_flags,
					(void *) clone_params,
					NULL, new_r13, &(clone_params->thread_tid));

		r = SYSCALL (rt_sigprocmask,
			     (SIG_SETMASK, &oldmask, NULL, sizeof (k_sigset_t)));
		sys_err_chk (r, "rt_sigprocmask(2)");
	} else { 
		/* mme mode */
		retval = pthread_create (&cur_thread_data->pthread_h, NULL, clone_main, clone_params);
	}

	DBPRINT (retval
		 && btl_prt_thread, "child pid=%ld<%lx>\n", retval, retval);
	if (IS_SYS_ERROR (retval)) {
		DBPRINT (TRUE, "clone failed=%ld\n", retval);
		/* undo allocate_thread_data */
		cleanup_thread (child, TRUE);
		return 0;
	}

	return (void *)(clone_params);
}

/******************************************************************************/
/* terminate a native (64 bit) thread
 */
/******************************************************************************/
void
btlib_terminate_native_thread (void * thread_params)
{
	btlib_native_clone_data_t *native_clone_params =
		(btlib_native_clone_data_t *) thread_params;
	
	BTGENERIC_NATIVE_THREAD_TERMINATE(thread_params);

	/* Grab the existance Semaphore */
	btlib_wait_semaphore((void *)(&(native_clone_params->thread_exist_sem)), 1);

	// Just in case we caught the existance semaphore BEFORE The thread started living, we 
	// mark "thread_should_die" so the thread wont begin running
	native_clone_params->thread_should_die = 1;

	btlib_signal_semaphore((void *)(&(native_clone_params->thread_exist_sem)));

	lll_wait_tid (native_clone_params->thread_tid);
}

/******************************************************************************/
/* terminate all native (64 bit) threads
 * no lock to protect native thread list, because it will conflict with a 
 * BTGeneric lock later.
 */
/******************************************************************************/
void
btlib_terminate_all_native_thread (void)
{
	btlib_native_clone_data_t* thread;
	
	for (thread = native_thread_list; thread != NULL; thread = thread->next ){
		btlib_terminate_native_thread(thread);
	}
}

/**************************************************************************/
/* 
 * group termination procedure
 * called when the translated process calls the exit_group syscall
 */
/**************************************************************************/
void
btlib_group_exit (const BTGENERIC_IA32_CONTEXT * bc32)
{
	int exit_code = IA32_SYSCALL_PARAM1 (bc32);

	
	BLOCK_ALL_SIGNALS ();

	/* if this syscall is called in a thread that is not a part
	   of a thread group ( as in linuxthreads ) then it should be
	   treated as exit() */
	DBPRINT (btl_prt_thread, "m_thread_group=%d\n", m_thread_group);
	if (!m_thread_group && btl_num_threads > 1) {
		before_thread_exit ();
		__btlib_exit (exit_code, cur_thread_data->stackblock,
			      __default_stack_size);
	}
	STATCODE (btl_stats, btlib_print_stats ());

	/* let btgeneric prepare stats on IA32-EL run */
	BTGENERIC_NOTIFY_EXIT ();

        detach_all_debugee ();
	SYSCALL (exit_group, (exit_code));
	RESTORE_SIGNALS ();	/* dummy - should never be reached */
}

/**************************************************************************/
/* 
 * normal termination procedure
 * called when the translated process calls the IA-32 exit() syscall
 */
/**************************************************************************/
void
btlib_thread_exit (BTGENERIC_IA32_CONTEXT * bc32)
{
	int exit_code;
	int signo;
	siginfo_t siginfo;
        int syscall_id;

	assert (cur_thread_data->tid == SYSCALL (gettid, ()));
	assert (bc32 == cur_thread_data->bc32);

	assert (!btlib_is_inside_lock (cur_thread_data));
	assert (!is_holding_critical_section (cur_thread_data));

	/* get syscall params */
	exit_code = IA32_SYSCALL_PARAM1 (bc32);

	DBPRINT (0, "exiting %d\n", exit_code);

	/* we dont want to get signals from this point and further */
	BLOCK_ALL_SIGNALS ();

	/* see if we have pending signals */
	signo = btlib_que_deque (&siginfo, &syscall_id);
	if (signo != 0) {
		/* the signal is not blocked. it needs to be handled. */

		/* 1 - set the IA-32 context to repeat the call to exit() */
		bc32->eip -= 2;
		bc32->eax = IA32_SYSCALL_EXIT;
		cur_thread_data->syscall_id = IA32_SYSCALL_NONE;
		/* 2 - deliver the signal. continue with signals masked -
		   it shouldn't matter much */
		btlib_run_ia32_handler (&siginfo, NULL, 0);
		assert (!"should never get here");
	}

	before_thread_exit ();

	if (g_is_mme_mode) {
		/* clear the child_tidptr field */
		assert (cur_thread_data->child_tidptr != 0);
		DBPRINT (0, "cur_thread_data->child_tidptr=%p *cur_thread_data->child_tidptr=%d\n",
			 cur_thread_data->child_tidptr,
			 (cur_thread_data->child_tidptr) ? *cur_thread_data->child_tidptr : -1);
		if (cur_thread_data->child_tidptr != 0) {
			*cur_thread_data->child_tidptr = 0;
			/* we know that the parent, and only the parent, is waiting */
			int r = SYSCALL (futex, (cur_thread_data->child_tidptr, FUTEX_WAKE, 1, NULL, NULL));
			if (IS_SYS_ERROR (r)) {
				DBPRINT (TRUE, "futex failed \n");
			}
		}
		/* in mme mode a 32-bit thread will always have 64-bit thread 
		 * resource attached, so we should return from there, and not
		 * call sys_exit() for this 32-bit thread
		 * also markup a flag so we can return to clone_main from restart_btg()
		 */
		cur_thread_data->ia32_thread_exited = 1;
		return;
	}

	/* all signals must be blocked! */
	__btlib_exit (exit_code, cur_thread_data->stackblock,
		      __default_stack_size);

	RESTORE_SIGNALS ();	/* dummy - never reached */
}

/******************************************************************************/
/* implement the IA-32 fork syscall */
/******************************************************************************/
long
btlib_fork (BTGENERIC_IA32_CONTEXT * bc32)
{
	long retval;

	assert (cur_thread_data->tid == SYSCALL (gettid, ()));

	/* do the actual work */
	retval = do_fork (cur_thread_data, SIGCHLD, bc32->esp, NULL, NULL, bc32);

	return retval;
}

/**************************************************************************/
/**************************************************************************/
static long
do_fork (btlib_thread_data_t * parent RELEASE_UNUSED,
	 unsigned clone_flags,
	 unsigned ia32_tos,
	 int *parent_tidptr, int *child_tidptr, BTGENERIC_IA32_CONTEXT * bc32)
{
	long retval;
	btlib_thread_data_t *p = NULL;
	BT_STATUS_CODE btg_retval = BT_STATUS_UNSUCCESSFUL;
	btlib_native_clone_data_t *temp_native_list = NULL;
#if ALL_THREAD_AFFINITY
	cpu_set_t current_affinity_mask;
	long aff_retval;
#endif	/* ALL_THREAD_AFFINITY */
	unsigned long parent_ptrace_data =
			(!g_is_mme_mode) ? *cur_thread_ptrace() : 0;
	int clear_child_ptrace_data = FALSE;

	assert (parent->tid == SYSCALL (gettid, ()));
	DBPRINT (btl_prt_thread, "clone_flags=0x%x ia32_tos=0x%x\n",
		 clone_flags, ia32_tos);

	if (g_has_smc) {
		btlib_disable_smc_opt(ia32_tos, PAGE_SIZE);
		btlib_disable_smc_opt(parent_tidptr, sizeof(int));
		btlib_disable_smc_opt(child_tidptr, sizeof(int));
	}
	/* 
	 * freeze ia32exec.bin to ensure that its state is
	 * consistent in the child process. 
	 */
	BTGENERIC_PREPARE_PROCESS_FORK ();

	/* 
	 * freeze btlib
	 * **********
	 * DONT EVER move it before the call to BTGENERIC_FREEZE
	 * **********
	 * (see the header of btlib_enter_critical_section)
	 */

	/* acquire all the sync objects used by btlib */
	btlib_freeze ();

	/* mask all the signals, so the child dont get one before it is ready */
	BLOCK_ALL_SIGNALS ();

#if ALL_THREAD_AFFINITY
	/* before we do actual fork...
	   if we narrowed the thread's affinity mask to allow only one processor
	   (due to threads\process affinity solution), we don't want this to be
	   inherited by the new process we'll craete.
	   So first change the affinity mask to be the original mask as before we 
	   narrowed it.
	   After the fork work is completed, then the narrowed affinity mask will 
	   be restored for the parent process only.

	   BUT, in case the current thread explicitly called set-affinity, the 
	   affinity mask he requested overrides the affinity-solution selected mask 
	   and the process's original affinity mask, therefore we'll let the child 
	   process inherit the current affinity mask. */
	if (btl_narrowed_affinity &&
	    __CPU_IS_ZERO (&cur_thread_data->explicit_affinity_mask)) {
		bt_memset (&current_affinity_mask, 0,
			   sizeof (current_affinity_mask));
		aff_retval = SYSCALL (sched_getaffinity,
			     	      (cur_thread_data->tid,
			      	      (unsigned int) sizeof (current_affinity_mask),
				      &current_affinity_mask.__bits[0]));
		sys_err_chk (aff_retval, "sys_sched_getaffinity");

		aff_retval = SYSCALL (sched_setaffinity,
				      (cur_thread_data->tid,
				       (unsigned int) sizeof (m_orig_process_affinity_mask),
				       &m_orig_process_affinity_mask.__bits[0]));
		sys_err_chk (aff_retval, "sys_sched_setaffinity");

	}
#endif	/* ALL_THREAD_AFFINITY */

	assert ((clone_flags & CLONE_VM) == 0);
	assert ((clone_flags & CLONE_SETTLS) == 0);
	retval = SYSCALL (clone2,
			  ((unsigned long) clone_flags,
			   (unsigned long) NULL, 0l,
			   (unsigned long) parent_tidptr,
			   (unsigned long) child_tidptr, (unsigned long) NULL));

	if (retval) {

		/************** we are in the parent */
		DBPRINT (0, "in parent: child tid=%ld\n", retval);

#if ALL_THREAD_AFFINITY
		/* just before returning, we must narrow again the affinity mask.
		   This is done only for the parent */
		if (btl_narrowed_affinity &&
		    __CPU_IS_ZERO (&cur_thread_data->explicit_affinity_mask)) {
			aff_retval = SYSCALL (sched_setaffinity,
					      (cur_thread_data->tid,
					       (unsigned int)
					       sizeof (current_affinity_mask),
					       &current_affinity_mask.
					       __bits[0]));
			sys_err_chk (aff_retval, "sys_sched_setaffinity");

		}
#endif	/* ALL_THREAD_AFFINITY */

		/* restore the signals and release the handlers table */
		RESTORE_SIGNALS1 ();

		/* release all the sync objects used by btlib */
		btlib_unfreeze (1l);

		/* btgeneric may now continue as usual */
		BTGENERIC_COMPLETE_PROCESS_FORK (cur_thread_data->tid, 0);

		/* return the child's tid */
		return retval;
	}

	/************** we are in the child process */

	/* 
	 * the child has a copy of its parent memory space, so their
	 * thread data structures reside at the same address
	 */
	assert (cur_thread_data == parent);
	cur_thread_data->tid = SYSCALL (gettid, ());

	/* re-init all critical sections */
	btlib_unfreeze (0l);

	btl_num_threads = 1;


	cur_thread_data->is_run_context_ready = cur_thread_data->tid;

	/* change the esp of the current process to the given stack */
	if (ia32_tos != 0) {
		bc32->esp = ia32_tos;
	}
#if ALL_THREAD_AFFINITY
	/*clear the global flags related to the affinity-ordering-solution */
	__CPU_ZERO (&m_orig_process_affinity_mask);
	m_did_clone = 0;
	btl_narrowed_affinity = 0;
#endif	
#if SHMEM_ORDER_SOLUTION || ALL_THREAD_AFFINITY
	m_did_shared_mem = 0;
#endif

	/* it's a single threaded app now */
	m_thread_group = 0;

	if (!g_is_mme_mode) {
		unsigned long mask = PTEMU_TRACE_FORK | PTEMU_TRACE_VFORK;
		*cur_thread_ptrace () = (*cur_thread_ptrace () & ~mask)
					| (parent_ptrace_data & mask);
	}

	if (clone_flags & CLONE_UNTRACED) {
		clear_child_ptrace_data = TRUE;
	} else if (clone_flags & CLONE_PTRACE) {
		clear_child_ptrace_data = FALSE;
	} else if (clone_flags & CLONE_VFORK) {
		/* if CLONE_VFORK is set we will set child be traced when
		   trace options contains PTEMU_TRACE_VFORK bit */
		if (parent_ptrace_data & PTEMU_TRACE_VFORK) {
			clear_child_ptrace_data = FALSE;
		} else {
			clear_child_ptrace_data = TRUE;
		}
	} else if (parent_ptrace_data & PTEMU_TRACE_FORK) {
		/* in this case CLONE_VFORK is not set, we will set child be
		   traced when trace options contains PTEMU_TRACE_FORK bit */
		clear_child_ptrace_data = FALSE;
	} else {
		clear_child_ptrace_data = TRUE;
	}

	if (clear_child_ptrace_data) {
		/* don't reset ptrace flag if the process is created with CLONE_PTRACE 
		   because its debug flag may be set by debugger */ 
		if (!g_is_mme_mode) {
			*cur_thread_ptrace () = 0ul;
			btl_is_ptraced = IS_PTRACED_NO;
		}
	}

	/* save original native_thread_list because BTGENERIC_COMPLETE_PROCESS_FORK may 
	   overwrite  native_thread_list */
	temp_native_list = native_thread_list;
	native_thread_list = NULL;

	/* 
	 * let btgeneric continue - it cannot do anything anyway
	 * because the current process is the only thread in the new
	 * app
	 */
	btg_retval = BTGENERIC_COMPLETE_PROCESS_FORK (cur_thread_data->tid, 1);
	assert (btg_retval == BT_STATUS_SUCCESS);

	/* 
	 * If the parent had some native threads, those were not cloned, but their
	 * memory structures (native_clone_data) are cloned. we need to free those.
	 * Also, the btlib_thread_data_t should be freed. 
	 */
	while (temp_native_list != NULL) {
		btlib_native_clone_data_t * next_thread = temp_native_list;
		temp_native_list = next_thread->next;

		btlib_destroy_semaphore ((void *)(&(next_thread->thread_exist_sem)));
		assert (next_thread->clone_data.blst->sighandlers->ref_count == 1);
		cleanup_thread (next_thread->clone_data.blst, TRUE);
		bt_free (next_thread);
	}

	/* 
	 * remove all the elements in the thread list 
	 * (no need to lock it since we are the only active thread )
	 */
	p = m_used_thread_list;
	while (p != NULL) {

		btlib_thread_data_t *next = p->next;
		if (p != cur_thread_data) {
			/* there is a chance that between the thread release the cs 
			   and in_critical_section is decreased, so there is possibility 
			   that in_critical_section is not 0 while the thread is not in
			   critical section */
			p->in_critical_section = 0; 
			btlib_cleanup_thread (p->tid);
		}
		p = next;
	}
	
	cur_thread_data->sighandlers->ref_count = 1; 

	/* from here on the current thread may receive signals */
	RESTORE_SIGNALS ();

	/* return success indication */
	return 0;
}

#if SHMEM_ORDER_SOLUTION || ALL_THREAD_AFFINITY
/**************************************************************************
set_affinity_to_all_threads - set the affinity mask of all the threads that
 didn't explicitly request a specific affinity mask

Return Value:
 Success: 0
 Failure: -1
**************************************************************************/
static long
set_affinity_to_all_threads (cpu_set_t * new_affinity_mask)
{
	long retval;
	btlib_thread_data_t *susp_thread;

	/* go over the existing threads list */
	btlib_acquire_lock (BTL_LOCK_IND_THREAD);

	for (susp_thread = m_used_thread_list; susp_thread != NULL;
	     susp_thread = susp_thread->next) {
		if (__CPU_IS_ZERO (&susp_thread->explicit_affinity_mask)) {
			/* this thread didn'e explicitly call set-affinity to 
			   request an affinity mask we can chagne his affinity 
			   mask */
			retval = SYSCALL (sched_setaffinity,
					  (susp_thread->tid,
					   (unsigned int)
					   sizeof (*new_affinity_mask),
					   &new_affinity_mask->__bits[0]));
			sys_err_chk (retval, "sys_sched_setaffinity");
		}
	}			/*for */

	btlib_release_lock (BTL_LOCK_IND_THREAD);

	return 0;
}
#endif	// SHMEM_PROC_AFFINITY || SHMEM_ALL_THREAD_AFFINITY

#if SHMEM_ORDER_SOLUTION
void
apply_shared_memory_ordering_solution ()
{
	EMULATION_POLICY emulation_policy;

	DBPRINT (btl_prt_mem || btl_prt_thread, "\n");

	if (btl_is_ptraced == IS_PTRACED_YES) {
		/* no sense in trying under debugger - required suspension will fail */
		DBPRINT (btl_prt_mem || btl_prt_thread,
			 "apply_shared_memory_ordering_solution not applied under debugger\n");
		return;
	}

	if (m_did_shared_mem == 0 && btl_did_ordered_code == 0
	    /* since can't put call to BTGENERIC_SET_EMULATION_POLICY in critical 
	       section, and since the call can theoretically fail => the two boolean's 
	       m_did_shared_mem and m_did_ordered_code are not set atomically 
	       (using __sync_bool_compare_and_swap_si before the call, since in that 
	       case if the call fails another thread might miss calling it again.  
	       having themnot set atomicaly merely means that two threads might call 
	       BTGENERIC_SET_EMULATION_POLICY, but since the change always goes in 
	       the same direction (to USE_ORDERED_LD_ST) and since there is a critical 
	       section in BTGENERIC_SET_EMULATION_POLICY - there is no correctness 
	       problem). */
#if ! RELEASE
	    && btl_no_affinity == 0
#endif
	    ) {
		emulation_policy.memory_order_policy = USE_ORDERED_LD_ST;
		emulation_policy.strict_state_policy = POLICY_UNCHANGED;

		BTGENERIC_SET_EMULATION_POLICY (&emulation_policy);
		assert (emulation_policy.memory_order_policy ==
			USE_ORDERED_LD_ST);
		btl_did_ordered_code = m_did_shared_mem = 1;
#if ALL_THREAD_AFFINITY
		if (GCC_BUILTIN_SYNC_BOOL_COMPARE_AND_SWAP (&btl_narrowed_affinity, 1, 0)) {
			assert (!__CPU_IS_ZERO (&m_orig_process_affinity_mask));
			/* give all the threads (that didn't explicitly change their affinity) 
			   the original process affinity */
			set_affinity_to_all_threads (&m_orig_process_affinity_mask);
		}
#endif
	}

	return;
}
#endif

/******************************************************************************
 This function will try to suspend all running threads, not including current. 
 If it fails to suspend an existing thread, it will still continue to the next 
 threads and attempt to suspend them.

 Note: After suspending other threads and before resuming them, calling 
       printf\DBPRINT is not allowed! This is because they might be holding 
       a lock that prinf needs this will cause a deadlock!!

Return Value:
 Success: 0 (even if not all threads were successfully suspended)
 Failure: -1
**************************************************************************/
long
suspend_all_threads (void)
{
	long retval;
	int status;
	btlib_thread_data_t *susp_thread;

	/* go over the existing threads list and try to suspend them */
	btlib_acquire_lock (BTL_LOCK_IND_THREAD);

	for (susp_thread = m_used_thread_list; susp_thread != NULL;
	     susp_thread = susp_thread->next) {
		if (susp_thread->tid == cur_thread_data->tid) {
			/* don't suspend current thread... */
			continue;
		}

		/* now attach to target thread and suspend it in the 64 bit level */
		retval = SYSCALL (ptrace, (PTRACE_ATTACH, susp_thread->tid, 0ul, 0ul));
		sys_err_chk (retval, "PTRACE_ATTACH failed");
		if (!IS_SYS_ERROR (retval)) {
			/* wait for the target thread to actually stop */
			do {
				DBPRINT (btl_prt_thread | btl_prt_pt,
					 "before-wait: pid=%d options=%x\n",
					 susp_thread->tid, WUNTRACED | __WALL);
				retval = SYSCALL (wait4, (susp_thread->tid, &status,
						  WUNTRACED | __WALL, NULL));
				DBPRINT (btl_prt_thread | btl_prt_pt,
					 "after-wait: ret=%ld status=%d\n",
					 retval, status);
			} while (retval == -EINTR);
			sys_err_chk (retval, "waitpid");
			assert ((int) retval == susp_thread->tid);

			if (__WIFSTOPPED (status)) {
				/* finally, the target thread is suspended, 
				   mark it as such */
				susp_thread->is_suspended = 1;
			}
		}
	}

	btlib_release_lock (BTL_LOCK_IND_THREAD);

	return 0;
}

/**************************************************************************
 This function will try to resume all threads
 that were suspended in the "suspend_all_threads" function.

Return Value:
 Success: 0 (all threads that were suspended were successfully resumed)
 Failure: -1
**************************************************************************/
long
resume_all_threads (void)
{
	long retval;
	btlib_thread_data_t *susp_thread;

	/* go over the existing threads list, look for suspended threads 
	   and try to resume them */
	btlib_acquire_lock (BTL_LOCK_IND_THREAD);

	for (susp_thread = m_used_thread_list; susp_thread != NULL;
	     susp_thread = susp_thread->next) {
		if (susp_thread->is_suspended) {
			/* ok, found a suspended thread */

			/* make sure a thread will not resume itself */
			assert (susp_thread->tid != cur_thread_data->tid);

			retval = SYSCALL (ptrace, (PTRACE_DETACH,
						   susp_thread->tid, 0ul, 0ul));
			sys_err_chk (retval, "PTRACE_DETACH failed");

			/* mark that the thread is no longer suspended */
			susp_thread->is_suspended = 0;
		}
	}			/* for */

	btlib_release_lock (BTL_LOCK_IND_THREAD);

	return 0;
}

/**************************************************************************/
/* 
 * suspends thread's execution in a consistent state
 * input:  tid - the thread to suspend
 *         max_count - max tries allowed
 * output: 0 on success ( thread suspended in a consistent state),
 *         -1 otherwise.
 * remark: this function is only for internal use, and the caller must take 
 * care of serilzation, that is, it should grarantee only one thread is doing 
 * suspendsion.
 */
/**************************************************************************/
long
btlib_suspend_thread (int tid, unsigned max_count)
{
	long retval;
        unsigned try_count = 0;
        btlib_thread_data_t *susp_thread;
        suspend_t *suspendeeP, *suspenderP = NULL;
        assert (cur_thread_data->tid == SYSCALL (gettid, ()));
        /* a thread should not suspend itself */
        assert (tid != cur_thread_data->tid);

        DBPRINT (btl_prt_thread, "suspend thread %d, max_count %d\n", tid, max_count);
        /* find the thread's data struct of the target thread */
        btlib_acquire_lock (BTL_LOCK_IND_THREAD);
        susp_thread = m_used_thread_list;
        while (susp_thread != NULL) {
                if (susp_thread->tid == tid) {
                        break;
                }
                susp_thread = susp_thread->next;
        }
        assert (susp_thread != NULL);
        btlib_release_lock (BTL_LOCK_IND_THREAD);

        suspendeeP = &(susp_thread->suspend_data);
	 suspendeeP->suspender_tid = cur_thread_data->tid; //my tid

        /* do't try to suspend a thread if the whole process is debugged*/
        if( btl_is_ptraced!=IS_PTRACED_NO || *thread_ptrace((void*)(susp_thread->__r13)) ){
                suspendeeP->suspender_tid = 0;
                return -1;
        }

        // get the suspender itself
        suspenderP = &(cur_thread_data->suspend_data);
        suspenderP->semaphore = 0;
        btlib_init_semaphore(&(suspenderP->semaphore),0);

repeat:
        suspendeeP->suspender_tid = cur_thread_data->tid; //my tid
        /* Here we also use sig_ptemu0, since suspend and debug don't accour and the same time  */
        retval = SYSCALL (tkill, ( tid, PTEMU_SIG0));
        if (IS_SYS_ERROR (retval)) {
                DBPRINT (btl_prt_thread, "tkill failed=%ld\n", retval);
                retval = -1;
                goto out;
	}
	/*wait suspendee to canonize itself */
        btlib_wait_semaphore ( &(suspenderP->semaphore), 1);
        switch( suspendeeP->canonize_result ) {
                case SUSPEND_STATE_CONSISTENT:
                case SUSPEND_STATE_CANONIZED:
                        //btlib_signal_semaphore( &(suspendeeP->semaphore) );
                        retval = 0;
                        break;
                case BAD_SUSPEND_STATE:
                        try_count ++;
                        if( try_count >= max_count ){
                                btlib_signal_semaphore( &(suspendeeP->semaphore) );
                                suspendeeP->suspender_tid = 0;
                                retval = -1;
                                break;;
                        }
                        else{
btlib_signal_semaphore( &(suspendeeP->semaphore) );
                                suspendeeP->suspender_tid = 0;
                                YIELD_THREAD_EXECUTION ( );
                                goto repeat;
                        }
                case SUSPEND_STATE_INACCESIBLE:
                        btlib_signal_semaphore( &(suspendeeP->semaphore) );
                        suspendeeP->suspender_tid = 0;
                        retval = -1;
                        break;;
                default:
                        assert(!"should not get this canonize result");
                        break;
        }

out:
        btlib_destroy_semaphore ( &(suspenderP->semaphore));
       	DBPRINT (btl_prt_thread, "suspend thread %d finish, retval = %ld\n", tid, retval);
        return retval;
}

/*
* the suspendee did canonize work in this handler and notify suspender the 
* canonize result.
*/
void btlib_suspend_handler(int signo UNUSED, siginfo_t* siginfo UNUSED,
        struct sigcontext* ctx){

        int suspender_tid = 0;
        btlib_thread_data_t *susp_thread;
        suspend_t *suspendeeP, *suspenderP;
        BT_THREAD_SUSPEND_STATE btg_retval;

        DBPRINT (btl_prt_thread, "suspendee tid %d\n", cur_thread_data->tid );
        // get the suspendee itself
        suspendeeP = &(cur_thread_data->suspend_data);
        suspendeeP->canonize_result = 0; //my tid
        suspendeeP->semaphore = 0;
        btlib_init_semaphore(&(suspendeeP->semaphore),0);

        /* find the suspender*/
        suspender_tid = suspendeeP->suspender_tid;
        assert(suspender_tid != 0);
 btlib_acquire_lock (BTL_LOCK_IND_THREAD);
        susp_thread = m_used_thread_list;
        while (susp_thread != NULL) {
                if (susp_thread->tid == suspender_tid) {
                        break;
                }
                susp_thread = susp_thread->next;
        }
        btlib_release_lock (BTL_LOCK_IND_THREAD);
        assert(susp_thread != NULL);

        suspenderP = &(susp_thread->suspend_data);



        BTGENERIC_IA64_CONTEXT bc64;
        sigctx2btgctx_64( ctx,&bc64);

  if( cur_thread_data->in_critical_section )
                btg_retval = BAD_SUSPEND_STATE;
        else
                btg_retval = BTGENERIC_CANONIZE_SUSPEND_CONTEXT(glstp_of_pctx(cur_thread_clst), &bc64,0);

        if( btg_retval == SUSPEND_STATE_CANONIZED ){
                assert ( !(cur_thread_data->in_critical_section) );
                btgctx2sigctx_64 (&bc64, ctx);
        }

        //btgctx2sigctx_64 (&bc64, ctx);
        suspendeeP->canonize_result = btg_retval;

        DBPRINT (btl_prt_thread, "canonize result = %d\n", btg_retval );
        /* wake up the suspender to check the canonize result*/


        btlib_signal_semaphore( &(suspenderP->semaphore));

        btlib_wait_semaphore ( &(suspendeeP->semaphore),1 );
        btlib_destroy_semaphore( &(suspendeeP->semaphore) );

        return;

}




/* resume execution of the given thread
 * input:  tid - the thread to resume
 * output: 0 on success, non zero value otherwise
 */
/**************************************************************************/
long
btlib_resume_thread (int tid)
{
        //DBPRINT(btl_prt_thread,"## resume 1 thread %d ##\n", suspendee->pid);
        /* a thread should not resume itself */
        btlib_thread_data_t *susp_thread;
        assert (tid != SYSCALL (gettid, ()));

        DBPRINT(btl_prt_thread,"resume thread %d\n", tid);
        suspend_t  *suspendee = NULL;

        btlib_acquire_lock (BTL_LOCK_IND_THREAD);
        susp_thread = m_used_thread_list;
        while (susp_thread != NULL) {
                suspendee = &(susp_thread->suspend_data);
                if( susp_thread->tid==tid )
                        if ( suspendee->suspender_tid == cur_thread_data->tid )
				break;
                   susp_thread = susp_thread->next;
        }
        btlib_release_lock (BTL_LOCK_IND_THREAD);
        assert(susp_thread != NULL);

        btlib_signal_semaphore( &(suspendee->semaphore));
        suspendee->suspender_tid = 0; // no suspender now
        DBPRINT(btl_prt_thread,"resume thread %d done\n", tid);
        return 0;

	

}

/**************************************************************************/
/* set the signal hanlders according to the inherited handlers.
 * input:   thread -
 * output:  none
 * remarks: after an execve - a process inherits from its parent the 
 *          SIG_IGN dispositions and the signal mask.
 */
/**************************************************************************/
static void
set_sig_inherit ()
{
	int j;
	long retval;
	struct sigaction act;

	assert (cur_thread_data->tid == SYSCALL (gettid, ()));

	DBPRINT (0, "\n");
	/* change all the inherited SIG_IGN signals */
	for (j = 1; j < _NSIG; j++) {
		retval = SYSCALL (rt_sigaction, (j, NULL, &act, sizeof (k_sigset_t)));
		sys_err_chk (retval, "rt_sigaction");
		if (act.sa_handler == SIG_IGN) {
			DBPRINT (0, "ignored %d\n", j);
			cur_thread_data->sighandlers->handlers[j].sa.handler =
				(unsigned) UL(SIG_IGN);
		}
	}
}

/*****************************************************************************/
/* init thread data */
/*****************************************************************************/
static void
init_thread_data (btlib_thread_data_t * blst,
		  void *clst, void *stackblock, unsigned long user_stack_size)
{
	DBPRINT (0, "blst=%p clst=%p stackblock=%lx stack_size=%x\n",
		 blst, clst, stackblock, user_stack_size);

	/* init some of the thread's fields */
	blst->stackblock = stackblock;
	blst->user_stack_size = user_stack_size;
	blst->syscall_id = IA32_SYSCALL_NONE;
	blst->bc32 = (BTGENERIC_IA32_CONTEXT *) ((uintptr_t) clst);
	blst->bc32->p_glst=(unsigned long)((unsigned long)(&(blst->bc32->p_glst)) + (unsigned long)sizeof(blst->bc32->p_glst) + ctx_align_for_glst);
	DBPRINT (0, "&(blst->bc32->p_glst)=%p sizeof(blst->bc32->p_glst)=%x\n,glst address=%lx\n",
		&(blst->bc32->p_glst), sizeof(blst->bc32->p_glst),blst->bc32->p_glst);
	blst->is_eax_changed = 0;
	blst->set_reg_addr = -1;
}

/*****************************************************************************/
/*****************************************************************************/
static void
list_add (btlib_thread_data_t * elem, btlib_thread_data_t ** list)
{
	elem->next = *list;
	elem->prev = NULL;

	if (*list != NULL) {
		(*list)->prev = elem;
	}

	*list = elem;
}

/*****************************************************************************/
/*****************************************************************************/
static void
list_del (btlib_thread_data_t * elem, btlib_thread_data_t ** list)
{
	if (*list == elem) {
		/* elem is head */
		*list = elem->next;
	}

	if (elem->next != NULL) {
		elem->next->prev = elem->prev;
	}
	if (elem->prev != NULL) {
		elem->prev->next = elem->next;
	}
}

/*****************************************************************************/
/* 
 * allocate memory for 64 bit thread stack, btlib thread data, and CLST+GLST. 
 * initialize btlib thread data
 */
/*****************************************************************************/
static void *
allocate_thread_data (void **new_r13, btlib_thread_data_t ** blst, void **clst)
{
	void *stackblock, *bos;	/* top and bottom of 64 bit stack */
	int ret;
	size_t stack_size;

	/* 
	 * allocate the stack - both the thread_data and clst+glst
	 * will reside on the memory chunk allocated here
	 */


	ret = btlib_allocate_stack (&stackblock, &stack_size, new_r13);
	DBPRINT (btl_prt_thread, "stackblock=%p stack_size=%lx new_r13=%p\n",
		 stackblock, stack_size, *new_r13);

	if (IS_SYS_ERROR (ret)) {
		DBPRINT (btl_prt_thread, "btlib_allocate_stack failed=%d\n",
			 ret);
		return (void *) (uintptr_t) ret;
	}
	assert ((uintptr_t) new_r13 ==
		ROUNDUP ((uintptr_t) new_r13, sizeof (void *)));

	bos = stackblock + stack_size - 1;

	/* allocate thread_data */
	bos -= sizeof (btlib_thread_data_t);
	ALIGN_STACK (bos);
	*blst = (btlib_thread_data_t *) bos;
	bt_memset (*blst, 0, sizeof (btlib_thread_data_t));

	/* allocate clst+glst */
	bos -= btl_clst_glst_size;
	ALIGN_STACK (bos);
	*clst = bos;
	DBPRINT (btl_prt_thread,, "blst=%p clst=%p\n", *blst, *clst);

	init_thread_data (*blst, *clst, stackblock,
			  (unsigned long) (*clst - stackblock));

	DBPRINT (btl_prt_thread, "bos=%p\n", bos);
	return bos;
}

/**************************************************************************/
static void
before_thread_exit ()
{
	/* 
	 * if this is the last thread then call this api function
	 * which prepares stats on IA32-EL run
	 */
	if ((cur_thread_data->next == NULL) && (cur_thread_data->prev == NULL)) {
		STATCODE (btl_stats, btlib_print_stats ());
		BTGENERIC_NOTIFY_EXIT ();
	}

	/* notify btgeneric */
	BTGENERIC_THREAD_TERMINATED ();

	/* mark as exited */
	cur_thread_data->is_exited = 1;

	/* last print to log */

	DBPRINT (btl_prt_thread, "bye\n");

	/* remove current thread from used thread list */
	btlib_acquire_lock (BTL_LOCK_IND_THREAD);
	list_del (cur_thread_data, &m_used_thread_list);
	btl_num_threads--;
	btlib_release_lock (BTL_LOCK_IND_THREAD);

	/* signal handler table and pending que will never be used, but thread 
	   stack should not be freed */
	cleanup_thread (cur_thread_data, FALSE);
}

#if DEBUG
/************************************************************************** 
* print the thread list starting from the current thread
**************************************************************************/
void
print_thread_list (char *header)
{
	btlib_thread_data_t *p;
	char list[4096] = { 0 };
	unsigned long len;
	unsigned long chars_left = sizeof (list) - 1;
	char th[32];
	int is_first = 1;

	assert (cur_thread_data->tid == SYSCALL (gettid, ()));

	len = bt_snprintf (list, chars_left,
			   "(%d) %s - thread list = ",
			   cur_thread_data->tid, header);
	chars_left -= len;
	btlib_acquire_lock (BTL_LOCK_IND_THREAD);
	for (p = m_used_thread_list; p != NULL; p = p->next) {
		bt_snprintf (th, sizeof (th) - 1, "%s(%d)",
			     is_first ? "" : "->", p->tid);
		is_first = 0;
		th[sizeof (th) - 1] = '\0';
		assert (chars_left > 0);
		len = MIN (chars_left, bt_strlen (th));
		bt_strncat (list, th, len);
		chars_left -= len;
	}
	btlib_release_lock (BTL_LOCK_IND_THREAD);
	bt_strncat (list, "\n", MIN (sizeof (list) - 1 - bt_strlen (list), 2));
	fprintf (g_log_file, list);
}
#endif	// DEBUG

#if ALL_THREAD_AFFINITY

/*
read the thread utilization onfo fro the /proc/<tid>/stat file - whoes format is:
pid %d
comm %s
state %c
ppid %d
pgrp %d
session %d
tty_nr %d
tpgid %d
flags %lu
minflt %lu
cminflt %lu
majflt %lu
cmajflt %lu
utime %lu
stime %lu
pid   comm state ppid  pgrp session  tty_nr  tpgid  flags   minflt  cminflt majflt cmajflt  utime   stime
6979 (hald) S    1     6979 6979      0      -1     4194624 9268    43946    2      1       771     1456 392
*/
static unsigned long
get_thread_utilization (int tid)
{
	char proc_tid_stat[PATH_MAX];
	char stat_str[STAT_STR_LENGTH];
	char *start_string, *p;
	int tid_stat_file, retval, i;
	int pgrp_id;
	unsigned long utime, stime;

	bt_snprintf (proc_tid_stat, sizeof (proc_tid_stat) - 1,
		     "/proc/%d/stat", tid);

	tid_stat_file = SYSCALL (open, (proc_tid_stat, O_RDONLY, 0));
	if (IS_SYS_ERROR (tid_stat_file)) {
		DBPRINT (btl_prt_thread, "Error: open of %s failed: %d\n",
			 proc_tid_stat, tid_stat_file);
		return 0;
	} else {
		retval = SYSCALL (read, (tid_stat_file, stat_str, STAT_STR_LENGTH));
		if (IS_SYS_ERROR (retval)) {
			DBPRINT (btl_prt_thread,
				 "Error: failed to read from %s: %d\n",
				 proc_tid_stat, retval);
			SYSCALL (close, (tid_stat_file));
			return (0);
		} else {

			p = stat_str;
			for (i = 0; i < NUM_BLANKS_TIL_PGRP; i++) {
				p = bt_strchr (p, ' ');
				if (p == NULL) {
					SYSCALL (close, (tid_stat_file));
					return (0);
				}
				p++;

			}
			start_string = p;
			p = bt_strchr (p, ' ');
			if (p == NULL) {
				SYSCALL (close, (tid_stat_file));
				return (0);
			}

			*p = '\0';
			pgrp_id = bt_atol (start_string);
			if (-1 == m_grpid) {
				m_grpid = SYSCALL (getpgid, (0));
			}
			if (pgrp_id != m_grpid) {
				/* not in this process */
				SYSCALL (close, (tid_stat_file));
				return (0);
			}
			p++;

			for (i = 0; i < NUM_BLANKS_TIL_UITME; i++) {
				p = bt_strchr (p, ' ');
				if (p == NULL) {
					SYSCALL (close, (tid_stat_file));
					return (0);
				}
				p++;
			}
			start_string = p;
			p = bt_strchr (p, ' ');
			if (p == NULL) {
				SYSCALL (close, (tid_stat_file));
				return (0);
			}
			*p = '\0';
			utime = bt_atol (start_string);
			p++;
			start_string = p;
			p = bt_strchr (p, ' ');
			if (p == NULL) {
				SYSCALL (close, (tid_stat_file));
				return (0);
			}
			*p = '\0';
			stime = bt_atol (start_string);
			SYSCALL (close, (tid_stat_file));
			return (stime + utime);
		}
	}
	return (0);
}

static unsigned int
count_ready_threads_by_sim_state ()
/*++

Routine Description:

    Go over all the threads of the process (on the thread list).
    If the thread is the current thread or is in CpuSimulate or it is in a non-blocking system call - then
    it is considered as ready to run at this point if it has had some cpu-utlization since it was last checked
	Note that once more than NUM_FOR_VERY_MULTI_THREADED are found to be ready - the counting stops

Arguments:

    unsigned long cur_timestamp_cntr - the time stamp counter at which this call is made

Return Value:

    unsigned int - Number of threads that appear ready to run - stop counting after reaching NUM_FOR_VERY_MULTI_THREADED

--*/
{
	unsigned int num_threads_active = 0;
	unsigned long cur_total_utilization;

	btlib_thread_data_t *thread_ptr;
	DBPRINT (btl_prt_thread, "\n\n\n");

	/* go over the existing threads list */
	btlib_acquire_lock (BTL_LOCK_IND_THREAD);

	for (thread_ptr = m_used_thread_list; thread_ptr != NULL;
	     thread_ptr = thread_ptr->next) {

		DBPRINT (btl_prt_thread,
			 "thread id %d (is cur thread %d  in simuation %d  in nonblock syscall %d)\n",
			 thread_ptr->tid, cur_thread_data == thread_ptr,
			 thread_ptr->is_in_simulation_session,
			 thread_ptr->in_nonblocking_syscall);

		if (cur_thread_data == thread_ptr	/* running thread always checked */
		    || thread_ptr->is_in_simulation_session	/* thread in simulation */
		    || thread_ptr->in_nonblocking_syscall	/* in a non-blocking system call */
		    ) {
			/* thread will be considered ready to run only if it has 
			   consumned some cpu cycles since it was last checked */
			cur_total_utilization = get_thread_utilization (thread_ptr->tid);
			DBPRINT (btl_prt_thread,
				 "  thread possibly active cur_total_utilization %u last_utlization %u\n",
				 cur_total_utilization,
				 thread_ptr->last_utlization);
			if (thread_ptr->last_utlization != 0 &&
			    /* some utilization recorded on last check */
			    cur_total_utilization != 0 &&
			    (cur_total_utilization != thread_ptr->last_utlization)) {
				num_threads_active++;
				if (num_threads_active > NUM_FOR_VERY_MULTI_THREADED) {
					break;	/* no need to count any more */
				}
			}
			thread_ptr->last_utlization = cur_total_utilization;
		}
	}
	btlib_release_lock (BTL_LOCK_IND_THREAD);

	return (num_threads_active);
}

void
btlib_thread_scalability_analysis ()
/*++

Routine Description:

    This function will check to see if btl_cycles_between_ts_checks have elapsed since the last time it checked 
    the states of the threads of the process.
      If btl_cycles_between_ts_checks have passed then it will count how many of the threads appear to be able
      to run. A history of how many threads appears to be able to run is maintained - if there are eough threads
      that appear ready to run for a sufficient number of consecutive checks - then the process is deemed to be thread
      scalable, and all translated code will use the ld.acq-st.rel and the affinity restriction will be removed from all
      threads.

Arguments:

    None

Return Value:

    None

--*/
{
	unsigned long my_in_thread_scalability_analysis;
	unsigned long cur_timestamp_cntr;
	unsigned int num_threads_in_paralled;
	EMULATION_POLICY emulation_poilcy;

	my_in_thread_scalability_analysis
	    = GCC_BUILTIN_SYNC_FETCH_AND_ADD (&m_in_thread_scalability_analysis, 1ul);
	if (0 == my_in_thread_scalability_analysis) {
		/* only one thread in this code */

		/* check for elapsed time only once every 
		   NUM_THREAD_SCALABILITY_CALLS_TO_CHECK_AFTER calls to this function */
		cur_timestamp_cntr = __btlib_read_timestamp_counter ();

		if (((cur_timestamp_cntr - m_last_ts_check_timestamp)
		     >= btl_cycles_between_ts_checks)
#if ! RELEASE
		    && btl_no_affinity == 0
#endif
		    ) {
			/* btl_cycles_between_ts_checks has passed since last check - 
			   time for a new check */

			num_threads_in_paralled = count_ready_threads_by_sim_state ();

			DBPRINT (btl_prt_thread,
				 "btlib_thread_scalability_analysis at elapsed %u num_threads_in_paralled %d\n",
				 (unsigned) ((cur_timestamp_cntr - btl_start_time) /
					     CYCLES_IN_MILI_SECOND),
				 num_threads_in_paralled);

			if (num_threads_in_paralled > 1) {
				if (m_last_num_threads_in_paralled > 1) {
					m_num_consec_multi_threaded++;
					if (m_last_num_threads_in_paralled >
					    NUM_FOR_VERY_MULTI_THREADED
					    && num_threads_in_paralled >
					    NUM_FOR_VERY_MULTI_THREADED) {
						m_num_consec_very_multi_threaded++;
					} else {
						m_num_consec_very_multi_threaded = 0;
					}
				} else {
					m_num_consec_multi_threaded = 0;
					m_num_consec_very_multi_threaded = 0;
				}
			}
			m_last_num_threads_in_paralled = num_threads_in_paralled;
			m_last_ts_check_timestamp = cur_timestamp_cntr;

			if (m_num_consec_multi_threaded > SCALABILITY_THRESHOLD ||
					m_num_consec_very_multi_threaded > HIGH_SCALABILITY_THRESHOLD) {

				DBPRINT (btl_prt_thread,
					 "switching to scalability mode at elapsed %u \n",
					 (unsigned) ((cur_timestamp_cntr - btl_start_time) /
						     CYCLES_IN_MILI_SECOND));

				emulation_poilcy.memory_order_policy = USE_ORDERED_LD_ST;
				emulation_poilcy.strict_state_policy = POLICY_UNCHANGED;

				BTGENERIC_SET_EMULATION_POLICY (&emulation_poilcy);
				assert (emulation_poilcy.memory_order_policy == USE_ORDERED_LD_ST);
				if (emulation_poilcy.memory_order_policy == USE_ORDERED_LD_ST) {
					/* give all the threads their original affinity */
					set_affinity_to_all_threads (&m_orig_process_affinity_mask);
					btl_narrowed_affinity = 0;
					btl_did_ordered_code = 1;
				}
			}

		}
	}			// end only one thread in this code
	GCC_BUILTIN_SYNC_FETCH_AND_ADD (&m_in_thread_scalability_analysis, -1l);
}

#endif
