/* 
 * ia32el_mme_api.h
 * interface of ia32el with for MME
 */
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
#ifndef _IA32EL_MME_API_H
#define _IA32EL_MME_API_H

#ifndef __ASSEMBLY__

#define IA32EL_NAME "libia32x.so"

#ifndef __IA32EL__ /* IA32EL does not use standard h files */
#include <signal.h>
#include <stdint.h>
#include <stdbool.h>
#endif // __IA32EL__ 


/*****************************************************************************/
/*  IA-32 execution layer specific context */

#define ia32_fp_stack_byte_size         80

#define ctx32_critical          (0x10001L) 
#define ctx32_gp                (0x10002L) 
#define ctx32_segment           (0x10004L) 
#define ctx32_floating_point    (0x10008L) 
#define ctx32_debug             (0x10010L) 
#define ctx32_xmm               (0x10020L) 
#define ctx32_bt_extended       (0x10040L) 



typedef struct {
    uint32_t   fs_base;
    uint32_t   gs_base;
    uint32_t   es_base; 
    uint32_t   gdt_base;
    uint32_t   idt_base;
    uint16_t   gdt_limit;
    uint16_t   idt_limit;
    uint16_t   ldt_selector;
} bt_context32_ext;
/* mask for significant buts in segment registers
 * (bits beyond this mask are ignored in any segment
 * registers comparison) */
#define bt_context32_segreg_mask 0xffff

#define bt_maximum_supported_extension (512 + sizeof(bt_context32_ext))
#define btext_index_in_ext_regs 512

typedef struct {
    uint32_t   control_word;
    uint32_t   status_word;
    uint32_t   tag_word;
    uint32_t   instruction_pointer_offset;
    uint32_t   instruction_pointer_selector;
    uint32_t   data_pointer_offset;
    uint32_t   data_pointer_selector;
    uint8_t    stack[ia32_fp_stack_byte_size];
    uint32_t   unused;
} floating_point_state32;

typedef struct  {
    uint32_t   padding1;
    uint32_t   composition_bits;
    uint32_t      dr0;
    uint32_t      dr1;
    uint32_t      dr2;
    uint32_t      dr3;
    uint32_t      dr6;
    uint32_t      dr7;
    floating_point_state32 fpu_state;
    uint32_t   gs, fs, es, ds;
    uint32_t   edi;
    uint32_t   esi;
    uint32_t   ebx;
    uint32_t   edx;
    uint32_t   ecx;
    uint32_t   eax;
    uint32_t   ebp;
    uint32_t   eip;
    uint32_t   cs;
    uint32_t   eflags;
    uint32_t   esp;
    uint32_t   ss;
    uint8_t    xmm_and_bt_ext[bt_maximum_supported_extension];
    uint64_t   p_glst;
} ia32el_context32;
/*****************************************************************************/




/* ia32el return value */
typedef enum {IA32EL_SUCCESS = 0, IA32EL_FAILURE = 1} ia32el_retval;

/* ia32el signal handling result */
typedef enum {MME_IA32_NATIVE_EXCEPTION = -1, MME_IA32_SIGNAL_NOT_HANDLED = 0, MME_IA32_SIGNAL_HANDLED = 1} ia32el_sighand_result;

/* version */
#define IA32EL_MME_VER_STR_LEN 64
#define IA32EL_MME_VER_MAJOR "0"
#define IA32EL_MME_VER_MINOR "1"
#ifdef __IA32EL__
#if MAJOR_VERSION_NUMBER != 7 || MINOR_VERSION_NUMBER != 2
/* if this error is triggered then -
 * check if the 32 bit context used by ia32el was changed
 * -> if yes : 1) make similar changes to the ia32el_context
 *                structure defined above 
 *             2) bump the IA32EL-MME interface version */
# error incompatible 32 bit context in ia32el_mme_api.h
#endif
#endif

/* functions provided by mme to ia32el */
extern int64_t mme_upcall (ia32el_context32 * bc32, uint32_t code);
extern int64_t mme_thread_at_clone (void * params);
extern void    mme_run_jvm_handler (void * siginfo);
struct _mme_api_table_tag {
	int64_t (*mme_upcall) (ia32el_context32 * bc32, uint32_t code);
	int64_t (*mme_thread_at_clone) (void * params);
	void    (*mme_run_jvm_handler) (void * siginfo);
};

/* functions exported by ia32el */
extern ia32el_retval ia32el_mme_version (char version[IA32EL_MME_VER_STR_LEN]);
extern ia32el_retval ia32el_mme_init (uint64_t flags, struct _mme_api_table_tag * mme_api, int64_t tls_offset);
extern ia32el_sighand_result ia32el_mme_signal_handler (int32_t signo, struct siginfo * si, struct sigcontext * sc);
extern ia32el_retval ia32el_init_new_thread (int32_t flag, void * params);
extern ia32el_retval ia32el_mme_thread_term (void);
extern void * ia32el_mme_memory_map (uint32_t addr, uint32_t size, uint32_t prot);
extern int32_t ia32el_mme_memory_unmap (void * addr, uint32_t size);
extern ia32el_retval ia32el_mme_load_and_exec (int32_t argc, char ** argv, char ** envp);
extern ia32el_retval ia32el_mme_run (ia32el_context32 * bc32);
extern ia32el_retval ia32el_mme_ia32_context_get (ia32el_context32 * bc32);
extern ia32el_retval ia32el_mme_complete_child_init (void * clone_params);
extern ia32el_retval ia32el_mme_complete_parent_thread_create (void *, int64_t retval, int64_t *cpid);
extern ia32el_retval ia32el_mme_query_memory_permission (void *addr, int64_t *pperm);

struct _mme_ia32el_exports_tag {
	ia32el_retval (*ia32el_version) (char version[IA32EL_MME_VER_STR_LEN]);
	ia32el_retval (*ia32el_init) (uint64_t flags, struct _mme_api_table_tag * mme_api, int64_t tls_offset);
	ia32el_sighand_result (*ia32el_signal_handler) (int32_t signo, struct siginfo * si, struct sigcontext * sc);
	ia32el_retval (*ia32el_complete_child_init) (void *);
	ia32el_retval (*ia32el_complete_parent_thread_create) (void *, int64_t, int64_t *);
	ia32el_retval (*ia32el_thread_term) (void);
	void * (*ia32el_memory_map) (uint32_t addr, uint32_t size, uint32_t prot);
	int32_t (*ia32el_memory_unmap) (void * addr, uint32_t size);
	ia32el_retval (*ia32el_load_and_exec) (int32_t argc, char ** argv, char ** envp);
	ia32el_retval (*ia32el_run) (ia32el_context32 * bc32);
	ia32el_retval (*ia32el_ia32_context_get) (ia32el_context32 * bc32);
	ia32el_retval (*ia32el_query_memory_permission) (void *addr, int64_t *pperm);
};

#endif // __ASSEMBLY__

/* ia32el trap values */
#define IA32EL_TRAP_AUX_APP_READY 0x10
#define IA32EL_TRAP_JNI_CALL 0x20
#define IA32EL_TRAP_JAVAVM_CALL 0x30
#define IA32EL_TRAP_RETURN_FROM_RUN 0x40
#define IA32EL_TRAP_THREAD_FUNC 0x50
#define IA32EL_TRAP_PHDR_ITER 0x60
#define IA32EL_TRAP_JNIVM_CALL 0x70

#endif // _IA32EL_MME_API_H
