/* 
 * permissions.c
 * check exec permissions of files
 */
 /*
 *
 * Copyright (C) 2005       Jakub Jelinek <jakub@redhat.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
#include <ia32x_utils.h>

static int check_exec_permissions (struct stat * st);
static long check_noexec_mount (struct stat *stp);
/*****************************************************************************/
 /*  returns 1 iff the given file resides on a noexec file */
/*****************************************************************************/
static long check_noexec_mount (struct stat *stp)
{
	int fd = SYSCALL (open, ("/proc/mounts", O_RDONLY, 0));
	size_t alen = 1024;
	ssize_t len;
	long ret = 0, r;
	char *buf, *rp, *p, *q, *path;
	struct stat st;

	if (fd < 0)
		return 0;
	buf = (char *) SYSCALL (mmap, (0ul, alen, PROT_READ|PROT_WRITE,
				MAP_PRIVATE|MAP_ANON, -1, 0ul));
	if (IS_SYS_ERROR (buf)) {
		buf = NULL;
		goto out;
	}
	p = rp = buf;
	while ((len = SYSCALL (read, (fd, rp, (unsigned long)(buf + alen - rp)))) > 0) {
		while (1) {
			/* Lack of memchr... */
			for (q = p; q < p + len && *q != '\n'; ++q);
			if (q == p + len)
				break;
			*q = '\0';
			len -= q - p + 1;
			while (p < q && *p != ' ')
				p++;
			while (*p == ' ')
				p++;
			if (p == q)
				goto out;
			path = p;
			while (p < q && *p != ' ')
				p++;
			if (p == q)
				goto out;
			*p++ = '\0';
			if (bt_strstr (p, "noexec")) {
				if (SYSCALL (stat, (path, &st)) >= 0
				    && st.st_dev == stp->st_dev) {
					ret = 1;
					goto out;
				}
			}
			p = q + 1;
		}
		if (p == buf) {
			if (rp + len == buf + alen) {
				char *new_buf = (char *)SYSCALL (mremap, ((unsigned long)buf,
							alen, 2 * alen, MREMAP_MAYMOVE, 0ul));
				if (IS_SYS_ERROR (new_buf)) {
					goto out;
				}
				rp = new_buf + (rp - buf);
				buf = new_buf;
				p = buf;
				alen *= 2;
				/* Should never happen.  */
				if (alen >= 0x8000000000000000)
					goto out;
			}
			rp += len;
		} else {
			bt_memmove (buf, p, (unsigned long)len);
			p = buf;
			rp = buf + len;
		}
	}
out:
	if (buf)
		(void) SYSCALL (munmap, ((unsigned long)buf, alen));
	r = SYSCALL (close, (fd));
	DBPRINT (IS_SYS_ERROR (r), "close failed: %ld\n", r);
	return ret;
}

/*****************************************************************************/
/* check execue permission of the given file. returns 1 if file has exec
 * permissions, -EACCESS if not and a negative value in case of fatal error. */
/*****************************************************************************/
int check_exec_permissions_by_name (char * name)
{
	struct stat st;
	int ret;
	/* To check executable perm's, we can't use access here, as that
	   checks uid/gid, not fsuid/fsgid.  Well, euid/egid is not
	   fsuid/fsgid either...  */
	ret = SYSCALL (stat, (name, &st));
	if (ret < 0)
		goto out;
	ret = check_exec_permissions (&st);
out:
	return ret;
}
int check_exec_permissions_by_fd (int fd)
{
	struct stat st;
	int ret;
	/* To check executable perm's, we can't use access here, as that
	   checks uid/gid, not fsuid/fsgid.  Well, euid/egid is not
	   fsuid/fsgid either...  */
	ret = SYSCALL (fstat, (fd, &st));
	if (ret < 0)
		goto out;
	ret = check_exec_permissions (&st);
out:
	return ret;
}

static int check_exec_permissions (struct stat * st)
{
	int ret;

	ret = -EACCES;
	if ((st->st_mode & 0111) != 0111) {
		uid_t uid = SYSCALL (geteuid, ());
		/* FIXME: this should actually be capable (CAP_DAC_OVERRIDE).  */

		if (uid == 0 && (st->st_mode & 0111)) {
		/* if the file is executable at all than root can exec it */
			/* do nothing */;
		} else if (uid == st->st_uid) {
			/* check owner perm */
			if ((st->st_mode & 0100) == 0) {
				goto out;
			}
		} else if (SYSCALL (getegid, ()) == st->st_gid) {
			/* check main group perm */
			if ((st->st_mode & 010) == 0) {
				goto out;
			}
		} else if ((st->st_mode & 011) != 011) {
			/* check secondary group perm */
			gid_t grps[64], *groups = grps;
			unsigned long groups_size = 0;
			long r;
			if ((st->st_mode & 011) == 0)
				goto out;
			r = SYSCALL (getgroups, (64, grps));
			if (r == -EINVAL) {
				r = SYSCALL (getgroups, (0, NULL));
				if (r < 0)
					goto out;
				groups_size = r * sizeof (gid_t);
				groups = (gid_t*)SYSCALL (mmap, (0ul, groups_size, 
							PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANON,
							-1, 0ul));
				if (IS_SYS_ERROR (groups)) {
					ret = -ENOMEM;
					goto out;
				}
				if (SYSCALL (getgroups, ((int)r, groups)) != r) {
					(void)SYSCALL (munmap, ((unsigned long)groups, groups_size));
					goto out;
				}
			} else if (r < 0)
				goto out;
			while (r > 0)
				if (groups[--r] == st->st_gid) {
					r = 3;
					break;
				}

			if (groups != grps)
				(void)SYSCALL (munmap, ((unsigned long)groups, groups_size));

			if ((st->st_mode & (1 << r)) == 0)
				goto out;
		}
	}
	if (check_noexec_mount (st))
		goto out;

	return 1;

out:
	return ret;
}
