/*  
 * ia32x_utils.c
 *  
 * main file for the utility lib
 * 
*/
 /*
 *
 * This file has been derived from GNU libc. Please see COPYING.LIBC
 * for more information
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include "../ia32x_utils.h"
#include "../include/ia32x_utils_internal.h"

#ifndef AT_SYSINFO
#define AT_SYSINFO 32
#endif // AT_SYSINFO
#ifndef AT_SECURE
#define AT_SECURE 23
#endif // AT_SECURE

/*the tick of 64 kernel*/
unsigned long g_tick64 = 0;

int g_log_file = 2; /* stderr */
/* fd of the binary to be emulated */
int g_is_at_execfd = 0;
long g_at_execfd = 0;
/* the address of __kernel_syscall_via_epc() */
unsigned long g_at_sysinfo = 0;
/* program headers */
void * g_at_phdr = NULL;
unsigned long g_at_phnum = 0;
/* the page size */
unsigned long g_at_pagesize = 0;
/* the real [gu]id */
unsigned long g_at_uid = 0, g_at_gid = 0;
/* the effective [gu]id */
unsigned long g_at_euid = 0, g_at_egid = 0;
/* security info */
int g_is_at_secure = 0;
unsigned long g_at_secure = 0;

static char *error_message[] = {"",
	"For IA-32 application compatability you should install the IA32 Execution Layer from the installation CD"
};

static long analyze_at_vector (unsigned long * elf_info);
long lib_init (unsigned long *, long, long);
long g_clstp_offset = 0;
long g_ptrace_offset = 0;

/*****************************************************************************/
/*****************************************************************************/
long lib_init (unsigned long * elf_info, long clstp_offset, long ptrace_offset)
{
	long ret = 0;

	DBPRINT (0, "elf_info=%p, clstp_offset=%lx, ptrace_offset=%lx\n", 
		 elf_info, clstp_offset, ptrace_offset);
	ret = analyze_at_vector (elf_info);
	if (ret) 
		goto out;
	ret = mutex_module_init ();
	if (ret) 
		goto out;
	ret = malloc_module_init ();
	if (ret) 
		goto out;

	g_clstp_offset = clstp_offset;
	g_ptrace_offset = ptrace_offset;
out:
	return ret;
}
/*****************************************************************************/
/*****************************************************************************/
static long analyze_at_vector (unsigned long * elf_info)
{
	DBPRINT (0, "elf_info=%p\n", elf_info);
	while (*(elf_info) != AT_NULL) {
		unsigned long id, val;
		id = *elf_info;
		val = *(elf_info+1);
		DBPRINT (0, "id=%lx val=%lx\n", id, val);
		switch (id) {
			case AT_PHDR:
				g_at_phdr = (void*)val;
				DBPRINT (0, "phdr=%p\n", g_at_phdr);
				break;
			case AT_PHNUM:
				g_at_phnum = val;
				DBPRINT (0, "phnum=%ld\n", g_at_phnum);
				break;
			case AT_UID:
				g_at_uid = val;
				DBPRINT (0, "uid=%ld\n", g_at_uid);
				break;
			case AT_GID:
				g_at_gid = val;
				DBPRINT (0, "gid=%ld\n", g_at_gid);
				break;
			case AT_EUID:
				g_at_euid = val;
				DBPRINT (0, "euid=%ld\n", g_at_euid);
				break;
			case AT_EGID:
				g_at_egid = val;
				DBPRINT (0, "egid=%ld\n", g_at_egid);
				break;
			case AT_SYSINFO:
				g_at_sysinfo = val;
				DBPRINT (0, "sysinfo=%ld\n", g_at_sysinfo);
				break;
			case AT_EXECFD:
				g_is_at_execfd = 1;
				g_at_execfd = val;
				DBPRINT (0, "execfd=%ld\n", g_at_execfd);
				break;
			case AT_PAGESZ:
				g_at_pagesize = val;
				DBPRINT (0, "pagesize=%ld\n", g_at_pagesize);
				break;
			case AT_SECURE:
				g_is_at_secure = 1;
				g_at_secure = val;
				DBPRINT (0, "secure=%ld\n", g_at_secure);
				break;
			case AT_CLKTCK:
				g_tick64 = val;
				DBPRINT (0, "tick64=val%ld\n", g_tick64);
				break;
		}
		elf_info += 2;
	}
	return 0;
}
/*****************************************************************************/
/*****************************************************************************/
int lcs_before_fork ()
{
	return malloc_lock ();
}
/*****************************************************************************/
/*****************************************************************************/
int lcs_after_fork (long is_parent)
{
	if (is_parent) {
		return malloc_unlock ();
	} else {
		return malloc_init ();
	}
}
/*****************************************************************************/
/* kill current thread-group */
/*****************************************************************************/
void abort ()
{
	SYSCALL (kill, (0, 9/*SIGKILL*/));
	/* dummy - supress the compiler warning about a noreturn function
	 * that does return */
	SYSCALL (exit_group, (0));
}
/*****************************************************************************/
/*****************************************************************************/
void lcs_get_ia32x_utils_version (char *version_str)
{
	long ret;
	ret = bt_snprintf(version_str, MAX_UNIT_VER_STR_LEN, "%s", BUILD_NO);
	assert (ret < MAX_UNIT_VER_STR_LEN);
	return;
}

/*****************************************************************************/
/* print the error message to the user and into adminstrator's log files     */
/*****************************************************************************/

void notify_error (int err)
{
	int fd;
	long retval;
	struct sockaddr sock;

	/* print the error message to the user */
	fprintf (stderr, "%s\n", error_message[err]);

	/* print the error message into the administrator's log files */
	fd = SYSCALL (socket, (AF_UNIX, SOCK_DGRAM, 0));
	if (IS_SYS_ERROR(fd)) {
		goto out;
	}
	retval = SYSCALL (fcntl, (fd, F_SETFD, (long)1));
	if (IS_SYS_ERROR(retval)) {
		goto out;
	}
	sock.sa_family = AF_UNIX;
	retval = bt_snprintf (sock.sa_data, sizeof (sock.sa_data) - 1, "%s", PATH_LOG);		
	assert (retval >= 0 && (unsigned long)retval < sizeof (sock.sa_data));
	retval = SYSCALL (connect, (fd, &sock, (socklen_t)sizeof(sock)));
	if (IS_SYS_ERROR(retval)) {
		goto out;
	}
	retval = SYSCALL (send, (fd, (void *)error_message[err],
				bt_strlen(error_message[err]) + 1, 0));
out:
	if (fd >= 0) {
		retval = SYSCALL (close, (fd));
	}
	return;
}


