/* 
 * bt_malloc.c
 * heap management functions
 * 
 *
 */
 /*
 *
 * This file has been derived from GNU libc. Please see COPYING.LIBC
 * for more information
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include "../ia32x_utils.h"
#include "../include/ia32x_utils_internal.h"

int g_is_mme_mode = 0;
extern void * malloc (size_t);
extern void free (void *);
extern void * calloc (size_t, size_t);
extern void * realloc (void *, size_t);

/* The allocator divides the heap into blocks of fixed size; large
   requests receive one or more whole blocks, and small requests
   receive a fragment of a block.  Fragment sizes are powers of two,
   and all fragments of a block are the same size.  When all the
   fragments in a block have been freed, the block itself is freed.
   The default definition here is 16K bytes. */
#define INT_BIT (CHAR_BIT * sizeof (int))
#define BLOCKLOG 12
#define BLOCKSIZE (unsigned long)(1 << BLOCKLOG)
#define BLOCKIFY(SIZE) (((SIZE) + BLOCKSIZE - 1) / BLOCKSIZE)

/* Determine the amount of memory spanned by the initial heap table
   (not an absolute limit). */
#define HEAP (10 * BLOCKSIZE)

/* Number of contiguous bt_free blocks allowed to build up at the end of
   memory before they will be returned to the system. */
#define FINAL_FREE_BLOCKS 8


/* Data structure giving per-block information. */
union info {
	struct {
		int type;	/* Zero for a large block, or positive
				   giving the logarithm to the base two
				   of the fragment size. */
		union {
			struct {
				unsigned nfree;	/* Free fragments in a fragmented block. */
				unsigned first;	/* First bt_free fragment of the block. */
			} frag;
			unsigned size;	/* Size (in blocks) of a large cluster. */
		} info;
	} busy;
	struct {
		unsigned size;	/* Size (in blocks) of a bt_free cluster. */
		unsigned next;	/* Index of next bt_free cluster. */
		unsigned prev;	/* Index of previous bt_free cluster. */
	} bt_free;
};
/* Address to block number and vice versa. */
#define BLOCK(A) (((char *) (A) - _heapbase) / BLOCKSIZE + 1)
#define ADDRESS(B) ((void *) (((B) - 1) * BLOCKSIZE + _heapbase))

/* Doubly linked lists of bt_free fragments. */
struct list {
	struct list *next;
	struct list *prev;
};

/* change the brk */
static void * _sbrk (long);

/* Default value of previous. */
static void *_default_morecore (unsigned long);

/* How to really get more memory. */
static void *(*_morecore) (unsigned long) = _default_morecore;

/* Pointer to the base of the first block. */
static char *_heapbase = NULL;

/* Block information table. */
static union info *_heapinfo = NULL;

/* Number of info entries. */
static unsigned heapsize = 0;

/* Search index in the info table. */
static unsigned _heapindex = 0;

/* Limit of valid info table indices. */
static unsigned _heaplimit = 0;

/* Count of large blocks allocated for each fragment size. */
static int _fragblocks[BLOCKLOG] = {0};

/* Free lists for each fragment size. */
static struct list _fraghead[BLOCKLOG] = {{0,0}};

/* Are we experienced? */
static int initialized = 0;

/* sync object */
static mutex_opaque m_malloc_sync;


/* change brk value */
/* we don't want to use sys_brk here because we
 * cannot afford to be limited by the RLIM_DATA
 * value the user picked. we will use mmap instead,
 * trying to emulate the brk behavior. */
static void * _sbrk (long increment)
{
	static unsigned long curbrk = 0;
	static unsigned long brktop = 0;
	unsigned long page_size = SYSCALL (getpagesize, ());
	unsigned long oldbrk;

	/* we assume that this function is called *only* for
	 * increasing the brk. it will make the task a bit
	 * easier */	 
	assert (increment >= 0);

	/* get the initial brk value */
	if (curbrk == 0) {
		long err;
		curbrk = SYSCALL (brk, (0ul));
		__asm__ __volatile__ ("mov %0=r10;;\n":"=r"(err));
		if (err) {
			LOG ("failed 0\n");
			return (void*)-1;
		}
		curbrk = ROUNDUP (curbrk, page_size);
		brktop = curbrk;
	}
	if (curbrk + increment > brktop) {
		/* map more memory */
		unsigned long size = ROUNDUP (increment, page_size);
		DBPRINT (0, "curbrk=%lx size=%lx\n", curbrk, size);
		unsigned long p = SYSCALL (mmap, 
				(brktop, size, PROT_READ|PROT_WRITE|PROT_EXEC,
				 MAP_PRIVATE|MAP_ANON|MAP_FIXED, -1, 0ul));
		if (IS_SYS_ERROR (p)) {
			LOG ("mmap failed: %ld\n", (long)p);
			return (void*)-1;
		}
		DBPRINT (0, "addr=%lx size=%lx FIXED\n", p, size);
		assert (p == brktop);
		/* assign value to brktop */
		brktop = p + size;
	}
	oldbrk = curbrk;
	curbrk += increment;

	return (void*)oldbrk;
}
/* Note that morecore has to take a signed argument so
   that negative values can return memory to the system. */
static void * _default_morecore (unsigned long size)
{
	void *result;

	result = (void*)_sbrk ((long)size);
	if (result == (void *) -1) {
		LOG ("failed\n");
		return NULL;
	}
	return result;
}
/* Aligned allocation. */
static void * align (size_t size)
{
	void *result;
	unsigned long adj;

	result = (*_morecore) (size);
	adj = (unsigned int) ((char *) result - (char *) NULL) % BLOCKSIZE;
	if (adj != 0) {
		adj = BLOCKSIZE - adj;
		(*_morecore) (adj);
		result = (char *) result + adj;
	}
	return result;
}

/* Set everything up and remember that we have. */
static int initialize ()
{
	heapsize = HEAP / BLOCKSIZE;
	_heapinfo = align (heapsize * sizeof (union info));
	if (!_heapinfo)
		return 0;
	bt_memset (_heapinfo, 0, heapsize * sizeof (union info));
	_heapinfo[0].bt_free.size = 0;
	_heapinfo[0].bt_free.next = _heapinfo[0].bt_free.prev = 0;
	_heapindex = 0;
	_heapbase = (char *) _heapinfo;
	initialized = 1;
	return 1;
}

/* Get neatly aligned memory, initializing or growing the
   heap info table as necessary. */
static void * morecore (size_t size)
{
	void *result;
	union info *newinfo, *oldinfo;
	unsigned newsize;

	result = align (size);
	if (!result) {
		LOG ("failed 1\n");
		return NULL;
	}

	/* Check if we need to grow the info table. */
	if (BLOCK ((char *) result + size) > heapsize) {
		newsize = heapsize;
		while (BLOCK ((char *) result + size) > newsize)
			newsize *= 2;
		newinfo = align (newsize * sizeof (union info));
		if (!newinfo) {
			(*_morecore) (-size);
			LOG ("failed 2\n");
			return NULL;
		}
		bt_memset (newinfo, 0, newsize * sizeof (union info));
		bt_memcpy (newinfo, _heapinfo, heapsize * sizeof (union info));
		oldinfo = _heapinfo;
		newinfo[BLOCK (oldinfo)].busy.type = 0;
		newinfo[BLOCK (oldinfo)].busy.info.size
		    = BLOCKIFY (heapsize * sizeof (union info));
		_heapinfo = newinfo;
		bt_free (oldinfo);
		heapsize = newsize;
	}

	_heaplimit = BLOCK ((char *) result + size);
	return result;
}

/* Allocate memory from the heap. */
void * bt_malloc (size_t size)
{
	if (g_is_mme_mode)
		return malloc (size);
	
	void *result = NULL;
	unsigned llog, block, blocks, i, lastblocks, start;
	struct list *next;

	//LOG ("size=%lu\n", size);

	malloc_lock();

	if (!initialized && !initialize ()) {
		LOG ("failed 1\n");
		goto out;
	}

	if (size == 0) {
		goto out;
	}

	if (size < sizeof (struct list))
		size = sizeof (struct list);

	/* Determine the allocation policy based on the request size. */
	if (size <= BLOCKSIZE / 2) {
		/* Small allocation to receive a fragment of a block. Determine
		   the logarithm to base two of the fragment size. */
		--size;
		for (llog = 1; (size >>= 1) != 0; ++llog) ;

		/* Look in the fragment lists for a bt_free fragment of the
		   desired size. */
		if ((next = _fraghead[llog].next) != 0) {
			/* There are bt_free fragments of this size.  Pop a fragment
			   out of the fragment list and return it.  Update the block's
			   nfree and first counters. */
			result = next;
			next->prev->next = next->next;
			if (next->next)
				next->next->prev = next->prev;
			block = BLOCK (result);
			if (--_heapinfo[block].busy.info.frag.nfree)
				_heapinfo[block].busy.info.frag.first
				    =
				    (unsigned int) ((char *) next->next -
						    (char *) NULL)
				    % BLOCKSIZE >> llog;
		} else {
			/* No bt_free fragments of the desired size, so get a new block
			   and break it into fragments, returning the first. */
			result = bt_malloc (BLOCKSIZE);
			if (!result) {
				LOG ("failed 2\n");
				goto out;
			}
			++_fragblocks[llog];

			/* Link all fragments but the first into the bt_free list. */
			for (i = 1; i < BLOCKSIZE >> llog; ++i) {
				next = (struct list *) (result + (i << llog));
				next->next = _fraghead[llog].next;
				next->prev = &_fraghead[llog];
				next->prev->next = next;
				if (next->next)
					next->next->prev = next;
			}

			/* Initialize the nfree and first counters for this block. */
			block = BLOCK (result);
			_heapinfo[block].busy.type = llog;
			_heapinfo[block].busy.info.frag.nfree = i - 1;
			_heapinfo[block].busy.info.frag.first = i - 1;
		}
	} else {
		/* Large allocation to receive one or more blocks.  Search
		   the bt_free list in a circle starting at the last place visited.
		   If we loop completely around without finding a large enough
		   space we will have to get more memory from the system. */
		blocks = BLOCKIFY (size);
		start = block = _heapindex;
		while (_heapinfo[block].bt_free.size < blocks) {
			block = _heapinfo[block].bt_free.next;
			if (block == start) {
				/* Need to get more from the system.  Check to see if
				   the new core will be contiguous with the final bt_free
				   block; if so we don't need to get as much. */
				block = _heapinfo[0].bt_free.prev;
				lastblocks = _heapinfo[block].bt_free.size;
				if (_heaplimit
				    && block + lastblocks == _heaplimit
				    && (*_morecore) (0ul) ==
				    ADDRESS (block + lastblocks)
				    && morecore ((blocks - lastblocks) *
						 BLOCKSIZE)) {
					/* Note that morecore() can change the location of
					   the final block if it moves the info table and the
					   old one gets coalesced into the final block. */
					block = _heapinfo[0].bt_free.prev;
					_heapinfo[block].bt_free.size +=
					    blocks - lastblocks;
					continue;
				}
				result = morecore (blocks * BLOCKSIZE);
				if (!result) {
					LOG ("failed 3\n");
					goto out;
				}
				block = BLOCK (result);
				_heapinfo[block].busy.type = 0;
				_heapinfo[block].busy.info.size = blocks;
				goto out;
			}
		}

		/* At this point we have found a suitable bt_free list entry.
		   Figure out how to remove what we need from the list. */
		result = ADDRESS (block);
		if (_heapinfo[block].bt_free.size > blocks) {
			/* The block we found has a bit left over, so relink the
			   tail end back into the bt_free list. */
			_heapinfo[block + blocks].bt_free.size
			    = _heapinfo[block].bt_free.size - blocks;
			_heapinfo[block + blocks].bt_free.next
			    = _heapinfo[block].bt_free.next;
			_heapinfo[block + blocks].bt_free.prev
			    = _heapinfo[block].bt_free.prev;
			_heapinfo[_heapinfo[block].bt_free.prev].bt_free.next
			    = _heapinfo[_heapinfo[block].bt_free.next].bt_free.prev
			    = _heapindex = block + blocks;
		} else {
			/* The block exactly matches our requirements, so
			   just remove it from the list. */
			_heapinfo[_heapinfo[block].bt_free.next].bt_free.prev
			    = _heapinfo[block].bt_free.prev;
			_heapinfo[_heapinfo[block].bt_free.prev].bt_free.next
			    = _heapindex = _heapinfo[block].bt_free.next;
		}

		_heapinfo[block].busy.type = 0;
		_heapinfo[block].busy.info.size = blocks;
	}

out:
	malloc_unlock ();
	//LOG ("result=%p\n", result);
	return result;
}
/* Allocate space for the given number of elements of the given
   size, initializing the whole region to binary zeroes. */
void * bt_calloc (size_t nelem, size_t size)
{
	if (g_is_mme_mode)
		return calloc (nelem, size);

	void *result;

	result = bt_malloc (size * nelem);
	if (result)
		bt_memset (result, 0, nelem * size);
	return result;
}

/* Return memory to the heap. */
void bt_free (void *ptr)
{
	if (g_is_mme_mode)
		return free (ptr);

	unsigned int block, blocks, i, type;
	struct list *prev, *next;

	if (!ptr)
		return;

	block = BLOCK (ptr);

	//LOG ("ptr=%p\n", ptr);
	malloc_lock ();

	switch (type = _heapinfo[block].busy.type) {
	case 0:
		/* Find the bt_free cluster previous to this one in the bt_free list.
		   Start searching at the last block referenced; this may benefit
		   programs with locality of allocation. */
		i = _heapindex;
		if (i > block)
			while (i > block)
				i = _heapinfo[i].bt_free.prev;
		else {
			do
				i = _heapinfo[i].bt_free.next;
			while (i > 0 && i < block);
			i = _heapinfo[i].bt_free.prev;
		}

		/* Determine how to link this block into the bt_free list. */
		if (block == i + _heapinfo[i].bt_free.size) {
			/* Coalesce this block with its predecessor. */
			_heapinfo[i].bt_free.size +=
			    _heapinfo[block].busy.info.size;
			block = i;
		} else {
			/* Really link this block back into the bt_free list. */
			_heapinfo[block].bt_free.size =
			    _heapinfo[block].busy.info.size;
			_heapinfo[block].bt_free.next = _heapinfo[i].bt_free.next;
			_heapinfo[block].bt_free.prev = i;
			_heapinfo[i].bt_free.next = block;
			_heapinfo[_heapinfo[block].bt_free.next].bt_free.prev = block;
		}

		/* Now that the block is linked in, see if we can coalesce it
		   with its successor (by deleting its successor from the list
		   and adding in its size). */
		if (block + _heapinfo[block].bt_free.size ==
		    _heapinfo[block].bt_free.next) {
			_heapinfo[block].bt_free.size +=
			    _heapinfo[_heapinfo[block].bt_free.next].bt_free.size;
			_heapinfo[block].bt_free.next =
			    _heapinfo[_heapinfo[block].bt_free.next].bt_free.next;
			_heapinfo[_heapinfo[block].bt_free.next].bt_free.prev = block;
		}

		/* Now see if we can return stuff to the system. */
		blocks = _heapinfo[block].bt_free.size;
		if (blocks >= FINAL_FREE_BLOCKS && block + blocks == _heaplimit
		    && (*_morecore) (0ul) == ADDRESS (block + blocks)) {
			_heaplimit -= blocks;
			(*_morecore) (-blocks * BLOCKSIZE);
			_heapinfo[_heapinfo[block].bt_free.prev].bt_free.next
			    = _heapinfo[block].bt_free.next;
			_heapinfo[_heapinfo[block].bt_free.next].bt_free.prev
			    = _heapinfo[block].bt_free.prev;
			block = _heapinfo[block].bt_free.prev;
		}

		/* Set the next search to begin at this block. */
		_heapindex = block;
		break;

	default:
		/* Get the address of the first bt_free fragment in this block. */
		prev = (struct list *) ( ADDRESS (block)
					+ (_heapinfo[block].busy.info.frag.first
					   << type));

		if (_heapinfo[block].busy.info.frag.nfree ==
		    (BLOCKSIZE >> type) - 1 && _fragblocks[type] > 1) {
			/* If all fragments of this block are bt_free, remove them
			   from the fragment list and bt_free the whole block. */
			--_fragblocks[type];
			for (next = prev, i = 1; i < BLOCKSIZE >> type; ++i)
				next = next->next;
			prev->prev->next = next;
			if (next)
				next->prev = prev->prev;
			_heapinfo[block].busy.type = 0;
			_heapinfo[block].busy.info.size = 1;
			bt_free (ADDRESS (block));
		} else if (_heapinfo[block].busy.info.frag.nfree) {
			/* If some fragments of this block are bt_free, link this fragment
			   into the fragment list after the first bt_free fragment of
			   this block. */
			next = ptr;
			next->next = prev->next;
			next->prev = prev;
			prev->next = next;
			if (next->next)
				next->next->prev = next;
			++_heapinfo[block].busy.info.frag.nfree;
		} else {
			/* No fragments of this block are bt_free, so link this fragment
			   into the fragment list and announce that it is the first
			   bt_free fragment of this block. */
			prev = (struct list *) ptr;
			_heapinfo[block].busy.info.frag.nfree = 1;
			_heapinfo[block].busy.info.frag.first
			    =
			    (unsigned int) ((char *) ptr -
					    (char *) NULL) % BLOCKSIZE >> type;
			prev->next = _fraghead[type].next;
			prev->prev = &_fraghead[type];
			prev->prev->next = prev;
			if (prev->next)
				prev->next->prev = prev;
		}
		break;
	}
	malloc_unlock ();
}

/* Resize the given region to the new size, returning a pointer
   to the (possibly moved) region.  This is optimized for speed;
   some benchmarks seem to indicate that greater compactness is
   achieved by unconditionally allocating and copying to a
   new region. */
void * bt_realloc (void *ptr, size_t size)
{
	if (g_is_mme_mode)
		return realloc (ptr, size);

	void *result, *previous;
	unsigned block, blocks, type;
	int oldlimit;
	void * ret;

	if (!ptr) {
		return bt_malloc (size);
	}
	if (!size) {
		bt_free (ptr);
		return bt_malloc (0ul);
	}

	block = BLOCK (ptr);

	//LOG ("ptr=%p size=%lx\n", ptr, size);
	malloc_lock ();

	switch (type = _heapinfo[block].busy.type) {
	case 0:
		/* Maybe reallocate a large block to a small fragment. */
		if (size <= BLOCKSIZE / 2) {
			if ((result = bt_malloc (size)) != NULL) {
				bt_memcpy (result, ptr, size);
				bt_free (ptr);
			}
			ret = result;
			goto out;
		}

		/* The new size is a large allocation as well; see if
		   we can hold it in place. */
		blocks = BLOCKIFY (size);
		if (blocks < _heapinfo[block].busy.info.size) {
			/* The new size is smaller; return excess memory
			   to the bt_free list. */
			_heapinfo[block + blocks].busy.type = 0;
			_heapinfo[block + blocks].busy.info.size
			    = _heapinfo[block].busy.info.size - blocks;
			_heapinfo[block].busy.info.size = blocks;
			bt_free (ADDRESS (block + blocks));
			ret = ptr;
			goto out;
		} else if (blocks == _heapinfo[block].busy.info.size) {
			/* No size change necessary. */
			ret = ptr;
			goto out;
		} else {
			/* Won't fit, so allocate a new region that will.  Free
			   the old region first in case there is sufficient adjacent
			   bt_free space to grow without moving. */
			blocks = _heapinfo[block].busy.info.size;
			/* Prevent bt_free from actually returning memory to the system. */
			oldlimit = _heaplimit;
			_heaplimit = 0;
			bt_free (ptr);
			_heaplimit = oldlimit;
			result = bt_malloc (size);
			if (!result) {
				/* Now we're really in trouble.  We have to unfree
				   the thing we just freed.  Unfortunately it might
				   have been coalesced with its neighbors. */
				if (_heapindex == block)
					bt_malloc (blocks * BLOCKSIZE);
				else {
					previous =
					    bt_malloc ((block -
						     _heapindex) * BLOCKSIZE);
					bt_malloc (blocks * BLOCKSIZE);
					bt_free (previous);
				}
				ret = NULL;
				goto out;
			}
			if (ptr != result)
				bt_memmove (result, ptr, blocks * BLOCKSIZE);
			ret = result;
			goto out;
		}
		break;

	default:
		/* Old size is a fragment; type is logarithm to base two of
		   the fragment size. */
		if ((size > 1ul << (type - 1)) && (size <= 1ul << type)) {
			/* New size is the same kind of fragment. */
			ret = ptr;
			goto out;
		} else {
			/* New size is different; allocate a new space, and copy
			   the lesser of the new size and the old. */
			result = bt_malloc (size);
			if (!result) {
				ret = NULL;
				goto out;
			}
			bt_memcpy (result, ptr, MIN (size, 1ul << type));
			bt_free (ptr);
			ret = result;
			goto out;
		}
		break;
	}
out:
	malloc_unlock ();
	return ret;
}

int malloc_module_init ()
{
	int ret;

	ret = malloc_init ();
	return ret;
}

/* lock functions */
int malloc_init () 
{
	return mutex_init (&m_malloc_sync);
}

int malloc_lock ()
{
	int ret = mutex_lock (&m_malloc_sync);
	DBPRINT (IS_SYS_ERROR (ret), "failed:%d\n", ret);
	return ret;
}
int malloc_unlock ()
{
	int ret = mutex_unlock (&m_malloc_sync);
	DBPRINT (IS_SYS_ERROR (ret), "failed:%d\n", ret);
	return ret;
}

