/*
 * mem_regions.c
 *
 * function to handle the list of memory regions that were 
 * mapped by the application
 *
 */
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 * Copyright (C) 2005       Xiaolan Huang <xiaolan.huang@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include "../ia32x_utils.h"
#include "../include/mem_regions.h"
#include "../include/kernel_subst32.h"
#include "../include/mutex.h"

#undef DBCODE
#if DEBUG
#define DBCODE(cond, expr)	do{if(cond){expr;}}while(0)
void HIDDEN print_mregion (list_handle region);
void HIDDEN print_mregion_manager(mem_regions_manager manager);
extern char btl_prt_mem_region;
#else
#define DBCODE(cond, expr)
#endif


#define MR_PAGE_SIZE(manager) 		((manager)->pagesize)

#define MR_PAGE_MASK(manager)		(~(MR_PAGE_SIZE(manager) - 1))
#define MR_PAGE_OFF(manager, addr)	((addr) & ~MR_PAGE_MASK(manager))
#define MR_PAGE_START(manager, addr)	(unsigned long)((addr) & MR_PAGE_MASK(manager))
#define MR_PAGE_ALIGN(manager, addr) 	(((addr) + MR_PAGE_SIZE(manager) - 1) & MR_PAGE_MASK(manager))


#ifndef PAGE_SIZE
extern unsigned long g_page_size;
#define PAGE_SIZE 		g_page_size
#define PAGE_MASK               (~(PAGE_SIZE - 1))
#define PAGE_ALIGN(addr)        (((addr) + PAGE_SIZE - 1) & PAGE_MASK)

#endif

unsigned long g_ia32_page_offset;
extern int g_is_mme_mode;

struct mem_region_attrib_t;
typedef struct mem_region_attrib_t * mem_region_attrib;

/* in the region list -
 *   key == region base addr
 *   data == region attrib
 */
struct mem_regions_manager_t{
	unsigned long minaddr;
	unsigned long maxaddr;
	unsigned long preminaddr;
	unsigned long pagesize;

	mutex_handle sync;
	struct list_t * regions;

	long changed;
};

struct mem_region_attrib_t {
	unsigned long len;
	unsigned prot;
	unsigned flags;
};

static long do_mregion_manager_init(mem_regions_manager manager, unsigned long minaddr, unsigned long maxaddr, unsigned long preminaddr, unsigned long pagesize);
static long remove_overlapping_mregion (mem_regions_manager manager, list_handle old_region, unsigned long start, unsigned long len);
static list_handle merge_two_mregions (mem_regions_manager manager, list_handle region1, list_handle region2);
static long do_remove_mregion (mem_regions_manager manager, unsigned long start, unsigned long len);
static long do_insert_mregion (mem_regions_manager manager, unsigned long start, unsigned long len,  unsigned prot, unsigned flags);
static long do_is_inside_mregions (mem_regions_manager manager, unsigned long start, unsigned long len);
static long do_find_mregion (mem_regions_manager manager, unsigned long addr, void **p_region);
static long do_get_mregion_prot (mem_regions_manager manager, unsigned long addr, unsigned *prot, unsigned long *start, unsigned long *len, unsigned *flags);
static long do_set_mregion_prot (mem_regions_manager manager, unsigned long start, unsigned long len, unsigned prot);
static unsigned long do_find_free_mregion (mem_regions_manager manager, unsigned long start, unsigned long len, unsigned flags, unsigned long maxaddr);
void HIDDEN print_mregion (list_handle region);


#define GET_MEMORY_REGION_LIST(mregion)	(&((mregion)->regions))
#define MEM_REGION_STARTADDR(region)	(get_element_key(region))
#define MEM_REGION_ATTRIBUTE(region)	((mem_region_attrib)get_element_data(region))
#define MEM_REGION_ENDADDR(p_region)					\
	({								\
	 	mem_region_attrib ra = (void*)get_element_data (region);	\
		unsigned long region_end;				\
		region_end = MEM_REGION_STARTADDR (p_region) + ra->len;	\
		region_end;						\
	})
#define INSERT_MEM_REGION(mregion, start, attribute)	(list_insert(GET_MEMORY_REGION_LIST(mregion), (start), (unsigned long)(attribute)))
#define REMOVE_MEM_REGION(mregion, region)		(list_remove (GET_MEMORY_REGION_LIST(mregion), (region)))
#define GET_FIRST_MEM_REGION(mregion)	(list_get_first(GET_MEMORY_REGION_LIST(mregion)))
#define NEXT_MEM_REGION(p_region)	((list_handle) get_element_next ((list_handle) p_region))
#define PREV_MEM_REGION(p_region)	((list_handle) get_element_prev ((list_handle) p_region))

#define MAX_MEM_REGION_ADDR(mregion)	((mregion)->maxaddr)
#define MIN_MEM_REGION_ADDR(mregion)	((mregion)->minaddr)
#define PRE_MIN_ADDR(mregion)		((mregion)->preminaddr)

/******************************
 * Wrapper for region managers
 ******************************/
#define IS_LEGAL_MREGION_MANAGER_ID(id)		(id < MREGION_MANAGER_NUM)
#define GET_REGION_MANAGER(id)			((mem_regions_manager) (&(mregion_managers[id])))

struct mem_regions_manager_t mregion_managers[MREGION_MANAGER_NUM];

#if RELEASE
long HIDDEN init_mregion_managers(void)
#else
long HIDDEN init_mregion_managers(int extended_user_mem)
#endif
{
	long ret1, ret2, retval;
	if (g_is_mme_mode
#if ! RELEASE
	    || (!extended_user_mem)
#endif
	) {
		g_ia32_page_offset = 0xc0000000;
	} else {
		g_ia32_page_offset = 0x100000000 - 2 * PAGE_SIZE; 
	}

	ret1 = do_mregion_manager_init(GET_REGION_MANAGER(MREGION_MANAGER_IA32), 
		0ul, IA32_MAX_USER_ADDR, IA32_TASK_UNMAPPED_BASE, IA32_PAGE_SIZE); 
	ret2 = do_mregion_manager_init(GET_REGION_MANAGER(MREGION_MANAGER_BTGENERIC), 
		BTGENERIC_MEM_BASE, BTGENERIC_MAX_ADDRESS, BTGENERIC_MEM_BASE, PAGE_SIZE);

	if (ret1 != 0 || ret2 != 0) {
		return -ENOMEM;
	}

	lock_mregion_manager(MREGION_MANAGER_IA32);
	retval = insert_mregion ( MREGION_MANAGER_IA32, 0ul, PAGE_SIZE, PROT_READ, MAP_PRIVATE | MAP_FIXED);
        assert(retval==0);
	unlock_mregion_manager(MREGION_MANAGER_IA32);


	DBCODE (btl_prt_mem_region, print_mregion_manager(GET_REGION_MANAGER(MREGION_MANAGER_IA32)));
	DBCODE (btl_prt_mem_region, print_mregion_manager(GET_REGION_MANAGER(MREGION_MANAGER_BTGENERIC)));
	return 0;
}


long HIDDEN lock_mregion_manager(MREGION_MANAGER_ID id)
{
	return (long) mutex_lock( GET_REGION_MANAGER(id)->sync );
}
long HIDDEN unlock_mregion_manager(MREGION_MANAGER_ID id)
{
	return (long) mutex_unlock( GET_REGION_MANAGER(id)->sync );
}

long HIDDEN init_mregion_manager_after_fork(MREGION_MANAGER_ID id, long is_parent)
{
	int ret;
	
	if (is_parent) {
		ret = mutex_unlock( GET_REGION_MANAGER(id)->sync );
		assert(ret ==0);
	} else {
		ret = mutex_init( GET_REGION_MANAGER(id)->sync );
		assert (ret == 0);
	}
	return (long) ret;
}

long HIDDEN is_inside_mregions (MREGION_MANAGER_ID id, unsigned long start, unsigned long len)
{
	long ret = 0;
	assert(IS_LEGAL_MREGION_MANAGER_ID(id));
	assert (mutex_islocked (GET_REGION_MANAGER(id)->sync));

	DBPRINT(btl_prt_mem_region, "start=0x%lx len=0x%lx\n", start, len);
	ret = do_is_inside_mregions (GET_REGION_MANAGER(id), start, len);
	DBCODE(btl_prt_mem_region, print_mregions(GET_REGION_MANAGER(id)));
	return ret;
}
long HIDDEN find_mregion (MREGION_MANAGER_ID id, unsigned long addr, void **p_region)
{
	long ret = 0;
	assert (IS_LEGAL_MREGION_MANAGER_ID(id));
	assert (mutex_islocked (GET_REGION_MANAGER(id)->sync));

	ret = do_find_mregion(GET_REGION_MANAGER(id), addr, p_region);
	return ret;
}
long HIDDEN remove_mregion (MREGION_MANAGER_ID id, unsigned long start, unsigned long len)
{
	long retval;
	assert (IS_LEGAL_MREGION_MANAGER_ID(id));
	assert (mutex_islocked (GET_REGION_MANAGER(id)->sync));

	DBPRINT(btl_prt_mem_region, "start=0x%lx len=0x%lx\n", start, len);
	GET_REGION_MANAGER(id)->changed = 1;
	retval = do_remove_mregion (GET_REGION_MANAGER(id), start, len);
	DBCODE(btl_prt_mem_region, print_mregions(GET_REGION_MANAGER(id)));
	return retval;
}

long HIDDEN insert_mregion (MREGION_MANAGER_ID id, unsigned long start, unsigned long len, 
                              unsigned prot, unsigned flags)
{
	long retval;
	assert (IS_LEGAL_MREGION_MANAGER_ID(id));
	assert (mutex_islocked (GET_REGION_MANAGER(id)->sync));

	DBPRINT(btl_prt_mem_region, "start=0x%lx len=0x%lx\n", start, len);
	GET_REGION_MANAGER(id)->changed = 1;
	retval = do_insert_mregion (GET_REGION_MANAGER(id), start, len, prot, flags);
	DBCODE(btl_prt_mem_region, print_mregions(GET_REGION_MANAGER(id)));
	return retval;
}
long HIDDEN get_mregion_prot (MREGION_MANAGER_ID id, unsigned long addr, unsigned *prot, unsigned long *start, unsigned long *len, unsigned *flags)
{
	long retval;
	assert (IS_LEGAL_MREGION_MANAGER_ID(id));
	assert (mutex_islocked (GET_REGION_MANAGER(id)->sync));

	retval = do_get_mregion_prot (GET_REGION_MANAGER(id), addr, prot, start, len, flags);
	return retval;
}
long HIDDEN set_mregion_prot (MREGION_MANAGER_ID id, unsigned long start, unsigned long len, unsigned prot)
{
	long retval;
	assert (IS_LEGAL_MREGION_MANAGER_ID(id));
	assert (mutex_islocked (GET_REGION_MANAGER(id)->sync));

	GET_REGION_MANAGER(id)->changed = 1;
	retval = do_set_mregion_prot(GET_REGION_MANAGER(id), start, len, prot);
	return retval;
}
unsigned long HIDDEN find_free_mregion (MREGION_MANAGER_ID id, unsigned long start, unsigned long len, unsigned flags)
{
	unsigned long retval;
	assert (IS_LEGAL_MREGION_MANAGER_ID(id));
	DBPRINT(btl_prt_mem_region, "start=0x%lx len=0x%lx\n", start, len);
	DBCODE(btl_prt_mem_region, print_mregions(GET_REGION_MANAGER(id)));
	assert (mutex_islocked (GET_REGION_MANAGER(id)->sync));

	retval = do_find_free_mregion(GET_REGION_MANAGER(id), start, len, flags, 0ul);
	DBPRINT(btl_prt_mem_region, "retval=0x%lx\n", retval);
	return retval;
	
}
unsigned long HIDDEN find_free_mregion_below(MREGION_MANAGER_ID id, unsigned long start, unsigned long len, unsigned flags, unsigned long maxaddr)
{
	unsigned long retval;
	assert (IS_LEGAL_MREGION_MANAGER_ID(id));
	DBPRINT(btl_prt_mem_region, "start=0x%lx len=0x%lx\n", start, len);
	DBCODE(btl_prt_mem_region, print_mregions(GET_REGION_MANAGER(id)));
	assert (mutex_islocked (GET_REGION_MANAGER(id)->sync));

	retval = do_find_free_mregion(GET_REGION_MANAGER(id), start, len, flags, maxaddr);
	DBPRINT(btl_prt_mem_region, "retval=0x%lx\n", retval);
	return retval;
}
long HIDDEN is_memory_maps_changed(MREGION_MANAGER_ID id)
{
	assert (IS_LEGAL_MREGION_MANAGER_ID(id));
	return GET_REGION_MANAGER(id)->changed;
}
long HIDDEN reset_memory_maps(MREGION_MANAGER_ID id)
{
	long ret = GET_REGION_MANAGER(id)->changed;
	assert (IS_LEGAL_MREGION_MANAGER_ID(id));

	GET_REGION_MANAGER(id)->changed = 0;
	return ret;
}

/****************************
 * Region 
 ****************************/
unsigned long HIDDEN get_mregion_startaddr(list_handle region)
{
	return MEM_REGION_STARTADDR(region);
}
unsigned long HIDDEN get_mregion_endaddr(list_handle region)
{
	return MEM_REGION_ENDADDR(region);
}

list_handle HIDDEN get_next_mregion(list_handle region)
{
	return NEXT_MEM_REGION(region);
}
list_handle HIDDEN get_prev_mregion(list_handle region)
{
	return PREV_MEM_REGION(region);
}

list_handle HIDDEN get_first_mregion(MREGION_MANAGER_ID id)
{
	assert (mutex_islocked (GET_REGION_MANAGER(id)->sync));
	return GET_FIRST_MEM_REGION(GET_REGION_MANAGER(id));
}
/*****************************************************************************/
/* check whether the given memory chunk overlapps with one
 * of the regions in the list
 */
/*****************************************************************************/
long do_mregion_manager_init(mem_regions_manager manager, unsigned long minaddr, unsigned long maxaddr, 
	unsigned long preminaddr, unsigned long pagesize)
{
	long retval = 0;

	assert (manager != NULL);

	manager->minaddr = minaddr;
	manager->maxaddr = maxaddr;
	manager->preminaddr = preminaddr;
	manager->pagesize = pagesize;
	manager->regions = NULL;

	manager->changed = 1;

	/* creat lock for region manager */
	manager->sync = mutex_create ();
	if (manager->sync == NULL) {
		retval = -ENOMEM;
		goto out;
	}
	retval = mutex_init (manager->sync);
	if (retval != 0) {
		goto out;
	}

out:
	return retval;
}

static long do_is_inside_mregions (mem_regions_manager manager, unsigned long start, unsigned long len)
{
	list_handle region;
	unsigned long end;
	unsigned long region_start, region_len, region_end;
	mem_region_attrib ra;
	long ret = 0;

	end = start + len;
	if (start < manager->minaddr || end >= manager->maxaddr ) {
		return -EINVAL;
	}

	region = GET_FIRST_MEM_REGION(manager);
	if (region == 0) {
		/* list empty */
		goto out;
	}
	region_start = MEM_REGION_STARTADDR (region);

	while (region_start < end) {
		ra = MEM_REGION_ATTRIBUTE (region);
		region_len = ra->len;
		region_end = region_start + region_len;

		if (region_end > start) {
			/* overlapping */
			ret = 1;
			break;
		}
		region = NEXT_MEM_REGION (region);
		if (region == 0) {
			/* end of list */
			break;
		}
		region_start = MEM_REGION_STARTADDR (region);
	}
out:
	return ret;
}

/*****************************************************************************/
/* find the region containing the addr 
 * out:
 *     p_region contains the found region's pointer
 * returns 0 on success and negative on failure */
/*****************************************************************************/
static long do_find_mregion (mem_regions_manager manager, unsigned long addr, void **p_region)
{
	list_handle region;
	mem_region_attrib ra;
	unsigned long region_end;
	unsigned long region_len;
	unsigned long region_start;
	
	region = GET_FIRST_MEM_REGION (manager);
	
	while (1) {
		if (region == NULL) {
			return -1;
		}

		region_start = MEM_REGION_STARTADDR(region);
		ra = MEM_REGION_ATTRIBUTE(region);
		region_len = ra->len;
		region_end = region_start + region_len;

		if (region_start > addr) {
			return -1;
		} else if (region_end >= addr) {
			/* here, region_start <= addr */
			if (p_region != NULL) {
				*p_region = region;
			}
			break;
		}
		
		region = NEXT_MEM_REGION(region);
	}

	return 0;
}

/*****************************************************************************/
/* remove a region from the list 
 * returns 0 on success and a negative value on failure */
/*****************************************************************************/
static long do_remove_mregion (mem_regions_manager manager, unsigned long start, unsigned long len)
{
	list_handle next = 0;
	unsigned long end;
	unsigned long next_len, next_start, next_end;
	long ret;

	end = start + len;
	len = MR_PAGE_ALIGN(manager, len);//IA32_PAGE_ALIGN (len);
	DBPRINT (0, "start=%lx len=%lx\n", start, len);

	next = GET_FIRST_MEM_REGION (manager);
	if (next == 0) {
		/* if the list is empty than no extra work is needed */
		DBPRINT (0, "try to remove a region from an empty list\n");
		ret = 0;
		goto out;
	}

	next_start = MEM_REGION_STARTADDR (next);
	while (next_start < end) {
		list_handle tmp = next;
		mem_region_attrib ra;

		ra = MEM_REGION_ATTRIBUTE(next);
		next_len = ra->len;
		next_end = next_start + next_len;

		/* get the next element before calling remove_overlapping_region
		 * because the current element  might be removed from the list
		 */
		next = NEXT_MEM_REGION (next);

		if (next_end > start) {
			ret = remove_overlapping_mregion (manager, tmp, start, len);
			if (ret < 0) {
				goto out;
			}
		}
		if (next == 0) {
			break;
		}
		next_start = MEM_REGION_STARTADDR (next);
	}

	ret = 0;

out:
	DBCODE (0, print_mregions (manager));
	return ret;
}

/*****************************************************************************/
/* insert a new region to the list 
 * returns 0 on success and a negative value on failure */
/*****************************************************************************/
static long do_insert_mregion (mem_regions_manager manager, unsigned long start, unsigned long len, 
                                 unsigned prot, unsigned flags)
{
	list_handle next_region, prev_region, new_region, tmp_region;
	mem_region_attrib ra = NULL;
	long ret;

	len = MR_PAGE_ALIGN(manager, len);//IA32_PAGE_ALIGN (len);
	DBPRINT (0, "start=%lx len=%lx prot=%x\n", start, len, prot);
	/* remove previously mapped region */
	ret = do_remove_mregion (manager, start, len);
	if (ret < 0) {
		goto out;
	}

	/* create a new region */
	ra = bt_malloc (sizeof (*ra));
	if (ra == NULL) {
		ret = -ENOMEM;
		goto out;
	}

	ra->len = len;
	ra->prot = prot;
	ra->flags = flags;
	new_region = INSERT_MEM_REGION(manager, start, ra);
	if ((long)new_region < 0) {
		bt_free (ra);
		ret = (long)new_region;
		goto out;
	}

	/* get the regions on both sides of start */
	next_region = NEXT_MEM_REGION (new_region);
	prev_region = PREV_MEM_REGION (new_region);

	/* try to merge with the existing regions */
	if (prev_region != 0) {

		tmp_region = merge_two_mregions (manager, prev_region, new_region);
		if ((long)tmp_region < 0) {
			ret = (long)tmp_region;
			goto out;
		} else if (tmp_region != 0) {
			new_region = tmp_region;
		}

	}

	if (next_region != 0) {

		tmp_region = merge_two_mregions (manager, next_region, new_region);
		if ((long)tmp_region < 0) {
			ret = (long)tmp_region;
			goto out;
		} else if (tmp_region != 0) {
			new_region = tmp_region;
		}

	}

	DBCODE (0, print_mregions (manager));

out:
	return ret;
}

/*****************************************************************************/
/* get the permissions of the region where the given address resides.
 * input:  manager - memory region manager
 *         addr -
 *         prot - will hold the permission at the given address
 *         start - will hold the region's base address
 *         len - will hold the region's size
 *         flags - will hold the region's flags 
 * output: 1 if the address is within a mapped memory area, 0 otherwise
 */
/*****************************************************************************/
static long do_get_mregion_prot (mem_regions_manager manager, unsigned long addr, unsigned *prot, 
                                unsigned long *start, unsigned long *len, 
			        unsigned *flags)
{
	unsigned p = 0, fl = 0;
	unsigned long s = 0, l = 0;
	int found = 0;
	list_handle region;

	DBPRINT (0, "addr=%lx\n", addr);

	region = GET_FIRST_MEM_REGION(manager);

	while (1) {

		mem_region_attrib ra;
		unsigned long region_end;
		unsigned long region_len;
		unsigned long region_start = 0;


		if (region == 0) {
			/* no more regions to check */
			break;
		}

		
		region_start = MEM_REGION_STARTADDR (region);
		if (region_start > addr) {
			/* the list is guranteed to be sorted, so there is no point
			 * in checking further
			 */
			break;
		}

		/* get the region's size */
		ra = MEM_REGION_ATTRIBUTE (region);
		region_len = ra->len;
		region_end = region_start + region_len;

		if (region_end > addr) {
			found = 1;
			p = ra->prot;
			s = region_start;
			l = region_len;
			fl = ra->flags;
			break;
		}
		region = NEXT_MEM_REGION (region);
	}

	if (!found) {
		p = 0;
		s = MR_PAGE_ALIGN(manager, addr);
		l = MR_PAGE_SIZE(manager);
		fl = 0;
	}

	DBPRINT (0, "addr=%lx found=%d perm=%x start=%lx len=%lx flags=%x\n", 
		 addr, found, p, s, l, fl);

	if (prot != NULL) {
		*prot = p;
	}
	if (start != NULL) {
		*start = s;
	}
	if (len != NULL) {
		*len = l;
	}
	if (flags != NULL) {
		*flags = fl;
	}

	return found;
}

/*****************************************************************************/
/* change the permissions of the given memory area
 * input:  manager - memory region manager
 *         start - base address of the area
 *         len - size of area
 *         prot - new permissions for the area
 * output: 0 on success, negative error value on error
 */
/*****************************************************************************/
static long do_set_mregion_prot (mem_regions_manager manager, unsigned long start, unsigned long len, unsigned prot)
{
	long ret;
	unsigned flags;

	DBPRINT (0, "start=%lx len=%lx prot=%x\n", start, len, prot);
	/* mprotect fails for non mapped memory, so we can simple remove the region
	 * and re-insert it with the new permissions
	 */
	ret = do_get_mregion_prot (manager, start, NULL, NULL, NULL, &flags);
	if (ret < 0) {
		goto out;
	}
	ret = do_remove_mregion (manager, start, len);
	if (ret < 0) {
		goto out;
	}
	ret = do_insert_mregion (manager, start, len, prot, flags); 
out:
	return ret;
}

/*****************************************************************************/
/* try to merge two regions 
 * input:  manager - memory region manager
 *         reg0 and reg1 are handles of the regions to be merged.
 * output: in case of error return a negative value.
 *         in case the regions cannot be merged return 0. 
 *         otherwise return a handle to a region that is the result of the
 *         merge operation of the two regions.
 * comments :the given regions can be destroyed, that is - removed from the
 *           list and get their attribute freed.
 */
/*****************************************************************************/
static list_handle merge_two_mregions (mem_regions_manager manager, list_handle reg0, list_handle reg1)
{
	mem_region_attrib ra0, ra1;
	unsigned long start0, start1, end0, end1;
	unsigned long new_len = 0, new_start = 0;
	list_handle new_reg = (list_handle)0;
	unsigned prot;
	unsigned flags;
	long ret = 0;

	/* finish with trivial cases right here */
	assert (reg0 != 0);
	assert (reg1 != 0);

	/* get info of the regions */

	start0 = MEM_REGION_STARTADDR(reg0);
	start1 = MEM_REGION_STARTADDR(reg1);

	ra0 = MEM_REGION_ATTRIBUTE(reg0); 
	ra1 = MEM_REGION_ATTRIBUTE(reg1);

	end0 = start0 + ra0->len;
	end1 = start1 + ra1->len;

	/* see whether the regions can be merged */

	if ((end0 < start1) || (end1 < start0)) {
		/* cannot merge */
		goto out;
	}

	if (ra1->prot != ra0->prot) {
		/* cannot merge */
		goto out;
	}
	prot = ra0->prot;

	if (ra1->flags != ra0->flags) {
		/* cannot merge */
		goto out;
	}
	flags = ra0->flags;
	
	/* the regions can be merged - there are few possible scenarios
	 * 1) the new region contains the base region:
	 * 	the base region should be removed
	 * 	and the new region should be returned as the result
	 * 	of the merge operation
	 * 2) the base region contains the new one:
	 * 	the new region has to be removed and the base
	 * 	region should be returned as the result of the merge
	 * 	operation
	 * 3) the regions overlap ( or touch ):
	 * 	both regions should be removed and a new region
	 * 	should be created with the appropriate settings
	 */ 
	if (start0 >= start1) {
		if (end0 <= end1) {
			/* reg1 contains reg0 */
			new_reg = reg1;
		} else { /* end0 > end1 */
			new_start = start1;
			new_len = end0 - new_start;
		}
	} else { /* start1 > start0 */

		if  (end1 <= end0) {
			/* reg0 contains reg1 */
			new_reg = reg0;
		} else { /* end1 > end0 */
			new_start = start0;
			new_len = end1 - new_start;
		}
	}

	if (new_reg != reg0) {
		/* remove region 0 */
		bt_free (ra0);
		ret = REMOVE_MEM_REGION(manager, reg0);
		if (ret < 0) {
			new_reg = (list_handle)ret;
		}
	}
	if (ret < 0) {
		goto out;
	}

	if (new_reg != reg1) {
		/* remove region 1 */
		bt_free (ra1);
		ret = REMOVE_MEM_REGION(manager, reg1);
		if (ret < 0) {
			new_reg = (list_handle)ret;
		}
	}
	if (ret < 0) {
		goto out;
	}

	if (new_reg == 0) {
        	mem_region_attrib ra = bt_malloc (sizeof (*ra)); 
        	if (ra == NULL) { 
            		new_reg = (list_handle)-ENOMEM;
        	} else {
	        	ra->len = new_len;
	        	ra->prot = prot; 
			ra->flags = flags;
			new_reg = INSERT_MEM_REGION(manager, new_start, ra);
			if ((long)new_reg < 0) {
				bt_free (ra);
			}
		}
	}

out:
	return new_reg;
}

/*****************************************************************************/
/* remove a memory region from an existing region
 * input:   manager - memory region manager
 *          old_region - existing region
 *          start - base address of the region to be removed
 *          len - size of region to be removed
 * output:  0 on success, negative value on error
 */
/*****************************************************************************/
static long remove_overlapping_mregion (mem_regions_manager manager, list_handle old_region, 
                                          unsigned long start, unsigned long len)
{
	unsigned long old_start, old_len, old_end, end;
	unsigned long new_start[2] = {0};
	unsigned long new_len[2] = {0};
	mem_region_attrib old_ra;
	long j;
	long ret;
	mem_region_attrib  new_ra[2] = {0};

	old_start = MEM_REGION_STARTADDR (old_region);
	old_ra = MEM_REGION_ATTRIBUTE (old_region);
	old_len = old_ra->len;
	old_end = old_start + old_len;
	end = start + len;

	assert (old_end > start);
	assert (end > old_start);

	/* the old region is removed in any case */
	ret = REMOVE_MEM_REGION (manager, old_region);
	if (ret < 0) {
		return ret;
	}

	if (old_start >= start) {
		if (old_end <= end) {
			/* the existing old_region is inside the tbr region 
			 * we have nothing more to do
			 */
		} else { /* old_end > end */
			/* shrink the old region */
			new_start[0] = end;
			new_len[0] = old_end - new_start[0];
		}
	} else { /* old_start < start */
		/* we will need at least one new region */
		new_start[0] = old_start;
		new_len[0] = start - old_start;
		if (old_end > end) {
			/* the tbr region is inside the old region. the
			 * old region has to split
			 */
			new_start[1] = end;
			new_len[1] = old_end - end;
		}
	}

	/* create the new regions */
	for (j = 0; j < 2; j ++) {
		if (new_len[j] != 0) {
			new_ra[j] = bt_malloc (sizeof (struct mem_region_attrib_t));
			if (new_ra[j] == NULL) {
				ret = -ENOMEM;
				if (j == 1) {
					bt_free (new_ra[0]);
				}
				goto out;
			}
			new_ra[j]->len = new_len[j];
			new_ra[j]->prot = old_ra->prot;
			new_ra[j]->flags = old_ra->flags;
			ret = (long)INSERT_MEM_REGION(manager, new_start[j], new_ra[j]);
			if (ret < 0) {
				long i;
				for (i = 0; i <= j; i ++) {
					bt_free (new_ra[i]);
				}
				goto out;
			}
		}
	}

out:
	bt_free (old_ra);
	return ret;
}

/*****************************************************************************/
/* look for an unmapped area in the given memory area
 * input:  manager - memory region manager
 *         start - the address from which to start the search
 *         len - size of requested area
 *         flags - flags to map the memory, also see mmap
 *         maxaddr - given maxaddr, 
 *                   it is meaningful only if maxaddr > MIN_MEM_REGION_ADDR(manager)
 * output: -ENOMEM if no free memory is found, address of area on success
 *         the returned address is aligned on system page-size for most
 *         cases, and on SHMLBA if MAP_SHARED is asserted in 'flags'.
 */
/*****************************************************************************/
static unsigned long do_find_free_mregion (mem_regions_manager manager, unsigned long start, unsigned long len, unsigned flags, unsigned long maxaddr)
{
#define SHMLBA_MINUS_ONE ((unsigned long)SHMLBA - 1ul)
	unsigned long unmapped_area = 0;
	int bak_found = 0;
	unsigned long tmp_unmapped_area = 0;
	unsigned shared, fixed;
	list_handle region;
	long retval;
	unsigned long prev_region_end = MIN_MEM_REGION_ADDR(manager);

	if (maxaddr <= MIN_MEM_REGION_ADDR(manager)) {
		maxaddr = MAX_MEM_REGION_ADDR(manager);
	}

	shared = flags & MAP_SHARED;
	fixed = flags & MAP_FIXED;

	if (start == 0) { 
		start = PRE_MIN_ADDR(manager);
	}


	/* if this area is going to be mapped than make sure the
	 * whole page is free. otherwise, if the end of the last
	 * page is allocated, its contents wil be erased once the
	 * page is mapped again */
	len = PAGE_ALIGN (len);


	if (shared && !fixed) {
		/* shared memory must be aligned on SHMLBA */
		start = (start + SHMLBA_MINUS_ONE) & ~SHMLBA_MINUS_ONE;
	}

	DBPRINT (btl_prt_mem_region, "start=%lx len=%lx flags=%x %lx-%lx\n", start, len, flags, MIN_MEM_REGION_ADDR(manager), maxaddr);
	/* go over the list of mapped memory regions and look for a free slot */
	region = GET_FIRST_MEM_REGION(manager); 

	while (1) {

		mem_region_attrib ra;
		unsigned long region_end;
		unsigned long region_len;
		unsigned long region_start = 0;
		unsigned long slot_size = 0;


		if (region == 0) {
			/* no more regions to check */
			break;
		}

		
		region_start = MEM_REGION_STARTADDR (region);
		assert (region_start >= prev_region_end);

		/* get the region's size */
		ra = MEM_REGION_ATTRIBUTE(region);
		region_len = ra->len;
		region_end = region_start + region_len;

		if (region_start >= start) {
			/* 1. don't get areas lower than 'start' */

			/* the memory needs to be aligned at the native page size */
			unmapped_area = PAGE_ALIGN (MAX(start, prev_region_end));
			if (shared) {
				/* shared memory must be aligned at SHMLBA */
				unmapped_area = (unmapped_area + SHMLBA_MINUS_ONE) & 
						~SHMLBA_MINUS_ONE;
			}
			
			/* the free slot stretches from 'unmapped_area' to 'region_start' .
			 * unmapped_area might be bigger than region_start because it
			 * is aligned on a page boundry, while region_start not */
			if (region_start >= unmapped_area) {
				slot_size = region_start - unmapped_area;
				if (slot_size >= len) {
					goto out;
				}
			}

		} else if (!bak_found && prev_region_end >= PRE_MIN_ADDR(manager)) {
			/* 2. get a backup unmapped area */ 

			/* the memory needs to be aligned at the native page size */
			tmp_unmapped_area = PAGE_ALIGN (prev_region_end);

			if (shared) {
				/* shared memory must be aligned at SHMLBA */
				tmp_unmapped_area = (tmp_unmapped_area + SHMLBA_MINUS_ONE) & 
						~SHMLBA_MINUS_ONE;
			}
			
			/* the free slot stretches from 'unmapped_area' to 'region_start' .
			 * unmapped_area might be bigger than region_start because it
			 * is aligned on a page boundry, while region_start not */
			if (region_start >= tmp_unmapped_area) {
				slot_size = region_start - tmp_unmapped_area;
				if (slot_size >= len) {
					bak_found = 1; /* free area was found */
				}
			}
		}
	
		prev_region_end = region_end;
		region = get_element_next (region);
	}

	/* there is no free slot between mapped areas - try the end of the list */
	/* the memory needs to be aligned at the native page size */
	unmapped_area = MAX( prev_region_end, PRE_MIN_ADDR(manager));
	unmapped_area = MAX( start, unmapped_area);

	if (shared) {
		/* shared memory must be aligned at SHMLBA */
		unmapped_area = (unmapped_area + SHMLBA_MINUS_ONE) & 
				~SHMLBA_MINUS_ONE;
	}
	DBPRINT(btl_prt_mem_region, "unmapped_area = 0x%lx len=0x%lx MAX=0x%lx\n", unmapped_area, len, maxaddr);

out:

	/* unmapped_area is the address we want to return */
	assert (unmapped_area >= start);

	/* if the first available page is too high then return error */
	if (unmapped_area + len >= maxaddr ) {
		if ((tmp_unmapped_area + len >= maxaddr) || (!bak_found)) {
			retval = -ENOMEM;
		} else {
			retval = tmp_unmapped_area;
		}	
	} else {
		retval = unmapped_area;
	}

	DBPRINT (btl_prt_mem_region, "retval=%lx\n", (unsigned long)retval);
	return (unsigned long)retval;
}

#if DEBUG
void HIDDEN print_mregion (list_handle region)
{
	unsigned long start, len, end;
	unsigned prot;
	unsigned flags;
	mem_region_attrib ra = MEM_REGION_ATTRIBUTE (region);

	start = MEM_REGION_STARTADDR (region);
	len = ra->len;
	prot = ra->prot;
	flags = ra->flags;
	end = start + len;
	fprintf (g_log_file, "%lx:%lx-%x-%x ", start, end, prot, flags);
}

void HIDDEN print_mregions (mem_regions_manager manager)
{
	list_handle h;

	h = GET_FIRST_MEM_REGION (manager);
	while (h != 0) {
		print_mregion (h);
		h = NEXT_MEM_REGION (h);
	}
	fprintf (g_log_file, "\n");
}

void HIDDEN print_mregion_manager(mem_regions_manager manager)
{
	DBPRINT(1, "Memory Region Manager [0x%lx - 0x%lx] unmmapedbase=0x%lx pagesize=0x%lx=%d\n", 
		manager->minaddr, manager->maxaddr, manager->preminaddr, manager->pagesize);
}
#endif

