/* 
 * mutex.c
 *
 * implement the mutex services as declared in mutex.h
 *
 */
 /*
 *
 * This file has been derived from GNU libc. Please see COPYING.LIBC
 * for more information
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include "../ia32x_utils.h"
#include "../include/mutex_internal.h"
#include "../include/ia32x_utils_internal.h"



static int m_has_futex = -1;



/*****************************************************************************/
/*****************************************************************************/
static void __lll_lock_wait (int * ftx)
{
	do {
		int	oldval = GCC_BUILTIN_SYNC_VAL_COMPARE_AND_SWAP (ftx, 1, 2);
		if (oldval != 0) {
			lll_futex_wait (ftx, 2);
		}
	} while (GCC_BUILTIN_SYNC_BOOL_COMPARE_AND_SWAP (ftx, 0, 2) != 1);
}




/*****************************************************************************/
/* return 0 on success */
/*****************************************************************************/
static inline int __attribute__ ((always_inline)) __lll_mutex_trylock (int *ftx)
{
	int is_free = GCC_BUILTIN_SYNC_BOOL_COMPARE_AND_SWAP (ftx, 0, 1);
	return ! is_free; /* is_free ? 0 : 1; */
}

/*****************************************************************************/
/*****************************************************************************/
static inline void __attribute__ ((always_inline)) __lll_mutex_lock (int *ftx)
{
	int is_free = GCC_BUILTIN_SYNC_BOOL_COMPARE_AND_SWAP (ftx, 0, 1);
	if ( ! is_free ) {
		__lll_lock_wait (ftx);
	}
}

/*****************************************************************************/
/*****************************************************************************/
static inline void __attribute__ ((always_inline)) __lll_mutex_unlock (int *ftx)
{
	int is_waiting;
	GCC_BUILTIN_SYNC_SYNCHRONIZE ();
	is_waiting = GCC_BUILTIN_SYNC_LOCK_TEST_AND_SET (ftx, 0);
	if (is_waiting > 1) {
		lll_futex_wake (ftx, 1);
	}
}

/*****************************************************************************/
/*****************************************************************************/
static int ftx_mutex_lock (mutex_t * mutex)
{
	register unsigned long _tp asm ("r13");
	/* Check whether we already hold the mutex.  */
	if (mutex->__data.__owner == _tp) {
		/* Just bump the counter.  */
		if (mutex->__data.__count + 1 == 0)
			/* Overflow of the counter.  */
			return -EAGAIN;

		++mutex->__data.__count;
	} else {
		/* We have to get the mutex.  */
		__lll_mutex_lock (&mutex->__data.__lock);

		/* Record the ownership.  */
		mutex->__data.__owner = _tp;
		mutex->__data.__count = 1;
	}
	return 0;
}

/*****************************************************************************/
/*****************************************************************************/
static int ftx_mutex_unlock (mutex_t * mutex)
{
	register unsigned long _tp asm ("r13");
	if (mutex->__data.__owner != _tp)
		return -EPERM;

	if (--mutex->__data.__count != 0)
		/* We still hold the mutex.  */
		return 0;
	/* Always reset the owner field.  */
	mutex->__data.__owner = 0ul;

	/* Unlock.  */
	__lll_mutex_unlock (&mutex->__data.__lock);

	return 0;
}

/*****************************************************************************/
/*****************************************************************************/
static int ftx_mutex_init (mutex_t * mutex)
{
	mutex->__data.__lock = 0;
	mutex->__data.__count = 0;
	mutex->__data.__owner = 0;
	return 0;
}

/*****************************************************************************/
/* return 0 on success */
/*****************************************************************************/
static int ftx_mutex_trylock (mutex_t * mutex)
{
	register unsigned long _tp asm ("r13");
	/* Check whether we already hold the mutex.  */
	if (mutex->__data.__owner == _tp) {
		/* Just bump the counter.  */
		if (mutex->__data.__count + 1 == 0)
			/* Overflow of the counter.  */
			return -EAGAIN;

		++mutex->__data.__count;
		return 0;
	}

	if (__lll_mutex_trylock (&mutex->__data.__lock) == 0) {
		/* Record the ownership.  */
		mutex->__data.__owner = _tp;
		mutex->__data.__count = 1;
		return 0;
	}
	return -EBUSY;
}

/*****************************************************************************/
/*****************************************************************************/
static int spl_mutex_lock (mutex_t * mutex)
{
	int cont = 1;
	unsigned long locked;
	register unsigned long _tp asm ("r13");

	do {
		locked = GCC_BUILTIN_SYNC_VAL_COMPARE_AND_SWAP (&mutex->__data.__owner, 0l, (long)_tp);
		GCC_BUILTIN_SYNC_SYNCHRONIZE ();
		if (locked == _tp) {
			/* recurese mutex */
			if (mutex->__data.__count + 1 == 0) {
				return -EAGAIN;
			}
			mutex->__data.__count ++;
			cont = 0;
		} else if (locked == 0) {
			mutex->__data.__count = 1;
			cont = 0;
		} else {
			/* mutex is busy */
			SYSCALL (sched_yield, ());
		}
	} while (cont);

	return 0;
}

/*****************************************************************************/
/*****************************************************************************/
static int spl_mutex_unlock (mutex_t * mutex)
{
	int count;
	register unsigned long _tp asm ("r13");

	if (mutex->__data.__owner != _tp) {
		return -EPERM;
	}
	count = -- mutex->__data.__count;
	GCC_BUILTIN_SYNC_SYNCHRONIZE ();
	if (count == 0) {
		mutex->__data.__owner = 0;
	}
	return 0;
}
/*****************************************************************************/
/*****************************************************************************/
static int spl_mutex_init (mutex_t * mutex)
{
	mutex->__data.__lock = 0;
	mutex->__data.__count = 0;
	mutex->__data.__owner = 0;
	return 0;
}
/*****************************************************************************/
/*****************************************************************************/
static int spl_mutex_trylock (mutex_t * mutex)
{
	unsigned long locked;
	int ret = 0;
	register unsigned long _tp asm ("r13");

	locked = GCC_BUILTIN_SYNC_VAL_COMPARE_AND_SWAP (&mutex->__data.__owner, 0l, (long)_tp);
	GCC_BUILTIN_SYNC_SYNCHRONIZE ();
	if (locked == _tp) {
		/* recurese mutex */
		if (mutex->__data.__count + 1 == 0) {
			ret = -EAGAIN;
		} else {
			mutex->__data.__count ++;
		}
	} else if (locked == 0) {
		mutex->__data.__count = 1;
	} else {
		/* mutex is busy */
		ret = -EBUSY;
	}

	return ret;
}

/*****************************************************************************/
/* export functions */
/*****************************************************************************/
int mutex_module_init ()
{
	int ret = SYSCALL (futex, ((int*)1, FUTEX_WAKE, 1, NULL, NULL));
	if (ret == -ENOSYS) {
		m_has_futex = 0;
	} else {
		m_has_futex = 1;
	}
	DBPRINT (0, "ret=%d m_has_futex=%d\n", ret, m_has_futex);
	return 0;
}
int mutex_lock (mutex_handle mh) 
{
	mutex_t * m = (mutex_t *)mh;
	DBPRINT (0, "enter: mh=%p lock=%d count=%d owner=%lx\n",
		 mh, m->__data.__lock, m->__data.__count, m->__data.__owner);
	if (m_has_futex) {
		return ftx_mutex_lock (m);
	} else {
		return spl_mutex_lock (m);
	}
}
int mutex_unlock (mutex_handle mh) 
{
	mutex_t * m = (mutex_t *)mh;
	DBPRINT (0, "enter: mh=%p lock=%d count=%d owner=%lx\n",
		 mh, m->__data.__lock, m->__data.__count, m->__data.__owner);
	if (m_has_futex) {
		return ftx_mutex_unlock (m);
	} else {
		return spl_mutex_unlock (m);
	}
}
int mutex_init (mutex_handle mh) 
{
	mutex_t * m = (mutex_t *)mh;
	if (m_has_futex) {
		return ftx_mutex_init (m);
	} else {
		return spl_mutex_init (m);
	}
}
int mutex_trylock (mutex_handle mh) 
{
	mutex_t * m = (mutex_t *)mh;
	DBPRINT (0, "enter: mh=%p lock=%d count=%d owner=%lx\n",
		 mh, m->__data.__lock, m->__data.__count, m->__data.__owner);
	if (m_has_futex) {
		return ftx_mutex_trylock (m);
	} else {
		return spl_mutex_trylock (m);
	}
}
int mutex_islocked (mutex_handle mh)
{
	mutex_t * m = (mutex_t *)mh;
	DBPRINT (0, "enter: mh=%p lock=%d count=%d owner=%lx\n",
		 mh, m->__data.__lock, m->__data.__count, m->__data.__owner);
	return (m->__data.__lock != 0);
}
unsigned long get_mutex_owner (mutex_handle mh) 
{
	mutex_t * m = (mutex_t *)mh;
	return m->__data.__owner;
}
mutex_handle mutex_create ()
{
	mutex_t * m = bt_malloc (sizeof (mutex_t));
	return (mutex_handle)m;
}
void mutex_destroy (mutex_handle mh)
{
	bt_free (mh);
}

int
btlib_init_semaphore(void * sem, __s32 start_size)
{
	long *new_semaphore; 
	long **target = (long **)sem; // cast semaphore pointer to correct type

	int futex_try = SYSCALL (futex, ((int*)1, FUTEX_WAKE, 1, NULL, NULL));
	if (futex_try == -ENOSYS) {
		return -1;// Futex is not supported - cannot implement semaphore
	}

	if (*target != NULL) {
		// new_semaphore is intialized already (usually due to Fork)
		// so we only need to set the correct start size
		**target = (long)start_size;
		return 0; 
	}

	new_semaphore = bt_malloc(sizeof(long)); 
	if (new_semaphore == NULL) { 
		return -1; // bt_malloc error
	}
	*new_semaphore = (long)start_size;
	*target = new_semaphore;
	return 0;
}


int
btlib_destroy_semaphore(void * sem)
{
	long **target = (long **)sem; // cast semaphore pointer to correct type
	if (*target != NULL)
		bt_free(*target);
	*target = NULL;
	return 0;
}


int
btlib_wait_semaphore(void * sem, int block)
{

	int loop = 0;
	long **target = (long **)sem; // cast semaphore pointer to correct type
	long * counterPtr = *target;
	long value;
	long status;
	
	if (block){
		while (loop == 0) {
			status = 0;
			value = *counterPtr;
			if (value <= 0) {
				status = lll_futex_wait(counterPtr, value); // sleep while this value is in the counter
			}
			value = GCC_BUILTIN_SYNC_FETCH_AND_SUB (counterPtr, 1ul); // decrement counter
			// check if futex was not waken by WAKE, or counter value changed since checked
			if (status != 0 || value <= 0) {
				value =	GCC_BUILTIN_SYNC_FETCH_AND_ADD (counterPtr, 1ul); // undo the decrement
			} else {
				loop = -1; // break loop
			}
		}
	} else {
		status = 0;
		value = *counterPtr;
		if (value <= 0) {
			return -1;	
		}
		value = GCC_BUILTIN_SYNC_FETCH_AND_SUB (counterPtr, 1ul); // decrement counter
		if (value <= 0) {
			value = GCC_BUILTIN_SYNC_FETCH_AND_ADD (counterPtr, 1ul); // undo the decrement
			return -1;
		}
		// if we got here - entering the semaphore was successful
	}
    return 0;
}


void 
btlib_signal_semaphore(void * sem)
{

    long **target = (long **)sem; // cast semaphore pointer to correct type
	long * counterPtr = *target;
	long value = GCC_BUILTIN_SYNC_FETCH_AND_ADD (counterPtr, 1ul); // increment counter

	if (value >= 0) {
		lll_futex_wake(counterPtr, 1); // wake a sleeping thread
	}
}

__s32
btlib_get_count_semaphore(void * sem)
{
	long **target = (long **)sem; // cast semaphore pointer to correct type
	return (__s32)(**target);
}

