/* 
 * ia32x_loader.c
 * custom loader for libia32x.so.
 *
 */
 /*
 *
 * This file has been derived from GNU libc. Please see COPYING.LIBC
 * for more information
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/


#include <ia32x_utils.h>
#include "utils.h"
#include "ia32x_loader.h"
#include "loader_tls.h"
#include "selinux.h"

#define NUMBER_OF_GROUPS 64
#define SUID_LOADER_NAME "suid_ia32x_loader"

/* functions exported by btlib */
static union {
	struct _fdesc_tag * fdesc;
	long (*fptr)(unsigned long *, long, long);
} m_lib_init = {0};

static union {
	struct _fdesc_tag * fdesc;
	long (*fptr)(u64, int, char **, char **);
} m_btlib_load_and_exec = {0};

struct _btlib_entry_points_tag g_btlib_entry_points[3] = {
	{"lib_init", &m_lib_init.fdesc},
	{"btlib_load_and_exec", &m_btlib_load_and_exec.fdesc},
	/* null entry must be last in the table */
	{"", NULL}
};

u64 g_pagesize = 0;
int g_log_file = 0; /* will be initialized later */
static u64 m_loader_ptrace_flag = 0;
#if GDB_INTERFACE
/* support for debugger */
static char m_libname[PATH_MAX];
struct link_map g_maps = {
	.l_name = m_libname,
	.l_next = NULL,
	.l_prev = NULL,
};
struct r_debug _r_debug = { /* this symbol name should not be changed */
	.r_version = 1,
	.r_map = &g_maps,
	.r_brk = (Elf64_Addr)&_dl_debug_state,
	.r_state = RT_CONSISTENT,
	.r_ldbase = 0x4000000000000000, /* load address of this loader */
};
#endif

/* place holder for the clst pointer */
__thread void * t_clstp  __attribute__ ((tls_model ("initial-exec"))) = {0L};
__thread unsigned long t_ptrace __attribute__ ((tls_model ("initial-exec"))) = {0L};

/* local funcs prototypes */
static int open_btlib (char ** envp);
static unsigned long * get_auxvec (int *, char **, char **);
static int get_execfd (unsigned long * auxvec);
static unsigned long * set_auxvec (unsigned long * auxvec, int bin32, unsigned long *);
#if SUID_LOADER
static long compute_creds (int bin32);
#else
static int invoke_suid_loader_if_needed (int bin32, char ** argv, char ** envp);
#endif


int main (int argc, char ** argv, char ** envp)
{
	long ret = -1;
	int btlib = 0, bin32 = 0;
	unsigned long * auxvec;
	unsigned long auxvec_size = 0;
	long clstp_offset, ptrace_offset;
	int free_auxvec = 0;
	int is_ver = 0;

	if (argc < 2) {
		DBPRINT (1, "Usage: %s <full path to app> [arguments]\n", argv[0]);
		return -1;
	}

	auxvec = get_auxvec (&argc, argv, envp);
	if (auxvec == NULL) {
		DBPRINT (1, "cannot find aux vector\n");
		goto cleanup;
	}

	if (argc == 2 && argv[1][0] == '-' && argv[1][1] == 'v' && argv[1][2] == '\0') {
		is_ver = 1;
	}

	if ( ! is_ver ) {
		bin32 = get_execfd (auxvec);
		if (bin32 < 0) {
#if ! SUID_LOADER
			bin32 = SYSCALL (open, (argv[1], O_RDONLY, 0));
			if (bin32 < 0) {
				DBPRINT (1, "open_binary failed\n");
				goto cleanup;
			}
			ret = invoke_suid_loader_if_needed (bin32, argv, envp);
			if (ret < 0) {
				DBPRINT (1, "invoke_suid_loader_if_needed failed\n");
				goto cleanup;
			}
#else // SUID_LOADER
			char new_argv1[PATH_MAX] = { 0 };
			char proc_fd_name[64];

			bin32 = bt_atol (argv[1]);
			ret = bt_snprintf (proc_fd_name, sizeof (proc_fd_name) - 1,
					"/proc/self/fd/%d", bin32);
			assert (ret > 0 && (unsigned long)ret < sizeof (proc_fd_name));
			assert (sizeof (new_argv1) < INT_MAX);
			ret = SYSCALL (readlink, (proc_fd_name, new_argv1, 
						(int)sizeof (new_argv1)));
			if (IS_SYS_ERROR (ret)) {
				DBPRINT (1, "readlink (%s) failed: %ld\n",
						proc_fd_name, ret);
			} else {
				argv[1] = new_argv1;
			}
			ret = compute_creds (bin32);
			if (ret < 0) {
				DBPRINT (1, "compute_creds failed\n");
				goto cleanup;
			}
#endif

			auxvec = set_auxvec (auxvec, bin32, &auxvec_size);
			if (auxvec == NULL) {
				DBPRINT (1, "set_auxvec failed\n");
				goto cleanup;
			}
		       	free_auxvec = 1;
		}
		/* check whether we're allowed to exec this binary */
		ret = check_exec_permissions_by_fd (bin32);
		if (IS_SYS_ERROR (ret)) {
			return ret;
		}
	} else { /* is_ver */
		fprintf (stdout, "IA32X Loader:\t\tversion=%s.%s\n", BUILD_NO, MODE);
	}

	g_pagesize = SYSCALL (getpagesize, ());
	DBPRINT (0, "pagesize=%lx\n", g_pagesize);
	g_log_file = 2;

	/* open btlib */
	btlib = open_btlib (envp);
	if (btlib < 0) {
		DBPRINT (1, "open_btlib failed:%d\n", btlib);
		ret = -1;
		goto cleanup;
	}

	/* load btlib */
	ret = load_btlib (btlib);
	if (ret) {
		DBPRINT (1, "load_btlib failed: %ld\n", ret);
		goto cleanup;
	}

	DBPRINT (0, "m_lib_init=%lx.%lx m_btlib_load_and_exec=%lx.%lx\n",
			m_lib_init.fdesc->ip,
			m_lib_init.fdesc->gp,
			m_btlib_load_and_exec.fdesc->ip,
			m_btlib_load_and_exec.fdesc->gp);
	/* we don't need this file open anymore */
	SYSCALL (close, (btlib));
#if GDB_INTERFACE
	/* notify the debugger about load complete */
	_dl_debug_state ();
#endif
	/* init btlib */

	clstp_offset = (void *)&t_clstp - __r13;
	ptrace_offset = (void *)&t_ptrace - __r13;

	DBPRINT (0, "auxvec=%p clstp_offset=%lx, ptrace_offset=%lx\n", 
		 auxvec, clstp_offset, ptrace_offset);
	ret = (*m_lib_init.fptr) (auxvec, clstp_offset, ptrace_offset);

	if (ret) {
		DBPRINT (1, "lib_init failed: %ld\n", ret);
		goto cleanup;
	}

	g_pagesize = SYSCALL (getpagesize, ());
	/* run app */
	ret = (*m_btlib_load_and_exec.fptr) (m_loader_ptrace_flag, argc, argv, envp);
	DBPRINT (ret, "btlib_load_and_exec failed:%d\n", ret);

	g_pagesize = SYSCALL (getpagesize, ());
cleanup:

	if (free_auxvec)
		SYSCALL (munmap, ((unsigned long)auxvec, auxvec_size));
	if (bin32)
		SYSCALL (close, (bin32));
	if (btlib)
		SYSCALL (close, (btlib));

	return ret;
}

static unsigned long * get_auxvec (int * argc_addr, char ** argv, char ** envp)
{
	char ** p;
	unsigned long * elf_info = NULL;

	/* skip the vector of pointers to env strings */
	if (envp != NULL) {
		for (p = envp; *p != NULL; p ++);
		elf_info = (void*)++p;
	} else if (argv != NULL) {
		/* skip the vector of pointers to args */
		for (p = argv; *p != NULL; p ++);
		elf_info = (void*)++p;
	} else {
		/* there might be alignment issue here - 
		 * it's not performance critical, so let the kernel handle it */
		elf_info = (unsigned long*)(void*)(argc_addr + 1);
		DBPRINT (ROUNDUP ((unsigned long)elf_info, 
				  __alignof(elf_info)) != (unsigned long)elf_info,
				  "elf_info is not well aligned\n");
	}

	return elf_info;
}

/* information for locating BTGeneric */
#define HOME		"HOME="
#define BTLIB_NAME	"libia32x.so"
static char * m_search_path[] = {
	"/usr/local/bin",
	"/usr/lib",
	"/lib",
	"/usr/lib/ia32el",
	""
};


/*
* tranist from unconfined_t to unconfined_execmem_t
* to let btlib alloc WX memory under selinux
*/
static int selinux_transition(char* libpath){
        security_context_t psc, fsc, newpsc;
        int ret;
	/* get the current security context context of the process*/
        ret = bt_getcon( &psc );
        if (ret<0)
                return -1;
	/*get the security context of btlib executable*/
        ret = bt_getfilecon( libpath, &fsc );
        if (ret<0){
        	bt_freecon( psc );
                return -1;
	}
	/* get the new security context of the process*/
        ret = bt_security_compute_create( psc, fsc, SECCLASS_PROCESS, &newpsc);
        if (ret<0){
		bt_freecon( psc );
        	bt_freecon( fsc );
                return -1;
	}
	/* set the contex to process*/
        ret = bt_setcon( newpsc );
        if (ret<0){
        	bt_freecon( psc );
        	bt_freecon( fsc );
        	bt_freecon( newpsc );
                return -1;
	}
        bt_freecon( psc );
        bt_freecon( fsc );
        bt_freecon( newpsc );
        return 0;
}



static int open_btlib (char ** envp UNUSED)
{
	int fd = -1;
	char filename[PATH_MAX];
#if ! SYSBTG
	u64 home_len = bt_strnlen (HOME, PATH_MAX);
	u64 ret;
	char ** ep = envp;

	filename[PATH_MAX - 1] = '\0';
	if (ep != NULL) {
		/* Go through the environment strings until we find "HOME=" */
		while (*ep) {
			if (bt_strncmp (*ep, HOME, home_len) == 0) {
				break;
			} else {
				ep++;
			}
		}

		/* If *ep is not NULL, then we found HOME - 
		 * create btlib's path */
		if (*ep) {
			ret = bt_snprintf (filename,
					   sizeof (filename) - 1,
					   "%s/" BTLIB_NAME,
					   *ep + home_len);
			fd = SYSCALL (open, (filename, O_RDONLY, 0));
			if ( ! IS_SYS_ERROR (fd) ) {
#if GDB_INTERFACE
				bt_strncpy (m_libname, filename, sizeof (m_libname) -1);
#endif
			}
		}
	}
#endif
	if (IS_SYS_ERROR (fd)) {
		int j;

		for (j = 0; *m_search_path[j]; j++) {
			bt_snprintf (filename, 
				     sizeof (filename) - 1, "%s/" BTLIB_NAME, m_search_path[j]);
			fd = SYSCALL (open, (filename, O_RDONLY, 0));
			if ( ! IS_SYS_ERROR (fd) ) {
#if GDB_INTERFACE
				bt_strncpy (m_libname, filename, sizeof (m_libname) -1);
#endif
				break;
			}
		}
	}

	/* selinux transition, if failed,
         * just fall through without  changing security context	
         */
        selinux_transition( filename );
	return fd;
}


/*****************************************************************************/
/* return 1 if execfd is found in the auxvec, 0 otherwise */
/*****************************************************************************/
static int get_execfd (unsigned long * auxvec)
{
	int bin32 = -1;
	while (*(auxvec) != AT_NULL) {
		unsigned long id, val;
		id = *auxvec;
		val = *(auxvec + 1);
		if (id == AT_EXECFD) {
			bin32 = val;
			break;
		}
		auxvec += 2;
	}
	return bin32;
}
/*****************************************************************************/
/* create a new auxvec that includes the given fd in a new AT_EXECFD entry */
/*****************************************************************************/
static unsigned long *
set_auxvec (unsigned long * auxvec, int bin32, unsigned long * size)
{
	long vec_size = 0; /* number of unsigned longs in the vector */
	unsigned long * p;
	unsigned long * new_auxvec;

	p = auxvec;
	while (*(p) != AT_NULL) {
		p += 2;
	}
	vec_size = p - auxvec;
	*size = (vec_size + 4) * sizeof (unsigned long);

	new_auxvec = (void *)SYSCALL (mmap, (0, *size, PROT_READ | PROT_WRITE,
				      MAP_ANON | MAP_PRIVATE, -1, 0));
	if (IS_SYS_ERROR(new_auxvec)) {
		DBPRINT (1, "mmap failed: %ld\n", (long)new_auxvec);
		new_auxvec = NULL;
		goto out;
	}
	DBPRINT (0, "addr=%p size=%lx\n", new_auxvec, *size);

	p = new_auxvec;
	bt_memcpy (p, auxvec, vec_size * sizeof (unsigned long));
	p += vec_size;
	*p++= AT_EXECFD; /* execfd entry */
	*p++ = bin32;
	*p++ = 0; /* terminating NULL entry */
	*p = 0;

#if SUID_LOADER
	{
		/* need to set the effective id's too */
		unsigned int euid = 0, egid = 0;
		p = new_auxvec;

		/* these syscalls are guaranteed not to fail */
		euid = SYSCALL (geteuid, ());
		egid = SYSCALL (getegid, ());

		while (*p != 0ul) {
			if (*p == AT_EUID) {
				*(p + 1) = euid;
			} else if (*p == AT_EGID) {
				*(p + 1) = egid;
			}
			p += 2;
		}
	}
#endif // SUID_LOADER


out:
	return new_auxvec;
}


#if ! SUID_LOADER
/*****************************************************************************/
/* if the binary is loaded then switch to suid_btlib */
/* input:  bin32 - the fd of the binary to run
 *         argv, envp - the arguments that btlib's main() got from the system
 * output: negative value in case of error. 0 if no need to change to 
 *         suid_btlib. does not return if suid_btlib is invoked */
/*****************************************************************************/
static int invoke_suid_loader_if_needed (int bin32, char ** argv, char ** envp)
{
	int is_setuid, is_setgid;
	struct stat buf;
	int ret;

	DBPRINT (0, "bin32=%d argv[0]='%s' argv[1]='%s'\n",
		 bin32, argv[0], argv[1]);

	/* get the setuid/setgid bits of the file */
	ret = SYSCALL (fstat, (bin32, &buf));
	if (IS_SYS_ERROR (ret)) {
		DBPRINT (1, "stat failed:%d\n", ret);
		goto out;
	}
	is_setuid = buf.st_mode & S_ISUID;
	is_setgid = buf.st_mode & S_ISGID;

	DBPRINT (0, "is_setuid=%d is_setgid=%d\n", is_setuid, is_setgid);

	/* if it's setuid then invoke suid_loader */
	if (is_setuid || is_setgid) {
		int flags;
		char ** search_path = m_search_path;
		char buff[PATH_MAX];
		char argv1[32];

		/* clear the close-on-exec flag */
		flags = SYSCALL (fcntl, (bin32, F_GETFD, 0ul));
		if (IS_SYS_ERROR (flags)) {
			DBPRINT (1, "getfd failed: %d\n", ret);
			goto out;
		}
		flags &= ~FD_CLOEXEC;
		ret = SYSCALL (fcntl, (bin32, F_SETFD, (unsigned long)flags));
		if (IS_SYS_ERROR (ret)) {
			DBPRINT (1, "fcntl failed: %d", ret);
			goto out;
		}
		buff [PATH_MAX - 1] = 0;
		/* search the suid-btlib, then invoke it */
		while (bt_strnlen(*search_path, PATH_MAX) != 0) {
			bt_snprintf (buff, PATH_MAX - 1, "%s/" SUID_LOADER_NAME, *search_path);
			ret = SYSCALL (stat, (buff, &buf));
			if (ret == 0){
				break;
			} else {
				ret = -1;
				search_path++;
			}	
		}
		if (ret == -1){
			DBPRINT (1, "%s could not be found.\n", SUID_LOADER_NAME);
			goto out;
		}
		/* replace file_name with the fd */
		ret = bt_snprintf (argv1, sizeof (argv1) - 1, "%d", bin32);
		assert (ret > 0 && (unsigned long)ret < sizeof (argv1));
		argv[1] = argv1;
		/* invoke suid_btlib */
		DBPRINT (0, "buff='%s' argv[1]='%s'\n", buff, argv[1]);
		ret = SYSCALL (execve, (buff, argv, envp));
		DBPRINT (1, "execve failed: %d\n", ret);
		goto out;
	}

	ret = 0;
out:
	return ret;
}
#else // SUID_LOADER
/*****************************************************************************/
/* set the crdentials for the process.
 * input:  bin32 - the file desc of the binary
 * output: 0 on success, a negative error value on failure
 *****************************************************************************/
static long compute_creds (int bin32)
{
	struct stat buf;
	long ret;
	unsigned ruid, euid, suid;
	unsigned rgid, egid, sgid;
	unsigned new_ruid, new_euid, new_suid;
	unsigned new_rgid, new_egid, new_sgid;
	int is_setuid, is_setgid;

	DBPRINT (0, "bin32=%d\n", bin32);

	/* get the user's details */
	ret = SYSCALL (getresuid, (&ruid, &euid, &suid));
	if (IS_SYS_ERROR (ret)) {
		DBPRINT (1, "getresuid failed:%ld\n", ret);
		return ret;
	}
	ret = SYSCALL (getresgid, (&rgid, &egid, &sgid));
	if (IS_SYS_ERROR (ret)) {
		DBPRINT (1, "getresgid failed:%ld\n", ret);
		return ret;
	}
	DBPRINT (0, "ruid=%d euid=%d suid=%d rgid=%d egid=%d sgid=%d\n",
		 ruid, euid, suid, rgid, egid, sgid);

 	/* get the details of the file */
	ret = SYSCALL (fstat, (bin32, &buf));
	if (IS_SYS_ERROR (ret)) {
		DBPRINT (1, "stat failed:%ld bin32=%d\n", ret, bin32);
		return ret;
	}

 	/* get the s bits */
 	is_setuid = buf.st_mode & S_ISUID;
 	is_setgid = ((buf.st_mode & (S_ISGID | S_IXGRP)) == (S_ISGID | S_IXGRP));

	/* we should get to suid-btlib only if the binary is setuid */
	assert (is_setuid || is_setgid);

	/* real remains the same */
	new_ruid = ruid;
	new_rgid = rgid;

	/* effective uid/gid depends on the set?id bit of the binary */
	new_euid = is_setuid ? buf.st_uid : ruid;
	new_egid = is_setgid ? buf.st_gid : rgid;

	/* forget the original root permissions of btlib */
	new_suid = new_euid;
	new_sgid = new_egid;

	DBPRINT (0, "new_ruid=%d new_euid=%d new_suid=%d "
		    "new_rgid=%d new_egid=%d new_sgid=%d\n",
		 new_ruid, new_euid, new_suid, new_rgid, new_egid, new_sgid);

	/* set the credentials */
	ret = SYSCALL (setresgid, (new_rgid, new_egid, new_sgid));
	if (IS_SYS_ERROR (ret)) {
		DBPRINT (1, "setresgid failed:%ld\n", ret);
		return ret;
	}

	ret = SYSCALL (setresuid, (new_ruid, new_euid, new_suid));
	if (IS_SYS_ERROR (ret)) {
		DBPRINT (1, "setresuid failed:%ld\n", ret);
		return ret;
	}

	/* make sure we have the right credentials */
	ret = SYSCALL (getresuid, (&ruid, &euid, &suid));
	if (IS_SYS_ERROR (ret)) {
		DBPRINT (1, "getresuid failed:%ld\n", ret);
		return ret;
	}
	ret = SYSCALL (getresgid, (&rgid, &egid, &sgid));
	if (IS_SYS_ERROR (ret)) {
		DBPRINT (1, "getresgid failed:%ld\n", ret);
		return ret;
	}

	if ((ruid != new_ruid) ||
	    (euid != new_euid) ||
	    (suid != new_suid) ||
	    (rgid != new_rgid) ||
	    (egid != new_egid) ||
	    (sgid != new_sgid)) {
		DBPRINT (1, "compute_creds failed\n");
		return -1;
	}

	return 0;
}
#endif


void * __tls_get_addr (u64 m UNUSED, u64 offset)
{
	void * addr;

	addr = __r13 + g_tls_offset + offset;

	return addr;
}

void __gmon_start__ (void)
{
	return;
}

#if GDB_INTERFACE
/* This function exists solely to have a breakpoint set on it by the
 * debugger.  The debugger is supposed to find this function's address by
 * examining the r_brk member of struct r_debug, but GDB 4.15 in fact looks
 * for this particular symbol name in the PT_INTERP file.  */
void _dl_debug_state (void) {}
#endif

u64 * get_loader_ptrace_flag_addr () {
	return &m_loader_ptrace_flag;
}


/* this is really ugly - but otherwise all static
   initializations in btlib need to go */
void * memset(void *s, int c, unsigned long n)
{
	return bt_memset (s, c, n);
}
void * memcpy(void *dest, const void *src, unsigned long n)
{
	return bt_memcpy (dest, src, n);
}

