/* 
 * loader.c
 * minimal elf loader.
 *
 */
 /*
 *
 * This file has been derived from GNU libc. Please see COPYING.LIBC
 * for more information
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include <ia32x_utils.h>

#include "ia32x_loader.h"
#include "dl-hash.h"
#include "utils.h"
#include "loader_tls.h"

#define FPTR_TABLE_SIZE 500
#define TEXT_SEGMENT_BASE_ADDR 0x500000000
#define BTLIB_TD_SIZE 16

/* extern vars */
Elf64_Dyn * _DYNAMIC = NULL;
Elf64_Addr g_load_addr = 0;
Elf64_Off g_tls_offset = 0;

/* local vars */
static int m_tls_allocated = 0;
static Elf64_Rel * m_rel_ptr = NULL;
static Elf64_Word m_rel_size = 0l;
static Elf64_Rela * m_rela_ptr = NULL;
static Elf64_Word m_rela_size = 0l;
static char * m_dyn_sym_strtab = NULL;
static Elf64_Sym * m_dyn_symtab = NULL;
static Elf64_Phdr * m_dyn_phdr = NULL;
static Elf64_Phdr * m_tls_phdr = NULL;
static struct _fdesc_tag m_fdesc_table[FPTR_TABLE_SIZE] = {{0, 0}};
static Elf64_Addr m_gp = 0;
static struct loadcmd {
	Elf64_Addr mapstart, mapend, dataend, allocend;
	Elf64_Off mapoff;
	int prot;
} m_loadcmds[2];
static long m_nloadcmds = 0;
static int m_has_holes = 0;
static Elf64_Addr m_map_start = 0, m_map_end = 0, m_text_end = 0;
static Elf64_Rela * m_plt = NULL;
static Elf64_Xword m_plt_size = 0;

/* a struct for hash table in dynamic section */
static struct hash_info {
#define SYSV_HASH 1
#define GNU_HASH 2
        int hash_type;

        Elf64_Word nbuckets;
        Elf64_Word gnu_bitmask_idxbits;
        Elf64_Word gnu_shift;
        const Elf64_Word *gnu_bitmask;
        union
        {
         const Elf64_Word *gnu_buckets;
         const Elf64_Word *buckets;
        };
        union
	{
         const Elf64_Word *gnu_chain_zero;
         const Elf64_Word *chain;
        };
} m_hash_info;




/* local funcs */
static inline struct _fdesc_tag * set_official_fptr (u64 ip, u64 gp, Elf64_Sym * sym);
static Elf64_Sym * lookup_symbol (char * name);
static int lookup_btlib_entry_points (void);
static int setup_hash (Elf64_Dyn * dh);
static int setup_gnu_hash (Elf64_Dyn * dh);
static int set_tls (Elf64_Phdr * tlshdr);
static Elf64_Addr do_fptr64lsb_reloc (Elf64_Rela * rela, Elf64_Addr * reloc_addr, Elf64_Sym * sym);
static Elf64_Addr do_rel64lsb_reloc (Elf64_Rela * rela, Elf64_Addr * reloc_addr);
static Elf64_Addr do_tprel64lsb_reloc (Elf64_Rela * rela, Elf64_Addr * reloc_addr, Elf64_Sym * sym);
static Elf64_Addr do_dtpmod64lsb_reloc (Elf64_Rela * rela, Elf64_Addr * reloc_addr);
static int read_reloc (void);
static int read_plt (void);
static int check_elf_header (Elf64_Ehdr * hdr);
static int load_program_headers (int fd, Elf64_Ehdr * elfhdr, Elf64_Phdr ** phdrp);
static int read_prog_hdrs (Elf64_Phdr * phdr, Elf64_Half phnum);
static int read_prog_entry (Elf64_Phdr * phdr);
static int read_dyn_section (Elf64_Phdr * phdr);
static int load_segments (int fd);

/* debug funcs and string tables */
#if DEBUG
static char * m_sym_strtab = NULL;
static char * m_sh_strtab = NULL;
static Elf64_Word m_dyn_symtab_entry_num = 0;

static int load_section_headers (int fd, Elf64_Ehdr * elfhdr, Elf64_Shdr ** shdrp);
static int load_strtab (int fd, Elf64_Shdr * shdr, long indx, char **strtabp);
static int read_section_hdrs (int fd, Elf64_Shdr * shdr, Elf64_Half shnum) UNUSED;
static void print_symtab (int fd, Elf64_Shdr * shdr) UNUSED;
static void print_dyn_section (void) UNUSED;
static void print_symbol (Elf64_Sym * sym, char * strtab) UNUSED;
static void print_section_entry (Elf64_Shdr * shdr) UNUSED;

Elf64_Sym * m_fdesc_table_syms[FPTR_TABLE_SIZE];
static inline char * get_reloc_type_str (Elf64_Word type)
{
	char * str = NULL;
	if (type == R_IA64_NONE)
		str = "R_IA64_NONE";
	if (type == R_IA64_IMM14)
		str = "R_IA64_IMM14";
	if (type == R_IA64_IMM22)
		str = "R_IA64_IMM22";
	if (type == R_IA64_IMM64)
		str = "R_IA64_IMM64";
	if (type == R_IA64_DIR32MSB)
		str = "R_IA64_DIR32MSB";
	if (type == R_IA64_DIR32LSB)
		str = "R_IA64_DIR32LSB";
	if (type == R_IA64_DIR64MSB)
		str = "R_IA64_DIR64MSB";
	if (type == R_IA64_DIR64LSB)
		str = "R_IA64_DIR64LSB";
	if (type == R_IA64_GPREL22)
		str = "R_IA64_GPREL22";
	if (type == R_IA64_GPREL64I)
		str = "R_IA64_GPREL64I";
	if (type == R_IA64_GPREL32MSB)
		str = "R_IA64_GPREL32MSB";
	if (type == R_IA64_GPREL32LSB)
		str = "R_IA64_GPREL32LSB";
	if (type == R_IA64_GPREL64MSB)
		str = "R_IA64_GPREL64MSB";
	if (type == R_IA64_GPREL64LSB)
		str = "R_IA64_GPREL64LSB";
	if (type == R_IA64_LTOFF22)
		str = "R_IA64_LTOFF22";
	if (type == R_IA64_LTOFF64I)
		str = "R_IA64_LTOFF64I";
	if (type == R_IA64_PLTOFF22)
		str = "R_IA64_PLTOFF22";
	if (type == R_IA64_PLTOFF64I)
		str = "R_IA64_PLTOFF64I";
	if (type == R_IA64_PLTOFF64MSB)
		str = "R_IA64_PLTOFF64MSB";
	if (type == R_IA64_PLTOFF64LSB)
		str = "R_IA64_PLTOFF64LSB";
	if (type == R_IA64_FPTR64I)
		str = "R_IA64_FPTR64I";
	if (type == R_IA64_FPTR32MSB)
		str = "R_IA64_FPTR32MSB";
	if (type == R_IA64_FPTR32LSB)
		str = "R_IA64_FPTR32LSB";
	if (type == R_IA64_FPTR64MSB)
		str = "R_IA64_FPTR64MSB";
	if (type == R_IA64_FPTR64LSB)
		str = "R_IA64_FPTR64LSB";
	if (type == R_IA64_PCREL60B)
		str = "R_IA64_PCREL60B";
	if (type == R_IA64_PCREL21B)
		str = "R_IA64_PCREL21B";
	if (type == R_IA64_PCREL21M)
		str = "R_IA64_PCREL21M";
	if (type == R_IA64_PCREL21F)
		str = "R_IA64_PCREL21F";
	if (type == R_IA64_PCREL32MSB)
		str = "R_IA64_PCREL32MSB";
	if (type == R_IA64_PCREL32LSB)
		str = "R_IA64_PCREL32LSB";
	if (type == R_IA64_PCREL64MSB)
		str = "R_IA64_PCREL64MSB";
	if (type == R_IA64_PCREL64LSB)
		str = "R_IA64_PCREL64LSB";
	if (type == R_IA64_LTOFF_FPTR22)
		str = "R_IA64_LTOFF_FPTR22";
	if (type == R_IA64_LTOFF_FPTR64I)
		str = "R_IA64_LTOFF_FPTR64I";
	if (type == R_IA64_LTOFF_FPTR32MSB)
		str = "R_IA64_LTOFF_FPTR32MSB";
	if (type == R_IA64_LTOFF_FPTR32LSB)
		str = "R_IA64_LTOFF_FPTR32LSB";
	if (type == R_IA64_LTOFF_FPTR64MSB)
		str = "R_IA64_LTOFF_FPTR64MSB";
	if (type == R_IA64_LTOFF_FPTR64LSB)
		str = "R_IA64_LTOFF_FPTR64LSB";
	if (type == R_IA64_SEGREL32MSB)
		str = "R_IA64_SEGREL32MSB";
	if (type == R_IA64_SEGREL32LSB)
		str = "R_IA64_SEGREL32LSB";
	if (type == R_IA64_SEGREL64MSB)
		str = "R_IA64_SEGREL64MSB";
	if (type == R_IA64_SEGREL64LSB)
		str = "R_IA64_SEGREL64LSB";
	if (type == R_IA64_SECREL32MSB)
		str = "R_IA64_SECREL32MSB";
	if (type == R_IA64_SECREL32LSB)
		str = "R_IA64_SECREL32LSB";
	if (type == R_IA64_SECREL64MSB)
		str = "R_IA64_SECREL64MSB";
	if (type == R_IA64_SECREL64LSB)
		str = "R_IA64_SECREL64LSB";
	if (type == R_IA64_REL32MSB)
		str = "R_IA64_REL32MSB";
	if (type == R_IA64_REL32LSB)
		str = "R_IA64_REL32LSB";
	if (type == R_IA64_REL64MSB)
		str = "R_IA64_REL64MSB";
	if (type == R_IA64_REL64LSB)
		str = "R_IA64_REL64LSB";
	if (type == R_IA64_LTV32MSB)
		str = "R_IA64_LTV32MSB";
	if (type == R_IA64_LTV32LSB)
		str = "R_IA64_LTV32LSB";
	if (type == R_IA64_LTV64MSB)
		str = "R_IA64_LTV64MSB";
	if (type == R_IA64_LTV64LSB)
		str = "R_IA64_LTV64LSB";
	if (type == R_IA64_PCREL21BI)
		str = "R_IA64_PCREL21BI";
	if (type == R_IA64_PCREL22)
		str = "R_IA64_PCREL22";
	if (type == R_IA64_PCREL64I)
		str = "R_IA64_PCREL64I";
	if (type == R_IA64_IPLTMSB)
		str = "R_IA64_IPLTMSB";
	if (type == R_IA64_IPLTLSB)
		str = "R_IA64_IPLTLSB";
	if (type == R_IA64_COPY)
		str = "R_IA64_COPY";
	if (type == R_IA64_SUB)
		str = "R_IA64_SUB";
	if (type == R_IA64_LTOFF22X)
		str = "R_IA64_LTOFF22X";
	if (type == R_IA64_LDXMOV)
		str = "R_IA64_LDXMOV";
	if (type == R_IA64_TPREL14)
		str = "R_IA64_TPREL14";
	if (type == R_IA64_TPREL22)
		str = "R_IA64_TPREL22";
	if (type == R_IA64_TPREL64I)
		str = "R_IA64_TPREL64I";
	if (type == R_IA64_TPREL64MSB)
		str = "R_IA64_TPREL64MSB";
	if (type == R_IA64_TPREL64LSB)
		str = "R_IA64_TPREL64LSB";
	if (type == R_IA64_LTOFF_TPREL22)
		str = "R_IA64_LTOFF_TPREL22";
	if (type == R_IA64_DTPMOD64MSB)
		str = "R_IA64_DTPMOD64MSB";
	if (type == R_IA64_DTPMOD64LSB)
		str = "R_IA64_DTPMOD64LSB";
	if (type == R_IA64_LTOFF_DTPMOD22)
		str = "R_IA64_LTOFF_DTPMOD22";
	if (type == R_IA64_DTPREL14)
		str = "R_IA64_DTPREL14";
	if (type == R_IA64_DTPREL22)
		str = "R_IA64_DTPREL22";
	if (type == R_IA64_DTPREL64I)
		str = "R_IA64_DTPREL64I";
	if (type == R_IA64_DTPREL32MSB)
		str = "R_IA64_DTPREL32MSB";
	if (type == R_IA64_DTPREL32LSB)
		str = "R_IA64_DTPREL32LSB";
	if (type == R_IA64_DTPREL64MSB)
		str = "R_IA64_DTPREL64MSB";
	if (type == R_IA64_DTPREL64LSB)
		str = "R_IA64_DTPREL64LSB";
	if (type == R_IA64_LTOFF_DTPREL22)
		str = "R_IA64_LTOFF_DTPREL22";

	return str;
}
char * d_tag_str[] = {
"DT_NULL",
"DT_NEEDED",
"DT_PLTRELSZ",
"DT_PLTGOT",
"DT_HASH",
"DT_STRTAB",
"DT_SYMTAB",
"DT_RELA",
"DT_RELASZ",
"DT_RELAENT",
"DT_STRSZ",
"DT_SYMENT",
"DT_INIT",
"DT_FINI",
"DT_SONAME",
"DT_RPATH",
"DT_SYMBOLIC",
"DT_REL",
"DT_RELSZ",
"DT_RELENT",
"DT_PLTREL",
"DT_DEBUG",
"DT_TEXTREL",
"DT_JMPREL",
"DT_BIND_NOW",
"DT_INIT_ARRAY",
"DT_FINI_ARRAY",
"DT_INIT_ARRAYSZ",
"DT_FINI_ARRAYSZ",
"DT_RUNPATH",
"DT_FLAGS",
"DT_ENCODING",
"DT_PREINIT_ARRAY",
"DT_PREINIT_ARRAYSZ",
"DT_NUM"
};

#define DT_TAG_HI_IND DT_VERSYM
char * d_tag_hi_str[] = {
"DT_VERSYM",
"",
"",
"",
"",
"",
"",
"",
"",
"DT_RELACOUNT",
"DT_RELCOUNT",
"DT_FLAGS_1",
"DT_VERDEF",
"DT_VERDEFNUM",
"DT_VERNEED",
"DT_VERNEEDNUM",
"DT_IA_64_PLT_RESERVE"
};

char *e_type_str[] = {
"ET_NONE",
"ET_REL",
"ET_EXEC",
"ET_DYN",
"ET_CORE",
"ET_NUM"};
char * e_osabi_str[] = {
"ELFOSABI_SYSV",
"ELFOSABI_HPUX",
"ELFOSABI_NETBSD",
"ELFOSABI_LINUX",
"ELFOSABI_SOLARIS",
"ELFOSABI_AIX",
"ELFOSABI_IRIX",
"ELFOSABI_FREEBSD",
"ELFOSABI_TRU64",
"ELFOSABI_MODESTO",
"ELFOSABI_OPENBSD"
};

char * p_type_str[] = {
"PT_NULL",
"PT_LOAD",
"PT_DYNAMIC",
"PT_INTERP",
"PT_NOTE",
"PT_SHLIB",
"PT_PHDR",
"PT_TLS",
"PT_NUM"
};

char * sh_type_str[] = {
"SHT_NULL",
"SHT_PROGBITS",
"SHT_SYMTAB",
"SHT_STRTAB",
"SHT_RELA",
"SHT_HASH",
"SHT_DYNAMIC",
"SHT_NOTE",
"SHT_NOBITS",
"SHT_REL",
"SHT_SHLIB",
"SHT_DYNSYM",
"SHT_INIT_ARRAY",
"SHT_FINI_ARRAY",
"SHT_PREINIT_ARRAY",
"SHT_GROUP",
"SHT_SYMTAB_SHNDX"
};
char * shf_str [] = {
"NONE",
"SHF_WRITE",
"SHF_ALLOC",
"SHF_EXECINSTR",
"SHF_MERGE",
"SHF_STRINGS",
"SHF_INFO_LINK",
"SHF_LINK_ORDER",
"SHF_OS_NONCONFORMING",
"SHF_GROUP",
"SHF_TLS"
};



#endif




int load_btlib (int fd)
{
	int error = 0;
	long ret;
	Elf64_Ehdr elfhdr; 
	Elf64_Phdr * phdr = NULL;
#if DEBUG
	Elf64_Shdr * shdr = NULL;
#endif


	/* read the elf header */
	ret = SYSCALL (pread, (fd, &elfhdr, sizeof (elfhdr), 0););
	if (IS_SYS_ERROR (ret)) {
		error = 20;
		goto cleanup;
	}
	assert (ret == sizeof (elfhdr));


	/* check header's validity */
	ret = check_elf_header (&elfhdr);
	if (ret) {
		error = 30;
		goto cleanup;
	}

	/* load program headers */
	ret = load_program_headers (fd, &elfhdr, &phdr);
	if (ret) {
		error = 40;
		goto cleanup;
	}
#if DEBUG
	ret = load_section_headers (fd, &elfhdr, &shdr);
	if (ret) {
		error = 45;
		goto cleanup;
	}
	/* load the section-name string table */
	ret = load_strtab (fd, shdr, elfhdr.e_shstrndx, &m_sh_strtab);
	if (ret) {
		error = 50;
		goto cleanup;
	}

	/* read the section headers */
	ret = read_section_hdrs (fd, shdr, elfhdr.e_shnum);
	if (ret) {
		error = 60;
		goto cleanup;
	}
#endif


	/* read the program headers */
	ret = read_prog_hdrs (phdr, elfhdr.e_phnum);
	if (ret) {
		error = 70;
		goto cleanup;
	}

	ret = load_segments (fd);
	if (ret) {
		error = 80;
		goto cleanup;
	}

	ret = read_dyn_section (m_dyn_phdr);
	if (ret) {
		error = 90;
		goto cleanup;
	}

#if DEBUG
	print_dyn_section ();
#endif

	ret = set_tls (m_tls_phdr);
	if (ret) {
		error = 95;
		goto cleanup;
	}


	ret = read_reloc ();
	if (ret) {
		error = 100;
		goto cleanup;
	}

	ret = lookup_btlib_entry_points ();
	if (ret) {
		error = 110;
		goto cleanup;
	}

	ret = read_plt ();
	if (ret) {
		error = 120;
		goto cleanup;
	}



cleanup:
#if DEBUG
	if (shdr)
		zfree (shdr);
	if (m_sh_strtab)
		zfree (m_sh_strtab);
	if (m_sym_strtab)
		zfree (m_sym_strtab);
#endif
	DBPRINT (0, "error=%d\n", error);
	return error;
}

/*******************************************************************************
 * sanity checks
 ******************************************************************************/
static int check_elf_header (Elf64_Ehdr * hdr)
{
	int error = 0;


	/* e_ident */
	DBPRINT (0, "e_ident[MAG]=\\%o%c%c%c\n", hdr->e_ident[EI_MAG0],
		hdr->e_ident[EI_MAG1],
		hdr->e_ident[EI_MAG2], hdr->e_ident[EI_MAG3]);
	DBPRINT (0, "e_ident[CLASS]=%d<%s>\n", hdr->e_ident[EI_CLASS],
			(hdr->e_ident[EI_CLASS] == ELFCLASSNONE) ? "ELFCLASSNONE" :
			(hdr->e_ident[EI_CLASS] == ELFCLASS32) ? "ELFCLASS32" :
			(hdr->e_ident[EI_CLASS] == ELFCLASS64) ? "ELFCLASS64" :
			"unknown");
	DBPRINT (0, "e_ident[DATA]=%d<%s>\n", hdr->e_ident[EI_DATA],
			(hdr->e_ident[EI_DATA] == ELFDATANONE) ? "ELFDATANONE" :
			(hdr->e_ident[EI_DATA] == ELFDATA2LSB) ? "ELFDATA2LSB" :
			(hdr->e_ident[EI_DATA] == ELFDATA2MSB) ? "ELFDATA2MSB" :
			"unknown");
	DBPRINT (0, "e_ident[VERSION]=%d<%s>\n", hdr->e_ident[EI_VERSION],
			(hdr->e_ident[EI_VERSION] == EV_CURRENT) ? "EV_CURRENT" :
			"unknown");
	DBPRINT (0, "e_ident[EI_OSABI]=%d<%s>\n", hdr->e_ident[EI_OSABI],
			(hdr->e_ident[EI_OSABI] <= ELFOSABI_OPENBSD) ? 
			e_osabi_str[hdr->e_ident[EI_OSABI]] :
			"unknown");
	DBPRINT (0, "e_ident[EI_ABIVERSION]=%d\n", hdr->e_ident[EI_ABIVERSION]);

	if (hdr->e_ident[EI_MAG0] != ELFMAG0 ||
			hdr->e_ident[EI_MAG1] != ELFMAG1 ||
			hdr->e_ident[EI_MAG2] != ELFMAG2 ||
			hdr->e_ident[EI_MAG3] != ELFMAG3 ||
			hdr->e_ident[EI_CLASS] != ELFCLASS64 ||
			hdr->e_ident[EI_DATA] != ELFDATA2LSB ||
			hdr->e_ident[EI_VERSION] != EV_CURRENT ||
			hdr->e_ident[EI_OSABI] != ELFOSABI_SYSV) {
		DBPRINT (1, "unexpected e_ident\n");
		error = 1;
	}



	/* e_type */
	DBPRINT (0, "e_type=%d<%s>\n", hdr->e_type,
			(hdr->e_type <= ET_NUM) ?  e_type_str[hdr->e_type] : "unkown");
	if (hdr->e_type != ET_DYN) {
		DBPRINT (1, "unexpected e_type %x\n", hdr->e_type);
		error = 1;
	}

	/* e_machine */
	DBPRINT (0, "e_machine=%d<%s>\n", hdr->e_machine,
			(hdr->e_machine == EM_IA_64) ? "EM_IA_64" :
			(hdr->e_machine == EM_386) ? "EM_386" :
			(hdr->e_machine == EM_NONE) ? "EM_NONE" :
			"unknown");
	if (hdr->e_machine != EM_IA_64) {
		DBPRINT (1, "unexpected e_machine\n");
		error = 1;
	}
	/* e_version */
	DBPRINT (0, "e_version=%d<%s>\n", hdr->e_version, 
			(hdr->e_version == EV_NONE) ? "EV_NONE" :
			(hdr->e_version == EV_CURRENT) ? "EV_CURRENT" :
			"unknown");
	if (hdr->e_version != EV_CURRENT) {
		DBPRINT (1, "unexpected e_version\n");
		error = 1;
	}


	/* print debug info */ 

	/* e_entry */
	DBPRINT (0, "e_entry=%lx\n", hdr->e_entry);
	/* e_phoff */
	DBPRINT (0, "e_phoff=%lx\n", hdr->e_phoff);
	/* e_shoff */
	DBPRINT (0, "e_shoff=%lx\n", hdr->e_shoff);
	/* e_flags */
	DBPRINT (0, "e_flags=%x\n", hdr->e_flags);
	/* e_ehsize */
	DBPRINT (0, "e_ehsize=%d\n", hdr->e_ehsize);
	/* e_phentsize */
	DBPRINT (0, "e_phentsize=%d\n", hdr->e_phentsize);
	/* e_phnum */
	DBPRINT (0, "e_phnum=%d\n", hdr->e_phnum);
	/* e_shentsize */
	DBPRINT (0, "e_shentsize=%d\n", hdr->e_shentsize);
	/* e_shnum */
	DBPRINT (0, "e_shnum=%d\n", hdr->e_shnum);
	/* e_shstrndx */
	DBPRINT (0, "e_shstrndx=%d\n", hdr->e_shstrndx);

	return error;
}
static int load_program_headers (int fd, Elf64_Ehdr * elfhdr, Elf64_Phdr ** phdrp)
{
	u64 phsize, size;
	void * p;

	phsize = elfhdr->e_phnum * elfhdr->e_phentsize;
	size = ROUNDUP (elfhdr->e_phoff + phsize, getpagesize());

	/* this mapping will be overwritten when the segments are loaded
	 * it doesn't really matters because it is RO anyway */
	p = (void*)SYSCALL (mmap, (TEXT_SEGMENT_BASE_ADDR,
			size, PROT_READ, MAP_FIXED | MAP_PRIVATE, fd, 0));

	if (IS_SYS_ERROR (p)) {
		DBPRINT (1, "elfhdr->e_phoff=%ld  pagesize=%ld size=%ld\n", elfhdr->e_phoff+phsize, getpagesize(), size);
		DBPRINT (1, "mmap failed returns %d\n\n", p);
		return -1;
	}
	DBPRINT (0, "addr=%p size=%lx FIXED\n", p, size);
	*phdrp = p + elfhdr->e_phoff;

	return 0;
}


static int read_prog_hdrs (Elf64_Phdr * phdr, Elf64_Half phnum)
{
	Elf64_Half j;
	int error = 0;

	for (j = 0; j < phnum; j ++) {
		error |= read_prog_entry (&phdr[j]);
	}
	return error;
}






static int read_prog_entry (Elf64_Phdr * phdr)
{
	int error = 0;

#if DEBUG
	DBPRINT (0, "phdr=%p:\n", phdr);
	/* p_type */
	DBPRINT (0, "\tp_type=%x<%s>\n", phdr->p_type,
			(phdr->p_type <= PT_NUM) ? p_type_str[phdr->p_type] :
			(phdr->p_type == PT_IA_64_UNWIND) ? "PT_IA_64_UNWIND" :
			(phdr->p_type == PT_GNU_STACK) ? "PT_GNU_STACK" :
			"unknown");

	/* p_flags */
	DBPRINT (0, "\tp_flags=%x\n", phdr->p_flags);

	/* p_offset */
	DBPRINT (0, "\tp_offset=%lx\n", phdr->p_offset);

	/* p_vaddr */
	DBPRINT (0, "\tp_vaddr=%lx\n", phdr->p_vaddr);

	/* p_filesz */
	DBPRINT (0, "\tp_filesz=%lx\n", phdr->p_filesz);

	/* p_memsz */
	DBPRINT (0, "\tp_memsz=%lx\n", phdr->p_memsz);

	/* p_align */
	DBPRINT (0, "\tp_align=%lx\n", phdr->p_align);
#endif
	switch (phdr->p_type) {
		case PT_LOAD:
			{
				struct loadcmd * c = &m_loadcmds[m_nloadcmds++];
				assert (m_nloadcmds < 3);
				if ((phdr->p_align & (getpagesize() - 1)) != 0) {
					DBPRINT (1, "ELF load command alignment not page-aligned");
					error = 1;
					break;
				}
				if (((phdr->p_vaddr - phdr->p_offset) & (phdr->p_align - 1)) != 0) {
					DBPRINT (1, "ELF load command address/offset not properly aligned");
					error = 1;
					break;
				}

				c->mapstart = phdr->p_vaddr & ~(phdr->p_align - 1);
				c->mapend = ((phdr->p_vaddr + phdr->p_filesz + getpagesize() - 1) & ~(getpagesize() - 1));
				c->dataend = phdr->p_vaddr + phdr->p_filesz;
				c->allocend = phdr->p_vaddr + phdr->p_memsz;
				c->mapoff = phdr->p_offset & ~(phdr->p_align - 1);

				/* Determine whether there is a gap between the last segment and this one */
				if (m_nloadcmds > 1 && c[-1].mapend != c->mapstart)
					m_has_holes = 1;

				c->prot = 0;
				if (phdr->p_flags & PF_R)
					c->prot |= PROT_READ;
				if (phdr->p_flags & PF_W)
					c->prot |= PROT_WRITE;
				if (phdr->p_flags & PF_X)
					c->prot |= PROT_EXEC;
			}
			break;
		case PT_DYNAMIC:
			m_dyn_phdr = phdr;
			break;
		case PT_TLS:
			m_tls_phdr = phdr;
			break;
		case PT_NULL:
		case PT_INTERP:
		case PT_NOTE:
		case PT_SHLIB:
		case PT_PHDR:
		case PT_NUM:
		case PT_IA_64_UNWIND:
		case PT_GNU_STACK:
			break;
		default:
			DBPRINT (1, "unknown program header type\n");
	}

	return error;
}

static int read_dyn_section (Elf64_Phdr * phdr)
{
	int error = 0;
	int stop = 0;
	Elf64_Half dyn_symtab_entry_size = 0;

	assert (g_load_addr);
	_DYNAMIC = (Elf64_Dyn *)(phdr->p_vaddr + g_load_addr);
#if GDB_INTERFACE
	g_maps.l_ld = _DYNAMIC;
#endif
	/* print contents */
	DBPRINT (0, "_DYNAMIC=%p .dynamic section:\n", &_DYNAMIC[0]);
	u64 j;
	for (j = 0; ! stop ; j ++) {
		switch (_DYNAMIC[j].d_tag) {
			case DT_NULL:
				stop = 1;
				break;
			case DT_REL:
				m_rel_ptr = (Elf64_Rel*)(_DYNAMIC[j].d_un.d_ptr + g_load_addr);
				break;
			case DT_RELA:
				m_rela_ptr = (Elf64_Rela*)(_DYNAMIC[j].d_un.d_ptr + g_load_addr);
				break;
			case DT_RELSZ:
				m_rel_size = _DYNAMIC[j].d_un.d_val;
				break;
			case DT_RELASZ:
				m_rela_size = _DYNAMIC[j].d_un.d_val;
				break;
			case DT_SYMTAB:
				m_dyn_symtab = (Elf64_Sym*)(_DYNAMIC[j].d_un.d_ptr + g_load_addr);
				DBPRINT (0, "m_dyn_symtab=%p\n", m_dyn_symtab);
				break;
			case DT_SYMENT:
				dyn_symtab_entry_size = _DYNAMIC[j].d_un.d_val;
				break;
			case DT_PLTGOT:
				m_gp = _DYNAMIC[j].d_un.d_ptr + g_load_addr;
				DBPRINT (0, "gp=%lx m_gp=%lx\n", _DYNAMIC[j].d_un.d_ptr, m_gp);
				break;
			case DT_HASH:
				setup_hash (&_DYNAMIC[j]);
				break;
			case DT_GNU_HASH:
				setup_gnu_hash (&_DYNAMIC[j]);
				break;
			case DT_STRTAB:
				m_dyn_sym_strtab = (char *)(_DYNAMIC[j].d_un.d_ptr + g_load_addr);
				break;
			case DT_JMPREL:
				m_plt = (Elf64_Rela *)(_DYNAMIC[j].d_un.d_ptr + g_load_addr);
				break;
			case DT_PLTREL:
				assert (_DYNAMIC[j].d_un.d_val == DT_RELA);
				break;
			case DT_PLTRELSZ:
				m_plt_size = _DYNAMIC[j].d_un.d_val;
				break;
		}


		/*
		DBPRINT (0, "\t_DYNAMIC[%ld]\n\t\ttag=%lx<%s>\n\t\tun=%lx\n",
				j, _DYNAMIC[j].d_tag,
				(_DYNAMIC[j].d_tag <= DT_NUM) ? d_tag_str[_DYNAMIC[j].d_tag] :
				(_DYNAMIC[j].d_tag <= DT_IA_64_PLT_RESERVE && _DYNAMIC[j].d_tag >= DT_VERSYM) ? d_tag_hi_str[_DYNAMIC[j].d_tag - DT_VERSYM] :
				"unknown", 
				_DYNAMIC[j].d_un.d_val);
				*/
	}
#if DEBUG
	/* get the number of dyn symbols */
	/* XXX Apparently the only way to find out the size of the dynamic
	 *symbol section is to assume that the string table follows right
	 *afterwards...  */
	assert (dyn_symtab_entry_size);
	m_dyn_symtab_entry_num = (m_dyn_sym_strtab - (char*)m_dyn_symtab) / dyn_symtab_entry_size;
	DBPRINT (0, "m_dyn_symtab_entry_num=%d\n", m_dyn_symtab_entry_num);
	assert (m_dyn_symtab_entry_num < FPTR_TABLE_SIZE);
#endif
	return error;
}


static int read_reloc ()
{
	Elf64_Word j, rel_ent_num, rela_ent_num;
	int error = 0;
	Elf64_Addr * reloc_addr, value;
	Elf64_Sym * sym;


	DBPRINT (0, "m_rel_ptr=%p m_rel_size=%d m_rela_ptr=%p m_rela_size=%d\n",
			m_rel_ptr, m_rel_size, m_rela_ptr, m_rela_size);
	rel_ent_num = m_rel_size / sizeof (Elf64_Rel);
	rela_ent_num = m_rela_size / sizeof (Elf64_Rela);
	DBPRINT (0, "rel_ent_num=%d rela_ent_num=%d\n", rel_ent_num, rela_ent_num);
	assert (rel_ent_num == 0);

	/* read rela */
	for (j = 0; j < rela_ent_num; j ++) {
		Elf64_Xword type = ELF64_R_TYPE (m_rela_ptr[j].r_info); 
		assert (type == R_IA64_REL64LSB ||
			type == R_IA64_NONE ||
			type == R_IA64_TPREL64LSB ||
			type == R_IA64_DTPMOD64LSB ||
			type == R_IA64_FPTR64LSB);


		sym = m_dyn_symtab + ELF64_R_SYM(m_rela_ptr[j].r_info);
#if DEBUG
		char * str UNUSED = get_reloc_type_str (type);

		DBPRINT (0, "\t\toffset=%lx sym=%lx", 
				m_rela_ptr[j].r_offset,
				ELF64_R_SYM (m_rela_ptr[j].r_info));



		DBPRINT (0, " type=%lx<%s> addend=%lx\n",
				type,
				(str == NULL) ? "unknown" : str,
				m_rela_ptr[j].r_addend);

		DBPRINT (0, "\t\t\tsymbol=");
		DBCODE (0, print_symbol (sym, m_dyn_sym_strtab));
		DBPRINT (0, "\n");

#endif
		reloc_addr = (void *)(g_load_addr + m_rela_ptr[j].r_offset);
		assert (type == R_IA64_NONE || (u64)reloc_addr > m_text_end);
		switch (type) {
			case R_IA64_REL64LSB:
				value = do_rel64lsb_reloc (&m_rela_ptr[j], reloc_addr);
				break;
			case R_IA64_FPTR64LSB:
				value = do_fptr64lsb_reloc (&m_rela_ptr[j], reloc_addr, sym);
				break;
			case R_IA64_TPREL64LSB:
				value = do_tprel64lsb_reloc (&m_rela_ptr[j], reloc_addr, sym);
				break;
			case R_IA64_DTPMOD64LSB:
				value = do_dtpmod64lsb_reloc (&m_rela_ptr[j], reloc_addr);
				break;
			case R_IA64_NONE:
				break;
			default:
				DBPRINT (1, "unknown reloc\n");
				goto out;
		}
		if (type != R_IA64_NONE) {
			DBPRINT (0, "\t\t\treloc_addr=%p value=%lx\n", reloc_addr, value);
			*reloc_addr = value;
		}

	}
out:
	return error;
}

/*****************************************************************************/
/* tls structure is as follows -
 *
 * -----------------------------------------------------
 *  ^       ^             ^ ^              ^
 *  |       |             |a|              |
 *  |       | t_clstp     |l| btlib's tls  |
 *  |       | t_ptrace    |i|              |
 *  |       |             |g|              |
 *  |       |             |n|              |
 *  r13     |                              |
 *   T C B  |              T L S           |
 * -------------------------------------------------
 * 
 *
 */ 
/*****************************************************************************/

static int set_tls (Elf64_Phdr * tlshdr)
{
	long ret;
	u64 btlib_tls_offset;
#if DEBUG
	void * tls_addr;
	u64 tls_size;

	char * p;
	/* we assume btlib's tls is initialized to zero */
	assert (g_load_addr);
	tls_addr = (void *)(tlshdr->p_vaddr + g_load_addr);
	tls_size = tlshdr->p_filesz;

	for (p = tls_addr; (u64)(p - (char*)tls_addr) < tls_size; p ++) {
		assert (*p == 0);
	}
#endif
 	/* btlib_tls_offset holds the offset of btlib's tls from the beginning of tls */
	assert (tlshdr->p_align);
	btlib_tls_offset = ROUNDUP (sizeof (t_clstp) + sizeof (t_ptrace), tlshdr->p_align);

	/* set the offset of btlib's tls from r13 */
	g_tls_offset = ROUNDUP (btlib_tls_offset + sizeof (tcbhead_t), tlshdr->p_align);
	assert ((g_tls_offset % tlshdr->p_align) == 0);

	DBPRINT (0, "btlib_tls_offset=%lx g_tls_offset=%lx\n", btlib_tls_offset, g_tls_offset);
	/* arrange the tls for the first thread */
	ret = allocate_initial_tls (tlshdr);
	DBPRINT (ret, "allocate_initial_tls failed:%ld\n", ret);
	if (ret == 0) {
		m_tls_allocated = 1;
	}
	return ret;
}

static Elf64_Addr do_tprel64lsb_reloc (Elf64_Rela * rela, Elf64_Addr * reloc_addr UNUSED, Elf64_Sym * sym)
{
	Elf64_Addr value;

	value = g_tls_offset + sym->st_value + rela->r_addend;

	return value;
}

static Elf64_Addr do_dtpmod64lsb_reloc (Elf64_Rela * rela UNUSED, Elf64_Addr * reloc_addr UNUSED)
{
	Elf64_Addr value;
	value = 1; /* just a dummy value. it will not be used */

	return value;
}
static Elf64_Addr do_rel64lsb_reloc (Elf64_Rela * rela UNUSED, Elf64_Addr * reloc_addr)
{
	Elf64_Addr value;

	value = *reloc_addr + g_load_addr; /* this is what glibc does */

	return value;
}
static inline struct _fdesc_tag * set_official_fptr (u64 ip, u64 gp, Elf64_Sym * sym)
{
	long indx;
	struct _fdesc_tag * fdesc;

	indx = sym - m_dyn_symtab;
#if DEBUG
	assert (m_dyn_symtab_entry_num);
	assert (indx >= 0 && indx < m_dyn_symtab_entry_num); 
#endif

	fdesc = &m_fdesc_table[indx];
	if (fdesc->ip == 0) {
		DBPRINT (0, "%s: fdesc[%ld] ip=%lx gp=%lx\n", __FUNCTION__, indx, ip, m_gp);
		fdesc->ip = ip;
		fdesc->gp = gp;
	} else {
		assert (fdesc->ip == ip);
	}

	return fdesc;
}


static Elf64_Addr do_fptr64lsb_reloc (Elf64_Rela * rela, Elf64_Addr * reloc_addr UNUSED, Elf64_Sym * sym)
{
	Elf64_Addr value, ip;
	struct _fdesc_tag * fdesc;

	ip = g_load_addr + sym->st_value + rela->r_addend;
	assert (m_gp);
	/* create a new official fptr */
	fdesc = set_official_fptr (ip, m_gp ,sym);


	value = (Elf64_Addr) fdesc;


	return value;
}

static int setup_hash (Elf64_Dyn * dh)
{
	Elf64_Word *hash;
        Elf64_Word nchain;

        m_hash_info.hash_type = SYSV_HASH;
        hash = (void *) dh->d_un.d_ptr + g_load_addr;


        m_hash_info.nbuckets = *hash++;
        nchain = *hash++;
        m_hash_info.buckets = hash;
        hash += m_hash_info.nbuckets;
        m_hash_info.chain = hash;

        return 0;
}

static int setup_gnu_hash (Elf64_Dyn * dh)
{
        Elf64_Word *hash;
        Elf64_Word symbias;
        Elf64_Word bitmask_nwords;

        m_hash_info.hash_type = GNU_HASH;
        hash = (void*) dh->d_un.d_ptr + g_load_addr;
        m_hash_info.nbuckets = *hash++;
        symbias = *hash++;
        bitmask_nwords = *hash++;
        m_hash_info.gnu_bitmask_idxbits = bitmask_nwords - 1;
        m_hash_info.gnu_shift = *hash++;
        m_hash_info.gnu_bitmask = hash;
        hash +=  (bitmask_nwords * 2);
        m_hash_info.gnu_buckets = hash;
        hash += m_hash_info.nbuckets;
        m_hash_info.gnu_chain_zero = hash - symbias;
        return 0;

}


static int load_segments (int fd)
{
	struct loadcmd * c;
	const char * errstring = NULL;
	u64 maplength;

	c = &m_loadcmds[0];

	/* Length of the sections to be loaded. */
	maplength = m_loadcmds[m_nloadcmds - 1].allocend - c->mapstart;

	/* This is a position-independent shared object. We can let the
	 * kernel map it anywhere it likes, but we must have space for all
	 * the segments in their specified positions relative to the first.
	 * So we map the first segment without MAP_FIXED, but with its
	 * extent increased to cover all the segments.	Then we remove
	 * access from excess portion, and there is known sufficient space
	 * there to remap from the later segments. */

	/* Remember which part of the address space this object uses.	*/
	m_map_start = (Elf64_Addr) SYSCALL (mmap, (TEXT_SEGMENT_BASE_ADDR,
			maplength, c->prot, MAP_FIXED | MAP_PRIVATE, fd, c->mapoff));
	if (IS_SYS_ERROR (m_map_start)) {
		errstring = "failed to map segment from shared object";
		goto return_error;
	}
	DBPRINT (0, "addr=%p size=%lx FIXED\n", m_map_start, maplength);

	m_map_end = m_map_start + maplength;
	assert (c->mapstart == 0);
	g_load_addr = m_map_start;

	if (m_has_holes)
		/* Change protection on the excess portion to disallow all access;
		 * the portions we do not remap later will be inaccessible as if
		 * unallocated. Then jump into the normal segment-mapping loop to
		 * handle the portion of the segment past the end of the file
		 * mapping.	*/
		SYSCALL (mprotect, ((u64) (g_load_addr + c->mapend),
				m_loadcmds[m_nloadcmds - 1].allocend -
				c->mapend, PROT_NONE));


	while (c < &m_loadcmds[m_nloadcmds]) {
		if (c != &m_loadcmds[0]) { /* first segment was mapped at the beginning of the func */
 			assert (c->mapend > c->mapstart);
			/* Map the segment contents from the file */
			void * seg_addr;
			seg_addr = (void*)SYSCALL (mmap, (g_load_addr + c->mapstart,
					c->mapend - c->mapstart, c->prot,
					MAP_FIXED | MAP_PRIVATE, fd, c->mapoff));
			if (IS_SYS_ERROR (seg_addr)) {
				errstring = "failed to map segment from shared object";
				goto return_error;
			}
			DBPRINT (0, "addr=%p size=%lx FIXED\n", seg_addr, c->mapend - c->mapstart);
		}

		if (c->prot & PROT_EXEC)
			m_text_end = g_load_addr + c->mapend;


		if (c->allocend > c->dataend) {
			/* Extra zero pages should appear at the end of this segment,
			 * after the data mapped from the file. */
			Elf64_Addr zero, zeroend, zeropage;

			zero = g_load_addr + c->dataend;
			zeroend = g_load_addr + c->allocend;
			zeropage = ((zero + getpagesize() - 1) & ~(getpagesize() - 1));

			if (zeroend < zeropage)
				/* All the extra data is in the last page of the segment.
				 * We can just zero it. */
				zeropage = zeroend;

			if (zeropage > zero) {
				/* Zero the final part of the last page of the segment. */
				if ((c->prot & PROT_WRITE) == 0) {
					/* Dag nab it.	*/
					s64 ret;
					ret = SYSCALL (mprotect, ((u64) (zero & ~ (getpagesize() - 1)), getpagesize(), c-> prot | PROT_WRITE)); 
					if (IS_SYS_ERROR (ret)) {
						errstring = "cannot change memory protections";
						goto return_error;
					}
				}
				bt_memset ((void *) zero, '\0', zeropage - zero);
				if ((c->prot & PROT_WRITE) == 0)
					SYSCALL (mprotect, ((u64) (zero & ~(getpagesize() - 1)), getpagesize(), c->prot));
			}

			if (zeroend > zeropage) {
				/* Map the remaining zero pages in from the zero fill FD.		*/
				u64 mapat;
				mapat = SYSCALL (mmap, ((u64) zeropage,
						zeroend - zeropage,
						c->prot,
						MAP_ANON | MAP_PRIVATE | MAP_FIXED,
						-1, 0));
				if (IS_SYS_ERROR (mapat)) {
					errstring = "cannot map zero-fill pages";
					goto return_error;
				}
				DBPRINT (0, "addr=%p size=%lx FIXED\n",
						mapat, zeroend - zeropage);
			}
		}

		++c;
	}

	DBPRINT (0, "%s: map_start=%lx map_end=%lx addr=%lx text_end=%lx\n",
			__FUNCTION__, m_map_start, m_map_end, g_load_addr, m_text_end);


#if GDB_INTERFACE
	g_maps.l_addr = g_load_addr;
#endif
	return 0;
return_error:

	DBPRINT (1, "%s: %s\n", __FUNCTION__, errstring);
	return -1;
}

static unsigned long int elf_hash( const char* string)
{
        return _dl_elf_hash(string);
}

/* gnu hash function */
static unsigned long int elf_gnu_hash(const char* string)
{
        u64 h = 5381;
        unsigned char c;
        for( c=*string; c!='\0';c=*++string )
                h = h*33 + c;
        return h & 0xffffffff;
}
static Elf64_Sym * lookup_symbol (char * name)
{
        Elf64_Word hash;
        Elf64_Word symidx;
        Elf64_Sym *sym;
        char* symname;

        if( m_hash_info.hash_type == GNU_HASH ){
             hash = elf_gnu_hash (name);
             Elf64_Word bucket = m_hash_info.gnu_buckets[hash % m_hash_info.nbuckets];
             if (bucket != 0) {
                  const Elf64_Word *hasharr = &(m_hash_info.gnu_chain_zero[bucket]);
                  do
                     if ((*hasharr & ~1u) == (hash & ~1u)){
                        symidx = hasharr - m_hash_info.gnu_chain_zero;
                        sym = &m_dyn_symtab[symidx];
                        symname = m_dyn_sym_strtab + sym->st_name;
                        if (bt_strcmp (symname, name) == 0) {
                                return sym;
                        }
                     }
                  while ((*hasharr++ & 1u) == 0);
             }
        }
        else{
            hash = elf_hash (name);
            for (symidx = m_hash_info.buckets[hash % m_hash_info.nbuckets];
                        symidx != STN_UNDEF;
                        symidx = m_hash_info.chain[symidx]) {
                sym = &m_dyn_symtab[symidx];
                symname = m_dyn_sym_strtab + sym->st_name;
                if (bt_strcmp (symname, name) == 0) {
                        return sym;
                }
            }

        }
        return NULL;
}



static int lookup_btlib_entry_points ()
{
	struct _btlib_entry_points_tag * bep;
	Elf64_Sym * sym;
	u64 ip;
	struct _fdesc_tag * fdesc;

	for (bep = &g_btlib_entry_points[0]; bep->name[0]; bep ++) {
		sym = lookup_symbol (bep->name);
		if (sym == NULL) {
			return -1;
		}

		assert (ELF64_ST_TYPE (sym->st_info) == STT_FUNC);
		ip = g_load_addr + sym->st_value;
		assert (m_gp);
		fdesc = set_official_fptr (ip, m_gp ,sym);
		*bep->fdesc = fdesc;
		DBPRINT (0, "%s: name=%s fdesc.ip=%lx\n",
				__FUNCTION__, bep->name, fdesc->ip);
	}

	return 0;
}

static int read_plt (void)
{
	Elf64_Word j, plt_ent_num;
	//Elf64_Addr * reloc_addr;
	struct _fdesc_tag * reloc_addr;
	struct _fdesc_tag * value = NULL;
	Elf64_Sym * sym;
	char * name;
	int is_ignore_sym;

	plt_ent_num = m_plt_size / sizeof (Elf64_Rela);
	DBPRINT (0, "plt_ent_num=%d\n", plt_ent_num);

	assert (m_gp);
	for (j = 0; j < plt_ent_num; j ++) {


		is_ignore_sym = 0;

		sym = m_dyn_symtab + ELF64_R_SYM(m_plt[j].r_info);
		name = sym->st_name + m_dyn_sym_strtab;
#if DEBUG
		Elf64_Xword type = ELF64_R_TYPE (m_plt[j].r_info); 
		assert (type == R_IA64_IPLTLSB);
		char * str UNUSED = get_reloc_type_str (type);
		DBPRINT (0, "\t\toffset=%lx sym=%lx", 
				m_plt[j].r_offset,
				ELF64_R_SYM (m_plt[j].r_info));



		DBPRINT (0, " type=%lx<%s> addend=%lx\n",
				type,
				(str == NULL) ? "unknown" : str,
				m_plt[j].r_addend);

		DBPRINT (0, "\t\t\tsymbol=");
		DBCODE (0, print_symbol (sym, m_dyn_sym_strtab));
		DBPRINT (0, "\n");

#endif

		/* do plt reloc */
		reloc_addr = (void *)(g_load_addr + m_plt[j].r_offset);

		if (bt_strcmp (name, "__tls_get_addr") == 0) {
			value = (void*)__tls_get_addr;
		} else if (bt_strcmp (name, "__gmon_start__") == 0) {
			value = (void*)__gmon_start__;
		} else if (bt_strcmp (name, "get_loader_ptrace_flag_addr") == 0) {
			value = (void*)get_loader_ptrace_flag_addr;
		} else if (bt_strcmp (name, "cleanup_tls") == 0) {
			value = (void*)cleanup_tls;
		} else if (bt_strcmp (name, "get_tls_static_info") == 0) {
			value = (void*)get_tls_static_info;
		} else if (bt_strcmp (name, "init_tls") == 0) {
			value = (void*)init_tls;
		} else if (bt_strcmp (name, "memset") == 0) {
			value = (void*)memset;
		} else if (bt_strcmp (name, "memcpy") == 0) {
			value = (void*)memcpy;
		} else if (
			/* ignored symbols */
			(bt_strcmp (name, "pthread_create") == 0) ||
			(bt_strcmp (name, "malloc") == 0) ||
			(bt_strcmp (name, "calloc") == 0) ||
			(bt_strcmp (name, "realloc") == 0) ||
			(bt_strcmp (name, "free") == 0) 
			  ) {
			is_ignore_sym = 1;
		} else {
			DBPRINT (1, "%s: unexpected symbol in plt:<%s>\n",
					__FUNCTION__, name);
			return -1;
		}

		DBPRINT (0, "reloc_addr=%p ip=%lx gp=%lx\n",
				reloc_addr, value->ip, value->gp);

		if ( ! is_ignore_sym ) {
			*reloc_addr = *value;
		}
	}

	return 0;
}
#if DEBUG
static void print_symtab (int fd, Elf64_Shdr * shdr)
{
	long ret;
	u64 j, ent_num;
	Elf64_Sym * symtab = NULL;

	symtab = zmalloc (shdr->sh_size);
	if (symtab == NULL) {
		DBPRINT (1, "%s: zmalloc failed\n", __FUNCTION__);
		return;
	}

	ret = SYSCALL (pread, (fd, symtab, shdr->sh_size, shdr->sh_offset));;
	if (IS_SYS_ERROR (ret)) {
		DBPRINT (1, "%s: pread failed\n", __FUNCTION__);
		return;
	}
	assert ((u64)ret == shdr->sh_size);

	ent_num = shdr->sh_size / shdr->sh_entsize;

	DBPRINT (0, "symbol table: size=%ld entsize=%ld ent_num=%lu\n",
			shdr->sh_size, shdr->sh_entsize, ent_num);
	for (j = 0; j < ent_num; j ++) {
		fprintf (stdout, "\tsymbol[%ld]=", j);
		DBCODE (0, print_symbol (&symtab[j], m_sym_strtab));
		fprintf (stdout, "\n");
	}

	zfree (symtab);

}
static void print_symbol (Elf64_Sym * sym, char * strtab)
{
	fprintf (stdout, "[name=%d<%s> info=%d bind=%d<%s> type=%d<%s> shndx=%d value=%lx size=%lx]",
			sym->st_name,
			strtab + sym->st_name,
			sym->st_info,
			ELF64_ST_BIND(sym->st_info),
			(ELF64_ST_BIND(sym->st_info) == STB_LOCAL) ? "STB_LOCAL" :
			(ELF64_ST_BIND(sym->st_info) == STB_GLOBAL) ? "STB_GLOBAL" :
			(ELF64_ST_BIND(sym->st_info) == STB_WEAK) ? "STB_WEAK" :
			"unknown",
			ELF64_ST_TYPE(sym->st_info),
			(ELF64_ST_TYPE(sym->st_info) == STT_NOTYPE) ? "STT_NOTYPE" :
			(ELF64_ST_TYPE(sym->st_info) == STT_OBJECT) ? "STT_OBJECT" :
			(ELF64_ST_TYPE(sym->st_info) == STT_FUNC) ? "STT_FUNC" :
			(ELF64_ST_TYPE(sym->st_info) == STT_SECTION) ? "STT_SECTION" :
			"unknown",
			sym->st_shndx,
			sym->st_value,
			sym->st_size);
}


static void print_section_entry (Elf64_Shdr * shdr)
{

	/* sh_name */
	fprintf (stdout, "\tsh_name=%s\n", m_sh_strtab + shdr->sh_name);
	/* sh_type */
	fprintf (stdout, "\tsh_type=%x<%s>\n",
			shdr->sh_type,
			(shdr->sh_type <= SHT_SYMTAB_SHNDX) ? sh_type_str[shdr->sh_type] :
			(shdr->sh_type == SHT_HIOS) ? "SHT_HIOS" :
			(shdr->sh_type == SHT_GNU_verdef) ? "SHT_GNU_verdef" :
			(shdr->sh_type == SHT_IA_64_UNWIND) ? "SHT_IA_64_UNWIND" :
			(shdr->sh_type == SHT_GNU_verneed) ? "SHT_GNU_verneed" :
			"unknown");
	/* sh_flags */
	fprintf (stdout, "\tsh_flags=%lx<", shdr->sh_flags);

	{
		unsigned j;
		int first = 1;
		int unknown = 0;
		for (j = 1; j < 11; j ++) {
			if (shdr->sh_flags & (1 << j)) {
				fprintf (stdout, "%s%s", first ? "" : " | ", shf_str[j]);
				first = 0;
			}
		}
		for (; j < sizeof (shdr->sh_flags) * BITS_PER_BYTE; j ++) {


			if (shdr->sh_flags & (1 << j)) {
				if ((1 << j) == SHF_IA_64_SHORT) {
					fprintf (stdout, "%s%s", first ? "" : " | ", "SHF_IA_64_SHORT");
				} else if ((1 << j) == SHF_IA_64_NORECOV) {
					fprintf (stdout, "%s%s", first ? "" : " | ", "SHF_IA_64_NORECOV");
				} else {
					unknown = 1;
				}
			}
		}
		if (unknown) {
			fprintf (stdout, "%s", first ? "UNKNOWN" : " | UNKNOWN");
		}
		if (unknown == 0 && first == 1) {
			fprintf (stdout, "NONE");
		}
		fprintf (stdout, ">\n");
	}
	/* sh_addr */
	fprintf (stdout, "\tsh_addr=%lx\n", shdr->sh_addr);
	/* sh_offset */
	fprintf (stdout, "\tsh_offset=%lx\n", shdr->sh_offset);
	/* sh_size */
	fprintf (stdout, "\tsh_size=%lx\n", shdr->sh_size);
	/* sh_link */
	fprintf (stdout, "\tsh_link=%d\n", shdr->sh_link);
	/* sh_info */
	fprintf (stdout, "\tsh_info=%d\n", shdr->sh_info);
	/* sh_addralign */
	fprintf (stdout, "\tsh_addralign=%lx\n", shdr->sh_addralign);
	/* sh_entsize */
	fprintf (stdout, "\tsh_entsize=%lx\n", shdr->sh_entsize);

}

static void print_dyn_section ()
{
	int stop = 0;
	u64 j;

	DBPRINT (0, ".dynamic section:\n");
	for (j = 0; ! stop ; j ++) {
		switch (_DYNAMIC[j].d_tag) {
			case DT_NULL:
				stop = 1;
				break;
		}


		DBPRINT (0, "\t_DYNAMIC[%ld]\n\t\ttag=%lx<%s>\n\t\tun=%lx\n",
				j, _DYNAMIC[j].d_tag,
				(_DYNAMIC[j].d_tag <= DT_NUM) ? d_tag_str[_DYNAMIC[j].d_tag] :
				(_DYNAMIC[j].d_tag <= DT_IA_64_PLT_RESERVE && _DYNAMIC[j].d_tag >= DT_VERSYM) ? d_tag_hi_str[_DYNAMIC[j].d_tag - DT_VERSYM] :
				"unknown", 
				_DYNAMIC[j].d_un.d_val);
	}
}
static int read_section_hdrs (int fd, Elf64_Shdr * shdr, Elf64_Half shnum)
{
	Elf64_Half j;
       	Elf64_Half symtab_indx, dyn_symtab_indx, hash_symtab_indx;
	int error = 0;

	for (j = 0; j < shnum; j ++) {

		DBPRINT (0, "shdr[%d]=%p\n", j, &shdr[j]);
		DBCODE (0, print_section_entry (&shdr[j]));
		if (shdr[j].sh_type == SHT_DYNSYM) {
			dyn_symtab_indx = j;
			/* load the symbol-name string table */
		} else if (shdr[j].sh_type == SHT_SYMTAB) {
			symtab_indx = j;
			/* load the symbol-name string table */
			error = load_strtab (fd, shdr, shdr[j].sh_link, &m_sym_strtab);
			DBCODE (0, print_symtab (fd, &shdr[j]));
		} else if (shdr[j].sh_type == SHT_HASH) {
			hash_symtab_indx = shdr[j].sh_link;
		} else if (shdr[j].sh_type == SHT_GNU_HASH) {
                        hash_symtab_indx = shdr[j].sh_link;
                }
	}
	assert (hash_symtab_indx = dyn_symtab_indx);

	return error;
}
static int load_strtab (int fd, Elf64_Shdr * shdr, long indx, char **strtabp)
{
	u64 size;
	u64 offset;
	long ret;

	size = shdr[indx].sh_size;
	offset = shdr[indx].sh_offset;
	DBPRINT (0, "%s: size=%lx offset=%lx addr=%lx\n",
			__FUNCTION__, size, offset,
			offset + g_load_addr);
	*strtabp= zmalloc (size);
	if (*strtabp== NULL) {
		DBPRINT (1, "%s: zmalloc failed\n", __FUNCTION__);
		return -1;
	}
	ret = SYSCALL (pread, (fd, *strtabp, size, offset));;
	if (IS_SYS_ERROR (ret)) {
		DBPRINT (1, "pread strtab failed:%d\n", ret);
		return -1;
	}
	assert ((u64)ret == size);
	return 0;
}
static int load_section_headers (int fd, Elf64_Ehdr * elfhdr, Elf64_Shdr ** shdrp)
{
	u64 shsize;
	long ret;
	shsize = elfhdr->e_shnum * elfhdr->e_shentsize;
	*shdrp = zmalloc (shsize);
	if (*shdrp == NULL) {
		DBPRINT (1, "%s: zmalloc failed\n", __FUNCTION__);
		return -1;
	}
	ret = SYSCALL (pread, (fd, *shdrp, shsize, elfhdr->e_shoff));;
	if (IS_SYS_ERROR (ret)) {
		DBPRINT (1, "pread phdr failed:%d\n", ret);
		return -1;
	}
	assert ((u64)ret == shsize);

	return 0;
}
#endif

